% PAL (Policy Analysis using Logic-Programming)
% Copyright (C) 2004 Beata Sarna-Starosta
% Copyright (C) 2004 Scott D. Stoller
% Copyright (c) 2004 Stony Brook University
% Copyright (c) 2004 Research Foundation of SUNY

% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 2 of the License, or
% (at your option) any later version.
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see http://www.gnu.org/licenses/.

%% Build entire policy model

:- ['~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_fs_use',
    '~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_genfs_contexts'].
    

:- import directory/2 from directry.
:- import parse_filename/4 from machine.


%% this is for debugging
:- debug_ctl(hide, [model_flask/1,
		    define_program_domains/0,
		    model_constraints/1,
		    model_mls/1,
		    model_te/1,
		    model_rbac/1,
		    model_users/1,
		    model_init_sid_contexts/1,
		    model_fs_use/1,
		    model_genfs_contexts/1,
		    model_net_contexts/1
		   ]).


%%
%% macro-expanded
%%
%% builds entire policy model from policy.conf;
%% constraint specification in policy.conf must be moved to the end of file
%%

build_policy_model :-
	build_policy_model('~bss/projects/selinux/policy-1.1/mls',
			   '~bss/projects/selinux/policy-1.1/policy.conf').


build_policy_model(MLSFile,PolicyFile) :-
	consult('~bss/projects/selinux/policy-1.1/build_model/macro_expanded/tokenize_policy'),
	consult('~bss/projects/selinux/policy-1.1/build_model/macro_expanded/model_policy_dcg'),
	model_policy(MLSFile,PolicyFile), !, abolish_all_tables.


%%
%% macro-preserving
%%
%% attempt to build entire policy model at one shot
%% unfortunately, fails with memory volation on the LAST call
%%

build_policy_model([FlaskDir, MLSFile, TEFile, RBACFile, UserFile, ConstrFile,
		    InitSIDFile, FSUseFile, GenFSFile, NetConsFile]) :-
	define_program_domains, !,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_constr_dcg'), 
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/tokenize_constr'), 
	model_constraints(ConstrFile), !, abolish_all_tables,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_mls_dcg'), 
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/tokenize'), 
	model_mls(MLSFile), !, abolish_all_tables,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_flask_dcg'), 
	model_flask(FlaskDir), !, abolish_all_tables,
       consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_te_dcg'), 
	model_te(TEFile), !, abolish_all_tables,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_rbac_dcg'),
	model_rbac(RBACFile), !, abolish_all_tables,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_users_dcg'),
	model_users(UserFile), !, abolish_all_tables,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_init_sid_contexts_dcg'),
	model_init_sid_contexts(InitSIDFile), !,
	model_fs_use(FSUseFile), !,
	model_genfs_contexts(GenFSFile), !,
	consult('~bss/projects/selinux/policy-1.1/build_model/non_macro_expanded/model_net_contexts_dcg').
%	model_net_contexts(NetConsFile), !.


define_program_domains :-
	retractall(program_domain(_)),
	directory('domains/program/', L1),
	get_domains(L1),
	directory('domains/misc/', L2),
	get_domains(L2).

get_domains([entry(_,_,_,_,_,_,Name)|Xs]) :-
	parse_filename(Name, _Dir, DomainName, Ext),
	(
	 Ext == te
	 -> assert(program_domain(DomainName))
	  ; true
	),
	get_domains(Xs). 
get_domains([]).
