
:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_cat/3 from string.
:- import fmt_write/2 from file_io.

:- auto_table.

:- debug_ctl(hide, [scan_constr_file/2,
		    name/3,
%		    setL/3,
		    setS/3,
		    expr/3,
		    upcase_var/2,
		    cat_three/4
		   ]).


model_constraints(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling constraints after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_constr_file(InFile, ConstrTokens),
	parse_constraints(ConstrTokens, ConstrList, []),
	nl, writeln(':- import member/2 from basics.'), nl,
	write_constr_rule(ConstrList),
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.



constraint_rule(Class,Perm,Expr) -->
	[constrain], setL(Class), setL(Perm), expr(Expr), [';'].



expr(E) --> ['('], expr(E1), [')'], {E = E1}
	  | [not], expr(NE), {str_cat('not ', NE, E1), E = [E1]}
	  | [r1, Op], {role_operator(Op)}, [r2], {cat_three('R1', Op, 'R2', E1),
						  E = [E1]}
	  | [u1, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('U1', Op, CNs, E1),
						 E = [E1]}
	  | [u2, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('U2', Op, CNs, E1),
						 E = [E1]}
	  | [r1, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('R1', Op, CNs, E1),
						 E = [E1]}
	  | [r2, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('R2', Op, CNs, E1),
						 E = [E1]}
	  | [t1, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 t_operator(Op, TOp),
						 E1 =.. [TOp, 'T1', CNs],
						 E = [E1]}
	  | [t2, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 t_operator(Op, TOp),
						 E1 =.. [TOp, 'T1', CNs],
						 E = [E1]}
	  | expr(E1), [BC], {b_conn(BC)}, expr(E2), {append(E1,[BC|E2], E)}
	  | expr(E1), [ifdef, '('], name(PD), expr_suf(E2), [')'],
	    {program_domain(PD) -> append(E1,E2,E)
	                         ; E = E1}.

expr_suf(CE) --> [BC], {b_conn(BC)}, expr(E), {append([BC], E, CE)}.



%
% Auxiliaries
%

setL(S) --> name(Id), {S = [Id]}
	 | ['{'], name_list(S), ['}'].


name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.


name_list(Ns) --> name(Id), {Ns = [Id]}
	        | name_list(NL), name(Id), {append(NL, [Id], Ns)}.



setS(N) --> name(N)
	  | ['{'], name_string(N), ['}'].


name_string(NS) --> name(NS)
 	          | name_string(Ns), [N], {str_cat(Ns, N, NS)}.



is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


operator('==').
operator('!=').

t_operator('==', 'type_eq').
t_operator('!=', 'type_neq').

t_operator('==', T1,T2, TOp) :-
	TOp =.. type_eq(T1,T2).
t_operator('!=', T1,T2, TOp) :-
	TOp =.. type_neq(T1,T2).

role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').

b_conn('and').
b_conn('or').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).


upcase_vars([],[]).
upcase_vars([X|Xs],[CX|CXs]) :-
	upcase_var(X,CX),
	upcase_vars(Xs,CXs).


upcase_var(X,CX) :-
	( X = 't1' ; X = 't2' ;
	  X = 'r1' ; X = 'r2' ;
	  X = 'u1' ; X = 'u2' )
	  -> !, atom_codes(X, [L|N]),
	     CL is L-32,
	     atom_codes(CX, [CL|N])
	   ; CX = X.



cat_three(S1,S2,S3,S) :-
	str_cat(S1, ' ', S1B),
	str_cat(S1B, S2, S12),
	str_cat(S12, ' ', S2B),
	str_cat(S2B, S3, S).




parse_constraints([],[],[]) :- !.
parse_constraints([Li|Lm], [(Class,Perm,Expr)|Cs], Lo) :-
	  constraint_rule(Class,Perm,Expr,Li,_),
	  parse_constraints(Lm, Cs, Lo).

write_constr_rule(CL) :-
	writeln('constraint(C,P, T1,R1,U1, T2,R2,U2) :- '),
	write_constraints(CL).

write_constraints([]) :-
	nl,nl.
write_constraints([(Class,Perm,Expr)|Cs]) :-
	tab(6), write('(macro_member(C, '), write(Class), writeln('),'),
	tab(7), write('member(P, '), write(Perm), writeln(')'),
	tab(7), write('-> ('), write_expr(Expr), writeln(')'),
	tab(8), write('; true'),
	abolish_table_pred(write_expr/1),
	(Cs \= [] -> writeln('),') ; writeln(').')),
	write_constraints(Cs).


write_expr([]).
write_expr([and|Es]) :-
	!, write(', '),
	write_expr(Es).
write_expr([or|Es]) :-
	!, write(' ; '),
	write_expr(Es).
write_expr([E|Es]) :-
	write(E),
	write_expr(Es).
