
:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_sub/3, string_substitute/4, str_cat/3,
	  term_to_codes/2, codes_to_term/2 from string.

:- auto_table.


:- debug_ctl(hide, [%name/3,
		    name_list/3,
%		    setL/3,
		    type_pref/3,
		    write_mls_rules/1,
		    write_mls_rule/3,
		    scan_file/1,
		    append/3,
		    model_security_classes/1,
		    model_initial_sids/1
		   ]).

'C'(X,Y,Z) :- word(Y,X,Z).



%%
% build model for FLASK configuration
% FlaskDir must end with a '/'
%% 
model_flask(FlaskDir) :-
	str_cat(FlaskDir, 'security_classes', SecInFile),
	model_security_classes(SecInFile),
	str_cat(FlaskDir, 'initial_sids', SidInFile),
	model_initial_sids(SidInFile),
	str_cat(FlaskDir, 'access_vectors', AVInFile),
	model_access_vectors(AVInFile).


% security class definition
model_security_classes(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/flask/', OutFileName, OutFile), 
	nl, write('Modeling FLASK security classes after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(InFile),
	parse_flask_security_class_rules(0,_F),nl,
%	write_rules(Rules),nl,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.



parse_flask_security_class_rules --> parse_flask_security_class_rule
	                           | parse_flask_security_class_rules,
				     opt_semicolon,
				     parse_flask_security_class_rule.

parse_flask_security_class_rule --> [class], name(C),
	                            {fmt_write("security_class(%s).\n",
					       args(C))}.




% initial sid definition
model_initial_sids(InFile) :-
	abolish_all_tables,
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/flask/', OutFileName, OutFile), 
	nl, write('Modeling FLASK initial SIDs after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(InFile),
	parse_flask_initial_sid_rules(0,_F),nl,
%	write_rules(Rules),nl,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.


parse_flask_initial_sid_rules --> parse_flask_initial_sid_rule
	                        | parse_flask_initial_sid_rules,
				  opt_semicolon,
				  parse_flask_initial_sid_rule.

parse_flask_initial_sid_rule --> [sid], name(SID),
	                         {fmt_write("initial_sid(%s).\n",
					    args(SID))}.



% access vector specification
model_access_vectors(InFile) :-
	abolish_all_tables,
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/flask/', OutFileName, OutFile), 
	nl, write('Modeling FLASK access vectors after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(InFile),
	parse_flask_access_vector_rules(0,_F),nl,
%	write_rules(Rules),nl,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.


parse_flask_access_vector_rules --> parse_flask_access_vector_rule
	                          | parse_flask_access_vector_rules,
				    opt_semicolon,
				    parse_flask_access_vector_rule.

parse_flask_access_vector_rule --> [common],name(C),['{'],name_list(As),['}'],
	                           {!, write(access(common,C,As)),writeln('.')}
			         | [class],name(C),['{'],name_list(As),['}'],
	                           {!, write(access(class,C,As)),writeln('.')}	
			         | [class], name(C), [inherits], name(IC),
			           opt_name_list(As),
				   {write(inherits(C,IC)), writeln('.'),
				    As = [] -> true
				             ; write(access(class,C,As)),
				               writeln('.')}.

opt_name_list(Ns) --> ['{'], name_list(Ns), ['}'],!
	            | [], {Ns = []}.






%
% Auxiliaries
%

opt_semicolon --> [';'], ! | [].

name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.


name_list(Ns) --> name(Id), {Ns = [Id]}
	        | name_list(NL), name(Id), {append(NL, [Id], Ns)}.


is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


operator('==').
operator('!=').

role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').
/*
b_conn('and').
b_conn('or').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).
*/