
:- import file_open/3, file_read_line/2 from file_io.
:- import str_sub/3, str_cat/3 from string.
:- import append/3 from basics.
:- import re_match/5, re_substring/4 from regmatch.
:- import intern_string/2, parse_filename/4 from machine.

:- debug_ctl(hide, [re_match/5, re_substring/4]).




model_fs_use(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling fs use configuration after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	file_open(InFile, r, IOPort),
	parse_fs_use(IOPort),nl,
	file_close(IOPort),
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.


parse_fs_use(IOPort) :-
	parse_fs_use_rule(IOPort),
	parse_fs_use(IOPort).
parse_fs_use(_).

parse_fs_use_rule(IOPort) :-
	file_read_line(IOPort, Line),
	(
	 str_sub('fs_use_', Line, 0),
	 re_match("[a-z_]+", Line, 7, _, [match(BUT,EUT)]),
	 re_substring(Line, BUT,EUT, UT), 
	 intern_string(UT, UseType), 

	 %% match fs type
	 re_match("[a-z_]+", Line, EUT, _, [match(BFS,EFS)]),
	 re_substring(Line, BFS,EFS, FS), 
	 intern_string(FS, FSType),

	 %% match context
	 re_match("([a-z_]+):([a-z_]+):([a-z_]+)", Line, EFS, _, 
		[match(_,_), match(BU,EU), match(BR,ER), match(BT,ET)]),
	 re_substring(Line, BU,EU, U), 
	 intern_string(U, User),
	 re_substring(Line, BR,ER, R), 
	 intern_string(R, Role),
	 re_substring(Line, BT,ET, T), 
	 intern_string(T, Type),

	 write(fs_use(UseType, FSType, (User,Role,Type))), writeln('.')
	;
	 true
	).
	 
