
:- import file_open/3, file_read_line/2 from file_io.
:- import str_sub/3, str_cat/3 from string.
:- import append/3 from basics.
:- import re_match/5, re_substring/4 from regmatch.
:- import intern_string/2, parse_filename/4 from machine.

:- debug_ctl(hide, [re_match/5, re_substring/4]).



model_genfs_contexts(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling genfs contexts after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	file_open(InFile, r, IOPort),
	parse_genfs_contexts(IOPort),nl,
	file_close(IOPort),
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.


parse_genfs_contexts(IOPort) :-
	parse_genfs_context(IOPort),
	parse_genfs_contexts(IOPort).
parse_genfs_contexts(_).

parse_genfs_context(IOPort) :-
	file_read_line(IOPort, Line),
	(
	 str_sub('ifdef', Line, 0),
	 parse_conditional_genfs_context(IOPort, Line)
	;
	 str_sub('genfscon', Line, 0),

	 %% match fstype 
	 re_match("[a-z_]+", Line, 9, _, [match(BFS,EFS)]),
	 re_substring(Line, BFS,EFS, FS), 
	 intern_string(FS, FSType), 

	 %% match directory pathname prefix
	 re_match("[a-z/_]+", Line, EFS, _, [match(BP,EP)]),
	 re_substring(Line, BP,EP, P), 
	 intern_string(P, Path),

	 %% match context
	 re_match("([a-z_]+):([a-z_]+):([a-z_]+)", Line, EP, _, 
		[match(_,_), match(BU,EU), match(BR,ER), match(BT,ET)]),
	 re_substring(Line, BU,EU, U), 
	 intern_string(U, User),
	 re_substring(Line, BR,ER, R), 
	 intern_string(R, Role),
	 re_substring(Line, BT,ET, T), 
	 intern_string(T, Type),

	 write_genfscon(FSType, Path, (User,Role,Type))
	;
	 true
	).



	 

parse_conditional_genfs_context(IOPort, Line) :-
	re_match("[a-z_]+", Line, 5, _, [match(X,Y)]), 
	re_substring(Line, X,Y, UD), 
	intern_string(UD,Domain), 
	parse_conditional_genfscon(IOPort, Cons),
	(
	 program_domain(Domain)
	 -> write_context_list(Cons)
	  ; true
	).
%	parse_conditional_genfscon(IOPort, Cons),
%	write(ifdef(Domain, Cons)), writeln('.').



parse_conditional_genfscon(IOPort, Context) :-
	file_read_line(IOPort, Line),
	(
	 re_match("')", Line, 0, _, _),
	 Context = []
	;
	 re_match("[a-z_]+", Line, 9, _, [match(BFS,EFS)]),
	 re_substring(Line, BFS,EFS, FS),
	 intern_string(FS, FSType),	

	 re_match("[a-z/_]+", Line, EFS, _, [match(BP,EP)]),
	 re_substring(Line, BP,EP, P),
	 intern_string(P, Path),

	 re_match("([a-z_]+):([a-z_]+):([a-z_]+)", Line, EP, _, 
		[match(_,_), match(BU,EU), match(BR,ER), match(BT,ET)]),
	 re_substring(Line, BU,EU, U), 
	 intern_string(U, User),
	 re_substring(Line, BR,ER, R), 
	 intern_string(R, Role),
	 re_substring(Line, BT,ET, T), 
	 intern_string(T, Type),

	 Con = (User, Role, Type),

	 parse_conditional_genfscon(IOPort, Cons),

	 Context = [genfscon(FSType,Path,Con)|Cons]
	).


write_context_list([]).
write_context_list([C|Cs]) :-
	write(C), writeln('.'),
	write_context_list(Cs).


write_genfscon(FSType, Path, (User,Role,Type)) :-
	fmt_write("genfscon(%s,'%s',(%s,%s,%s)).\n",
		  args(FSType,Path,User,Role,Type)).
