
:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_sub/3, string_substitute/4, str_cat/3,
	  term_to_codes/2, codes_to_term/2 from string.

:- auto_table.


:- debug_ctl(hide, [name/3,
		    name_list/3,
%		    setL/3,
		    type_pref/3,
%		    write_mls_rules/1,
		    write_mls_rule/3,
		    scan_file/1,
		    append/3
		   ]).

'C'(X,Y,Z) :- word(Y,X,Z).



model_mls(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling mls permission map after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(InFile),
	mls_rules(Rules, 0,_F),
	write_mls_rules(Rules),nl,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.



%
% permission direction definition rules
%
mls_rules(RS) --> mls_rule(RS)
	        | mls_rules(Rs), mls_rule(R), {append(Rs,R,RS)}.




mls_rule(R) --> [sensitivity, N], {s_cat(N)}, opt_alias_list(As), [';'],
	        {R = [sensitivity(N,As)]}
	      | [dominance], setL(Order),
	        {R = [dominance(Order)]}
	      | [category], name(N), opt_alias_list(As), [';'],
	        {R = [category(N,As)]}
	      | [level], name(N), opt_cat_list(Cs), [';'],
	        {R = [level(N,Cs)]}
	      | type_pref(TypePref), name(TypeName), pair_set(Ps),
	        {R = [perm_map(TypePref, TypeName, Ps)]}.



s_cat(unclassified).
s_cat(confidential).
s_cat(secret).
s_cat(top_secret).

	      
opt_alias_list(AliasList) --> [alias], setL(AliasList)
	               | [], {AliasList = []}.


opt_cat_list(CatList) --> [':'], coma_list(CatList)
	                | [], {CatList = []}.

	      
type_pref(TP) --> [common], {TP = common}
	        | [class], {TP = class}.


perm_map(PS) --> name(ClassPerm), [':'], setL(BasePerm),
	         {PS = [(ClassPerm,BasePerm)]}
	       | perm_map(Ps), name(ClassPerm), [':'], name(BasePerm),
		 {append(Ps, [(ClassPerm,BasePerm)], PS)}.



%
% Auxiliaries
%


setL(S) --> name(Id), {S = Id}
	 | ['{'], name_list(S), ['}'].

/*
coma_list(XS) --> setL(XS)
	        | coma_list(Xs), [','], setL(X),
		  {append(Xs,X,XS)}.
*/

coma_list(XS) --> name(X), {XS = [X]}
	        | coma_list(Xs), [','], name(X),
		  {append(Xs,[X],XS)}.



pair_set(S) --> ['{'], pairs(S), ['}'], !
	      | [], {S = []}.


pairs(PS) --> pair(PS)
	    | pairs(Ps), pair(P), {append(Ps,P,PS)}.

	    
pair([(C,BP)]) --> name(C), [':'], setL(BP).


opt_name(N) --> name(N)
	      | ['`', CQ], {c_quote(CQ)}.


name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.

name_list(Ns) --> name(Id), {Ns = [Id]}
	        | name_list(NL), name(Id), {append(NL, [Id], Ns)}.


is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


operator('==').
operator('!=').

role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').

b_conn('and').
b_conn('or').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).



write_mls_rules([]).
write_mls_rules([perm_map(Pref,Class,Ps)|Rs]) :- !,
	write_mls_perm_map(Pref,Class,Ps),
	write_mls_rules(Rs).
write_mls_rules([R|Rs]) :-
	write(R), writeln('.'),
	write_mls_rules(Rs).

/*
write_mls_rule(_,_, []).
write_mls_rule(Pref, Class, [(P,BP)|Ps]) :- 
	  write(perm_map(Pref,Class,P,BP)), writeln('.'),
	  write_mls_rule(Pref,Class,Ps).
*/

/*
write_mls_rule(perm_map(Pref,Class,Ps)) :- !,
	write_mls_perm_map(Pref,Class,Ps).
write_mls_rule(R) :-
	write(R), writeln('.').
*/

write_mls_perm_map(_,_, []).
write_mls_perm_map(Pref, Class, [(_P,BP)|Ps]) :-
	BP = none,!,
	write_mls_perm_map(Pref,Class,Ps).
write_mls_perm_map(Pref, Class, [(P,BP)|Ps]) :-
	atomic(BP),!,
	write(perm_map(Pref,Class,P,BP)), writeln('.'),
	write_mls_perm_map(Pref,Class,Ps).
write_mls_perm_map(Pref, Class, [(P,BP)|Ps]) :-
	BP = [R,W],
	write(perm_map(Pref,Class,P,R)), writeln('.'),
	write(perm_map(Pref,Class,P,W)), writeln('.'),
	write_mls_perm_map(Pref,Class,Ps).
