
:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_sub/3, string_substitute/4, str_cat/3,
	  term_to_codes/2, codes_to_term/2 from string.

:- auto_table.


:- debug_ctl(hide, [name/3,
		    name_list/3,
		    setL/3,
		    ifdef_con/3,
		    cond_defs/3,
		    scan_file/1,
		    c_quote/1,
		    write_mls_rules/1,
		    write_mls_rule/3
		   ]).

'C'(X,Y,Z) :- word(Y,X,Z).



model_net_contexts(InFile) :-
	assert(program_domain('courier.te')),
	assert(net_macro(0)),
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling net contexts after '),
	write(InFile), writeln(' ...'),
	tell(OutFile),
	scan_file(InFile),
	net_contexts(Cs, 0,_F),
	write_net_contexts(Cs),nl,
	told,
	retractall(net_macro(_)),
	write('Done ... Output file: '), write(OutFile), writeln('.'), nl.



%
% definition rules
%
net_contexts(CS) --> net_context(CS)
  	           | net_contexts(Cs), net_context(C), {append(Cs,C,CS)}.



net_context(C) --> [portcon], name(ID), name(Num), s_con((U,R,T)),
		   {C = [portcon(ID, Num, (U,R,T))]}
		 | [netifcon], name(ID), s_con((U1,R1,T1)),
		   s_con((U2,R2,T2)),
		   {C = [netifcon(ID, (U1,R1,T1), (U2,R2,T2))]}
		 | [nodecon], name(A1), name(A2), s_con((U,R,T)),
		   {fmt_write_string(SC,
				     "nodecon('%s','%s',(%s,%s,%s))",
				     args(A1,A2,U,R,T)),
		    C = [SC]}
		 | ifdef_con(C).



s_con((U,R,T)) --> name(U), [':'], name(R), [':'], name(T).



% if-def
ifdef_con(R) --> [ifdef, '('], name(PD),  [','], cond_defs(TRs), [')'],
	         {is_net_macro(PD) -> (TRs = [define(N)]
					 -> assert(net_macro(N)), R = []
				          ; R = TRs)
		                 ; R = []}.


cond_defs(Cs) --> ['`'], net_contexts(Cs), [CQ], {c_quote(CQ)}
	        | ['`', define, '('], name(N), [')'],  [CQ], {c_quote(CQ),
							      Cs = [define(N)]}.




is_net_macro(N) :-
	program_domain(N).
is_net_macro(N) :-
	net_macro(N).

%
% auxiliaries
%
name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.


is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).





write_net_contexts([]).
write_net_contexts([C|Cs]) :-
	write(C), writeln('.'),
	write_net_contexts(Cs).
