
:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_sub/3, string_substitute/4, str_cat/3,
	  term_to_codes/2, codes_to_term/2 from string.

:- auto_table.


:- debug_ctl(hide, [name/3,
		    name_list/3,
%		    setL/3,
		    type_pref/3,
		    write_mls_rules/1,
		    write_mls_rule/3,
		    scan_file/1,
		    append/3
		   ]).

'C'(X,Y,Z) :- word(Y,X,Z).



model_rbac(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling RBAC configuration after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(InFile),
	rbac_rules(Rules, 0,_F),
	write_rbac_rules(Rules),nl,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.



%
% rbac rules
%
rbac_rules(RS) --> rbac_rule(RS)
	         | rbac_rules(Rs), rbac_rule(R), {append(Rs,R,RS)}.




rbac_rule(R) --> [allow], name_list(CurrentRoles), name_list(NewRoles), [';'],
	         {R = [role_allow(CurrentRoles,NewRoles)]}
	       | [ifdef, '('], name(PD), [','], cond_rbac_rules(Rs), [')'],
	         {program_domain(PD) -> R = Rs; R = []}.



cond_rbac_rules(Rules) --> rbac_rules(Rules)
	                 | ['`'], rbac_rules(Rules), [CQ], {c_quote(CQ), !}
			 | ['`',CQ], {c_quote(CQ), Rules = []}
			 | [], {Rules = []}.



%
% Auxiliaries
%

name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.


name_list(Ns) --> name(Id), {Ns = [Id]}
	        | name_list(NL), name(Id), {append(NL, [Id], Ns)}.


is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


operator('==').
operator('!=').

role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').

b_conn('and').
b_conn('or').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).



write_rbac_rules([]).
write_rbac_rules([R|Rs]) :-
	write(R), writeln('.'),
	write_rbac_rules(Rs).
