

:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_sub/3, string_substitute/4, str_cat/3, substring/4 from string.

:- auto_table.
:- table compose_id/3.

:- debug_ctl(hide, [name/3,
		    opt_name/3,
		    name_list/3,
		    c_quote/1,
		    set/3,
		    setL/3,
		    arg_list/3,
		    coma_list/3,
		    opt_semicolon/2,
		    opt_alias_def/3,
		    opt_attr_list/3,
		    opt_var_attr_list/3,
		    scan_file/1,
		    undef_macro/3,
		    set_macro/3,
%		    te_rules/3,
		    attribute_rule/3,
		    type_decl_rule/3,
		    te_av_rule/3,
		    te_assertion_rule/3,
		    type_transition_rule/3,
		    type_change_rule/3,
		    role_decl_rule/3,
		    role_allow_rule/3,
		    ifdef_rule/3,
		    ifelse_rule/3,
%		    macro_def/3,
		    macro_call/3,
		    rule_macro_undef/3,
%		    write_te_rules/1,
		    write_macro_then_rules/3,
		    write_macro_else_rules/3,
%		    write_macro_def_rev/2,
		    write_list/2,
%		    write_rule/1,
%		    write_rule/2,
%		    get_args/3,
		    get_atomic_args/2,
		    get_compound_arg_rules/2,
		    l_append_no_rep/3,
		    append_no_rep/3
		   ]).

'C'(X,Y,Z) :- word(Y,X,Z).

model_te(InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile),
	str_cat(FileName, '.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling TE specification after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(InFile),
	te_rules(Rules, 0,_F),
	write_te_rules(Rules),nl,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. ... Output file: '),
	write(OutFile), writeln('.'), nl.


te_rules(RS) --> rules(RS).


% RL is a list of rules s.t. each element is of the form (Rule,Args)
rules(RL) --> rule(RL)%, {write_rule(RL,AL)}
	    | rules(Rs), rule(R),
	      {
%	       write_rule(R,A),
	       append(Rs, R, RL)}.


% RL contains tuples (Rule,Args)
rule(RL) --> ['`'], rule(RL), [CQ], {!, c_quote(CQ)}
	      | attribute_rule(R), {!, RL = [R]}
	      | type_decl_rule(R), {!, RL = [R]}
	      | te_av_rule(RL), !
	      | te_assertion_rule(R), {!, RL = [R]}
	      | type_transition_rule(R), {!, RL = [R]}
	      | type_change_rule(R), {!, RL = [R]}
	      | role_decl_rule(R), {!, RL = [R]}
	      | role_allow_rule(R), {!, RL = [R]}
	      | ifdef_rule(RL), !
	      | ifelse_rule(RL), !
	      | macro_def(RL), !
	      | macro_call(R), {!, RL = [R]}.

write_rule([],[]) :- !.
write_rule(R,A) :-
	writeln('RULE:'),
	write_list(R,2),nl,
	writeln('ARGS:'),
	write(A),
	nl,nl.


%
% macro definition rules
%
macro_def(Rules) --> undef_macro(Head), {Rules = []}
	           | set_macro(Head), {Rules = [(Head,[])]}
	           | rule_macro(Rules).


% undefine macro - appearing sometimes before giving new definition
undef_macro(undefine(MName)) --> [undefine, '('], name(MName), [')'].
%	              {MDef = undefine(MName)}.


% set macros define aliases for sets of permissions of commonly used together
set_macro(set_macro(SName,NDefSet)) --> [define, '('], name(SName),[',','`'],
	                                setL(DefSet),[CQ],{c_quote(CQ)},[')'],
					{get_args(DefSet, _, NDefSet)}.
%		     MDef = set_macro(SName,NDefSet)}.


% rule macros may have one of two functions:
% - representing set of rules  giving permission of certain kind, or
% - defining a domain (domain macro); 
% they may contain variable parameters which need to be extracted and 
% put as arguments to macro names
rule_macro(Rules) --> def_stmt(Head,Body), {Rules = [(Head,Body)]}
	            | ['`'], def_stmt(Head,Body), [CQ],{c_quote(CQ),
							Rules = [(Head,Body)]}
	            | cond_rule_macro(Rules).


% leave arguments handling for the writing predicate
def_stmt(Head,Body) --> [define, '('], name(Head), [',', '`'],
			rule_def(Body), [CQ],{c_quote(CQ)},[')'].



cond_rule_macro(Rules) --> [ifdef,'('], name(PD), [','],
			   cond_macro_def_rules(Then),
			   opt_cond_macro_def_rules(Else), [')'],
			   {program_domain(PD) -> Rules = Then; Rules = Else}.


cond_macro_def_rules(Rules) --> ['`'], te_rules(Rules), [CQ],
	                        {c_quote(CQ)}, !
	                      | te_rules(Rules)
			      | ['`',CQ], {c_quote(CQ), Rules = []}.

opt_cond_macro_def_rules(Rules) --> [','], cond_macro_def_rules(Rules)
	                          | [], {Rules = []}.



get_compound_arg_rules([],[]).
get_compound_arg_rules([A|As],Rules) :-
	str_sub('Id', A, 0), !,
	str_sub('Arg', A, VarPos), PrefE is VarPos-1,
	substring(A, 3,PrefE, Pref),
	substring(A, VarPos, -1, VarPart),
	get_compound_arg_rules([VarPart], R),
	append(R, [compose_id(Pref,VarPart,A)], ARs),
	get_compound_arg_rules(As, Rs),
	append(ARs,Rs,Rules).
get_compound_arg_rules([A|As],Rules) :-
	str_sub('_', A, Pos),
	substring(A, 0, Pos, Arg),
	SufPos is Pos+1,
	substring(A, SufPos, -1, Sufix),
	get_compound_arg_rules([Sufix], SRs),
	get_compound_arg_rules(As, Rs),
	(
	 var_string(Arg)
	 -> !, append(SRs, [compose_id(Arg,Sufix,A)|Rs], Rules)
	  ; append(SRs, Rs, Rules)
	).
get_compound_arg_rules([_A|As],Rules) :-
	get_compound_arg_rules(As, Rules).
	


rule_def(Rules) --> rules(Rules)
	               | [], {Rules = []}.




%
% TE rules
%

% attribute declaration
attribute_rule(attr(Id)) --> [attribute], !, name(Id), [';'].


% attribute declaration
%attribute_rule((R,Args)) --> [attribute], !, name(Id), [';'],
%	                   {R = attr(Id), Args = []}.


% type declaration
type_decl_rule((R,Args)) --> [type], name(Id),
        	             opt_alias_def(Aliases),
			     opt_attr_list(Atts),
			     opt_var_attr_list(VAtts), [';'],
			     {get_args([Id],A1,[NId]),
			      get_args(Aliases,A2,NAliases),
			      l_append_no_rep(A1,A2,A3),
			      get_args(Atts,A4,NAtts),
			      l_append_no_rep(A3,A4,A5),
			      get_args(VAtts,A6,NVAtts),
			      l_append_no_rep(A5,A6,Args),
%			      append(NAtts,NVAtts, AllAtts),
%			      R = type_decl(NId,NAliases,AllAtts)}.
			      R = type_decl(NId,NAliases,NAtts,NVAtts)}.


opt_alias_def(Aliases) --> [alias], setL(Aliases)
	                 | [], {Aliases = []}.  % empty string

opt_attr_list(Atts) --> [','], attr_list(Atts)
	              | [], {Atts = []}.

attr_list(Atts) --> name(A), {Atts = [A]}
	          | attr_list(As), [','], name(A), {append(As,[A],Atts)}.

%opt_var_attr_list(VAtts) --> name_list(VAtts), !
%	                   | [], {VAtts = []}.

opt_var_attr_list(VAtts) --> setL(VAtts), !
	                   | [], {VAtts = []}.



% TE access vector
te_av_rule(Rules) --> [AV], {av_kind(AV)}, set(STs), set(TTs),
	              [':'], set(Cs), set(Ps), [';'],
		      {is_list(STs)
		       -> separate_source_ts(AV,STs,TTs,Cs,Ps, Rules)
		        ; is_list(TTs)
		          -> separate_target_ts(AV,STs,TTs,Cs,Ps, Rules)
		           ; is_list(Cs)
		             -> separate_classes(AV,STs,TTs,Cs,Ps, Rules)
		              ; is_list(Ps)
		                -> separate_perms(AV,STs,TTs,Cs,Ps, Rules)
		                 ; get_args(STs,A1,NSs),
		                   get_args(TTs,A2,NTs),
		                   l_append_no_rep(A1,A2,A3),
			           get_args(Cs,A4,NCs),
			           l_append_no_rep(A3,A4,A5),
			           get_args(Ps,A6,NPs),
			           l_append_no_rep(A5,A6,Args),
			           R = access_vector(AV,NSs,NTs,NCs,NPs),
		                   Rules = [(R,Args)]}.


av_kind(allow).
av_kind(auditallow).
av_kind(dontaudit).


set(S) --> ['*'], {S = '_'}
	 | setL(S)
	 | ['~'], setL(S1), {S = except(S1)}.


separate_source_ts(_AV,[],_TTs,_Cs,_Ps, []).
separate_source_ts(AV,[ST|STs],TTs,Cs,Ps, Rules) :-
	separate_target_ts(AV,ST,TTs,Cs,Ps, Rules1),
	separate_source_ts(AV,STs,TTs,Cs,Ps, Rules2),
	append(Rules1,Rules2,Rules).

separate_target_ts(_AV,_ST,[],_Cs,_Ps, []).
separate_target_ts(AV,ST,TT,Cs,Ps, Rules) :-
	atomic(TT), TT \= [],
	separate_classes(AV,ST,TT,Cs,Ps, Rules).
separate_target_ts(AV,ST,[TT|TTs],Cs,Ps, Rules) :-
	separate_classes(AV,ST,TT,Cs,Ps, Rules1),
	separate_target_ts(AV,ST,TTs,Cs,Ps, Rules2),
	append(Rules1,Rules2,Rules).

separate_classes(_AV,_ST,_TT,[],_Ps, []).
separate_classes(AV,ST,TT,C,Ps, Rules) :-
	atomic(C), C \= [],
	separate_perms(AV,ST,TT,C,Ps,Rules).
separate_classes(AV,ST,TT,[C|Cs],Ps, Rules) :-
	separate_perms(AV,ST,TT,C,Ps,Rules1),
	separate_classes(AV,ST,TT,Cs,Ps,Rules2),
	append(Rules1,Rules2,Rules).

separate_perms(_AV,_ST,_TT,_C,[],[]).
separate_perms(AV,ST,TT,C,P,[(access_vector(AV,NST,NTT,NC,NP),Args)]) :-
	atomic(P), P \= [],
	get_args(ST,A1,NST),
	get_args(TT,A2,NTT),
	l_append_no_rep(A1,A2,A3),
	get_args(C,A4,NC),
	l_append_no_rep(A3,A4,A5),
	get_args(P,A6,NP),
	l_append_no_rep(A5,A6,Args).
separate_perms(AV,ST,TT,C,[P|Ps],[(access_vector(AV,NST,NTT,NC,NP),Args)|Rs]) :-
	get_args(ST,A1,NST),
	get_args(TT,A2,NTT),
	l_append_no_rep(A1,A2,A3),
	get_args(C,A4,NC),
	l_append_no_rep(A3,A4,A5),
	get_args(P,A6,NP),
	l_append_no_rep(A5,A6,Args),
	separate_perms(AV,ST,TT,C,Ps, Rs).


% TE assertion
te_assertion_rule((R,Args)) --> [neverallow], set(SourceTs), set(TargetTs),
	               [':'], set(Classes), set(Perms), [';'],
 	               {get_args(SourceTs,A1,NSs),
			get_args(TargetTs,A2,NTs),
			l_append_no_rep(A1,A2,A3),
			get_args(Classes,A4,NCs),
			l_append_no_rep(A3,A4,A5),
			get_args(Perms,A6,NPs),
			l_append_no_rep(A5,A6,Args),
			R = neverallow(NSs,NTs,NCs,NPs)}.



% type transition
type_transition_rule((R,Args)) --> [type_transition],set(SourceTs),set(TargetTs),
	                         [':'], set(Classes), name(NewType), [';'],
				 {get_args(SourceTs,A1,NSs),
				  get_args(TargetTs,A2,NTs),
				  l_append_no_rep(A1,A2,A3),
				  get_args(Classes,A4,NCs),
				  l_append_no_rep(A3,A4,A5),
				  get_args([NewType],A6,[NNT]),
				  l_append_no_rep(A5,A6,Args),
				  R = type_trans(NSs,NTs,NCs,NNT)}.


% type change
type_change_rule((R,Args)) --> [type_change], set(SourceTs), set(TargetTs),
	                     [':'], set(Classes), name(NewType), [';'],
			     {get_args(SourceTs,A1,NSs),
			      get_args(TargetTs,A2,NTs),
			      l_append_no_rep(A1,A2,A3),
			      get_args(Classes,A4,NCs),
			      l_append_no_rep(A3,A4,A5),
			      get_args([NewType],A6,[NNT]),
			      l_append_no_rep(A5,A6,Args),
			      R = type_change(NSs,NTs,NCs,NNT)}.


%
% RBAC rules
%
% role declaration
role_decl_rule((R,Args)) --> [role], name(Id), [types], setL(Types), [';'],
	                   {get_args([Id],A1,[NId]),
			    get_args(Types,A2,NTypes),
			    l_append_no_rep(A1,A2,Args),
			    R = role_decl(NId,NTypes)}.



% role allow
role_allow_rule((R,Args)) --> [allow], set(CurrentRs), set(NewRs), [';'],
	                    {get_args(CurrentRs,A1,NCRs),
			     get_args(NewRs,A2,NNRs),
			     l_append_no_rep(A1,A2,Args),
			     R = role_allow(NCRs,NNRs)}.


%
% conditional rules
%

% if-def
ifdef_rule(R) --> [ifdef, '('], name(Cond),  [','],
	          cond_rules(TRs), 
	          opt_cond_rules(ERs), [')'],
		  {program_domain_name(Cond,PD)
		   -> (
		       program_domain(PD) -> R = TRs
		                             ; R = ERs
		      )
		    ; get_args([Cond],CArgs,[NCond]),
		      R = [ifdef((NCond,CArgs),TRs,ERs)]}.


program_domain_name(S) :-
	str_sub('.te', S, Pos),
	str_length(S,N),
	Pos is N-3.


program_domain_name(S,PD) :-
	str_sub('.te', S, Pos),
	str_length(S,N),
	Pos is N-3,
	substring(S,0,Pos,PD).



% if-else
ifelse_rule(R) --> [ifelse, '('], name(CArg1), [','], opt_name(CArg2),
		     [','], cond_rules(TRs), opt_cond_rules(ERs), [')'],
		     {get_args([CArg1],Args1,[NCArg1]),
		      get_args([CArg2],Args2,[NCArg2]),
		      R = [ifelse((NCArg1,Args1),(NCArg2,Args2),TRs,ERs)]}.
%		     {(CArg1 = CArg2 -> R = Then; R = Else)}.



% need a cut here: if there is only one rule in Rules, it matches both
% first and second case
cond_rules(Rules) --> rules(Rules), !
	                 | ['`'], rules(Rules), [CQ], {c_quote(CQ), !}
			 | ['`',CQ], {c_quote(CQ), Rules = []}
			 | [], {Rules = []}.



opt_cond_rules(Rules) --> [','], cond_rules(Rules)
	                     | [], {Rules = []}.


% macro invocation
macro_call(R) --> name(MName), ['('], arg_list(Pars), [')'], opt_semicolon,
	            {get_args(Pars,A,NPars),
		     append([MName],NPars,Args), Rule =.. Args, R = (Rule,A)}
		  | builtin_macro(R).



opt_semicolon --> [';'], ! | [].



%arg_list(Args) --> coma_list(As), opt_set_suffix(Rest),
%	           {append(As, Rest, Args)}.

arg_list(Args) --> coma_list(As), opt_set_suffix(Rest),
	           {append(As, Rest, Args)}.


coma_list(XS) --> setL(X), {XS = [X]}
	        | coma_list(Xs), [','], setL(X),
		  {append(Xs,[X],XS)}.


%opt_set_suffix(S) --> [',', '`', ','],coma_list(S),[CQ],{c_quote(CQ)}
%                    | [], {S = []}. 

opt_set_suffix(S) --> [',', '`', ','],coma_list(L),[CQ],{c_quote(CQ), S=[L]}
                    | [], {S = []}. 



% builtin m4 macros - some have odd syntax, easiest to parse this way
builtin_macro((R,A)) --> [errprint, '('], err_msg(Msg), [')'],
	               {fmt_write_string(R, "errprint('%s')", args(Msg)),
			A = []}
                     | [define, '('], name(MName), [',', '`'],
			set_rule_def((Body,Args)), [CQ],{c_quote(CQ)},[')'],
			{get_args([MName], MA, [NMName]),
			 get_args(Body, BA, NBody),
			 R = defined(NMName, NBody),
			 append(MA,BA,As),
			 append(As, Args, A)}.


		     
err_msg(Msg) --> opt_err_msg_pref,
	         ['`'], string(Msg), [CQ], {c_quote(CQ)}.

opt_err_msg_pref --> ['`', m4, ':', CQ], {c_quote(CQ)},
	         name(_File), [':'],
		 name(_Line), [':'].


set_rule_def((Body,Args)) --> setL(Body), {Args = []}
	                  | rule_def((Body, Args)).



%
% Auxiliaries
%

% This one replaces any '$' in an identifier with a string 'Arg' which 
% is a better way to make Prolog treat it as a variable;
% sometimes id has more than one '$' in it, and then need to handle more
% than one variable (this is done in the 3rd clause) - not sure whether
% properly, may need correction
get_args([],[],[]) :-!.
get_args([L|R], Args, [NL|NR]) :-
	is_list(L), !,
	get_args(L, ArgsL, NL),
	get_args(R, ArgsR, NR),
	l_append_no_rep(ArgsL,ArgsR,Args).
get_args([X|R], Args, [NX|NR]) :-
	get_args_one(X, ArgsX, NX),
	get_args(R, ArgsR, NR),
	l_append_no_rep(ArgsX,ArgsR,Args).
get_args(Id, Args, NId) :-
	atomic(Id), !,
	get_args_one(Id, Args, NId).
get_args(except(Xs), Args, except(NXs)) :-
	!,
	get_args(Xs, Args, NXs).

get_args_one(except(Xs), Args, except(NXs)) :-
	!,
	get_args(Xs, Args, NXs).
get_args_one(Id, Args, NId) :-
	str_sub('$',Id,VarS), VarE is VarS+1, !,
	substring(Id, 0, VarS, IdPref),
	substring(Id, VarE, -1, VarPart),
	str_cat('Arg', VarPart, IdSuf),
	get_args(IdSuf, IdAs, NIdSuf),
	(\+var_string(IdPref), IdPref \= ''
	 -> str_cat('Id_', IdPref, VarIdPref)
	  ; VarIdPref = IdPref),
	str_cat(VarIdPref, NIdSuf, NId),
	append_no_rep(NId,IdAs,Args).
get_args_one(Id, [], Id).




compose_id(Pref, Suf, Id) :-
	var_string(Pref), !,
	str_cat(Pref, '_', Pref_),
	str_cat(Pref_, Suf, Id).


var_string(S) :-
	atom_codes(S, [C|_]),
	C >= 0'A,
	C =< 0'Z.


l_append_no_rep([],L,L).
l_append_no_rep(L,[],L) :- !.
l_append_no_rep([X|Xs],L1,L) :-
	l_append_no_rep(Xs,L1,L2),
	append_no_rep(X,L2,L).


append_no_rep(X, [], [X]) :- !.
append_no_rep(X, L, L) :-
	var_member(X,L), !.
append_no_rep(X, L, [X|L]).


var_member(X, [Y|Ys]) :-
	X == Y
	;
	var_member(X,Ys).


setL(S) --> name(Id), {S = Id}%{var_string(Id) -> !, S = Id; S = [Id]}
	 | ['{'], name_list(Ns), opt_cond_suffix(Suf), ['}'],
	   {append(Ns, Suf, S)}
	 | ['{'], opt_cond_suffix(Suf), name_list(Ns), ['}'],
	   {append(Suf, Ns, S)}.


opt_cond_suffix(S) --> [ifdef, '('], name(N), [','], name_list(Xs), [')'],
	               {(program_domain_name(N,D),program_domain(D))
			 -> S = Xs; S = []}
		     | [], {S = []}.


opt_name(N) --> name(N)
	      | ['`', CQ], {c_quote(CQ), N = empty}.


name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.

name_list(Ns) --> name(Id), {Ns = [Id]}
	        | name_list(NL), name(Id), {append(NL, [Id], Ns)}.



string(S) --> name(S)
	    | string(Ns), name(N), {str_cat(Ns, ' ', S1),
			            str_cat(S1, N, S)}.


is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


operator('==').
operator('!=').

role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').

b_conn('and').
b_conn('or').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).





write_te_rules([]).
write_te_rules([R|Rs]) :-
	write_te_rule(R),
	write_te_rules(Rs).


write_te_rule((H,B)) :-
	atomic(H), !,
	get_atomic_args(B,Args),
	sort(Args,SArgs),
	Head =.. [H|SArgs],
	write_macro_def_rev(Head,B).
write_te_rule((H,B)) :-
	write_macro_def_rev(H,B).
write_te_rule(ifdef((Cond,[]),TRs,ERs)) :-
	write_te_then_rules(Cond,TRs),
	write_te_else_rules(Cond,ERs).
write_te_rule(R) :-
	write(R), writeln('.').
	     

write_macro_def_rev(MName, []) :-
%	write_base_case(MName).
	write(MName), writeln('.').
write_macro_def_rev(Head, [ifdef(Cond,TRs,ERs)|Cs]) :-
	!,
	write_macro_then_rules(Head,Cond,TRs),		
	write_macro_else_rules(Head,Cond,ERs),		
	(Cs = [] -> ground_head(Head,GHead)
	          ; GHead = Head),
	write_macro_def_rev(GHead,Cs).
write_macro_def_rev(Head, [ifelse(Cond1,Cond2,TRs,ERs)|Cs]) :-
	!,
	write_macro_then_rules(Head,Cond1,Cond2,TRs),		
	write_macro_else_rules(Head,Cond1,Cond2,ERs),		
	(Cs = [] -> ground_head(Head,GHead)
	          ; GHead = Head),
	write_macro_def_rev(GHead,Cs).
write_macro_def_rev(Head, [(Rule,Args)|Cs]) :-
	abolish_table_pred(write_list/2),
	get_compound_arg_rules(Args,ArgRules),
	append([Head],ArgRules, Body),
        write(Rule), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
%	write_macro_def_rev(Head, Cs).
%	(Cs \= [] -> write_macro_def_rev(Head, Cs)
%	            ; true).	
	(Cs = [] -> ground_head(Head,GHead)
	          ; GHead = Head),
	write_macro_def_rev(GHead,Cs).


% for ifdef
write_macro_then_rules(_Head,_Cond,[]).
write_macro_then_rules(Head,(Cond,CArgs),[(TR,TR_Args)|TRs]) :-
	abolish_table_pred(write_list/2),
	append(CArgs,TR_Args,Args),
	get_compound_arg_rules(Args,ArgRules),
	append([defined(Cond),Head],ArgRules,Body),
	write(TR), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_then_rules(Head,(Cond,CArgs),TRs).

write_macro_else_rules(_Head,_Cond,[]).
write_macro_else_rules(Head,(Cond,CArgs),[(ER,ER_Args)|ERs]) :-
	abolish_table_pred(write_list/2),
	append(CArgs,ER_Args,Args),
	get_compound_arg_rules(Args,ArgRules),
	append([not_defined(Cond),Head],ArgRules,Body),
	write(ER), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_else_rules(Head,(Cond,CArgs),ERs).


% for ifdef inside another conditional - need to carry condition statements
write_macro_then_rules(_Head,_Cond,[],_MoreRules).
write_macro_then_rules(Head,(Cond,CArgs),[(TR,TR_Args)|TRs],Rs) :-
	abolish_table_pred(write_list/2),
	append(CArgs,TR_Args,Args),
	get_compound_arg_rules(Args,ArgRules),
	append([defined(Cond)|Rs],ArgRules,Body),
	write(TR), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_then_rules(Head,(Cond,CArgs),TRs,Rs).

write_macro_else_rules(_Head,_Cond,[],_MoreRules).
write_macro_else_rules(Head,(Cond,CArgs),[(ER,ER_Args)|ERs],Rs) :-
	abolish_table_pred(write_list/2),
	append(CArgs,ER_Args,Args),
	get_compound_arg_rules(Args,ArgRules),
	append([not_defined(Cond)|Rs],ArgRules,Body),
	write(ER), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_else_rules(Head,(Cond,CArgs),ERs,Rs).


% for ifelse
write_macro_then_rules(_Head,_Cond1,_Cond2,[]) :- !.
write_macro_then_rules(Head,(C1,CArgs1),(C2,CArgs2),
		       [ifelse(Cond1,Cond2,TRules,ERules)|TRs]) :- !,
	write_macro_then_rules(Head,Cond1,Cond2,TRules),
	write_macro_else_rules(Head,Cond1,Cond2,ERules),
	write_macro_then_rules(Head,(C1,CArgs1),(C2,CArgs2),TRs).
write_macro_then_rules(Head,(C1,CArgs1),(C2,CArgs2),TRs) :-
	(
	 (C1 == empty, C2 = Arg, CArgs2 = CArgs)
	 ;
	 (C2 == empty, C1 = Arg, CArgs1 = CArgs)
	), !,
	Head =.. [HF|HArgs],
	remove_last_element(HArgs,HArg,NHArgs),
	(
	 HArg == Arg
	 -> NHead =.. [HF|NHArgs]
	  ; NHead = Head
	),
%	write_base_case(Head),
%	write_base_case(NHead),
	write_macro_then_rules1(NHead,CArgs,TRs).
write_macro_then_rules(Head,(C1,CArgs1),(C2,CArgs2),TRs) :-
	append(CArgs1,CArgs2,CArgs),
	get_compound_arg_rules(CArgs,CArgRules),
	append([C1=C2,Head],CArgRules,BodyRules),
	write_macro_then_rules2(Head,TRs,BodyRules).

write_macro_then_rules1(_Head,_CArgs,[]).
write_macro_then_rules1(Head,CArgs,[(TR,TR_Args)|TRs]) :-
	abolish_table_pred(write_list/2),
	append(CArgs,TR_Args,Args),
	get_compound_arg_rules(Args,ArgRules),
	append([Head],ArgRules,Body),
	write(TR), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_then_rules1(Head,CArgs,TRs).


write_macro_then_rules2(_Head,[],_BodyRules).
write_macro_then_rules2(Head,[(TR,TR_Args)|TRs],BodyRules) :-
	abolish_table_pred(write_list/2),
	get_compound_arg_rules(TR_Args,TR_ArgRules),
	append(BodyRules,TR_ArgRules,Body),
	write(TR), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_then_rules2(Head,TRs,BodyRules).



write_macro_else_rules(_Head,_Cond1,_Cond2,[]).
write_macro_else_rules(Head,(C1,CArgs1),(C2,CArgs2),[ER|ERs]) :-
	(
	 (C1 == empty, C2 = Arg, CArgs2 = CArgs)
	 ;
	 (C2 == empty, C1 = Arg, CArgs1 = CArgs)
	), !,
	Head =.. [HF|HArgs],
	remove_last_element(HArgs,HArg,NHArgs),
	(
	 HArg == Arg
	 -> NHead =.. [HF|HArgs]
	  ; append(HArgs,[Arg],NHArgs),
	    NHead = Head
	),
%	write_base_case(Head),
%	write_base_case(NHead),
	write_macro_else_rules1(NHead,CArgs,[ER|ERs]).
write_macro_else_rules(Head,(C1,CArgs1),(C2,CArgs2),[ER|ERs]) :-
	append(CArgs1,CArgs2,CArgs),
	get_compound_arg_rules(CArgs,CArgRules),
	append([C1\=C2,Head],CArgRules,BodyRules),
	write_macro_else_rules2(Head,[ER|ERs],BodyRules).



write_macro_else_rules1(_Head,_CArgs,[]).
write_macro_else_rules1(Head,CArgs,[ifdef(Cond,TRules,ERules)|ERs]) :-
	write_macro_then_rules(_,Cond,TRules,[Head]),
	write_macro_else_rules(_,Cond,ERules,[Head]),
	write_macro_else_rules1(Head,CArgs,ERs).
write_macro_else_rules1(Head,CArgs,[ifelse(Cond1,Cond2,TRules,ERules)|ERs]) :-
	write_macro_then_rules(Head,Cond1,Cond2,TRules),
	write_macro_else_rules(Head,Cond1,Cond2,ERules),
	write_macro_else_rules1(Head,CArgs,ERs).
write_macro_else_rules1(Head,CArgs,[(ER,ER_Args)|ERs]) :-
	abolish_table_pred(write_list/2),
	append(CArgs,ER_Args,Args),
	get_compound_arg_rules(Args,ArgRules),
	append([Head],ArgRules,Body),
	write(ER), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_else_rules1(Head,CArgs,ERs).


write_macro_else_rules2(_Head,[],_BodyRules).
write_macro_else_rules2(Head,[ifdef(Cond,TRules,ERules)|ERs],BRs) :-
	write_macro_then_rules(_,Cond,TRules,BRs),
	write_macro_else_rules(_,Cond,ERules,BRs),
	write_macro_else_rules2(Head,ERs,BRs).
write_macro_else_rules2(Head,[ifelse(Cond1,Cond2,TRules,ERules)|ERs],BRs) :-
	write_macro_then_rules(Head,Cond1,Cond2,TRules),
	write_macro_else_rules(Head,Cond1,Cond2,ERules),
	write_macro_else_rules2(Head,ERs,BRs).
write_macro_else_rules2(Head,[(ER,ER_Args)|ERs],BRs) :-
	abolish_table_pred(write_list/2),
	get_compound_arg_rules(ER_Args,ER_ArgRules),
	append(ER_ArgRules,BRs,Body),
	write(ER), writeln(' :- '),
	write_list(Body,6), writeln('.'), nl,
	write_macro_else_rules2(Head,ERs,BRs).




remove_last_element([X|Xs],X,[]) :-
	Xs = [], !.
remove_last_element([X|Xs],E,[X|NXs]) :-
	remove_last_element(Xs,E,NXs).


write_list([S],Tab) :-
	!, tab(Tab), write(S).
write_list([S|Ss],Tab) :-
	var(S), !,
	write_list(Ss,Tab).
write_list([S|Ss],Tab) :-
	tab(Tab), write(S), writeln(','),
	write_list(Ss,Tab).



write_te_then_rules(_Cond,[]).
write_te_then_rules(Cond,[(TR,[])|TRs]) :-
	(atomic(TR) -> R = defined(TR) ; R = TR),
	write(R), writeln(' :-'),
	tab(6), write(defined(Cond)), writeln('.'), nl,
	write_te_then_rules(Cond,TRs).

write_te_else_rules(_Cond,[]).
write_te_else_rules(Cond,[(ER,[])|ERs]) :-
	(atomic(ER) -> R = defined(ER) ; R = ER),
	write(R), writeln(' :-'),
	tab(6), write(not_defined(Cond)), writeln('.'), nl,
	write_te_else_rules(Cond,ERs).

/*
write_base_case(Head) :-
	ground_and_var_head(Head,GH,VH),
	(
	 base_case(GH)
	 -> true
	  ; assert(base_case(GH)),
	    write(VH), writeln('.'), nl
	).
*/

write_base_case(Head) :-
	ground_head(Head,GH),
	write(GH),writeln('.'),nl.

ground_head(H,GH) :-
	functor(H,F,A),
	make_elem_list(A,x,L),
	GH =.. [F|L].


ground_and_var_head(H,GH,VH) :-
	functor(H,F,A),
	make_elem_list(A,x,XL),
	GH =.. [F|XL],
	make_elem_list(A,'_',UL),
	VH =.. [F|UL].


make_elem_list(0,_E,[]) :- !.
make_elem_list(N,E,[E|L]) :-
	M is N-1,
	make_elem_list(M,E,L).



% get_all_args(ALs, Args)
% ALs is a list of argument lists;
% Args is an ordered set of atomic arguments (of the form Argn)
% appearing in elements of ALs
get_atomic_args([],[]).
get_atomic_args([ifdef(_,TRs,ERs)|ALs], Args) :-
	get_atomic_args(TRs, TR_Args),
	get_atomic_args(ERs, ER_Args),
	get_atomic_args(ALs, ALs_Args),	
	l_append_no_rep(TR_Args, ER_Args, AL_Args),
	l_append_no_rep(AL_Args, ALs_Args, Args).
get_atomic_args([ifelse(_,_,TRs,ERs)|ALs], Args) :-
	get_atomic_args(TRs,TR_Args),
	get_atomic_args(ERs,ER_Args),
	get_atomic_args(ALs,ALs_Args),
	l_append_no_rep(TR_Args, ER_Args, AL_Args),
	l_append_no_rep(AL_Args, ALs_Args, Args).
get_atomic_args([(_,AL)|ALs], Args) :-
	get_atomic_args_from_list(AL, AL_Args),
	get_atomic_args(ALs, ALs_Args),
	l_append_no_rep(AL_Args, ALs_Args, Args).

get_atomic_args_from_list([],[]).
get_atomic_args_from_list([A|As], AtomicArgs) :-
	str_sub('Id', A, 0), !,
	get_atomic_args_from_list(As, Args),
	(
	 str_sub('Arg',A,Pos)
	 -> EPos is Pos+4,
	    substring(A,Pos,EPos,Arg),
	    AtomicArgs = [Arg|Args]
	  ; AtomicArgs = Args
	).
get_atomic_args_from_list([A|As], AtomicArgs) :-
	str_sub('_', A, Pos), !,
	substring(A, 0, Pos, Arg),
	get_atomic_args_from_list(As, AtArgs),
	append_no_rep(Arg, AtArgs, AtomicArgs).
get_atomic_args_from_list([A|As], [A|AtomicArgs]) :-
	get_atomic_args_from_list(As, AtomicArgs).


