%
% tokenizer for SELinux policy:
% reads in type enforcement (TE) specification and converts to 
% a list of tokens ready to be parsed;
% removes all comments;
%

/*
:- import expand_filename/2 from machine.
:- import word/3 from usermod.
*/

:- debug_ctl(hide, [atom_codes/2,
		    file_to_string/2,
		    scan_ident/4,
		    assert_words/3,
		    skip_line/2,
		    scan_te_file/1
		   ]).


%'C'(X,Y,Z) :- word(Y,X,Z).

scan_file(Filename) :-
	expand_filename(Filename, Afile),
	seeing(Ofile),
	see(Afile),
	file_to_string(String,[]),
	seen,
	see(Ofile),
	set_string(String, _Length).


%
% convert the current input stream into a list of character codes
%
file_to_string(Li,Lo) :-
	get0(C),
	(
	 C < 0			% end-of-file
	 -> Li = Lo
	  ; Li = [C|Lm],
	    file_to_string(Lm,Lo)
	).

set_string(String,Length) :-
	retractall(word(_,_,_)),
	scan(String,[],Words,[]),
	assert_words(Words,0,Length).

set_string(String,Length,Words) :-
	retractall(word(_,_,_)),
	scan(String,[],Words,[]),
	assert_words(Words,0,Length).

%
% scan list of character codes and separate into tokens
%

% empty list
scan([],[],W,W) :- !.
% comment mark - skip until end-of-line
scan([C|S0],S,W0,W) :-
	hash(C), !, 
	skip_line(S0, S1),
	scan(S1,S,W0,W).
% brackets & end-of-rule mark - append to token list
scan([C|S0],S,W0,W) :-
	(coma(C);
	 colon(C);
	 semicolon(C); 
	 C =:= 0'`; C =:= 0''; % opening or closing quote
	 C =:= 0'~;
	 wildchar(C);
	 l_par(C); r_par(C);
	 l_brace(C); r_brace(C)), !, 
	atom_codes(Wo, [C]),
	W0=[Wo|W1],
	scan(S0, S, W1, W).
% regular character
scan([C|S0],S,W0,W) :- 
	(C >= 0'a, C =< 0'z ; % lower case letter
         C >= 0'A, C =< 0'Z ; % upper case letter
	 C >= 0'0, C =< 0'9 ; % digit
	 C =:= 0'. ; % full stop
	 C =:= 0'_ ; % underscore
	 C =:= 0'- ; % hyphen
	 C =:= 0'$ ; % dollar sign
	 C =:= 0'= ; % equal sign
	 C =:= 0'!)  % exclamation mark
	-> !, scan_ident(S0,S1,Ws,[]),
	   atom_codes(Wo,[C|Ws]), 
	   % 'dnl' is another comment mark - skip till the new line 
	   (nonvar(Ws),
	    Wo = 'dnl'
	    -> skip_line(S1,S2),
	       scan(S2,S,W0,W)
	     ; Tok = Wo,
	       W0 = [Tok|W1],
	       scan(S1,S,W1,W)).
% ignore everything else
scan([_C|S0],S,W0,W) :-
	scan(S0,S,W0,W). 

new_line(10).
hash(35).
l_par(40).
r_par(41).
wildchar(42).
coma(44).
colon(58).
semicolon(59).
l_brace(123).
r_brace(125).


%
% scan a single token
%
scan_ident([],[],W,W) :- !.
scan_ident([C|Cs0],Cs,[C|Wcs0],Wcs) :-
	(C >= 0'a, C =< 0'z ;
         C >= 0'A, C =< 0'Z ;
	 C >= 0'0, C =< 0'9 ;
	 C =:= 0'. ; % full stop
	 C =:= 0'_ ; % underscore
	 C =:= 0'- ; % hyphen
	 C =:= 0'$ ; % dollar sign
%	 C =:= 0'~ ; % tylda
	 C =:= 0'= ; % equal sign
	 C =:= 0'!  % exclamation mark
	),!,
	scan_ident(Cs0,Cs,Wcs0,Wcs).
scan_ident(Cs,Cs,Wcs,Wcs).


%
% skip input until end-of-line character is seen
%
skip_line([],[]).
skip_line([C|L], L) :-
	new_line(C), !.
skip_line([_|L1], L2) :-
	skip_line(L1,L2).


:- index(word/3,2,101).
assert_words([],N,N).
assert_words([Word|String],N,M) :-
        N1 is N+1,
        assert(word(Word,N,N1)),
        assert_words(String,N1,M).

get_sentence_list([], []).
get_sentence_list(L, [S|Ss]) :-
	get_sentence(L, S, SR),
	get_sentence_list(SR, Ss).

get_sentence([W|Ws], [W], Ws) :-
	atom_codes(W, [C]), semicolon(C), !.
get_sentence([W|Ws], [W|S], R) :-
	get_sentence(Ws, S, R).
