% PAL (Policy Analysis using Logic-Programming)
% Copyright (C) 2004 Beata Sarna-Starosta
% Copyright (C) 2004 Scott D. Stoller
% Copyright (c) 2004 Stony Brook University
% Copyright (c) 2004 Research Foundation of SUNY

% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 2 of the License, or
% (at your option) any later version.
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see http://www.gnu.org/licenses/.

%
% auxiliary predicates for verification system and formula 
%


:- import append/3, member/2 from basics.
:- import str_cat/3 from string.

:- table compose_id/3.

:- debug_ctl(hide, [append/3,
		    init/2,
		    inv/3,
		    trans/7,
		    merge/4,
		    final/3,
		    member/2]).


% formatting the output
write_transitions :-
	write_transition,
	fail.
write_transitions.

write_transition :-
	trans_taken(Ls,Vs, Act, Cs, Ld,Vd),
	nl, writeln('TRANSITION:'),
	write('source location: '), writeln(Ls),
	write('source variables: '), writeln(Vs),nl,
	write('action taken: '), writeln(Act),nl,
	write('target loction: '), writeln(Ld),
	write('target variables: '), writeln(Vd),nl,
	write('the constraints: '), writeln(Cs),nl,
	writeln('***********************************************').
	


% merging constraint lists L1,L2 w.r.t. set of variables Vs into list L
merge([],[], _, []) :- !.
merge(L1,L2, Vs, L) :-
	append(L1,L2, L3),
	var_project(L3, Vs, L4),
	clean(L4,L).


% remove from L1 constraints that involve variables not in the set Vs
var_project([], _, []).

var_project([X|L1], Vs, L2) :-
	X = neq(V1,V2),
	atomic(V1), atomic(V2), !,
	V1 \== V2,
	var_project(L1, Vs, L2).
var_project([X|L1], Vs, [X|L2]) :-
	X = neq(V1,V2),
	(var_member(V1,Vs) ; var_member(V2,Vs)), !,
	V1 \== V2,
	var_project(L1, Vs, L2).

var_project([X|L1], Vs, [X|L2]) :-
	X = in(E,S),
	var_member(E,Vs),
	var_member(S,Vs), !,
	\+ constr_member(nin(E,S),L1),
	var_project(L1, Vs, L2).
var_project([X|L1], Vs, [univ(S)|L2]) :-
	X = in(E,S),
	var_member(S,Vs),!,
	\+ constr_member(nin(E,S),L1),
	var_project(L1, Vs, L2).

var_project([X|L1], Vs, [X|L2]) :-
	X = nin(E,S),
	var_member(E,Vs),
	var_member(S,Vs), !,
	\+ constr_member(in(E,S),L1),
	var_project(L1, Vs, L2).

var_project([_X|L1], Vs, L2) :-
	var_project(L1, Vs, L2).


% set membership on variables
var_member(V, [X|_Xs]) :-
	V == X.
var_member(V, [X|Xs]) :-
	V \== X,
	var_member(V,Xs).


% set membership on constraints
constr_member(neq(X1,X2), [neq(Y1,Y2)|Cs]) :-
	(X1 == Y1, X2 == Y2)
	;
	(X1 == Y2, X2 == Y1)
	;
	constr_member(neq(X1,X2), Cs).

constr_member(in(E1,S1), [in(E2,S2)|Cs]) :-
	(E1 == E2, S1 == S2)
	;
	constr_member(in(E1,S1), Cs).

constr_member(nin(E1,S1), [nin(E2,S2)|Cs]) :-
	(E1 == E2, S1 == S2)
	;
	constr_member(nin(E1,S1), Cs).

constr_member(C, [_|Cs]) :-
	constr_member(C, Cs).


% remove duplicate constraints from the list
clean([],[]).
clean([C|Cs], L) :-
	satisfiable(C),
	((ground(C) ; constr_member(C, Cs))
	  -> L = Cs
	   ; L = [C|Cs]
	).

satisfiable(eq(_,_)).
satisfiable(neq(X,Y)) :-
	X \== Y.
satisfiable(in(_,_)).
satisfiable(nin(_,[])).
satisfiable(nin(_,Cs)) :-
	var(Cs), !.
satisfiable(nin(X,[(A,_,_)|Cs])) :-
	X \== A,
	satisfiable(nin(X,Cs)).

ground(neq(X,Y)) :-
	atom(X),
	atom(Y).


type_eq(T1,T2) :-
	T1 = T2
	;
	(type_decl(T1,_Aliases,Atts,VAtts),
	 append(Atts,VAtts,AllAtts),
	 member(T2,AllAtts)).



compose_id(Pref, Suf, Id) :-
	atomic(Pref),
	str_cat(Pref, '_', Pref_),
	str_cat(Pref_, Suf, Id).



var_string(S) :-
	atom_codes(S, [C|_]),
	C >= 0'A,
	C =< 0'Z.


not_defined(X) :- \+ defined(X).
