% PAL (Policy Analysis using Logic-Programming)
% Copyright (C) 2004 Beata Sarna-Starosta
% Copyright (C) 2004 Scott D. Stoller
% Copyright (c) 2004 Stony Brook University
% Copyright (c) 2004 Research Foundation of SUNY

% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 2 of the License, or
% (at your option) any later version.
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see http://www.gnu.org/licenses/.

%
% in reachability analysis there are two options, the first to use 
% FORWARD reachability, the second to use BACKWARD reachability;
% few predicates have slightly different versions for each analysis
% direction - check comments for backward/forward;
% in case of SLAT fsadm_t example, backward analysis seems to work
% better, so all predicates are set for this case;
%

%
% before running for particural query, (un)comment proper version of equiv_type
%


:- import append/3, member/2 from basics.

:- table auth/8.
:- table flow_trans/4.
:- table role_type_perm/2.
:- table user_role_perm/2.
:- table read_like/2.
:- table write_like/2.
:- auto_table.

:- assert(defined(0)).

%
% verification
%

% construct a product of A1 and A2 with initial location (Ls1,Ls2)
% and the set of final locations {(Ld1,Ld2)},
% and look for a run from (Ls1,Ls2) to a (Ld1i,Ld2i) in {(Ld1,Ld2)}
% Constr contains the constraints that must hold for the run to exists

% (un)comment proper call to reach for forward or backward reachability
verify(A1,V_s1, A2,V_s2, C_in, C_out) :-
	assert(trans_taken(0, 0, 0, 0, 0)),
	assert(visited_node(0)),
	init((A1,A2), (V_s1,V_s2), C_s),
	append(V_s1,V_s2, V_s),
	merge(C_in,C_s, V_s, Constr_in),
	final((A1,A2), (V_d1,V_d2), C_d),
%	reach((A1,A2), (V_s1,V_s2), (V_d1,V_d2), Constr_in, Constr_out),
	b_reach((A1,A2), (V_d1,V_d2), (V_s1,V_s2), Constr_in, Constr_out),
	append(V_d1,V_d2, V_d),
	merge(C_d,Constr_out, V_d, C_out),
	write_transitions.



write_transitions :-
	retract(trans_taken(0,0,0,0,0)),
	fail.
write_transitions :-
	write_transition,
	fail.
write_transitions :-
	retractall(trans_taken(_,_,_,_,_)),
	assert(trans_taken(0,0,0,0,0)).


write_transition :-
	trans_taken(_A, Vs, Act, Vd, Cs),
	nl, writeln('TRANSITION:'),
	write('source variables: '), writeln(Vs),nl,
	write('action taken: '), writeln(Act),nl,
	write('target variables: '), writeln(Vd),nl,
	(Cs \== []
	 -> write('constraint set: '), writeln(Cs),nl
	  ; true),
	writeln('***********************************************').

write_transition(Vs, Act, Vd) :-
	nl, writeln('TRANSITION:'),
	write('source variables: '), writeln(Vs),nl,
	write('action taken: '), writeln(Act),nl,
	write('target variables: '), writeln(Vd),nl,
	writeln('***********************************************').


% forward reachability:
% look for a run from start state Vs to destination state Vd
reach(_A,V,V, C,C).

reach(A, Vs, Vd, C_in, C_out) :-
	trans(A, Vs, Act, Vm, C_trans),
	(
	 visited_node(Vm)
	 -> fail
	  ; assert(visited_node(Vm)),
	    reach(A, Vm, Vd, C_in, Cm),
	    merge(C_trans, Cm, Vd, C_out),
	    (
	     trans_taken(A, Vs, Act, Vm, C_trans)
	     -> true
	      ; asserta(trans_taken(A, Vs, Act, Vm, C_trans))
	    )
	).


% backward reachability:
% look for a run from destination state Vd to start state Vs
b_reach(_A, V, V, C, C).

b_reach(A, Vd, Vs, C_in, C_out) :-
	trans(A, Vm, Act, Vd, C_trans),
	(
	 visited_node(Vm)
	 -> fail
	  ; assert(visited_node(Vm))
	),
	b_reach(A, Vm, Vs, C_in, Cm),
	merge(Cm, C_trans, Vd, C_out),
	(
	 trans_taken(A, Vm, Act, Vd, C_trans)
	 -> true
	  ; assertz(trans_taken(A, Vm, Act, Vd, C_trans))
	).



% information flow from T1 to T2 is permitted by the policy
% flow-perm corresponds to alpha_t in [GHR03]

/*
% version for forward reachability
flow_perm(T1,T2,C,P) :-
	(AVKind = allow ; AVKind = auditallow),
	access_vector(AVKind, SourceType, TargetType, C, Perm),
	equiv_perm(C,P,Perm),
	equiv_type(T1,SourceType),
	(
	 TargetType = self
	 -> T2 = T1
	  ; equiv_type(T2,TargetType)
	).
*/

% version for backward reachability
flow_perm(T1,T2,C,P) :-
	(AVKind = allow ; AVKind = auditallow),
	(
	 (T2 \= self, equiv_type_member(T2,TargetTypes))
	 ;
	 (T2 = self, T1 = T2)
	),
	access_vector(AVKind, SourceTypes, TargetTypes, Class, Perm),
	equiv_type_member(T1,SourceTypes),
	valid_type(T1),
	valid_type(T2),
	macro_member(C,Class),
	equiv_perm(C,P,Perm).

valid_type(T) :-
	T \= 'x',
	\+ str_sub('x_', T, 0),
	\+ str_sub('_x_', T, _).



equiv_type(T1,T2) :-
	type_decl(T1,_Aliases,Atts,VAtts),
	(
	 T1 = T2
	;
	 append(Atts,VAtts,AllAtts),
	 member(T2,AllAtts)
	).

equiv_type_member(T1,TL) :-
	nonvar(TL), TL = [T2|Ts],
	(
	equiv_type(T1,T2)
	;
	equiv_type_member(T1,Ts)
	).
equiv_type_member(T1,T2) :-
	equiv_type(T1,T2).



macro_member(X,Y) :-
	set_macro(Y,L),!,
	member(X,L)
	;
	X = Y.

equiv_perm(C,P,Ps) :-
	macro_member(P,Ps)
	;
	Ps = except(PL),
	perm_set_complement(C,PL,ValidPs),
	member(P,ValidPs).


perm_set_complement(Class,ExcludedPs,ValidPs) :-
	all_access_perms(Class,AllPs),
	list_difference(AllPs,ExcludedPs,ValidPs).

all_access_perms(Class,AllPs) :-
	access(class,Class,Ps1),
	(
	 inherits(Class,Common)
	 -> access(common,Common,Ps2)
	  ; Ps2 = []
	),
	append(Ps1,Ps2,AllPs).


list_difference([],_,[]).
list_difference(L,[],L) :- !.
list_difference([X|Xs],LSub,XRs) :-
	member(X,LSub),!,
	list_difference(Xs,LSub,XRs).
list_difference([X|Xs],LSub,[X|XRs]) :-
	list_difference(Xs,LSub,XRs).



% role R can associate with type T
% role_type_perm corresponds to rho in [GHR03]

role_type_perm(R,T) :-
	var(R),
	var(T), !.
role_type_perm(object_r,_T) :- !.
role_type_perm(R,T) :-
	role_decl(R,RT),
	valid_role(RT),
	equiv_type(T,RT).
role_type_perm(R,T) :-
	role_decl(R,RL),
	member(RT,RL),
	equiv_type(T,RT).



valid_role(R) :-
	\+ is_list(R),
	R \== 'x_r'.



% user U can assume role R
% user_role_perm corresponds to mu in [GHR03]

user_role_perm(U,R) :-
	var(U),
	var(R), !.
user_role_perm(_U,object_r) :- !.
user_role_perm(U,R) :-
	user(U,R).


% *authorization predicate* specifying conditions for an authorized
% information flow from a valid security context <T1,R1,U1> 
% to a valid security context <T2,R2,U2>;
% auth corresponds to Delta_{cp} in [GHR03]

auth(C,P, T1,R1,U1, T2,R2,U2) :-
	flow_perm(T1,T2, C,P),
	constraint(C,P, T1,R1,U1, T2,R2,U2),

	(
	 C = process, P = transition
	 -> role_allow(R1,R2)
	  ; true
	),
	valid_role(R1), valid_role(R2).


% *information flow predicate* for information flow from
% <T1,R1,U1> to <T2,R2,U2> via event (C,P)
% flow_trans corresponds to \Phi_{cp} in [GHR03]

flow_trans([T1,R1,U1], C,P, [T2,R2,U2]) :-
	auth(C,P, T1,R1,U1, T2,R2,U2), write_like(C,P)
	;
	auth(C,P, T2,R2,U2, T1,R1,U1), read_like(C,P).


write_like(C,P) :-
	perm_map(C,P,write).


read_like(C,P) :-
	perm_map(C,P,read).


perm_map(C,P,Ps) :-
	(
	 perm_map(class,C,P,Ps1) ; Ps1 = []
	),
	(
	 inherits(C,Common)
	 -> perm_map(common,Common,P,Ps2),
	    append(Ps1,Ps2,Ps)
	  ; Ps = Ps1
	).	



% policy *information flow graph*.

% the states are consistent security contexts <T,R,U>

consistent([T,R,U]) :-
	role_type_perm(R,T),
	user_role_perm(U,R).


% the initial state is any <T,R,U> consistent without any
% role or type transitions in the policy specification

init(pg, [T,R,U], []) :-
	role_type_perm(R,T),
	user_role_perm(U,R).

% the set of final states is the set of all states of the graph:

final(pg, _, []).


% transitions in the graph are defined in terms of flow_trans predicate,
% with checking invariants to ensure consistency:

trans(pg, Vs, (C,P), Vd, []) :-
	consistent(Vs),
	flow_trans(Vs, C,P, Vd),
	consistent(Vd).


%
% automaton for security goal from SLAT example in info/slat.info
%
%
% Description:
%
%	``We will now provide an example of SLAT usage. One goal for 
%	the example SE Linux policy included in the distribution is to 
%	restrict access to raw disk data. The data is labeled with type 
%	`fixed_disk_device_t', and the type `fsadm_t' was created to 
%	track which programs required access to the data. A reasonable 
%	security goal, then, is to ensure that any information flowing 
%	into `fixed_disk_device_t' has passed through the "gateway" type 
%	`fsadm_t' first.''
%
% Negated property:
%
%	There exists information flow into type `fixed_disk_device_t'
%	that does not pass through type `fsadm_t' first:


init(p, [user_t,user_r,U], [neq(U,jadmin)]).

trans(p, [T0,_R0,_U0], (_C,_P), [_T1,_R1,_U1], [neq(T0,fsadm_t)]).
trans(p, [T0,_R0,_U0],(_C,_P),[fixed_disk_device_t,_R1,_U1],[neq(T0,fsadm_t)]).

final(p, [fixed_disk_device_t,_R,_U], []).





%
% product construction
%

% initial location

init((A_s,A_p), (V_s,V_p), Constr) :-
	init(A_s, V_s, C_s),
	init(A_p, V_p, C_p),
	
	% comment the following line for forward reachability,
	% uncomment for backward reachability
	V_s = V_p,

	% uncomment the following line for forward reachability,
	% comment for backward reachability
%	unifiable(V_s,V_p),

	append(C_s,C_p, Constr).



% transition relation

trans((A_s,A_p),
      (Vs_s,Vs_p),
      Act,
      (Vd_s,Vd_p),
      Constr) :-
	% comment the following line for forward reachability,
	% uncomment for backward reachability
	Vs_s = Vs_p,
	
	% uncomment the following line for forward reachability,
	% comment for backward reachability
%	type_unifiable(Vs_s,Vs_p),

	trans(A_p, Vs_p, Act, Vd_p, Constr_p),
	trans(A_s, Vs_s, Act, Vd_s, Constr_s),
	% comment the following line for forward reachability,
	% uncomment for backward reachability
	Vd_s = Vd_p,
	
	% uncomment the following line for forward reachability,
	% comment for backward reachability
%	type_unifiable(Vd_s,Vd_p),

	append(Constr_s,Constr_p, Constr).


% set of final locations

final((A_s,A_p), (V_s,V_p), Cs) :-
	final(A_s,V_s,C_s),
	final(A_p,V_p,C_p),
	
	% comment the following line for forward reachability,
	% uncomment for backward reachability
	V_s = V_p,

	% uncomment the following line for forward reachability,
	% comment for backward reachability
%	type_unifiable(V_s,V_p),

	append(C_s,C_p, Cs).


type_unifiable([T,_,_],[T,_,_]) :- !.


unifiable([],[]).
unifiable(L1,L2) :-
	(
	  var(L1)
	  ;
	  var(L2)
	), !.
unifiable([X|Xs],[Y|Ys]) :-
	(
	  X == Y
	  ;
	  var(X)
	  ;
	  var(Y)
	), !,
	unifiable(Xs,Ys).



%%
%% counterexample automaton
%%

init(ce, [user_t,user_r,jdoe], []).

trans(ce, [user_t,user_r,jdoe],
          (netif,rawip_send),
          [netif_ipsec2_t,object_r,jdoe],
          []).
trans(ce, [netif_ipsec2_t,object_r,jdoe],
          (netif,udp_recv),
          [dpkg_t,system_r,system_u],
          []).
trans(ce, [dpkg_t,system_r,system_u],
          (fifo_file,append),
          [fixed_disk_device_t,object_r,jdoe],
          []).

final(ce, [fixed_disk_device_t,object_r,jdoe], []).
