% PAL (Policy Analysis using Logic-Programming)
% Copyright (C) 2004 Beata Sarna-Starosta
% Copyright (C) 2004 Scott D. Stoller
% Copyright (c) 2004 Stony Brook University
% Copyright (c) 2004 Research Foundation of SUNY

% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 2 of the License, or
% (at your option) any later version.
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see http://www.gnu.org/licenses/.


% policy without dpkg.te
%:- load_dyn('/home/bss/projects/selinux/policy-1.1/policy_model/model/macro_expanded/policy_nm').

% policy with dpkg.te
:- load_dyn('/home/bss/projects/selinux/policy-1.1/policy_model/model/macro_expanded/policy_nm_dpkg').

:- ['/home/bss/projects/selinux/policy-1.1/policy_model/analysis/inflow/verify_aux',
    '/home/bss/projects/selinux/policy-1.1/policy_model/analysis/verify'].

:- auto_table.
:- use_subsumptive_tabling transitive_flow/2.

:- import member/2, length/2 from basics.
:- import str_sub/3 from string.


%% simple (non-transitive closure) queries for policy analysis

% integrity
:- table integrity_violation/2.
integrity_violation(TCBType, OutType) :-
	tcb(TCB),
	member(TCBType,TCB),
	type_trans(OutType,TCBType),
	\+ member(OutType,TCB).



tcb([init_t, kernel_t, initrc_t, sysadm_t, getty_t, mount_t, fsadm_t,
     load_policy_t, dpkg_t, check_policy_t, setfiles_t, syslogd_t, klogd_t,
     automount_t, sshd_t, sshd_login_t, local_login_t, quota_t, ldconfig_t,
     useradd_t, hwclock_t, apt_t, cardmgr_t, ipsec_mgmt_t, admin_passwd_exec_t,
     bootloader_t, logrotate_t, newrole_t, snmpd_t, passwd_t]).

:- table type_trans/2.
type_trans(OutType,TCBType) :-
	flow_trans([OutType,_OutRole,_OutUser],
		   _C,_P,
		   [TCBType,_TCBRole,_TCBUser]).


equiv_type_member(T,Ts) :-
	member(T1,Ts),
	equiv_type(T,T1).


iv_test(InType,OutType) :-
	statistics(0),
	integrity_violation(InType,OutType),
	statistics.

:- table iv_test_all/1.
iv_test_all(Tuples) :-
	statistics(0),
	findall((InType,OutType),
		integrity_violation(InType,OutType),
		Tuples),
	statistics.


iv_test_all(Tuples,L) :-
	statistics(0),
	tfindall((InType,OutType),
		integrity_violation(InType,OutType),
		Tuples),
	statistics,
	length(Tuples,L),
	tell('iv_output.P'),
	write_iv_list(Tuples),
	told.



write_iv_list([]).
write_iv_list([X|Xs]) :-
	write(iv(X)),writeln('.'),
	write_list(Xs).

iv((TCBType,OutType)) :-
	tcb(TCB),
	equiv_type_member(TCBType,TCB),
	flow_trans([OutType,_OutRole,_OutUser],
		   _C,_P,
		   [TCBType,_TCBRole,_TCBUser]),
	\+ equiv_type_member(OutType,TCB).


:- hilog iv_types.
iv_types(Ts) :- iv(Ts).

:- import bagCount/2, bagReduce/4 from aggregs.

:- hilog successor.
successor(X,_Y,Z) :- Z is X+1.




% separation of duty 
daemon_can_execute_and_write(DaemonType,FileType) :-
	daemon_type(DaemonType),
	auth(Class, execute, DaemonType,_R,_U, FileType,_R,_U),
	auth(Class, Perm, DaemonType,_R,_U, FileType,_R,_U),
	write_like(Class,Perm).


daemon_type(T) :-
	type_decl(T,_,_),
	is_daemon_type(T).

is_daemon_type(Type) :-
	is_daemon_suf(Type), !.
is_daemon_type(Type) :-
	str_sub('crond_t', Type, _), !.
is_daemon_type(Type) :-
	str_sub('klogd_t', Type, _), !.
is_daemon_type(Type) :-
	str_sub('sshd_t', Type, _), !.
is_daemon_type(Type) :-
	str_sub('syslogd_t', Type, _), !.

is_daemon_suf('crond_t').
is_daemon_suf('klogd_t').
is_daemon_suf('sshd_t').
is_daemon_suf('syslogd_t').

d_test(DaemonType,FileType) :-
	statistics(0),
	(
	 daemon_can_execute_and_write(DaemonType,FileType)
	 ;
	 (nl,writeln('FAILED'),nl)
	),
	statistics.


% completness
unspecified_permission(T1,T2,P) :-
	type_decl(T1,_,_),
	type_decl(T2,_,_),
	perm_valid_for_class(P,C),
	\+ access_vector(_,T1,T2,C,P),
	\+ neverallow1(T1,T2,C,P).


perm_valid_for_class(P,C) :-
	access(class,C,Ps),
	member(P,Ps).
perm_valid_for_class(P,C) :-
	access(common,CC,Ps),
	inherits(C,CC),
	member(P,Ps).


neverallow1(T1,T2,C,P) :-
	neverallow(T1s,T2s,Cs,Ps),
	equivalent(T1,T1s),
	equivalent(T2,T2s),
	equiv_class(C,Cs),
	equiv_perm(P,Ps).

% assertion check
assertion_violation(A_T1,A_T2,C,A_P,
		    AVKind,AV_T1,AV_T2,AV_P) :-
	neverallow(A_T1,A_T2,C,A_P),
	(AVKind = allow ; AVKind = auditallow),
	access_vector(AVKind,AV_T1,AV_T2,C,AV_P),
	equivalent(A_T1,AV_T1),
	(
	 (AV_T2 \= self, equivalent(A_T2,AV_T2))
	 ;
	 (AV_T2 = self, equivalent(A_T2,AV_T1))
	),
	equiv_perm(C,A_P,AV_P).


equivalent(X,Y) :-
	nonvar(X), X = except(T),
	(
	 is_list(T)
	 -> \+ equiv_member(Y,T)
	  ; equivalent(Y,Z),
	    \+ equivalent(T,Z)
	).
equivalent(X,Y) :-
	nonvar(Y), Y = except(T),
	(
	 is_list(T)
	 -> \+ equiv_member(X,T)
	  ; equivalent(X,Z),
	    \+ equivalent(T,Z)
	).
equivalent(X,Y) :-
	equiv_type(X,Y).



equiv_member(X,[Y|Ys]) :-
	equiv_type(X,Y)
	;
	equiv_member(X,Ys).



av_test1 :-
	statistics(0),
	(
	 (assertion_violation(T1,T2,C,P1,K,T3,T4,P2),
	 nl,write('Assertion: '),writeln(neverallow(T1,T2,C,P1)),
	    write('violated by rule: '), writeln(access_vector(K,T3,T4,C,P2)))
	 ;
	 (nl,writeln('No violated assertions found in the policy.'),nl)
	),
	statistics.


av_test :-
	statistics(0),
	findall((T1,T2,C,P1,K,T3,T4,P2),
	        assertion_violation(T1,T2,C,P1,K,T3,T4,P2),
		RuleArgs),
	(
	 RuleArgs = []
	 -> nl,writeln('No violated assertions found in the policy.'),nl
	  ; write_rule_args(RuleArgs)
	),
	statistics.

write_rule_args([]).
write_rule_args([(T1,T2,C,P1,K,T3,T4,P2)|Args]) :-
	nl,
	write('Assertion: '),
	writeln(neverallow(T1,T2,C,P1)),
	write('violated by rule: '),
	write(access_vector(K,T3,T4,C,P2)),
	writeln('...'),
	write_rule_args(Args).



% find the security context from which information can flow into shadow_t
transitive_flow(SC1,SC2) :-
	flow_trans(SC1,_C,_P,SC2).
transitive_flow(SC1,SC2) :-
	flow_trans(SC,_C,_P,SC2),
	transitive_flow(SC1,SC).


%
% to run any of the following tests, comment the second and third clauses
% of equiv_type in verify.P
%

tf_test(Context) :-
	transitive_flow(Context,[shadow_t,_R,_U]).

tf_test_all :-
	statistics(0),
	findall(Context,
		transitive_flow(Context,[shadow_t,_R,_U]),
		Contexts),
	(
	 Contexts = []
	 -> nl,writeln('No specified information flow found.'),nl
	  ; nl,write('Information flow into shadow_t found from '),
	    length(Contexts,N),write(N), writeln(' security contexts:'),
	    nl,write_list(Contexts)
	),
	statistics.


write_list([]).
write_list([X|Xs]) :-
	writeln(X),
	write_list(Xs).
