#line 1 "tmp/all.te"
#
# Declarations for type attributes.
# 

# A type attribute can be used to identify a set of types with a similar
# property.  Each type can have any number of attributes, and each
# attribute can be associated with any number of types.  Attributes are
# explicitly declared here, and can then be associated with particular
# types in type declarations.  Attribute names can then be used throughout 
# the configuration to express the set of types that are associated with 
# the attribute.  Except for the MLS attributes, attributes have no implicit
# meaning to SELinux.  The meaning of all other attributes are completely 
# defined through their usage within the configuration, but should be 
# documented here as comments preceding the attribute declaration.  

#####################
# Attributes for MLS:
#

# The mlstrustedreader attribute identifies every domain that can
# override the normal MLS restrictions on reading (i.e. domains
# that can read up).  
attribute mlstrustedreader;

# The mlstrustedwriter attribute identifies every domain that can
# override the normal MLS restrictions on writing (i.e. domains
# that can write down).  
attribute mlstrustedwriter;

# The mlstrustedobject attribute identifies every type that can
# be accessed without normal MLS restrictions (i.e. processes can
# read or write objects with this type regardless of MLS level).  
# Examples:  /dev/null, descriptors created by login
attribute mlstrustedobject;


#########################
# Attributes for domains:
#

# The domain attribute identifies every type that can be 
# assigned to a process.  This attribute is used in TE rules 
# that should be applied to all domains, e.g. permitting 
# init to kill all processes or permitting all processes
# to read a particular file.
attribute domain;

# The privuser attribute identifies every domain that can 
# change its SELinux user identity.  This attribute is used 
# in the constraints configuration.  NOTE:  This attribute
# is not required for domains that merely change the Linux
# uid attributes, only for domains that must change the
# SELinux user identity.
attribute privuser;

# The privrole attribute identifies every domain that can 
# change its SELinux role.  This attribute is used in the 
# constraints configuration.
attribute privrole;

# The privowner attribute identifies every domain that can 
# assign a different SELinux user identity to a file.  This 
# attribute is used in the constraints configuration.
attribute privowner;

# The privlog attribute identifies every domain that can 
# communicate with syslogd through its Unix domain socket.
# This attribute is used in the TE rules in 
# domains/program/syslogd.te to grant such access.  
# XXX If you want to mandate the use of this attribute for all 
# XXX domains that can log, then you should also write corresponding 
# XXX assertions in assert.te to enforce this restriction.  Otherwise,
# XXX it is just an optional convenience for domain writers.
attribute privlog;

# The privmem attribute identifies every domain that can 
# access kernel memory devices.
# This attribute is used in the TE assertions to verify
# that such access is limited to domains that are explicitly
# tagged with this attribute.
attribute privmem;

# The privfd attribute identifies every domain that should have
# file handles inherited widely (IE sshd_t and getty_t).
attribute privfd;

# The privhome attribute identifies every domain that can create files under
# regular user home directories in the regular context (IE act on behalf of
# a user in writing regular files)
attribute privhome;

# The auth attribute identifies every domain that needs
# to perform user authentication and requires access to
# the corresponding authentication data.  
# XXX This attribute is no longer in use except in type declarations.
# XXX It was introduced for the original attempt to put /etc/shadow into
# XXX a separate type and to limit read access to certain domains.  
# XXX Doing so transparently to applications is problematic due to
# XXX the fact that both /etc/passwd and /etc/shadow are recreated
# XXX by the same processes and live in the same directory, so file type
# XXX transition rules cannot specify different default types.  Relocating
# XXX /etc/shadow to a separate subdirectory was tried, but required
# XXX modifications to a lot of applications even after changing
# XXX the pwdb shared library.
attribute auth;

# The admin attribute identifies every administrator domain.
# It is used in TE assertions when verifying that only administrator 
# domains have certain permissions.  
# This attribute is presently associated with sysadm_t and 
# certain administrator utility domains.  
# XXX The use of this attribute should be reviewed for consistency.
# XXX Might want to partition into several finer-grained attributes 
# XXX used in different assertions within assert.te.
attribute admin;

# The userdomain attribute identifies every user domain, presently
# user_t and sysadm_t.  It is used in TE rules that should be applied
# to all user domains.
attribute userdomain;

# attribute for all non-administrative devpts types
attribute userpty_type;

# The user_crond_domain attribute identifies every user_crond domain, presently
# user_crond_t and sysadm_crond_t.  It is used in TE rules that should be
# applied to all user domains.
attribute user_crond_domain;

# The unpriv_userdomain identifies non-administrative users (default user_t)
attribute unpriv_userdomain;

# This attribute is for the main user home directory for unpriv users
attribute user_home_dir_type;

# The gphdomain attribute identifies every gnome-pty-helper derived
# domain.  It is used in TE rules to permit inheritance and use of
# descriptors created by these domains.
attribute gphdomain;


############################
# Attributes for file types:
#

# The file_type attribute identifies all types assigned to files 
# in persistent filesystems.  It is used in TE rules to permit
# the association of all such file types with persistent filesystem
# types, and to permit certain domains to access all such types as 
# appropriate.
attribute file_type;

# The sysadmfile attribute identifies all types assigned to files 
# that should be completely accessible to administrators.  It is used
# in TE rules to grant such access for administrator domains.
attribute sysadmfile;

# The fs_type attribute identifies all types assigned to filesystems
# (not limited to persistent filesystems).
# It is used in TE rules to permit certain domains to mount
# any filesystem and to permit most domains to obtain the
# overall filesystem statistics.
attribute fs_type;

# The root_dir_type attribute identifies all types assigned to 
# root directories of filesystems (not limited to persistent
# filesystems).
# XXX This attribute was used to grant mountassociate permission,
# XXX but this permission is no longer defined.  We can likely
# XXX remove this attribute.
attribute root_dir_type;

# The exec_type attribute identifies all types assigned
# to entrypoint executables for domains.  This attribute is 
# used in TE rules and assertions that should be applied to all 
# such executables.
attribute exec_type;

# The tmpfile attribute identifies all types assigned to temporary 
# files.  This attribute is used in TE rules to grant certain 
# domains the ability to remove all such files (e.g. init, crond).
attribute tmpfile;

# The user_tmpfile attribute identifies all types associated with temporary
# files for unpriv_userdomain domains.
attribute user_tmpfile;

# for the user_xserver_tmp_t etc
attribute xserver_tmpfile;

# The tmpfsfile attribute identifies all types defined for tmpfs 
# type transitions. 
# It is used in TE rules to grant certain domains the ability to
# access all such files.
attribute tmpfsfile;

# The home_type attribute identifies all types assigned to home
# directories.  This attribute is used in TE rules to grant certain
# domains the ability to access all home directory types.
attribute home_type;

# This attribute is for the main user home directory /home/user, to
# distinguish it from sub-dirs.  Often you want a process to be able to
# read the user home directory but not read the regular directories under it.
attribute home_dir_type;

# The ttyfile attribute identifies all types assigned to ttys.
# It is used in TE rules to grant certain domains the ability to
# access all ttys.
attribute ttyfile;

# The ptyfile attribute identifies all types assigned to ptys.
# It is used in TE rules to grant certain domains the ability to
# access all ptys.
attribute ptyfile;

# The pidfile attribute identifies all types assigned to pid files.
# It is used in TE rules to grant certain domains the ability to
# access all such files.
attribute pidfile;


############################
# Attributes for network types:
#

# The socket_type attribute identifies all types assigned to 
# kernel-created sockets.  Ordinary sockets are assigned the 
# domain of the creating process.
# XXX This attribute is unused.  Remove?
attribute socket_type;

# Identifies all types assigned to port numbers to control binding.
attribute port_type;

# Identifies all types assigned to network interfaces to control
# operations on the interface (XXX obsolete, not supported via LSM) 
# and to control traffic sent or received on the interface.
attribute netif_type;

# Identifies all default types assigned to packets received 
# on network interfaces.  
attribute netmsg_type;

# Identifies all types assigned to network nodes/hosts to control
# traffic sent to or received from the node.
attribute node_type;

# Identifier for log files or directories that only exist for log files.
attribute logfile;

# Identifier for lock files (/var/lock/*) or directories that only exist for
# lock files.
attribute lockfile;



##############################
# Attributes for security policy types:
#

# The login_contexts attribute idenitifies the files used
# to define default contexts for login types (e.g., login, cron).
attribute login_contexts;

# Identifier for a domain used by "sendmail -t" (IE user_mail_t,
# sysadm_mail_t, etc)
attribute user_mail_domain;

# Identifies domains that can transition to system_mail_t
attribute privmail;

# Type for non-sysadm home directory
attribute user_home_type;

# For domains that are part of a mail server and need to read user files and
# fifos, and inherit file handles to enable user email to get to the mail
# spool
attribute mta_user_agent;

# For domains that are part of a mail server for delivering messages to the
# user
attribute mta_delivery_agent;

# For web clients such as netscape and squid
attribute web_client_domain;

# For a mail server process that takes TCP connections on port 25
attribute mail_server_domain;



























#
# Macros for all admin domains.
#

#
# admin_domain(domain_prefix)
#
# Define derived types and rules for an administrator domain.
#
# The type declaration and role authorization for the domain must be
# provided separately.  Likewise, domain transitions into this domain
# must be specified separately.  If the every_domain() rules are desired,
# then these rules must also be specified separately.
#

#line 456

##############################
#
# Global macros for the type enforcement (TE) configuration.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# Modified: Howard Holm (NSA), <hdholm@epoch.ncsc.mil>
#           System V IPC added
#

#################################
# 
# Macros for groups of classes and 
# groups of permissions.
#

#
# All directory and file classes
#


#
# All non-directory file classes.
#


#
# Non-device file classes.
#


#
# Device file classes.
#


#
# All socket classes.
#


#
# Datagram socket classes.
# 


#
# Stream socket classes.
#


#
# Unprivileged socket classes (exclude rawip, netlink, packet).
#



# 
# Permissions for getting file attributes.
#


# 
# Permissions for executing files.
#


# 
# Permissions for reading files and their attributes.
#


# 
# Permissions for reading and executing files.
#


# 
# Permissions for reading and writing files and their attributes.
#


# 
# Permissions for reading and appending to files.
#


#
# Permissions for linking, unlinking and renaming files.
# 


#
# Permissions for creating and using files.
# 


# 
# Permissions for reading directories and their attributes.
#


# 
# Permissions for reading and writing directories and their attributes.
#


# 
# Permissions for reading and adding names to directories.
#



#
# Permissions for creating and using directories.
# 


#
# Permissions to mount and unmount file systems.
#


#
# Permissions for using sockets.
# 


#
# Permissions for creating and using sockets.
# 


#
# Permissions for using stream sockets.
# 


#
# Permissions for creating and using stream sockets.
# 



#
# Permissions for sending all signals.
#


#
# Permissions for sending and receiving network packets.
#


#
# Permissions for using System V IPC
#










#################################
# 
# Macros for type transition rules and
# access vector rules.
#

#
# Simple combinations for reading and writing both
# directories and files.
# 
#line 638


#line 643


#line 648


#line 653


#line 658


#line 663


#line 668


##################################
#
# can_getsecurity(domain)
#
# Authorize a domain to get security policy decisions.
#
#line 680


##################################
#
# can_setenforce(domain)
#
# Authorize a domain to set the enforcing flag.
# Due to its sensitivity, always audit this permission.
#
#line 694


##################################
#
# can_loadpol(domain)
#
# Authorize a domain to load a policy configuration.
# Due to its sensitivity, always audit this permission.
#
#line 708


#################################
#
# domain_trans(parent_domain, program_type, child_domain)
#
# Permissions for transitioning to a new domain.
#

#line 760


#################################
#
# domain_auto_trans(parent_domain, program_type, child_domain)
#
# Define a default domain transition and allow it.
#
#line 771


##################################
#
# can_setexec(domain)
#
# Authorize a domain to set its exec context
# (via /proc/pid/attr/exec).
#
#line 786


##################################
#
# can_setfscreate(domain)
#
# Authorize a domain to set its fscreate context
# (via /proc/pid/attr/fscreate).
#
#line 801


#################################
#
# uses_shlib(domain)
#
# Permissions for using shared libraries.
#
#line 820


#################################
#
# can_ptrace(domain, domain)
#
# Permissions for running ptrace (strace or gdb) on another domain
#
#line 830


#################################
#
# can_exec(domain, type)
#
# Permissions for executing programs with
# a specified type without changing domains.
#
#line 841


#################################
#
# can_exec_any(domain)
#
# Permissions for executing a variety
# of executable types.
#
#line 859



#################################
#
# file_type_trans(domain, dir_type, file_type)
#
# Permissions for transitioning to a new file type.
#

#line 890


#################################
#
# file_type_auto_trans(creator_domain, parent_directory_type, file_type, object_class)
#
# the object class will default to notdevfile_class_set if not specified as
# the fourth parameter
#
# Define a default file type transition and allow it.
#
#line 911


#################################
#
# can_network(domain)
#
# Permissions for accessing the network.
# See types/network.te for the network types.
# See net_contexts for security contexts for network entities.
#
#line 989


#################################
#
# can_unix_connect(client, server)
#
# Permissions for establishing a Unix stream connection.
#
#line 999


#################################
#
# can_unix_send(sender, receiver)
#
# Permissions for sending Unix datagrams.
#
#line 1009


#################################
#
# can_tcp_connect(client, server)
#
# Permissions for establishing a TCP connection.
#
#line 1022


#################################
#
# can_udp_send(sender, receiver)
#
# Permissions for sending/receiving UDP datagrams.
#
#line 1033


#################################
#
# can_sysctl(domain)
#
# Permissions for modifying sysctl parameters.
#
#line 1057



##################################
#
# base_pty_perms(domain_prefix)
#
# Base permissions used for can_create_pty() and can_create_other_pty()
#
#line 1077


##################################
#
# can_create_pty(domain_prefix, attributes)
#
# Permissions for creating ptys.
#
#line 1097



##################################
#
# can_create_other_pty(domain_prefix,other_domain)
#
# Permissions for creating ptys for another domain.
#
#line 1113


##################################
#
# read_locale(domain)
#
# Permissions for reading the locale data,
# /etc/localtime and the files that it links to
#
#line 1125



################################################
#
# The following macros are an attempt to start
# partitioning every_domain into finer-grained subsets
# that can be used by individual domains.
#

#
# general_domain_access(domain)
#
# Grant permissions within the domain.
# This includes permissions to processes, /proc/PID files,
# file descriptors, pipes, Unix sockets, and System V IPC objects
# labeled with the domain.
#
#line 1169


#
# general_proc_read_access(domain)
#
# Grant read/search permissions to most of /proc, excluding
# the /proc/PID directories and the /proc/kmsg and /proc/kcore files.
# The general_domain_access macro grants access to the domain /proc/PID
# directories, but not to other domains.  Only permissions to stat
# are granted for /proc/kmsg and /proc/kcore, since these files are more
# sensitive.
# 
#line 1204


#
# base_file_read_access(domain)
#
# Grant read/search permissions to a few system file types.
#
#line 1228


#
# general_file_read_access(domain)
#
# Grant read/search permissions to many system file types.
#
#line 1367


#
# general_file_write_access(domain)
#
# Grant write permissions to a small set of system file types, e.g. 
# /dev/tty, /dev/null, etc.
#
# For shared directories like /tmp, each domain should have its own derived
# type (with a file_type_auto_trans rule) for files created in the shared
# directory.
#
#line 1388


#
# every_test_domain(domain)
#
# Grant permissions common to the test domains.
#
#line 1421


################################
#
# every_domain(domain)
#
# Grant permissions common to most domains.
#
# This macro replaces the rules formerly located in domains/every.te.
# An every_domain macro has been inserted into each domain .te file
# for each domain defined within that file.  If you want a new domain
# to inherit these rules, then you can likewise use this macro in
# your new domain .te file.  However, for least privilege purposes, you 
# may want to consider using macros or individual rules that only include 
# a subset of these permissions for your new domain.  This macro has already 
# been partitioned into a few subsets, with corresponding macros defined 
# above and used in defining this macro.  
#
#line 1448


#######################
# daemon_base_domain(domain_prefix, attribs)
#
# Define a daemon domain with a base set of type declarations
# and permissions that are common to most daemons.
# attribs is the list of attributes which must start with "," if it is not empty
#
# Author:  Russell Coker <russell@coker.com.au>
#
#line 1484

# allow a domain to create its own files under /var/run and to create files
# in directories that are created for it.  For sock_file etc you need to add
# your own rules
#line 1492


#line 1510

#line 1516


# define a sub-domain, $1_t is the parent domain, $2 is the name
# of the sub-domain.
#
#line 1543



#line 1549


#line 1556


#line 1561


#line 1566


#line 1571


#######################
# application_domain(domain_prefix)
#
# Define a domain with a base set of type declarations
# and permissions that are common to simple applications.
#
# Author:  Russell Coker <russell@coker.com.au>
#
#line 1587


#line 1593


#line 1601


#
# Macros for all admin domains.
#

#
# admin_domain(domain_prefix)
#
# Define derived types and rules for an administrator domain.
#
# The type declaration and role authorization for the domain must be
# provided separately.  Likewise, domain transitions into this domain
# must be specified separately.  If the every_domain() rules are desired,
# then these rules must also be specified separately.
#

#line 1632


#line 1645


#
# Macros for all user login domains.
#

#
# user_domain(domain_prefix)
#
# Define derived types and rules for an ordinary user domain.
#
# The type declaration and role authorization for the domain must be
# provided separately.  Likewise, domain transitions into this domain
# must be specified separately.  
#

#line 1920



###########################################################################
#
# Domains for ordinary users.
#

#line 2019



#line 2025


#
# Macros for chkpwd domains.
#

#
# chkpwd_domain(domain_prefix)
#
# Define a derived domain for the *_chkpwd program when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/su.te. 
#

#line 2076

#line 2076

#line 2076

#line 2076


# macro for chroot environments
# Author Russell Coker

# chroot(initial_domain, basename, role, tty_device_type)
#line 2206

#
# Macros for clamscan
#
# Author:  Brian May <bam@snoopy.apana.org.au>
#

#
# clamscan_domain(domain_prefix)
#
# Define a derived domain for the clamscan program when executed
#
#line 2241


#line 2254


#line 2286


#line 2296


#
# Macros for crond domains.
#

#
# Authors:  Jonathan Crowley (MITRE) <jonathan@mitre.org>,
#	    Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser
#

#
# crond_domain(domain_prefix)
#
# Define a derived domain for cron jobs executed by crond on behalf 
# of a user domain.  These domains are separate from the top-level domain
# defined for the crond daemon and the domain defined for system cron jobs,
# which are specified in domains/program/crond.te.
#

#line 2391


# When system_crond_t domain executes a type $1 executable then transition to
# domain $2, allow $2 to interact with crond_t as well.
#line 2401

#
# Macros for crontab domains.
#

#
# Authors:  Jonathan Crowley (MITRE) <jonathan@mitre.org>
# Revised by Stephen Smalley <sds@epoch.ncsc.mil>
#

#
# crontab_domain(domain_prefix)
#
# Define a derived domain for the crontab program when executed by
# a user domain.  
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/crontab.te. 
#

#line 2487

#
# Macro for fingerd
#
# Author:  Russell Coker <russell@coker.com.au>
#

#
# fingerd_macro(domain_prefix)
#
# allow fingerd to create a fingerlog file in the user home dir
#
#line 2502

#
# Macros for gpg and pgp
#
# Author:  Russell Coker <russell@coker.com.au>
#
# based on the work of:
# Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser
#

#
# gpg_domain(domain_prefix)
#
# Define a derived domain for the gpg/pgp program when executed by
# a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/gpg.te.
#
#line 2587

#
# Macros for gnome-pty-helper domains.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser 
#

#
# gph_domain(domain_prefix)
#
# Define a derived domain for the gnome-pty-helper program when
# executed by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/gnome-pty-helper.te. 
#
# The *_gph_t domains are for the gnome_pty_helper program.
# This program is executed by gnome-terminal to handle
# updates to utmp and wtmp.  In this regard, it is similar
# to utempter.  However, unlike utempter, gnome-pty-helper
# also creates the pty file for the terminal program.
# There is one *_gph_t domain for each user domain.  
#

#line 2659


#
# Macros for irc domains.
#

#
# Author:  Russell Coker <russell@coker.com.au>
#

#
# irc_domain(domain_prefix)
#
# Define a derived domain for the irc program when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/irc.te. 
#

#line 2747

#line 2747

#line 2747

#line 2747

#line 2747

#
# Macros for lpr domains.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser 
#

#
# lpr_domain(domain_prefix)
#
# Define a derived domain for the lpr/lpq/lprm programs when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/lpr.te. 
#

#line 2854

#
# Macros for mount
#
# Author:  Brian May <bam@snoopy.apana.org.au>
# Extended by Russell Coker <russell@coker.com.au>
#

#
# mount_domain(domain_prefix,dst_domain_prefix)
#
# Define a derived domain for the mount program for anyone.
#
#line 2903

#
# Macros for MTA domains.
#

#
# Author:   Russell Coker <russell@coker.com.au>
# Based on the work of: Stephen Smalley <sds@epoch.ncsc.mil>
#                       Timothy Fraser 
#

#
# mail_domain(domain_prefix)
#
# Define a derived domain for the sendmail program when executed by
# a user domain to send outgoing mail.  These domains are separate and
# independent of the domain used for the sendmail daemon process.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/mta.te. 
#

#line 3010

#
# Macros for netscape/mozilla (or other browser) domains.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser 
#

#
# netscape_domain(domain_prefix)
#
# Define a derived domain for the netscape/mozilla program when executed by
# a user domain.  
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/netscape.te. 
#
#line 3068

# $1 is the source domain (or domains), $2 is the source role (or roles) and $3
# is the base name for the domain to run.  $1 is normally sysadm_t, and $2 is
# normally sysadm_r.  $4 is the type of program to run and $5 is the domain to
# transition to.
# sample usage:
# run_program(sysadm_t, sysadm_r, init, etc_t, initrc_t)
#
# if you have several users who run the same run_init type program for
# different purposes (think of a run_db program used by several database
# administrators to start several databases) then you can list all the source
# domains in $1, all the source roles in $2, but you may not want to list all
# types of programs to run in $4 and target domains in $5 (as that may permit
# entering a domain from the wrong type).  In such a situation just specify
# one value for each of $4 and $5 and have some rules such as the following:
# domain_trans(run_whatever_t, whatever_exec_t, whatever_t)

#line 3123

#
# Macros for screen domains.
#

#
# Author: Russell Coker <russell@coker.com.au>
# Based on the work of Stephen Smalley <sds@epoch.ncsc.mil>
# and Timothy Fraser
#

#
# screen_domain(domain_prefix)
#
# Define a derived domain for the screen program when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/screen.te. 
#

#line 3217

#line 3217

#line 3217

#line 3217

#line 3217

#
# Macros for sendmail domains.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser 
#           Russell Coker <russell@coker.com.au>
#

#
# sendmail_user_domain(domain_prefix)
#
# Define a derived domain for the sendmail program when executed by
# a user domain to send outgoing mail.  These domains are separate and
# independent of the domain used for the sendmail daemon process.
#

#line 3254


#
# Macros for ssh domains.
#

#
# Author:  Stephen Smalley <sds@epoch.ncsc.mil>
#

# 
# ssh_domain(domain_prefix)
#
# Define a derived domain for the ssh program when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/ssh.te. 
#

#line 3424

#line 3424

#line 3424

#line 3424

#
# Macros for su domains.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser
#

#
# su_domain(domain_prefix)
#
# Define a derived domain for the su program when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/su.te. 
#


#line 3517

#line 3517

#line 3517

#line 3517

#
# Macros for xauth domains.
#

#
# Author:  Russell Coker <russell@coker.com.au>
#

#
# xauth_domain(domain_prefix)
#
# Define a derived domain for the xauth program when executed
# by a user domain.
#
# The type declaration for the executable type for this program is
# provided separately in domains/program/xauth.te. 
#

#line 3605

#line 3605

#line 3605

#line 3605

#line 3605

#
# Macros for X client programs ($2 etc)
#

#
# Author: Russell Coker <russell@coker.com.au>
# Based on the work of Stephen Smalley <sds@epoch.ncsc.mil>
# and Timothy Fraser 
#

#
# x_client_domain(domain_prefix)
#
# Define a derived domain for an X program when executed by
# a user domain.  
#
# The type declaration for the executable type for this program ($2_exec_t)
# must be provided separately!
#
# The first parameter is the base name for the domain/role (EG user or sysadm)
# The second parameter is the program name (EG $2)
# The third parameter is the attributes for the domain (if any)
#
#line 3768
#
# Macros for X server domains.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser
#

#################################
#
# xserver_domain(domain_prefix)
#
# Define a derived domain for the X server when executed
# by a user domain (e.g. via startx).  See the xdm_t domain
# in domains/program/xdm.te if using an X Display Manager.
#
# The type declarations for the executable type for this program 
# and the log type are provided separately in domains/program/xserver.te. 
#
# FIXME!  The X server requires far too many privileges.
#

#line 3944

#line 3944

#line 3944

#line 3944

#line 3944


#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

############################################
#
# Device types
#

#
# device_t is the type of /dev.
#
type device_t, file_type;

#
# null_device_t is the type of /dev/null.
#
type null_device_t, file_type, mlstrustedobject;

#
# zero_device_t is the type of /dev/zero.
#
type zero_device_t, file_type, mlstrustedobject;

#
# console_device_t is the type of /dev/console.
#
type console_device_t, file_type;

#
# memory_device_t is the type of /dev/kmem,
# /dev/mem, and /dev/port.
#
type memory_device_t, file_type;

#
# random_device_t is the type of /dev/urandom
# and /dev/random.
#
type random_device_t, file_type;

#
# devtty_t is the type of /dev/tty.
#
type devtty_t, file_type, mlstrustedobject;

#
# tty_device_t is the type of /dev/*tty*
#
type tty_device_t, file_type;

#
# fixed_disk_device_t is the type of 
# /dev/hd* and /dev/sd*.
#
type fixed_disk_device_t, file_type;

#
# removable_device_t is the type of
# /dev/scd* and /dev/fd*.
#
type removable_device_t, file_type;

#
# clock_device_t is the type of
# /dev/rtc.
#
type clock_device_t, file_type;

#
# tun_tap_device_t is the type of /dev/net/tun/* and /dev/net/tap/*
#
type tun_tap_device_t, file_type;

#
# misc_device_t is the type of miscellaneous devices.
# XXX:  FIXME!  Appropriate access to these devices need to be identified.
#
type misc_device_t, file_type;

#
# psaux_t is the type of the /dev/psaux mouse device.
# This type is deprecated and will be removed.
#
type psaux_t, file_type;

#
# A more general type for mouse devices.
#
type mouse_device_t, file_type;

#
# Not sure what these devices are for, but X wants access to them.
#
type agp_device_t, file_type;
type dri_device_t, file_type;

# Type for sound devices.
type sound_device_t, file_type;

# Type for /dev/ppp.
type ppp_device_t, file_type;

# Type for frame buffer /dev/fb/*
type framebuf_device_t, file_type;

# Type for /dev/.devfsd
type devfs_control_t, file_type;

# Type for /dev/cpu/mtrr
type mtrr_device_t, file_type;


# Type for /dev/apm_bios
type apm_bios_t, file_type;

# Type for v4l
type v4l_device_t, file_type;
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

############################################
#
# Devpts types
#

#
# ptmx_t is the type for /dev/ptmx.
#
type ptmx_t, file_type, sysadmfile;

#
# devpts_t is the type of the devpts file system and 
# the type of the root directory of the file system.
#
type devpts_t, fs_type, root_dir_type;


#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

#######################################
#
# General file-related types
#

#
# unlabeled_t is the type of unlabeled objects.
# Eventually, once all objects are labeled and 
# controlled by the security policy, unlabeled_t 
# will not be used.
#
type unlabeled_t, fs_type, file_type, root_dir_type, sysadmfile;

#
# fs_t is the default type of an unlabeled
# file system.  When an unlabeled file system is
# mounted read-write, the file system is labeled
# with this type.  Except for the root file system,
# other types may be specified using the fs_contexts
# or fs_contexts.mls file.
#
type fs_t, fs_type;

#
# file_t is the default type of an unlabeled
# file.  When an unlabeled file system is mounted
# read-write, all files in the file system are labeled
# with this type.   Except for the root file system,
# other types may be specified using the fs_contexts
# or fs_contexts.mls file.
#
type file_t, file_type, root_dir_type, sysadmfile;

#
# root_t is the type for the root directory.
#
type root_t, file_type, sysadmfile;

#
# home_root_t is the type for the directory where user home directories
# are created
#
type home_root_t, file_type, sysadmfile;

#
# lost_found_t is the type for the lost+found directories.
#
type lost_found_t, file_type, sysadmfile;

#
# boot_t is the type for files in /boot,
# including the kernel.
#
type boot_t, file_type, root_dir_type, sysadmfile;
# system_map_t is for the system.map files in /boot
type system_map_t, file_type, sysadmfile;

#
# boot_runtime_t is the type for /boot/kernel.h,
# which is automatically generated at boot time.
# only for red hat
type boot_runtime_t, file_type, sysadmfile;

#
# tmp_t is the type of /tmp and /var/tmp.
#
type tmp_t, file_type, sysadmfile, tmpfile;

#
# etc_t is the type of the system etc directories.
#
type etc_t, file_type, sysadmfile;
#
# shadow_t is the type of the /etc/shadow file
#
type shadow_t, file_type;
allow auth shadow_t:file { getattr read };

#
# ld_so_cache_t is the type of /etc/ld.so.cache.
#
type ld_so_cache_t, file_type, sysadmfile;

#
# etc_runtime_t is the type of various
# files in /etc that are automatically
# generated during initialization.
#
type etc_runtime_t, file_type, sysadmfile;

#
# etc_aliases_t is the type of the aliases database.
# etc_mail_t is the type of /etc/mail.
# sendmail needs write access to these files.
#
type etc_aliases_t, file_type, sysadmfile;
type etc_mail_t, file_type, sysadmfile;

# resolv_conf_t is the type of the /etc/resolv.conf file.
# pump needs write access to this file.
type resolv_conf_t, file_type, sysadmfile;

#
# lib_t is the type of files in the system lib directories.
#
type lib_t, file_type, sysadmfile;

#
# shlib_t is the type of shared objects in the system lib
# directories.
#
type shlib_t, file_type, sysadmfile;

#
# ld_so_t is the type of the system dynamic loaders.
#
type ld_so_t, file_type, sysadmfile;

#
# bin_t is the type of files in the system bin directories.
#
type bin_t, file_type, sysadmfile;

#
# ls_exec_t is the type of the ls program.
#
type ls_exec_t, file_type, exec_type, sysadmfile;

#
# shell_exec_t is the type of user shells such as /bin/bash.
#
type shell_exec_t, file_type, exec_type, sysadmfile;

#
# sbin_t is the type of files in the system sbin directories.
#
type sbin_t, file_type, sysadmfile;

#
# usr_t is the type for /usr.
#
type usr_t, file_type, root_dir_type, sysadmfile;

#
# src_t is the type of files in the system src directories.
#
type src_t, file_type, sysadmfile;

#
# var_t is the type for /var.
#
type var_t, file_type, root_dir_type, sysadmfile;

#
# Types for subdirectories of /var.
#
type var_run_t, file_type, sysadmfile;
type var_log_t, file_type, sysadmfile, logfile;
type faillog_t, file_type, sysadmfile, logfile;
type var_lock_t, file_type, sysadmfile, lockfile;
type var_lib_t, file_type, sysadmfile;
# for /var/{spool,lib}/texmf index files
type tetex_data_t, file_type, sysadmfile, tmpfile;
type var_spool_t, file_type, sysadmfile;
type var_yp_t, file_type, sysadmfile;

# Type for /var/log/sa.
type var_log_sa_t, file_type, sysadmfile, logfile;

# Type for /var/log/ksyms.
type var_log_ksyms_t, file_type, sysadmfile, logfile;

# Type for /var/log/lastlog.
type lastlog_t, file_type, sysadmfile, logfile;

# Type for /var/lib/nfs.
type var_lib_nfs_t, file_type, sysadmfile;

#
# wtmp_t is the type of /var/log/wtmp.
#
type wtmp_t, file_type, sysadmfile, logfile;

#
# catman_t is the type for /var/catman.
#
type catman_t, file_type, sysadmfile, tmpfile;

#
# at_spool_t is the type for /var/spool/at.
#
type at_spool_t, file_type, sysadmfile;

#
# cron_spool_t is the type for /var/spool/cron.
#
type cron_spool_t, file_type, sysadmfile;

#
# print_spool_t is the type for /var/spool/lpd and /var/spool/cups.
#
type print_spool_t, file_type, sysadmfile;

#
# mail_spool_t is the type for /var/spool/mail.
#
type mail_spool_t, file_type, sysadmfile;

#
# mqueue_spool_t is the type for /var/spool/mqueue.
#
type mqueue_spool_t, file_type, sysadmfile;

#
# man_t is the type for the man directories.
#
type man_t, file_type, sysadmfile;

#
# readable_t is a general type for
# files that are readable by all domains.
#
type readable_t, file_type, sysadmfile;

# 
# Base type for the tests directory.
# 
type test_file_t, file_type, sysadmfile;

#
# poly_t is the type for the polyinstantiated directories.
#
type poly_t, file_type, sysadmfile;

#
# swapfile_t is for swap files
#
type swapfile_t, file_type, sysadmfile;

#
# locale_t is the type for system localization
# 
type locale_t, file_type;

#
# Allow each file type to be associated with 
# the default file system type.
#
allow file_type fs_t:filesystem associate;

# Allow the pty to be associated with the file system.
allow devpts_t devpts_t:filesystem associate;

type tmpfs_t, file_type, sysadmfile, fs_type, root_dir_type;
allow { tmpfs_t tmp_t } tmpfs_t:filesystem associate;

type usbdevfs_t, fs_type, root_dir_type, sysadmfile;
allow usbdevfs_t usbdevfs_t:filesystem associate;
type usbdevfs_device_t, file_type, sysadmfile;
allow usbdevfs_device_t usbdevfs_t:filesystem associate;

type sysfs_t alias driverfs_t, fs_type, root_dir_type, sysadmfile;
allow sysfs_t sysfs_t:filesystem associate;

type iso9660_t, fs_type, root_dir_type, sysadmfile;
allow iso9660_t iso9660_t:filesystem associate;

type dosfs_t, fs_type, root_dir_type, sysadmfile;
allow dosfs_t dosfs_t:filesystem associate;
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

# Modified by Reino Wallin <reino@oribium.com>
# Multi NIC, and IPSEC features

# Modified by Russell Coker
# Move port types to their respective domains, add ifdefs, other cleanups.

############################################
#
# Network types
#

# 
# any_socket_t is the default destination
# socket type for UDP traffic.  Unless a 
# destination socket type is explicitly specified
# using sendto_secure/sendmsg_secure, this type
# is used for the udp_socket sendto permission check.
#
type any_socket_t, socket_type;
role system_r types any_socket_t;

#
# igmp_packet_t is the type of kernel-generated IGMP packets.
# icmp_socket_t is the type of the kernel socket used to send ICMP messages.
# tcp_socket_t is the type of the kernel socket used to send TCP resets.
# scmp_packet_t is the type for SCMP packets used by the SELOPT packet labeling.
#
type igmp_packet_t;
role system_r types igmp_packet_t;
type icmp_socket_t, socket_type;
role system_r types icmp_socket_t;
type tcp_socket_t, socket_type;
role system_r types tcp_socket_t;
type scmp_packet_t;
role system_r types scmp_packet_t;

#
# port_t is the default type of INET port numbers.
# The *_port_t types are used for specific port
# numbers in net_contexts or net_contexts.mls.
#
type port_t, port_type;

#
# netif_t is the default type of network interfaces.
# The netif_*_t types are used for specific network
# interfaces in net_contexts or net_contexts.mls.
#
type netif_t, netif_type;
type netif_eth0_t, netif_type;
type netif_eth1_t, netif_type;
type netif_eth2_t, netif_type;
type netif_lo_t, netif_type;
type netif_ippp0_t, netif_type;

type netif_ipsec0_t, netif_type;
type netif_ipsec1_t, netif_type;
type netif_ipsec2_t, netif_type;

#
# netmsg_t is the default type of unlabeled received messages.
# The netmsg_*_t types are used for specific network
# interfaces in net_contexts or net_contexts.mls.
#
type netmsg_t, netmsg_type;
type netmsg_eth0_t, netmsg_type;
type netmsg_eth1_t, netmsg_type;
type netmsg_eth2_t, netmsg_type;
type netmsg_lo_t, netmsg_type;
type netmsg_ippp0_t, netmsg_type;

type netmsg_ipsec0_t, netmsg_type;
type netmsg_ipsec1_t, netmsg_type;
type netmsg_ipsec2_t, netmsg_type;

#
# node_t is the default type of network nodes.
# The node_*_t types are used for specific network
# nodes in net_contexts or net_contexts.mls.
#
type node_t, node_type;
type node_lo_t, node_type;
type node_internal_t, node_type;

#
# Permissions for the kernel-generated IGMP packets.
#
allow igmp_packet_t netif_type:netif { rawip_send rawip_recv };
allow igmp_packet_t node_type:node { rawip_send rawip_recv };

#
# Permissions for the kernel ICMP socket.
#
allow icmp_socket_t netif_type:netif { rawip_send rawip_recv };
allow icmp_socket_t node_type:node { rawip_send rawip_recv };
allow icmp_socket_t netmsg_type:rawip_socket recvfrom;

#
# Permissions for the kernel TCP reset socket.
#
allow tcp_socket_t netif_type:netif { tcp_send tcp_recv };
allow tcp_socket_t netmsg_type:tcp_socket recvfrom;
allow tcp_socket_t node_type:node { tcp_send tcp_recv };

#
# Allow network messages to be received.
#
allow netmsg_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_eth0_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_eth0_t netif_eth0_t:netif { tcp_recv udp_recv rawip_recv };
allow netmsg_eth1_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_eth1_t netif_eth1_t:netif { tcp_recv udp_recv rawip_recv };
allow netmsg_eth2_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_eth2_t netif_eth2_t:netif { tcp_recv udp_recv rawip_recv };
allow netmsg_lo_t node_lo_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_lo_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_lo_t netif_lo_t:netif { tcp_recv udp_recv rawip_recv };
allow netmsg_ippp0_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_ippp0_t netif_ippp0_t:netif { tcp_recv udp_recv rawip_recv };
allow netmsg_ipsec0_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_ipsec0_t netif_ipsec0_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_ipsec1_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_ipsec1_t netif_ipsec1_t:netif { tcp_recv udp_recv rawip_recv };
allow netmsg_ipsec2_t node_t:node { tcp_recv udp_recv rawip_recv };
allow netmsg_ipsec2_t netif_ipsec2_t:netif { tcp_recv udp_recv rawip_recv };

#
# Allow ICMP echo requests to be sent and received, and echo replies to 
# be received (when packets are labeled)
#
#line 4495

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

#############################################
#
# NFS types
#

#
# nfs_t is the default type for NFS file systems 
# and their files.  
# The nfs_*_t types are used for specific NFS
# servers in net_contexts or net_contexts.mls.
#
type nfs_t, fs_type, root_dir_type;

#
# Allow NFS files to be associated with an NFS file system.
#
allow nfs_t nfs_t:filesystem associate;
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

############################################
#
# Procfs types
#

#
# proc_t is the type of /proc.
# proc_kmsg_t is the type of /proc/kmsg.
# proc_kcore_t is the type of /proc/kcore.
#
type proc_t, fs_type, root_dir_type;
type proc_kmsg_t;
type proc_kcore_t;

#
# sysctl_t is the type of /proc/sys.
# sysctl_fs_t is the type of /proc/sys/fs.
# sysctl_kernel_t is the type of /proc/sys/kernel.
# sysctl_modprobe_t is the type of /proc/sys/kernel/modprobe.
# sysctl_net_t is the type of /proc/sys/net.
# sysctl_net_unix_t is the type of /proc/sys/net/unix.
# sysctl_vm_t is the type of /proc/sys/vm.
# sysctl_dev_t is the type of /proc/sys/dev.
#
# These types are applied to both the entries in
# /proc/sys and the corresponding sysctl parameters.
#
type sysctl_t;
type sysctl_fs_t;
type sysctl_kernel_t;
type sysctl_modprobe_t;
type sysctl_net_t;
type sysctl_net_unix_t;
type sysctl_vm_t;
type sysctl_dev_t;


#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

############################################
#
# Security types
#

# 
# security_t is the target type when checking
# the permissions in the security class.  It is also
# applied to selinuxfs inodes.
#
type security_t, fs_type;

#
# policy_config_t is the type of /ss_policy,
# the security server policy configuration.
#
type policy_config_t, file_type;

#
# policy_src_t is the type of the policy source
# files.
#
type policy_src_t, file_type;

#
# default_context_t is the type applied to 
# /etc/security/default_context
#
type default_context_t, file_type, sysadmfile, login_contexts;

#
# file_labels_t is the type of the persistent
# label mapping stored in each file system.
# The mapping files are in the ...security
# subdirectory at the root of each file system.
#
type file_labels_t, file_type, sysadmfile;

#
# no_access_t is the type for objects that should
# only be accessed administratively. 
#
type no_access_t, file_type, sysadmfile;


#DESC Admin - Domains for administrators.
#
#################################

# sysadm_t is the system administrator domain.
type sysadm_t, domain, privlog, privowner, admin, userdomain, web_client_domain, privhome;
allow privhome home_root_t:dir search;

# system_r is authorized for sysadm_t for single-user mode.
role system_r types sysadm_t; 

# sysadm_r is authorized for sysadm_t for the initial login domain.
role sysadm_r types sysadm_t;


#line 4621
# Read system information files in /proc.
#line 4621
allow sysadm_t proc_t:dir { read getattr lock search ioctl };
#line 4621
allow sysadm_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4621

#line 4621
# Stat /proc/kmsg and /proc/kcore.
#line 4621
allow sysadm_t proc_kmsg_t:file { getattr };
#line 4621
allow sysadm_t proc_kcore_t:file { getattr };
#line 4621

#line 4621
# Read system variables in /proc/sys.
#line 4621
allow sysadm_t sysctl_modprobe_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_t:dir { read getattr lock search ioctl };
#line 4621
allow sysadm_t sysctl_fs_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_fs_t:dir { read getattr lock search ioctl };
#line 4621
allow sysadm_t sysctl_kernel_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_kernel_t:dir { read getattr lock search ioctl };
#line 4621
allow sysadm_t sysctl_net_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_net_t:dir { read getattr lock search ioctl };
#line 4621
allow sysadm_t sysctl_vm_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_vm_t:dir { read getattr lock search ioctl };
#line 4621
allow sysadm_t sysctl_dev_t:file { read getattr lock ioctl };
#line 4621
allow sysadm_t sysctl_dev_t:dir { read getattr lock search ioctl };
#line 4621


#line 4622
# Access other processes in the same domain.
#line 4622
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4622
# separately if desired.
#line 4622
allow sysadm_t self:process ~{ptrace setexec setfscreate};
#line 4622

#line 4622
# Access /proc/PID files for processes in the same domain.
#line 4622
allow sysadm_t self:dir { read getattr lock search ioctl };
#line 4622
allow sysadm_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4622

#line 4622
# Access file descriptions, pipes, and sockets
#line 4622
# created by processes in the same domain.
#line 4622
allow sysadm_t self:fd *;
#line 4622
allow sysadm_t self:fifo_file { ioctl read getattr lock write append };
#line 4622
allow sysadm_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4622
allow sysadm_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4622

#line 4622
# Allow the domain to communicate with other processes in the same domain.
#line 4622
allow sysadm_t self:unix_dgram_socket sendto;
#line 4622
allow sysadm_t self:unix_stream_socket connectto;
#line 4622

#line 4622
# Access System V IPC objects created by processes in the same domain.
#line 4622
allow sysadm_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4622
allow sysadm_t self:msg  { send receive };
#line 4622
allow sysadm_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4622
allow sysadm_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4622


# sysadm_t is also granted permissions specific to administrator domains.

#line 4625
# Inherit rules for ordinary users.
#line 4625

#line 4625
# Use capabilities
#line 4625
allow sysadm_t self:capability { setgid chown fowner };
#line 4625
dontaudit sysadm_t self:capability { sys_nice fsetid };
#line 4625

#line 4625
# Type for home directory.
#line 4625

#line 4625
type sysadm_home_dir_t, file_type, sysadmfile, home_dir_type, home_type;
#line 4625
type sysadm_home_t, file_type, sysadmfile, home_type;
#line 4625

#line 4625
type sysadm_tmp_t, file_type, sysadmfile, tmpfile ;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_t sysadm_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_t sysadm_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_t tmp_t:dir sysadm_tmp_t;
#line 4625
type_transition sysadm_t tmp_t:{ file lnk_file sock_file fifo_file } sysadm_tmp_t;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# allow ptrace
#line 4625
#can_ptrace(sysadm_t, sysadm_t)
#line 4625

#line 4625
# Create, access, and remove files in home directory.
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_t sysadm_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_t sysadm_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_t sysadm_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_t sysadm_home_dir_t:dir sysadm_home_t;
#line 4625
type_transition sysadm_t sysadm_home_dir_t:{ file lnk_file sock_file fifo_file } sysadm_home_t;
#line 4625

#line 4625

#line 4625

#line 4625
allow sysadm_t sysadm_home_t:{ dir file lnk_file sock_file fifo_file chr_file blk_file } { relabelfrom relabelto };
#line 4625

#line 4625
# Bind to a Unix domain socket in /tmp.
#line 4625
allow sysadm_t sysadm_tmp_t:unix_stream_socket name_bind;
#line 4625

#line 4625
# Type for tty devices.
#line 4625
type sysadm_tty_device_t, file_type, sysadmfile, ttyfile;
#line 4625
# Access ttys.
#line 4625
allow sysadm_t privfd:fd use;
#line 4625
allow sysadm_t sysadm_tty_device_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 4625
# Use the type when relabeling terminal devices.
#line 4625
type_change sysadm_t tty_device_t:chr_file sysadm_tty_device_t;
#line 4625

#line 4625
# Debian login is from shadow utils and does not allow resetting the perms.
#line 4625
# have to fix this!
#line 4625
type_change sysadm_t ttyfile:chr_file sysadm_tty_device_t;
#line 4625

#line 4625
# allow reading /var/lib/dpkg etc
#line 4625
allow sysadm_t var_lib_t:dir { read getattr lock search ioctl };
#line 4625

#line 4625
# Allow reading dpkg origins file
#line 4625

#line 4625
allow sysadm_t etc_dpkg_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_t etc_dpkg_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625

#line 4625
# read localization information
#line 4625
allow sysadm_t locale_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_t locale_t:{file lnk_file} { read getattr lock ioctl };
#line 4625

#line 4625
# for running TeX programs
#line 4625

#line 4625
allow sysadm_t tetex_data_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_t tetex_data_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625
allow sysadm_t tetex_data_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
# Type and access for pty devices.
#line 4625

#line 4625

#line 4625

#line 4625
# Access the pty master multiplexer.
#line 4625
allow sysadm_t ptmx_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625
allow sysadm_t devpts_t:filesystem getattr;
#line 4625

#line 4625
# allow searching /dev/pts
#line 4625
allow sysadm_t devpts_t:dir { getattr read search };
#line 4625

#line 4625
type sysadm_devpts_t, file_type, sysadmfile, ptyfile ;
#line 4625

#line 4625
# Allow the pty to be associated with the file system.
#line 4625
allow sysadm_devpts_t devpts_t:filesystem associate;
#line 4625

#line 4625
# Label pty files with a derived type.
#line 4625
type_transition sysadm_t devpts_t:chr_file sysadm_devpts_t;
#line 4625

#line 4625
# Read and write my pty files.
#line 4625
allow sysadm_t sysadm_devpts_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 4625

#line 4625

#line 4625

#line 4625
# Use the type when relabeling pty devices.
#line 4625

#line 4625

#line 4625
type_change sysadm_t sshd_devpts_t:chr_file sysadm_devpts_t;
#line 4625

#line 4625
# Connect to sshd.
#line 4625

#line 4625
allow sysadm_t sshd_t:tcp_socket { connectto recvfrom };
#line 4625
allow sshd_t sysadm_t:tcp_socket { acceptfrom recvfrom };
#line 4625
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625
allow sysadm_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625

#line 4625

#line 4625
# Connect to ssh proxy.
#line 4625

#line 4625
allow sysadm_t sysadm_ssh_t:tcp_socket { connectto recvfrom };
#line 4625
allow sysadm_ssh_t sysadm_t:tcp_socket { acceptfrom recvfrom };
#line 4625
allow sysadm_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625
allow sysadm_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625

#line 4625

#line 4625
allow sysadm_t sshd_t:fd use;
#line 4625
allow sysadm_t sshd_t:tcp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625
# Use a Unix stream socket inherited from sshd.
#line 4625
allow sysadm_t sshd_t:unix_stream_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625

#line 4625
allow sysadm_t boot_t:dir { getattr search };
#line 4625
dontaudit sysadm_t boot_t:dir read;
#line 4625
dontaudit sysadm_t boot_t:lnk_file getattr;
#line 4625
dontaudit sysadm_t boot_t:file { getattr read };
#line 4625
allow sysadm_t system_map_t:file { getattr read };
#line 4625

#line 4625

#line 4625
type sysadm_tmpfs_t, file_type, sysadmfile, tmpfsfile;
#line 4625
# Use this type when creating tmpfs/shm objects.
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_t tmpfs_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_t sysadm_tmpfs_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_t sysadm_tmpfs_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_t tmpfs_t:dir sysadm_tmpfs_t;
#line 4625
type_transition sysadm_t tmpfs_t:{ file lnk_file sock_file fifo_file } sysadm_tmpfs_t;
#line 4625

#line 4625

#line 4625

#line 4625
allow sysadm_tmpfs_t tmpfs_t:filesystem associate;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Read and write /var/catman.
#line 4625
allow sysadm_t catman_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625
allow sysadm_t catman_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625

#line 4625
# Modify mail spool file.
#line 4625
allow sysadm_t mail_spool_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_t mail_spool_t:file { ioctl read getattr lock write append };
#line 4625
allow sysadm_t mail_spool_t:lnk_file read;
#line 4625

#line 4625
#
#line 4625
# Allow the query of filesystem quotas
#line 4625
#
#line 4625
allow sysadm_t fs_type:filesystem quotaget;
#line 4625

#line 4625
# Run helper programs.
#line 4625

#line 4625
allow sysadm_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4625

#line 4625
allow sysadm_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_t device_t:dir search;
#line 4625
allow sysadm_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625
allow sysadm_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
# Run programs developed by other users in the same domain.
#line 4625

#line 4625
allow sysadm_t sysadm_home_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_t sysadm_tmp_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
# Run user programs that require different permissions in their own domain.
#line 4625
# These rules were moved into the individual program domains.
#line 4625

#line 4625
# Instantiate derived domains for a number of programs.
#line 4625
# These derived domains encode both information about the calling
#line 4625
# user domain and the program, and allow us to maintain separation
#line 4625
# between different instances of the program being run by different
#line 4625
# user domains.
#line 4625

#line 4625

#line 4625

#line 4625
# Derived domain based on the calling user domain and the program.
#line 4625
type sysadm_su_t, domain, privlog, auth;
#line 4625

#line 4625
# Transition from the user domain to this domain.
#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow sysadm_t sysadm_su_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit sysadm_t sysadm_su_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow sysadm_t su_exec_t:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_su_t sysadm_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_su_t sysadm_t:fd use;
#line 4625
allow sysadm_t sysadm_su_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_su_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_su_t su_exec_t:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_su_t su_exec_t:file entrypoint;
#line 4625

#line 4625
type_transition sysadm_t su_exec_t:process sysadm_su_t;
#line 4625

#line 4625

#line 4625

#line 4625
allow sysadm_su_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_su_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_su_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_su_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_su_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_su_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_su_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_su_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_su_t device_t:dir search;
#line 4625
allow sysadm_su_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
allow sysadm_su_t etc_t:file { getattr read };
#line 4625

#line 4625
allow sysadm_su_t etc_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_su_t locale_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_su_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625
allow sysadm_su_t self:unix_dgram_socket { connect create write };
#line 4625
allow sysadm_su_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625
allow sysadm_su_t proc_t:dir search;
#line 4625
allow sysadm_su_t proc_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_su_t self:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_su_t self:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625
allow sysadm_su_t self:process setsched;
#line 4625
allow sysadm_su_t device_t:dir search;
#line 4625
allow sysadm_su_t self:process { fork sigchld };
#line 4625

#line 4625
# Use capabilities.
#line 4625
allow sysadm_su_t self:capability { setuid setgid net_bind_service chown dac_override sys_nice sys_resource };
#line 4625

#line 4625
# Revert to the user domain when a shell is executed.
#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow sysadm_su_t sysadm_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit sysadm_su_t sysadm_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow sysadm_su_t shell_exec_t:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_t sysadm_su_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_t sysadm_su_t:fd use;
#line 4625
allow sysadm_su_t sysadm_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_t sysadm_su_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_t shell_exec_t:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_t shell_exec_t:file entrypoint;
#line 4625

#line 4625
type_transition sysadm_su_t shell_exec_t:process sysadm_t;
#line 4625

#line 4625
allow sysadm_su_t bin_t:dir search;
#line 4625
allow sysadm_su_t bin_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_su_t privfd:fd use;
#line 4625

#line 4625
# Write to utmp.
#line 4625
allow sysadm_su_t { var_t var_run_t }:dir search;
#line 4625
allow sysadm_su_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625

#line 4625
# Run chkpwd.
#line 4625

#line 4625
allow sysadm_su_t chkpwd_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625

#line 4625
# Inherit and use descriptors from gnome-pty-helper.
#line 4625

#line 4625

#line 4625
# The user role is authorized for this domain.
#line 4625
role sysadm_r types sysadm_su_t;
#line 4625

#line 4625
# Write to the user domain tty.
#line 4625
allow sysadm_su_t sysadm_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_su_t sysadm_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
allow sysadm_su_t { home_root_t sysadm_home_dir_t }:dir search;
#line 4625

#line 4625
# Modify .Xauthority file (via xauth program).
#line 4625

#line 4625

#line 4625
# Access sshd cookie files.
#line 4625
allow sysadm_su_t sshd_tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625
allow sysadm_su_t sshd_tmp_t:file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_su_t sshd_tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_su_t sysadm_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_su_t sysadm_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_su_t sshd_tmp_t:dir sysadm_tmp_t;
#line 4625
type_transition sysadm_su_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } sysadm_tmp_t;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# stop su complaining if you run it from a directory with restrictive perms
#line 4625
dontaudit sysadm_su_t file_type:dir search;
#line 4625

#line 4625

#line 4625
# Derived domain based on the calling user domain and the program.
#line 4625
type sysadm_chkpwd_t, domain, privlog, auth;
#line 4625

#line 4625
# Transition from the user domain to this domain.
#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow sysadm_t sysadm_chkpwd_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit sysadm_t sysadm_chkpwd_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow sysadm_t chkpwd_exec_t:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_chkpwd_t sysadm_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_chkpwd_t sysadm_t:fd use;
#line 4625
allow sysadm_t sysadm_chkpwd_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_chkpwd_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_chkpwd_t chkpwd_exec_t:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_chkpwd_t chkpwd_exec_t:file entrypoint;
#line 4625

#line 4625
type_transition sysadm_t chkpwd_exec_t:process sysadm_chkpwd_t;
#line 4625

#line 4625

#line 4625
# The user role is authorized for this domain.
#line 4625
role sysadm_r types sysadm_chkpwd_t;
#line 4625

#line 4625

#line 4625
allow sysadm_chkpwd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_chkpwd_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_chkpwd_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_chkpwd_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_chkpwd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_chkpwd_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_chkpwd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_chkpwd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_chkpwd_t device_t:dir search;
#line 4625
allow sysadm_chkpwd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
allow sysadm_chkpwd_t etc_t:file { getattr read };
#line 4625
allow sysadm_chkpwd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_chkpwd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625

#line 4625
allow sysadm_chkpwd_t etc_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_chkpwd_t locale_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_chkpwd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625

#line 4625
# Use capabilities.
#line 4625
allow sysadm_chkpwd_t self:capability setuid;
#line 4625

#line 4625
# Inherit and use descriptors from gnome-pty-helper.
#line 4625

#line 4625

#line 4625
# Inherit and use descriptors from newrole.
#line 4625
allow sysadm_chkpwd_t newrole_t:fd use;
#line 4625

#line 4625
# Write to the user domain tty.
#line 4625
allow sysadm_chkpwd_t sysadm_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_chkpwd_t sysadm_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Derived domain based on the calling user domain and the program.
#line 4625
type sysadm_crontab_t, domain, privlog;
#line 4625

#line 4625
# Transition from the user domain to the derived domain.
#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow sysadm_t sysadm_crontab_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit sysadm_t sysadm_crontab_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow sysadm_t crontab_exec_t:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_crontab_t sysadm_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_crontab_t sysadm_t:fd use;
#line 4625
allow sysadm_t sysadm_crontab_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_crontab_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_crontab_t crontab_exec_t:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_crontab_t crontab_exec_t:file entrypoint;
#line 4625

#line 4625
type_transition sysadm_t crontab_exec_t:process sysadm_crontab_t;
#line 4625

#line 4625

#line 4625
# The user role is authorized for this domain.
#line 4625
role sysadm_r types sysadm_crontab_t;
#line 4625

#line 4625

#line 4625
allow sysadm_crontab_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crontab_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crontab_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_crontab_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_crontab_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crontab_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_crontab_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crontab_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_crontab_t device_t:dir search;
#line 4625
allow sysadm_crontab_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
allow sysadm_crontab_t etc_t:file { getattr read };
#line 4625
allow sysadm_crontab_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_crontab_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625

#line 4625
allow sysadm_crontab_t etc_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_crontab_t locale_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crontab_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625

#line 4625
# Use capabilities
#line 4625
allow sysadm_crontab_t sysadm_crontab_t:capability { setuid setgid chown };
#line 4625

#line 4625
# Type for temporary files.
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_crontab_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_crontab_t sysadm_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_crontab_t sysadm_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_crontab_t tmp_t:dir sysadm_tmp_t;
#line 4625
type_transition sysadm_crontab_t tmp_t:{ file lnk_file sock_file fifo_file } sysadm_tmp_t;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Type of user crontabs once moved to cron spool.
#line 4625
type sysadm_cron_spool_t, file_type, sysadmfile;
#line 4625
# Use the type when creating files in /var/spool/cron.
#line 4625
allow sysadm_crontab_t sysadm_cron_spool_t:file { getattr read };
#line 4625
allow sysadm_crontab_t { var_t var_spool_t }:dir search;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_crontab_t cron_spool_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_crontab_t sysadm_cron_spool_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_crontab_t sysadm_cron_spool_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_crontab_t cron_spool_t:dir sysadm_cron_spool_t;
#line 4625
type_transition sysadm_crontab_t cron_spool_t:{ file lnk_file sock_file fifo_file } sysadm_cron_spool_t;
#line 4625

#line 4625

#line 4625

#line 4625
allow sysadm_crontab_t self:process { fork { sigchld sigkill sigstop signull signal } };
#line 4625

#line 4625
# crontab signals crond by updating the mtime on the spooldir
#line 4625
allow sysadm_crontab_t cron_spool_t:dir setattr;
#line 4625
# Allow crond to read those crontabs in cron spool.
#line 4625
allow crond_t sysadm_cron_spool_t:file { read getattr lock ioctl };
#line 4625

#line 4625
# Run helper programs as sysadm_t
#line 4625
allow sysadm_crontab_t { bin_t sbin_t }:dir search;
#line 4625
allow sysadm_crontab_t bin_t:lnk_file read;
#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow sysadm_crontab_t sysadm_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit sysadm_crontab_t sysadm_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow sysadm_crontab_t { bin_t sbin_t shell_exec_t }:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_t sysadm_crontab_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_t sysadm_crontab_t:fd use;
#line 4625
allow sysadm_crontab_t sysadm_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_t sysadm_crontab_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_t { bin_t sbin_t shell_exec_t }:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_t { bin_t sbin_t shell_exec_t }:file entrypoint;
#line 4625

#line 4625
type_transition sysadm_crontab_t { bin_t sbin_t shell_exec_t }:process sysadm_t;
#line 4625

#line 4625

#line 4625
# Read user crontabs 
#line 4625
allow sysadm_crontab_t { sysadm_home_t sysadm_home_dir_t }:dir { read getattr lock search ioctl };  
#line 4625
allow sysadm_crontab_t sysadm_home_t:file { read getattr lock ioctl };  
#line 4625
dontaudit sysadm_crontab_t sysadm_home_dir_t:dir write;
#line 4625

#line 4625
# Access the cron log file.
#line 4625
allow sysadm_crontab_t cron_log_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_crontab_t cron_log_t:file { append };
#line 4625

#line 4625
# Access terminals.
#line 4625
allow sysadm_crontab_t device_t:dir search;
#line 4625
allow sysadm_crontab_t sysadm_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_crontab_t sysadm_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
# Inherit and use descriptors from gnome-pty-helper.
#line 4625

#line 4625
allow sysadm_crontab_t privfd:fd use;
#line 4625

#line 4625

#line 4625
# Derived domain based on the calling user domain and the program.
#line 4625
type sysadm_ssh_t, domain, privlog;
#line 4625

#line 4625
# Transition from the user domain to the derived domain.
#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow sysadm_t sysadm_ssh_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit sysadm_t sysadm_ssh_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow sysadm_t ssh_exec_t:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_ssh_t sysadm_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_ssh_t sysadm_t:fd use;
#line 4625
allow sysadm_t sysadm_ssh_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_ssh_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_ssh_t ssh_exec_t:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_ssh_t ssh_exec_t:file entrypoint;
#line 4625

#line 4625
type_transition sysadm_t ssh_exec_t:process sysadm_ssh_t;
#line 4625

#line 4625

#line 4625
# The user role is authorized for this domain.
#line 4625
role sysadm_r types sysadm_ssh_t;
#line 4625

#line 4625
# Grant permissions within the domain.
#line 4625

#line 4625
# Access other processes in the same domain.
#line 4625
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4625
# separately if desired.
#line 4625
allow sysadm_ssh_t self:process ~{ptrace setexec setfscreate};
#line 4625

#line 4625
# Access /proc/PID files for processes in the same domain.
#line 4625
allow sysadm_ssh_t self:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Access file descriptions, pipes, and sockets
#line 4625
# created by processes in the same domain.
#line 4625
allow sysadm_ssh_t self:fd *;
#line 4625
allow sysadm_ssh_t self:fifo_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_ssh_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_ssh_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625

#line 4625
# Allow the domain to communicate with other processes in the same domain.
#line 4625
allow sysadm_ssh_t self:unix_dgram_socket sendto;
#line 4625
allow sysadm_ssh_t self:unix_stream_socket connectto;
#line 4625

#line 4625
# Access System V IPC objects created by processes in the same domain.
#line 4625
allow sysadm_ssh_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4625
allow sysadm_ssh_t self:msg  { send receive };
#line 4625
allow sysadm_ssh_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4625
allow sysadm_ssh_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4625

#line 4625

#line 4625
# Use descriptors created by sshd
#line 4625
allow sysadm_ssh_t privfd:fd use;
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_ssh_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_ssh_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_ssh_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t device_t:dir search;
#line 4625
allow sysadm_ssh_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t etc_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_ssh_t locale_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625

#line 4625
# Get attributes of file systems.
#line 4625
allow sysadm_ssh_t fs_type:filesystem getattr;
#line 4625

#line 4625

#line 4625
# Read /.
#line 4625
allow sysadm_ssh_t root_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Read /home.
#line 4625
allow sysadm_ssh_t home_root_t:dir { read getattr lock search ioctl };
#line 4625

#line 4625
# Read /usr.
#line 4625
allow sysadm_ssh_t usr_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Read bin and sbin directories.
#line 4625
allow sysadm_ssh_t bin_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t sbin_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625
# Read the devpts root directory.
#line 4625
allow sysadm_ssh_t devpts_t:dir { read getattr lock search ioctl };
#line 4625

#line 4625
# Read /var.
#line 4625
allow sysadm_ssh_t var_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t var_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Read /var/run, /var/log.
#line 4625
allow sysadm_ssh_t var_run_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t var_run_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t var_log_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t var_log_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Read /etc.
#line 4625
allow sysadm_ssh_t etc_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t resolv_conf_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Read the linker, shared library, and executable types.
#line 4625
allow sysadm_ssh_t ld_so_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t shlib_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Read /dev directories and any symbolic links.
#line 4625
allow sysadm_ssh_t device_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t device_t:lnk_file { read getattr lock ioctl };
#line 4625

#line 4625
# Read /dev/random.
#line 4625
allow sysadm_ssh_t random_device_t:chr_file { read getattr lock ioctl };
#line 4625

#line 4625
# Read and write /dev/tty and /dev/null.
#line 4625
allow sysadm_ssh_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_ssh_t { null_device_t zero_device_t }:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
# Grant permissions needed to create TCP and UDP sockets and
#line 4625
# to access the network.
#line 4625

#line 4625
#
#line 4625
# Allow the domain to create and use UDP and TCP sockets.
#line 4625
# Other kinds of sockets must be separately authorized for use.
#line 4625
allow sysadm_ssh_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_ssh_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send UDP packets.
#line 4625
# Since the destination sockets type is unknown, the generic
#line 4625
# any_socket_t type is used as a placeholder.
#line 4625
#
#line 4625
allow sysadm_ssh_t any_socket_t:udp_socket sendto;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send using any network interface.
#line 4625
# netif_type is a type attribute for all network interface types.
#line 4625
#
#line 4625
allow sysadm_ssh_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4625

#line 4625
#
#line 4625
# Allow packets sent by the domain to be received on any network interface.
#line 4625
#
#line 4625
allow sysadm_ssh_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to receive packets from any network interface.
#line 4625
# netmsg_type is a type attribute for all default message types.
#line 4625
#
#line 4625
allow sysadm_ssh_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to initiate or accept TCP connections 
#line 4625
# on any network interface.
#line 4625
#
#line 4625
allow sysadm_ssh_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4625

#line 4625
#
#line 4625
# Receive resets from the TCP reset socket.
#line 4625
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4625
#
#line 4625
allow sysadm_ssh_t tcp_socket_t:tcp_socket recvfrom;
#line 4625

#line 4625
dontaudit sysadm_ssh_t tcp_socket_t:tcp_socket connectto;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send to any node.
#line 4625
# node_type is a type attribute for all node types.
#line 4625
#
#line 4625
allow sysadm_ssh_t node_type:node { tcp_send udp_send rawip_send };
#line 4625

#line 4625
#
#line 4625
# Allow packets sent by the domain to be received from any node.
#line 4625
#
#line 4625
allow sysadm_ssh_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send NFS client requests via the socket
#line 4625
# created by mount.
#line 4625
#
#line 4625
allow sysadm_ssh_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625

#line 4625
#
#line 4625
# Bind to the default port type.
#line 4625
# Other port types must be separately authorized.
#line 4625
#
#line 4625
allow sysadm_ssh_t port_t:udp_socket name_bind;
#line 4625
allow sysadm_ssh_t port_t:tcp_socket name_bind;
#line 4625

#line 4625

#line 4625
# for sshing to a ssh tunnel
#line 4625

#line 4625
allow sysadm_ssh_t sysadm_ssh_t:tcp_socket { connectto recvfrom };
#line 4625
allow sysadm_ssh_t sysadm_ssh_t:tcp_socket { acceptfrom recvfrom };
#line 4625
allow sysadm_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625
allow sysadm_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625

#line 4625

#line 4625
# Use capabilities.
#line 4625
allow sysadm_ssh_t self:capability { setuid setgid dac_override dac_read_search };
#line 4625

#line 4625
# Run helpers.
#line 4625

#line 4625
allow sysadm_ssh_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4625

#line 4625
allow sysadm_ssh_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_ssh_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_ssh_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_ssh_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_ssh_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_ssh_t device_t:dir search;
#line 4625
allow sysadm_ssh_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625

#line 4625
# Read the ssh key file.
#line 4625
allow sysadm_ssh_t sshd_key_t:file { read getattr lock ioctl };
#line 4625

#line 4625
# Access the ssh temporary files.
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_ssh_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_ssh_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_ssh_t sshd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_ssh_t tmp_t:dir sshd_tmp_t;
#line 4625
type_transition sysadm_ssh_t tmp_t:{ file lnk_file sock_file fifo_file } sshd_tmp_t;
#line 4625

#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t sysadm_tmp_t:dir { read getattr lock search ioctl };
#line 4625

#line 4625
# for rsync
#line 4625
allow sysadm_ssh_t sysadm_t:unix_stream_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625

#line 4625
# Access the users .ssh directory.
#line 4625
type sysadm_home_ssh_t, file_type, sysadmfile;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow { sysadm_ssh_t sysadm_ssh_t } sysadm_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow { sysadm_ssh_t sysadm_ssh_t } sysadm_home_ssh_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow { sysadm_ssh_t sysadm_ssh_t } sysadm_home_ssh_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition { sysadm_ssh_t sysadm_ssh_t } sysadm_home_dir_t:dir sysadm_home_ssh_t;
#line 4625
type_transition { sysadm_ssh_t sysadm_ssh_t } sysadm_home_dir_t:{ file lnk_file sock_file fifo_file } sysadm_home_ssh_t;
#line 4625

#line 4625

#line 4625

#line 4625
allow { sysadm_ssh_t sysadm_ssh_t } sysadm_home_ssh_t:lnk_file { getattr read };
#line 4625
dontaudit sysadm_ssh_t sysadm_home_t:dir search;
#line 4625

#line 4625
allow sshd_t sysadm_home_ssh_t:dir { read getattr lock search ioctl };
#line 4625
allow sshd_t sysadm_home_ssh_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625
allow sysadm_t sysadm_home_ssh_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625
allow sysadm_t sysadm_home_ssh_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625

#line 4625

#line 4625
# for /bin/sh used to execute xauth
#line 4625
dontaudit sysadm_ssh_t proc_t:dir search;
#line 4625

#line 4625
# Inherit and use descriptors from gnome-pty-helper.
#line 4625

#line 4625

#line 4625
# Connect to sshd.
#line 4625

#line 4625

#line 4625
allow sysadm_ssh_t sshd_t:tcp_socket { connectto recvfrom };
#line 4625
allow sshd_t sysadm_ssh_t:tcp_socket { acceptfrom recvfrom };
#line 4625
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625
allow sysadm_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625

#line 4625

#line 4625

#line 4625
# Write to the user domain tty.
#line 4625
allow sysadm_ssh_t sysadm_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_ssh_t sysadm_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625
# Allow the user shell to signal the ssh program.
#line 4625
allow sysadm_t sysadm_ssh_t:process signal;
#line 4625
# allow ps to show ssh
#line 4625
allow sysadm_t sysadm_ssh_t:dir { search getattr read };
#line 4625
allow sysadm_t sysadm_ssh_t:{ file lnk_file } { read getattr };
#line 4625

#line 4625
# Allow the ssh program to communicate with ssh-agent.
#line 4625
allow sysadm_ssh_t sysadm_tmp_t:sock_file write;
#line 4625
allow sysadm_ssh_t sysadm_t:unix_stream_socket connectto;
#line 4625
allow sysadm_ssh_t sshd_t:unix_stream_socket connectto;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Instantiate a derived domain for user cron jobs.
#line 4625

#line 4625
# Derived domain for user cron jobs, user user_crond_domain if not system
#line 4625

#line 4625
type sysadm_crond_t, domain, user_crond_domain;
#line 4625

#line 4625

#line 4625
# Permit a transition from the crond_t domain to this domain.
#line 4625
# The transition is requested explicitly by the modified crond 
#line 4625
# via execve_secure.  There is no way to set up an automatic
#line 4625
# transition, since crontabs are configuration files, not executables.
#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to transition to the new domain.
#line 4625
#
#line 4625
allow crond_t sysadm_crond_t:process transition;
#line 4625

#line 4625
#
#line 4625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4625
#
#line 4625
dontaudit crond_t sysadm_crond_t:process noatsecure;
#line 4625

#line 4625
#
#line 4625
# Allow the process to execute the program.
#line 4625
# 
#line 4625
allow crond_t shell_exec_t:file { read { getattr execute } };
#line 4625

#line 4625
#
#line 4625
# Allow the process to reap the new domain.
#line 4625
#
#line 4625
allow sysadm_crond_t crond_t:process sigchld;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to inherit and use file 
#line 4625
# descriptions from the creating process and vice versa.
#line 4625
#
#line 4625
allow sysadm_crond_t crond_t:fd use;
#line 4625
allow crond_t sysadm_crond_t:fd use;
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to write back to the old domain via a pipe.
#line 4625
#
#line 4625
allow sysadm_crond_t crond_t:fifo_file { ioctl read getattr lock write append };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to read and execute the program.
#line 4625
#
#line 4625
allow sysadm_crond_t shell_exec_t:file { read getattr lock execute ioctl };
#line 4625

#line 4625
#
#line 4625
# Allow the new domain to be entered via the program.
#line 4625
#
#line 4625
allow sysadm_crond_t shell_exec_t:file entrypoint;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# The user role is authorized for this domain.
#line 4625
role sysadm_r types sysadm_crond_t;
#line 4625

#line 4625
# This domain is granted permissions common to most domains.
#line 4625

#line 4625
#
#line 4625
# Allow the domain to create and use UDP and TCP sockets.
#line 4625
# Other kinds of sockets must be separately authorized for use.
#line 4625
allow sysadm_crond_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_crond_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send UDP packets.
#line 4625
# Since the destination sockets type is unknown, the generic
#line 4625
# any_socket_t type is used as a placeholder.
#line 4625
#
#line 4625
allow sysadm_crond_t any_socket_t:udp_socket sendto;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send using any network interface.
#line 4625
# netif_type is a type attribute for all network interface types.
#line 4625
#
#line 4625
allow sysadm_crond_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4625

#line 4625
#
#line 4625
# Allow packets sent by the domain to be received on any network interface.
#line 4625
#
#line 4625
allow sysadm_crond_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to receive packets from any network interface.
#line 4625
# netmsg_type is a type attribute for all default message types.
#line 4625
#
#line 4625
allow sysadm_crond_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to initiate or accept TCP connections 
#line 4625
# on any network interface.
#line 4625
#
#line 4625
allow sysadm_crond_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4625

#line 4625
#
#line 4625
# Receive resets from the TCP reset socket.
#line 4625
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4625
#
#line 4625
allow sysadm_crond_t tcp_socket_t:tcp_socket recvfrom;
#line 4625

#line 4625
dontaudit sysadm_crond_t tcp_socket_t:tcp_socket connectto;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send to any node.
#line 4625
# node_type is a type attribute for all node types.
#line 4625
#
#line 4625
allow sysadm_crond_t node_type:node { tcp_send udp_send rawip_send };
#line 4625

#line 4625
#
#line 4625
# Allow packets sent by the domain to be received from any node.
#line 4625
#
#line 4625
allow sysadm_crond_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send NFS client requests via the socket
#line 4625
# created by mount.
#line 4625
#
#line 4625
allow sysadm_crond_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625

#line 4625
#
#line 4625
# Bind to the default port type.
#line 4625
# Other port types must be separately authorized.
#line 4625
#
#line 4625
allow sysadm_crond_t port_t:udp_socket name_bind;
#line 4625
allow sysadm_crond_t port_t:tcp_socket name_bind;
#line 4625

#line 4625
allow sysadm_crond_t resolv_conf_t:file { getattr read };
#line 4625

#line 4625
allow sysadm_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_crond_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t device_t:dir search;
#line 4625
allow sysadm_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t self:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crond_t self:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625
allow sysadm_crond_t self:fifo_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_crond_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_crond_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_crond_t etc_runtime_t:file { getattr read };
#line 4625
allow sysadm_crond_t self:process { fork { sigchld sigkill sigstop signull signal } };
#line 4625
allow sysadm_crond_t proc_t:dir { getattr search read };
#line 4625
allow sysadm_crond_t proc_t:file { getattr read };
#line 4625

#line 4625
allow sysadm_crond_t etc_t:lnk_file read;
#line 4625

#line 4625
allow sysadm_crond_t locale_t:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crond_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t sysctl_kernel_t:dir search;
#line 4625
allow sysadm_crond_t sysctl_kernel_t:file { getattr read };
#line 4625
allow sysadm_crond_t var_spool_t:dir search;
#line 4625
allow sysadm_crond_t fs_t:filesystem getattr;
#line 4625

#line 4625
allow sysadm_crond_t devtty_t:chr_file { read write };
#line 4625

#line 4625
# Use capabilities.
#line 4625
allow sysadm_crond_t sysadm_crond_t:capability dac_override;
#line 4625

#line 4625
# Inherit and use descriptors from initrc.
#line 4625
allow sysadm_crond_t initrc_t:fd use;
#line 4625

#line 4625
# 
#line 4625
# Since crontab files are not directly executed,
#line 4625
# crond must ensure that the crontab file has
#line 4625
# a type that is appropriate for the domain of
#line 4625
# the user cron job.  It performs an entrypoint
#line 4625
# permission check for this purpose.
#line 4625
#
#line 4625
allow sysadm_crond_t sysadm_cron_spool_t:file entrypoint;
#line 4625

#line 4625
# Access user files and dirs.
#line 4625
allow sysadm_crond_t home_root_t:dir search;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_crond_t sysadm_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_crond_t sysadm_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_crond_t sysadm_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_crond_t sysadm_home_dir_t:dir sysadm_home_t;
#line 4625
type_transition sysadm_crond_t sysadm_home_dir_t:{ file lnk_file sock_file fifo_file } sysadm_home_t;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_crond_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625
allow sysadm_crond_t sysadm_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625
allow sysadm_crond_t sysadm_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4625

#line 4625

#line 4625
type_transition sysadm_crond_t tmp_t:dir sysadm_tmp_t;
#line 4625
type_transition sysadm_crond_t tmp_t:{ file lnk_file sock_file fifo_file } sysadm_tmp_t;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Run helper programs.
#line 4625

#line 4625
allow sysadm_crond_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crond_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4625

#line 4625
allow sysadm_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4625
allow sysadm_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_crond_t ld_so_t:file execute_no_trans;
#line 4625
allow sysadm_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4625
allow sysadm_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4625
allow sysadm_crond_t device_t:dir search;
#line 4625
allow sysadm_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625

#line 4625
# Run scripts in user home directory.
#line 4625

#line 4625
allow sysadm_crond_t sysadm_home_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4625

#line 4625

#line 4625
# ps does not need to access /boot when run from cron
#line 4625
dontaudit sysadm_crond_t boot_t:dir search;
#line 4625
# quiet other ps operations
#line 4625
dontaudit sysadm_crond_t domain:dir getattr;
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Read the mouse.
#line 4625
allow sysadm_t mouse_device_t:chr_file { read getattr lock ioctl };
#line 4625
# Access other miscellaneous devices.
#line 4625
allow sysadm_t misc_device_t:{ file lnk_file sock_file fifo_file chr_file blk_file } { ioctl read getattr lock write append };
#line 4625

#line 4625
# Use the network.
#line 4625

#line 4625
#
#line 4625
# Allow the domain to create and use UDP and TCP sockets.
#line 4625
# Other kinds of sockets must be separately authorized for use.
#line 4625
allow sysadm_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625
allow sysadm_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send UDP packets.
#line 4625
# Since the destination sockets type is unknown, the generic
#line 4625
# any_socket_t type is used as a placeholder.
#line 4625
#
#line 4625
allow sysadm_t any_socket_t:udp_socket sendto;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send using any network interface.
#line 4625
# netif_type is a type attribute for all network interface types.
#line 4625
#
#line 4625
allow sysadm_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4625

#line 4625
#
#line 4625
# Allow packets sent by the domain to be received on any network interface.
#line 4625
#
#line 4625
allow sysadm_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to receive packets from any network interface.
#line 4625
# netmsg_type is a type attribute for all default message types.
#line 4625
#
#line 4625
allow sysadm_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to initiate or accept TCP connections 
#line 4625
# on any network interface.
#line 4625
#
#line 4625
allow sysadm_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4625

#line 4625
#
#line 4625
# Receive resets from the TCP reset socket.
#line 4625
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4625
#
#line 4625
allow sysadm_t tcp_socket_t:tcp_socket recvfrom;
#line 4625

#line 4625
dontaudit sysadm_t tcp_socket_t:tcp_socket connectto;
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send to any node.
#line 4625
# node_type is a type attribute for all node types.
#line 4625
#
#line 4625
allow sysadm_t node_type:node { tcp_send udp_send rawip_send };
#line 4625

#line 4625
#
#line 4625
# Allow packets sent by the domain to be received from any node.
#line 4625
#
#line 4625
allow sysadm_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4625

#line 4625
#
#line 4625
# Allow the domain to send NFS client requests via the socket
#line 4625
# created by mount.
#line 4625
#
#line 4625
allow sysadm_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4625

#line 4625
#
#line 4625
# Bind to the default port type.
#line 4625
# Other port types must be separately authorized.
#line 4625
#
#line 4625
allow sysadm_t port_t:udp_socket name_bind;
#line 4625
allow sysadm_t port_t:tcp_socket name_bind;
#line 4625

#line 4625
allow sysadm_t resolv_conf_t:file { getattr read };
#line 4625
# for perl
#line 4625
dontaudit sysadm_t resolv_conf_t:file ioctl;
#line 4625

#line 4625
#
#line 4625
# connect_secure and sendmsg_secure calls with a 
#line 4625
# peer or destination socket SID can be enforced
#line 4625
# when using the loopback interface.  Enforcement
#line 4625
# for real network interfaces will be possible when
#line 4625
# a packet labeling mechanism is integrated.
#line 4625
#
#line 4625
allow sysadm_t node_lo_t:node enforce_dest;
#line 4625

#line 4625
# Communicate within the domain.
#line 4625

#line 4625
allow sysadm_t sysadm_t:udp_socket { sendto };
#line 4625
allow sysadm_t sysadm_t:udp_socket { recvfrom };
#line 4625

#line 4625

#line 4625
allow sysadm_t sysadm_t:tcp_socket { connectto recvfrom };
#line 4625
allow sysadm_t sysadm_t:tcp_socket { acceptfrom recvfrom };
#line 4625
allow sysadm_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625
allow sysadm_t tcp_socket_t:tcp_socket { recvfrom };
#line 4625

#line 4625

#line 4625
# Connect to inetd.
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
# Connect data port to ftpd.
#line 4625

#line 4625

#line 4625
# Connect to portmap.
#line 4625

#line 4625

#line 4625
# Inherit and use sockets from inetd
#line 4625

#line 4625

#line 4625
# Allow system log read
#line 4625
#allow sysadm_t kernel_t:system syslog_read;
#line 4625
# else do not log it
#line 4625
dontaudit sysadm_t kernel_t:system syslog_read;
#line 4625

#line 4625
# Very permissive allowing every domain to see every type.
#line 4625
allow sysadm_t kernel_t:system { ipc_info };
#line 4625

#line 4625
# When the user domain runs ps, there will be a number of access
#line 4625
# denials when ps tries to search /proc.  Do not audit these denials.
#line 4625
dontaudit sysadm_t domain:dir { read getattr lock search ioctl };
#line 4625
dontaudit sysadm_t domain:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4625

#line 4625
# Some shells ask for w access to utmp, but will operate
#line 4625
# correctly without it.  Do not audit write denials to utmp.
#line 4625
dontaudit sysadm_t initrc_var_run_t:file { getattr read write };
#line 4625

#line 4625
# do not audit getattr on tmpfile, otherwise ls -l /tmp fills the logs
#line 4625
dontaudit sysadm_t tmpfile:{ dir file lnk_file sock_file fifo_file chr_file blk_file } getattr;
#line 4625

#line 4625
# do not audit getattr on disk devices, otherwise KDE fills the logs
#line 4625
dontaudit sysadm_t { removable_device_t fixed_disk_device_t }:blk_file getattr;
#line 4625

#line 4625

#line 4625
# Access the sound device.
#line 4625
allow sysadm_t sound_device_t:chr_file { getattr read write ioctl };
#line 4625

#line 4625

#line 4625

#line 4625
# Get attributes of file systems.
#line 4625
allow sysadm_t fs_type:filesystem getattr;
#line 4625

#line 4625
# Read and write /dev/tty and /dev/null.
#line 4625
allow sysadm_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4625
allow sysadm_t zero_device_t:chr_file { { ioctl read getattr lock write append } execute };
#line 4625
allow sysadm_t random_device_t:chr_file { getattr read };
#line 4625

#line 4625
# Read the devpts root directory.
#line 4625
allow sysadm_t devpts_t:dir { read getattr lock search ioctl };
#line 4625

#line 4625

#line 4625
# Violates the goal of limiting write access to checkpolicy.
#line 4625
#rw_dir_create_file(sysadm_t, policy_config_t)
#line 4625

#line 4625

#line 4625
allow sysadm_crond_t var_log_t:file { read getattr lock ioctl };
#line 4625

#line 4625

#line 4625
# Create files in /tmp/orbit-* and /tmp/.ICE-unix
#line 4625
# with our derived tmp type rather than user_tmp_t.
#line 4625

#line 4625

#line 4625

#line 4625

#line 4625
#
#line 4625
# Allow the process to modify the directory.
#line 4625
#
#line 4625
allow sysadm_t user_tmpfile:dir { read getattr lock search ioctl add_name remove_name write };
#line 4625

#line 4625
#
#line 4625
# Allow the process to create the file.
#line 4625
#
#line 4625

#line 4625

#line 4625
allow sysadm_t sysadm_tmp_t:sock_file { create ioctl read getattr lock write setattr append link unlink rename };
#line 4625

#line 4625

#line 4625
type_transition sysadm_t user_tmpfile:sock_file sysadm_tmp_t;
#line 4625

#line 4625

#line 4625

#line 4625


# Allow administrator domains to set the enforcing flag.

#line 4628
allow admin security_t:dir { read search getattr };
#line 4628
allow admin security_t:file { getattr read write };
#line 4628
allow admin security_t:security setenforce;
#line 4628
auditallow admin security_t:security setenforce;
#line 4628





#DESC Httpd admin -  Domain for httpd administrators
########################################
# httpd_admin domain macro and rules for 
# httpd_admin_t
########################################

# Uses some types defined in apache.te, so conditionally defined this macro.
# The intent is to remove this file in the future.
#
#line 4777

#DESC User - Domains for ordinary users.
#
#################################

#line 4781

#line 4781
# user_t/user_t is an unprivileged users domain.
#line 4781
type user_t, domain, userdomain, unpriv_userdomain, web_client_domain;
#line 4781

#line 4781
# user_r is authorized for user_t for the initial login domain.
#line 4781
role user_r types user_t;
#line 4781
allow system_r user_r;
#line 4781

#line 4781
# Grant permissions within the domain.
#line 4781

#line 4781
# Access other processes in the same domain.
#line 4781
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4781
# separately if desired.
#line 4781
allow user_t self:process ~{ptrace setexec setfscreate};
#line 4781

#line 4781
# Access /proc/PID files for processes in the same domain.
#line 4781
allow user_t self:dir { read getattr lock search ioctl };
#line 4781
allow user_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Access file descriptions, pipes, and sockets
#line 4781
# created by processes in the same domain.
#line 4781
allow user_t self:fd *;
#line 4781
allow user_t self:fifo_file { ioctl read getattr lock write append };
#line 4781
allow user_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781

#line 4781
# Allow the domain to communicate with other processes in the same domain.
#line 4781
allow user_t self:unix_dgram_socket sendto;
#line 4781
allow user_t self:unix_stream_socket connectto;
#line 4781

#line 4781
# Access System V IPC objects created by processes in the same domain.
#line 4781
allow user_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4781
allow user_t self:msg  { send receive };
#line 4781
allow user_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4781
allow user_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4781
;
#line 4781

#line 4781
# Grant read/search permissions to some of /proc.
#line 4781
allow user_t proc_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t proc_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
# Read /.
#line 4781
allow user_t root_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /home.
#line 4781
allow user_t home_root_t:dir { read getattr lock search ioctl };
#line 4781

#line 4781
# Read /usr.
#line 4781
allow user_t usr_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read bin and sbin directories.
#line 4781
allow user_t bin_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781
allow user_t sbin_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
# Read directories and files with the readable_t type.
#line 4781
# This type is a general type for "world"-readable files.
#line 4781
allow user_t readable_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t readable_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Stat /...security and lost+found.
#line 4781
allow user_t file_labels_t:dir getattr;
#line 4781
allow user_t lost_found_t:dir getattr;
#line 4781

#line 4781

#line 4781

#line 4781
# Read the /tmp directory and any /tmp files with the base type.
#line 4781
# Temporary files created at runtime will typically use derived types.
#line 4781
allow user_t tmp_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t tmp_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /var, /var/spool, /var/run.
#line 4781
allow user_t var_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t var_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781
allow user_t var_spool_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t var_spool_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781
allow user_t var_run_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t var_run_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
allow user_t sysctl_kernel_t:dir search;
#line 4781
allow user_t sysctl_kernel_t:file { getattr read };
#line 4781

#line 4781
# Read /etc.
#line 4781
allow user_t etc_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781
allow user_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
# for running depmod as part of the kernel packaging process
#line 4781
allow user_t modules_conf_t:file { getattr read };
#line 4781

#line 4781
# Read man directories and files.
#line 4781
allow user_t man_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t man_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /dev directories and any symbolic links.
#line 4781
allow user_t device_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t device_t:lnk_file { read getattr lock ioctl };
#line 4781

#line 4781
# Do not audit write denials to /etc/ld.so.cache.
#line 4781
dontaudit user_t ld_so_cache_t:file write;
#line 4781

#line 4781
# Execute from the system shared libraries.
#line 4781

#line 4781
allow user_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_t ld_so_t:file execute_no_trans;
#line 4781
allow user_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_t device_t:dir search;
#line 4781
allow user_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781
;
#line 4781

#line 4781
# user_t is also granted permissions specific to user domains.
#line 4781

#line 4781
# Use capabilities
#line 4781
allow user_t self:capability { setgid chown fowner };
#line 4781
dontaudit user_t self:capability { sys_nice fsetid };
#line 4781

#line 4781
# Type for home directory.
#line 4781

#line 4781
type user_home_dir_t, file_type, sysadmfile, home_dir_type, user_home_dir_type, home_type, user_home_type;
#line 4781
type user_home_t, file_type, sysadmfile, home_type, user_home_type;
#line 4781
# do not allow privhome access to sysadm_home_dir_t
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow privhome user_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow privhome user_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow privhome user_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition privhome user_home_dir_t:dir user_home_t;
#line 4781
type_transition privhome user_home_dir_t:{ file lnk_file sock_file fifo_file } user_home_t;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
type user_tmp_t, file_type, sysadmfile, tmpfile , user_tmpfile;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_t user_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_t user_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_t tmp_t:dir user_tmp_t;
#line 4781
type_transition user_t tmp_t:{ file lnk_file sock_file fifo_file } user_tmp_t;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# allow ptrace
#line 4781
#can_ptrace(user_t, user_t)
#line 4781

#line 4781
# Create, access, and remove files in home directory.
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_t user_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_t user_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_t user_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_t user_home_dir_t:dir user_home_t;
#line 4781
type_transition user_t user_home_dir_t:{ file lnk_file sock_file fifo_file } user_home_t;
#line 4781

#line 4781

#line 4781

#line 4781
allow user_t user_home_t:{ dir file lnk_file sock_file fifo_file chr_file blk_file } { relabelfrom relabelto };
#line 4781

#line 4781
# Bind to a Unix domain socket in /tmp.
#line 4781
allow user_t user_tmp_t:unix_stream_socket name_bind;
#line 4781

#line 4781
# Type for tty devices.
#line 4781
type user_tty_device_t, file_type, sysadmfile, ttyfile;
#line 4781
# Access ttys.
#line 4781
allow user_t privfd:fd use;
#line 4781
allow user_t user_tty_device_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 4781
# Use the type when relabeling terminal devices.
#line 4781
type_change user_t tty_device_t:chr_file user_tty_device_t;
#line 4781

#line 4781
# Debian login is from shadow utils and does not allow resetting the perms.
#line 4781
# have to fix this!
#line 4781
type_change user_t ttyfile:chr_file user_tty_device_t;
#line 4781

#line 4781
# allow reading /var/lib/dpkg etc
#line 4781
allow user_t var_lib_t:dir { read getattr lock search ioctl };
#line 4781

#line 4781
# Allow reading dpkg origins file
#line 4781

#line 4781
allow user_t etc_dpkg_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t etc_dpkg_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781

#line 4781
# read localization information
#line 4781
allow user_t locale_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t locale_t:{file lnk_file} { read getattr lock ioctl };
#line 4781

#line 4781
# for running TeX programs
#line 4781

#line 4781
allow user_t tetex_data_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t tetex_data_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
allow user_t tetex_data_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
# Type and access for pty devices.
#line 4781

#line 4781

#line 4781

#line 4781
# Access the pty master multiplexer.
#line 4781
allow user_t ptmx_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781
allow user_t devpts_t:filesystem getattr;
#line 4781

#line 4781
# allow searching /dev/pts
#line 4781
allow user_t devpts_t:dir { getattr read search };
#line 4781

#line 4781
type user_devpts_t, file_type, sysadmfile, ptyfile , userpty_type;
#line 4781

#line 4781
# Allow the pty to be associated with the file system.
#line 4781
allow user_devpts_t devpts_t:filesystem associate;
#line 4781

#line 4781
# Label pty files with a derived type.
#line 4781
type_transition user_t devpts_t:chr_file user_devpts_t;
#line 4781

#line 4781
# Read and write my pty files.
#line 4781
allow user_t user_devpts_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 4781

#line 4781

#line 4781

#line 4781
# Use the type when relabeling pty devices.
#line 4781

#line 4781

#line 4781
type_change user_t sshd_devpts_t:chr_file user_devpts_t;
#line 4781

#line 4781
# Connect to sshd.
#line 4781

#line 4781
allow user_t sshd_t:tcp_socket { connectto recvfrom };
#line 4781
allow sshd_t user_t:tcp_socket { acceptfrom recvfrom };
#line 4781
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781
allow user_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781

#line 4781

#line 4781
# Connect to ssh proxy.
#line 4781

#line 4781
allow user_t user_ssh_t:tcp_socket { connectto recvfrom };
#line 4781
allow user_ssh_t user_t:tcp_socket { acceptfrom recvfrom };
#line 4781
allow user_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781
allow user_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781

#line 4781

#line 4781
allow user_t sshd_t:fd use;
#line 4781
allow user_t sshd_t:tcp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781
# Use a Unix stream socket inherited from sshd.
#line 4781
allow user_t sshd_t:unix_stream_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781

#line 4781
allow user_t boot_t:dir { getattr search };
#line 4781
dontaudit user_t boot_t:dir read;
#line 4781
dontaudit user_t boot_t:lnk_file getattr;
#line 4781
dontaudit user_t boot_t:file { getattr read };
#line 4781
allow user_t system_map_t:file { getattr read };
#line 4781

#line 4781

#line 4781
type user_tmpfs_t, file_type, sysadmfile, tmpfsfile;
#line 4781
# Use this type when creating tmpfs/shm objects.
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_t tmpfs_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_t user_tmpfs_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_t user_tmpfs_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_t tmpfs_t:dir user_tmpfs_t;
#line 4781
type_transition user_t tmpfs_t:{ file lnk_file sock_file fifo_file } user_tmpfs_t;
#line 4781

#line 4781

#line 4781

#line 4781
allow user_tmpfs_t tmpfs_t:filesystem associate;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Read and write /var/catman.
#line 4781
allow user_t catman_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781
allow user_t catman_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781

#line 4781
# Modify mail spool file.
#line 4781
allow user_t mail_spool_t:dir { read getattr lock search ioctl };
#line 4781
allow user_t mail_spool_t:file { ioctl read getattr lock write append };
#line 4781
allow user_t mail_spool_t:lnk_file read;
#line 4781

#line 4781
#
#line 4781
# Allow the query of filesystem quotas
#line 4781
#
#line 4781
allow user_t fs_type:filesystem quotaget;
#line 4781

#line 4781
# Run helper programs.
#line 4781

#line 4781
allow user_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4781

#line 4781
allow user_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_t ld_so_t:file execute_no_trans;
#line 4781
allow user_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_t device_t:dir search;
#line 4781
allow user_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781
allow user_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
# Run programs developed by other users in the same domain.
#line 4781

#line 4781
allow user_t user_home_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_t user_tmp_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
# Run user programs that require different permissions in their own domain.
#line 4781
# These rules were moved into the individual program domains.
#line 4781

#line 4781
# Instantiate derived domains for a number of programs.
#line 4781
# These derived domains encode both information about the calling
#line 4781
# user domain and the program, and allow us to maintain separation
#line 4781
# between different instances of the program being run by different
#line 4781
# user domains.
#line 4781

#line 4781

#line 4781

#line 4781
# Derived domain based on the calling user domain and the program.
#line 4781
type user_su_t, domain, privlog, auth;
#line 4781

#line 4781
# Transition from the user domain to this domain.
#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow user_t user_su_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit user_t user_su_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow user_t su_exec_t:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_su_t user_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_su_t user_t:fd use;
#line 4781
allow user_t user_su_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_su_t user_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_su_t su_exec_t:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_su_t su_exec_t:file entrypoint;
#line 4781

#line 4781
type_transition user_t su_exec_t:process user_su_t;
#line 4781

#line 4781

#line 4781

#line 4781
allow user_su_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_su_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_su_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_su_t ld_so_t:file execute_no_trans;
#line 4781
allow user_su_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_su_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_su_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_su_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_su_t device_t:dir search;
#line 4781
allow user_su_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
allow user_su_t etc_t:file { getattr read };
#line 4781

#line 4781
allow user_su_t etc_t:lnk_file read;
#line 4781

#line 4781
allow user_su_t locale_t:dir { read getattr lock search ioctl };
#line 4781
allow user_su_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
allow user_su_t self:unix_dgram_socket { connect create write };
#line 4781
allow user_su_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781
allow user_su_t proc_t:dir search;
#line 4781
allow user_su_t proc_t:lnk_file read;
#line 4781

#line 4781
allow user_su_t self:dir { read getattr lock search ioctl };
#line 4781
allow user_su_t self:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
allow user_su_t self:process setsched;
#line 4781
allow user_su_t device_t:dir search;
#line 4781
allow user_su_t self:process { fork sigchld };
#line 4781

#line 4781
# Use capabilities.
#line 4781
allow user_su_t self:capability { setuid setgid net_bind_service chown dac_override sys_nice sys_resource };
#line 4781

#line 4781
# Revert to the user domain when a shell is executed.
#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow user_su_t user_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit user_su_t user_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow user_su_t shell_exec_t:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_t user_su_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_t user_su_t:fd use;
#line 4781
allow user_su_t user_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_t user_su_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_t shell_exec_t:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_t shell_exec_t:file entrypoint;
#line 4781

#line 4781
type_transition user_su_t shell_exec_t:process user_t;
#line 4781

#line 4781
allow user_su_t bin_t:dir search;
#line 4781
allow user_su_t bin_t:lnk_file read;
#line 4781

#line 4781
allow user_su_t privfd:fd use;
#line 4781

#line 4781
# Write to utmp.
#line 4781
allow user_su_t { var_t var_run_t }:dir search;
#line 4781
allow user_su_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781

#line 4781
# Run chkpwd.
#line 4781

#line 4781
allow user_su_t chkpwd_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781

#line 4781
# Inherit and use descriptors from gnome-pty-helper.
#line 4781

#line 4781

#line 4781
# The user role is authorized for this domain.
#line 4781
role user_r types user_su_t;
#line 4781

#line 4781
# Write to the user domain tty.
#line 4781
allow user_su_t user_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_su_t user_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
allow user_su_t { home_root_t user_home_dir_t }:dir search;
#line 4781

#line 4781
# Modify .Xauthority file (via xauth program).
#line 4781

#line 4781

#line 4781
# Access sshd cookie files.
#line 4781
allow user_su_t sshd_tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781
allow user_su_t sshd_tmp_t:file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_su_t sshd_tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_su_t user_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_su_t user_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_su_t sshd_tmp_t:dir user_tmp_t;
#line 4781
type_transition user_su_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } user_tmp_t;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# stop su complaining if you run it from a directory with restrictive perms
#line 4781
dontaudit user_su_t file_type:dir search;
#line 4781

#line 4781

#line 4781
# Derived domain based on the calling user domain and the program.
#line 4781
type user_chkpwd_t, domain, privlog, auth;
#line 4781

#line 4781
# Transition from the user domain to this domain.
#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow user_t user_chkpwd_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit user_t user_chkpwd_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow user_t chkpwd_exec_t:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_chkpwd_t user_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_chkpwd_t user_t:fd use;
#line 4781
allow user_t user_chkpwd_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_chkpwd_t user_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_chkpwd_t chkpwd_exec_t:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_chkpwd_t chkpwd_exec_t:file entrypoint;
#line 4781

#line 4781
type_transition user_t chkpwd_exec_t:process user_chkpwd_t;
#line 4781

#line 4781

#line 4781
# The user role is authorized for this domain.
#line 4781
role user_r types user_chkpwd_t;
#line 4781

#line 4781

#line 4781
allow user_chkpwd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_chkpwd_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_chkpwd_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_chkpwd_t ld_so_t:file execute_no_trans;
#line 4781
allow user_chkpwd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_chkpwd_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_chkpwd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_chkpwd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_chkpwd_t device_t:dir search;
#line 4781
allow user_chkpwd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
allow user_chkpwd_t etc_t:file { getattr read };
#line 4781
allow user_chkpwd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_chkpwd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781

#line 4781
allow user_chkpwd_t etc_t:lnk_file read;
#line 4781

#line 4781
allow user_chkpwd_t locale_t:dir { read getattr lock search ioctl };
#line 4781
allow user_chkpwd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781

#line 4781
# Use capabilities.
#line 4781
allow user_chkpwd_t self:capability setuid;
#line 4781

#line 4781
# Inherit and use descriptors from gnome-pty-helper.
#line 4781

#line 4781

#line 4781
# Inherit and use descriptors from newrole.
#line 4781
allow user_chkpwd_t newrole_t:fd use;
#line 4781

#line 4781
# Write to the user domain tty.
#line 4781
allow user_chkpwd_t user_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_chkpwd_t user_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Derived domain based on the calling user domain and the program.
#line 4781
type user_crontab_t, domain, privlog;
#line 4781

#line 4781
# Transition from the user domain to the derived domain.
#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow user_t user_crontab_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit user_t user_crontab_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow user_t crontab_exec_t:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_crontab_t user_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_crontab_t user_t:fd use;
#line 4781
allow user_t user_crontab_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_crontab_t user_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_crontab_t crontab_exec_t:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_crontab_t crontab_exec_t:file entrypoint;
#line 4781

#line 4781
type_transition user_t crontab_exec_t:process user_crontab_t;
#line 4781

#line 4781

#line 4781
# The user role is authorized for this domain.
#line 4781
role user_r types user_crontab_t;
#line 4781

#line 4781

#line 4781
allow user_crontab_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_crontab_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crontab_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_crontab_t ld_so_t:file execute_no_trans;
#line 4781
allow user_crontab_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crontab_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_crontab_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crontab_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_crontab_t device_t:dir search;
#line 4781
allow user_crontab_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
allow user_crontab_t etc_t:file { getattr read };
#line 4781
allow user_crontab_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_crontab_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781

#line 4781
allow user_crontab_t etc_t:lnk_file read;
#line 4781

#line 4781
allow user_crontab_t locale_t:dir { read getattr lock search ioctl };
#line 4781
allow user_crontab_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781

#line 4781
# Use capabilities
#line 4781
allow user_crontab_t user_crontab_t:capability { setuid setgid chown };
#line 4781

#line 4781
# Type for temporary files.
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_crontab_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_crontab_t user_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_crontab_t user_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_crontab_t tmp_t:dir user_tmp_t;
#line 4781
type_transition user_crontab_t tmp_t:{ file lnk_file sock_file fifo_file } user_tmp_t;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Type of user crontabs once moved to cron spool.
#line 4781
type user_cron_spool_t, file_type, sysadmfile;
#line 4781
# Use the type when creating files in /var/spool/cron.
#line 4781
allow sysadm_crontab_t user_cron_spool_t:file { getattr read };
#line 4781
allow user_crontab_t { var_t var_spool_t }:dir search;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_crontab_t cron_spool_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_crontab_t user_cron_spool_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_crontab_t user_cron_spool_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_crontab_t cron_spool_t:dir user_cron_spool_t;
#line 4781
type_transition user_crontab_t cron_spool_t:{ file lnk_file sock_file fifo_file } user_cron_spool_t;
#line 4781

#line 4781

#line 4781

#line 4781
allow user_crontab_t self:process { fork { sigchld sigkill sigstop signull signal } };
#line 4781

#line 4781
# crontab signals crond by updating the mtime on the spooldir
#line 4781
allow user_crontab_t cron_spool_t:dir setattr;
#line 4781
# Allow crond to read those crontabs in cron spool.
#line 4781
allow crond_t user_cron_spool_t:file { read getattr lock ioctl };
#line 4781

#line 4781
# Run helper programs as user_t
#line 4781
allow user_crontab_t { bin_t sbin_t }:dir search;
#line 4781
allow user_crontab_t bin_t:lnk_file read;
#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow user_crontab_t user_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit user_crontab_t user_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow user_crontab_t { bin_t sbin_t shell_exec_t }:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_t user_crontab_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_t user_crontab_t:fd use;
#line 4781
allow user_crontab_t user_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_t user_crontab_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_t { bin_t sbin_t shell_exec_t }:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_t { bin_t sbin_t shell_exec_t }:file entrypoint;
#line 4781

#line 4781
type_transition user_crontab_t { bin_t sbin_t shell_exec_t }:process user_t;
#line 4781

#line 4781

#line 4781
# Read user crontabs 
#line 4781
allow user_crontab_t { user_home_t user_home_dir_t }:dir { read getattr lock search ioctl };  
#line 4781
allow user_crontab_t user_home_t:file { read getattr lock ioctl };  
#line 4781
dontaudit user_crontab_t user_home_dir_t:dir write;
#line 4781

#line 4781
# Access the cron log file.
#line 4781
allow user_crontab_t cron_log_t:file { read getattr lock ioctl };
#line 4781
allow user_crontab_t cron_log_t:file { append };
#line 4781

#line 4781
# Access terminals.
#line 4781
allow user_crontab_t device_t:dir search;
#line 4781
allow user_crontab_t user_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_crontab_t user_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
# Inherit and use descriptors from gnome-pty-helper.
#line 4781

#line 4781
allow user_crontab_t privfd:fd use;
#line 4781

#line 4781

#line 4781
# Derived domain based on the calling user domain and the program.
#line 4781
type user_ssh_t, domain, privlog;
#line 4781

#line 4781
# Transition from the user domain to the derived domain.
#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow user_t user_ssh_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit user_t user_ssh_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow user_t ssh_exec_t:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_ssh_t user_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_ssh_t user_t:fd use;
#line 4781
allow user_t user_ssh_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_ssh_t user_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_ssh_t ssh_exec_t:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_ssh_t ssh_exec_t:file entrypoint;
#line 4781

#line 4781
type_transition user_t ssh_exec_t:process user_ssh_t;
#line 4781

#line 4781

#line 4781
# The user role is authorized for this domain.
#line 4781
role user_r types user_ssh_t;
#line 4781

#line 4781
# Grant permissions within the domain.
#line 4781

#line 4781
# Access other processes in the same domain.
#line 4781
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4781
# separately if desired.
#line 4781
allow user_ssh_t self:process ~{ptrace setexec setfscreate};
#line 4781

#line 4781
# Access /proc/PID files for processes in the same domain.
#line 4781
allow user_ssh_t self:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Access file descriptions, pipes, and sockets
#line 4781
# created by processes in the same domain.
#line 4781
allow user_ssh_t self:fd *;
#line 4781
allow user_ssh_t self:fifo_file { ioctl read getattr lock write append };
#line 4781
allow user_ssh_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_ssh_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781

#line 4781
# Allow the domain to communicate with other processes in the same domain.
#line 4781
allow user_ssh_t self:unix_dgram_socket sendto;
#line 4781
allow user_ssh_t self:unix_stream_socket connectto;
#line 4781

#line 4781
# Access System V IPC objects created by processes in the same domain.
#line 4781
allow user_ssh_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4781
allow user_ssh_t self:msg  { send receive };
#line 4781
allow user_ssh_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4781
allow user_ssh_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4781

#line 4781

#line 4781
# Use descriptors created by sshd
#line 4781
allow user_ssh_t privfd:fd use;
#line 4781

#line 4781

#line 4781
allow user_ssh_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_ssh_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_ssh_t ld_so_t:file execute_no_trans;
#line 4781
allow user_ssh_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_ssh_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_ssh_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_ssh_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_ssh_t device_t:dir search;
#line 4781
allow user_ssh_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781
allow user_ssh_t etc_t:lnk_file read;
#line 4781

#line 4781
allow user_ssh_t locale_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781

#line 4781
# Get attributes of file systems.
#line 4781
allow user_ssh_t fs_type:filesystem getattr;
#line 4781

#line 4781

#line 4781
# Read /.
#line 4781
allow user_ssh_t root_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /home.
#line 4781
allow user_ssh_t home_root_t:dir { read getattr lock search ioctl };
#line 4781

#line 4781
# Read /usr.
#line 4781
allow user_ssh_t usr_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read bin and sbin directories.
#line 4781
allow user_ssh_t bin_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781
allow user_ssh_t sbin_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
# Read the devpts root directory.
#line 4781
allow user_ssh_t devpts_t:dir { read getattr lock search ioctl };
#line 4781

#line 4781
# Read /var.
#line 4781
allow user_ssh_t var_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t var_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /var/run, /var/log.
#line 4781
allow user_ssh_t var_run_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t var_run_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781
allow user_ssh_t var_log_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t var_log_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /etc.
#line 4781
allow user_ssh_t etc_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781
allow user_ssh_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781
allow user_ssh_t resolv_conf_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read the linker, shared library, and executable types.
#line 4781
allow user_ssh_t ld_so_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781
allow user_ssh_t shlib_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781
allow user_ssh_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Read /dev directories and any symbolic links.
#line 4781
allow user_ssh_t device_t:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t device_t:lnk_file { read getattr lock ioctl };
#line 4781

#line 4781
# Read /dev/random.
#line 4781
allow user_ssh_t random_device_t:chr_file { read getattr lock ioctl };
#line 4781

#line 4781
# Read and write /dev/tty and /dev/null.
#line 4781
allow user_ssh_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_ssh_t { null_device_t zero_device_t }:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
# Grant permissions needed to create TCP and UDP sockets and
#line 4781
# to access the network.
#line 4781

#line 4781
#
#line 4781
# Allow the domain to create and use UDP and TCP sockets.
#line 4781
# Other kinds of sockets must be separately authorized for use.
#line 4781
allow user_ssh_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_ssh_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send UDP packets.
#line 4781
# Since the destination sockets type is unknown, the generic
#line 4781
# any_socket_t type is used as a placeholder.
#line 4781
#
#line 4781
allow user_ssh_t any_socket_t:udp_socket sendto;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send using any network interface.
#line 4781
# netif_type is a type attribute for all network interface types.
#line 4781
#
#line 4781
allow user_ssh_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4781

#line 4781
#
#line 4781
# Allow packets sent by the domain to be received on any network interface.
#line 4781
#
#line 4781
allow user_ssh_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to receive packets from any network interface.
#line 4781
# netmsg_type is a type attribute for all default message types.
#line 4781
#
#line 4781
allow user_ssh_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to initiate or accept TCP connections 
#line 4781
# on any network interface.
#line 4781
#
#line 4781
allow user_ssh_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4781

#line 4781
#
#line 4781
# Receive resets from the TCP reset socket.
#line 4781
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4781
#
#line 4781
allow user_ssh_t tcp_socket_t:tcp_socket recvfrom;
#line 4781

#line 4781
dontaudit user_ssh_t tcp_socket_t:tcp_socket connectto;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send to any node.
#line 4781
# node_type is a type attribute for all node types.
#line 4781
#
#line 4781
allow user_ssh_t node_type:node { tcp_send udp_send rawip_send };
#line 4781

#line 4781
#
#line 4781
# Allow packets sent by the domain to be received from any node.
#line 4781
#
#line 4781
allow user_ssh_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send NFS client requests via the socket
#line 4781
# created by mount.
#line 4781
#
#line 4781
allow user_ssh_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781

#line 4781
#
#line 4781
# Bind to the default port type.
#line 4781
# Other port types must be separately authorized.
#line 4781
#
#line 4781
allow user_ssh_t port_t:udp_socket name_bind;
#line 4781
allow user_ssh_t port_t:tcp_socket name_bind;
#line 4781

#line 4781

#line 4781
# for sshing to a ssh tunnel
#line 4781

#line 4781
allow user_ssh_t user_ssh_t:tcp_socket { connectto recvfrom };
#line 4781
allow user_ssh_t user_ssh_t:tcp_socket { acceptfrom recvfrom };
#line 4781
allow user_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781
allow user_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781

#line 4781

#line 4781
# Use capabilities.
#line 4781
allow user_ssh_t self:capability { setuid setgid dac_override dac_read_search };
#line 4781

#line 4781
# Run helpers.
#line 4781

#line 4781
allow user_ssh_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4781

#line 4781
allow user_ssh_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_ssh_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_ssh_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_ssh_t ld_so_t:file execute_no_trans;
#line 4781
allow user_ssh_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_ssh_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_ssh_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_ssh_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_ssh_t device_t:dir search;
#line 4781
allow user_ssh_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781
allow user_ssh_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_ssh_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_ssh_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_ssh_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_ssh_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781

#line 4781
# Read the ssh key file.
#line 4781
allow user_ssh_t sshd_key_t:file { read getattr lock ioctl };
#line 4781

#line 4781
# Access the ssh temporary files.
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_ssh_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_ssh_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_ssh_t sshd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_ssh_t tmp_t:dir sshd_tmp_t;
#line 4781
type_transition user_ssh_t tmp_t:{ file lnk_file sock_file fifo_file } sshd_tmp_t;
#line 4781

#line 4781

#line 4781

#line 4781
allow user_ssh_t user_tmp_t:dir { read getattr lock search ioctl };
#line 4781

#line 4781
# for rsync
#line 4781
allow user_ssh_t user_t:unix_stream_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781

#line 4781
# Access the users .ssh directory.
#line 4781
type user_home_ssh_t, file_type, sysadmfile;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow { sysadm_ssh_t user_ssh_t } user_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow { sysadm_ssh_t user_ssh_t } user_home_ssh_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow { sysadm_ssh_t user_ssh_t } user_home_ssh_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition { sysadm_ssh_t user_ssh_t } user_home_dir_t:dir user_home_ssh_t;
#line 4781
type_transition { sysadm_ssh_t user_ssh_t } user_home_dir_t:{ file lnk_file sock_file fifo_file } user_home_ssh_t;
#line 4781

#line 4781

#line 4781

#line 4781
allow { sysadm_ssh_t user_ssh_t } user_home_ssh_t:lnk_file { getattr read };
#line 4781
dontaudit user_ssh_t user_home_t:dir search;
#line 4781

#line 4781
allow sshd_t user_home_ssh_t:dir { read getattr lock search ioctl };
#line 4781
allow sshd_t user_home_ssh_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
allow user_t user_home_ssh_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781
allow user_t user_home_ssh_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781

#line 4781

#line 4781
# for /bin/sh used to execute xauth
#line 4781
dontaudit user_ssh_t proc_t:dir search;
#line 4781

#line 4781
# Inherit and use descriptors from gnome-pty-helper.
#line 4781

#line 4781

#line 4781
# Connect to sshd.
#line 4781

#line 4781

#line 4781
allow user_ssh_t sshd_t:tcp_socket { connectto recvfrom };
#line 4781
allow sshd_t user_ssh_t:tcp_socket { acceptfrom recvfrom };
#line 4781
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781
allow user_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781

#line 4781

#line 4781

#line 4781
# Write to the user domain tty.
#line 4781
allow user_ssh_t user_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_ssh_t user_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781
# Allow the user shell to signal the ssh program.
#line 4781
allow user_t user_ssh_t:process signal;
#line 4781
# allow ps to show ssh
#line 4781
allow user_t user_ssh_t:dir { search getattr read };
#line 4781
allow user_t user_ssh_t:{ file lnk_file } { read getattr };
#line 4781

#line 4781
# Allow the ssh program to communicate with ssh-agent.
#line 4781
allow user_ssh_t user_tmp_t:sock_file write;
#line 4781
allow user_ssh_t user_t:unix_stream_socket connectto;
#line 4781
allow user_ssh_t sshd_t:unix_stream_socket connectto;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Instantiate a derived domain for user cron jobs.
#line 4781

#line 4781
# Derived domain for user cron jobs, user user_crond_domain if not system
#line 4781

#line 4781
type user_crond_t, domain, user_crond_domain;
#line 4781

#line 4781

#line 4781
# Permit a transition from the crond_t domain to this domain.
#line 4781
# The transition is requested explicitly by the modified crond 
#line 4781
# via execve_secure.  There is no way to set up an automatic
#line 4781
# transition, since crontabs are configuration files, not executables.
#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to transition to the new domain.
#line 4781
#
#line 4781
allow crond_t user_crond_t:process transition;
#line 4781

#line 4781
#
#line 4781
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4781
#
#line 4781
dontaudit crond_t user_crond_t:process noatsecure;
#line 4781

#line 4781
#
#line 4781
# Allow the process to execute the program.
#line 4781
# 
#line 4781
allow crond_t shell_exec_t:file { read { getattr execute } };
#line 4781

#line 4781
#
#line 4781
# Allow the process to reap the new domain.
#line 4781
#
#line 4781
allow user_crond_t crond_t:process sigchld;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to inherit and use file 
#line 4781
# descriptions from the creating process and vice versa.
#line 4781
#
#line 4781
allow user_crond_t crond_t:fd use;
#line 4781
allow crond_t user_crond_t:fd use;
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to write back to the old domain via a pipe.
#line 4781
#
#line 4781
allow user_crond_t crond_t:fifo_file { ioctl read getattr lock write append };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to read and execute the program.
#line 4781
#
#line 4781
allow user_crond_t shell_exec_t:file { read getattr lock execute ioctl };
#line 4781

#line 4781
#
#line 4781
# Allow the new domain to be entered via the program.
#line 4781
#
#line 4781
allow user_crond_t shell_exec_t:file entrypoint;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# The user role is authorized for this domain.
#line 4781
role user_r types user_crond_t;
#line 4781

#line 4781
# This domain is granted permissions common to most domains.
#line 4781

#line 4781
#
#line 4781
# Allow the domain to create and use UDP and TCP sockets.
#line 4781
# Other kinds of sockets must be separately authorized for use.
#line 4781
allow user_crond_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_crond_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send UDP packets.
#line 4781
# Since the destination sockets type is unknown, the generic
#line 4781
# any_socket_t type is used as a placeholder.
#line 4781
#
#line 4781
allow user_crond_t any_socket_t:udp_socket sendto;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send using any network interface.
#line 4781
# netif_type is a type attribute for all network interface types.
#line 4781
#
#line 4781
allow user_crond_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4781

#line 4781
#
#line 4781
# Allow packets sent by the domain to be received on any network interface.
#line 4781
#
#line 4781
allow user_crond_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to receive packets from any network interface.
#line 4781
# netmsg_type is a type attribute for all default message types.
#line 4781
#
#line 4781
allow user_crond_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to initiate or accept TCP connections 
#line 4781
# on any network interface.
#line 4781
#
#line 4781
allow user_crond_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4781

#line 4781
#
#line 4781
# Receive resets from the TCP reset socket.
#line 4781
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4781
#
#line 4781
allow user_crond_t tcp_socket_t:tcp_socket recvfrom;
#line 4781

#line 4781
dontaudit user_crond_t tcp_socket_t:tcp_socket connectto;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send to any node.
#line 4781
# node_type is a type attribute for all node types.
#line 4781
#
#line 4781
allow user_crond_t node_type:node { tcp_send udp_send rawip_send };
#line 4781

#line 4781
#
#line 4781
# Allow packets sent by the domain to be received from any node.
#line 4781
#
#line 4781
allow user_crond_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send NFS client requests via the socket
#line 4781
# created by mount.
#line 4781
#
#line 4781
allow user_crond_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781

#line 4781
#
#line 4781
# Bind to the default port type.
#line 4781
# Other port types must be separately authorized.
#line 4781
#
#line 4781
allow user_crond_t port_t:udp_socket name_bind;
#line 4781
allow user_crond_t port_t:tcp_socket name_bind;
#line 4781

#line 4781
allow user_crond_t resolv_conf_t:file { getattr read };
#line 4781

#line 4781
allow user_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_crond_t ld_so_t:file execute_no_trans;
#line 4781
allow user_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_crond_t device_t:dir search;
#line 4781
allow user_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781
allow user_crond_t self:dir { read getattr lock search ioctl };
#line 4781
allow user_crond_t self:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781
allow user_crond_t self:fifo_file { ioctl read getattr lock write append };
#line 4781
allow user_crond_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_crond_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_crond_t etc_runtime_t:file { getattr read };
#line 4781
allow user_crond_t self:process { fork { sigchld sigkill sigstop signull signal } };
#line 4781
allow user_crond_t proc_t:dir { getattr search read };
#line 4781
allow user_crond_t proc_t:file { getattr read };
#line 4781

#line 4781
allow user_crond_t etc_t:lnk_file read;
#line 4781

#line 4781
allow user_crond_t locale_t:dir { read getattr lock search ioctl };
#line 4781
allow user_crond_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4781

#line 4781

#line 4781
allow user_crond_t sysctl_kernel_t:dir search;
#line 4781
allow user_crond_t sysctl_kernel_t:file { getattr read };
#line 4781
allow user_crond_t var_spool_t:dir search;
#line 4781
allow user_crond_t fs_t:filesystem getattr;
#line 4781

#line 4781
allow user_crond_t devtty_t:chr_file { read write };
#line 4781

#line 4781
# Use capabilities.
#line 4781
allow user_crond_t user_crond_t:capability dac_override;
#line 4781

#line 4781
# Inherit and use descriptors from initrc.
#line 4781
allow user_crond_t initrc_t:fd use;
#line 4781

#line 4781
# 
#line 4781
# Since crontab files are not directly executed,
#line 4781
# crond must ensure that the crontab file has
#line 4781
# a type that is appropriate for the domain of
#line 4781
# the user cron job.  It performs an entrypoint
#line 4781
# permission check for this purpose.
#line 4781
#
#line 4781
allow user_crond_t user_cron_spool_t:file entrypoint;
#line 4781

#line 4781
# Access user files and dirs.
#line 4781
allow user_crond_t home_root_t:dir search;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_crond_t user_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_crond_t user_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_crond_t user_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_crond_t user_home_dir_t:dir user_home_t;
#line 4781
type_transition user_crond_t user_home_dir_t:{ file lnk_file sock_file fifo_file } user_home_t;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
#
#line 4781
# Allow the process to modify the directory.
#line 4781
#
#line 4781
allow user_crond_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4781

#line 4781
#
#line 4781
# Allow the process to create the file.
#line 4781
#
#line 4781

#line 4781
allow user_crond_t user_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4781
allow user_crond_t user_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4781

#line 4781

#line 4781
type_transition user_crond_t tmp_t:dir user_tmp_t;
#line 4781
type_transition user_crond_t tmp_t:{ file lnk_file sock_file fifo_file } user_tmp_t;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Run helper programs.
#line 4781

#line 4781
allow user_crond_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_crond_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4781

#line 4781
allow user_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4781
allow user_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4781
allow user_crond_t ld_so_t:file execute_no_trans;
#line 4781
allow user_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4781
allow user_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4781
allow user_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4781
allow user_crond_t device_t:dir search;
#line 4781
allow user_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781

#line 4781

#line 4781
allow user_crond_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_crond_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_crond_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_crond_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
allow user_crond_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781

#line 4781
# Run scripts in user home directory.
#line 4781

#line 4781
allow user_crond_t user_home_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4781

#line 4781

#line 4781
# ps does not need to access /boot when run from cron
#line 4781
dontaudit user_crond_t boot_t:dir search;
#line 4781
# quiet other ps operations
#line 4781
dontaudit user_crond_t domain:dir getattr;
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Read the mouse.
#line 4781
allow user_t mouse_device_t:chr_file { read getattr lock ioctl };
#line 4781
# Access other miscellaneous devices.
#line 4781
allow user_t misc_device_t:{ file lnk_file sock_file fifo_file chr_file blk_file } { ioctl read getattr lock write append };
#line 4781

#line 4781
# Use the network.
#line 4781

#line 4781
#
#line 4781
# Allow the domain to create and use UDP and TCP sockets.
#line 4781
# Other kinds of sockets must be separately authorized for use.
#line 4781
allow user_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781
allow user_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send UDP packets.
#line 4781
# Since the destination sockets type is unknown, the generic
#line 4781
# any_socket_t type is used as a placeholder.
#line 4781
#
#line 4781
allow user_t any_socket_t:udp_socket sendto;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send using any network interface.
#line 4781
# netif_type is a type attribute for all network interface types.
#line 4781
#
#line 4781
allow user_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4781

#line 4781
#
#line 4781
# Allow packets sent by the domain to be received on any network interface.
#line 4781
#
#line 4781
allow user_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to receive packets from any network interface.
#line 4781
# netmsg_type is a type attribute for all default message types.
#line 4781
#
#line 4781
allow user_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to initiate or accept TCP connections 
#line 4781
# on any network interface.
#line 4781
#
#line 4781
allow user_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4781

#line 4781
#
#line 4781
# Receive resets from the TCP reset socket.
#line 4781
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4781
#
#line 4781
allow user_t tcp_socket_t:tcp_socket recvfrom;
#line 4781

#line 4781
dontaudit user_t tcp_socket_t:tcp_socket connectto;
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send to any node.
#line 4781
# node_type is a type attribute for all node types.
#line 4781
#
#line 4781
allow user_t node_type:node { tcp_send udp_send rawip_send };
#line 4781

#line 4781
#
#line 4781
# Allow packets sent by the domain to be received from any node.
#line 4781
#
#line 4781
allow user_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4781

#line 4781
#
#line 4781
# Allow the domain to send NFS client requests via the socket
#line 4781
# created by mount.
#line 4781
#
#line 4781
allow user_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4781

#line 4781
#
#line 4781
# Bind to the default port type.
#line 4781
# Other port types must be separately authorized.
#line 4781
#
#line 4781
allow user_t port_t:udp_socket name_bind;
#line 4781
allow user_t port_t:tcp_socket name_bind;
#line 4781

#line 4781
allow user_t resolv_conf_t:file { getattr read };
#line 4781
# for perl
#line 4781
dontaudit user_t resolv_conf_t:file ioctl;
#line 4781

#line 4781
#
#line 4781
# connect_secure and sendmsg_secure calls with a 
#line 4781
# peer or destination socket SID can be enforced
#line 4781
# when using the loopback interface.  Enforcement
#line 4781
# for real network interfaces will be possible when
#line 4781
# a packet labeling mechanism is integrated.
#line 4781
#
#line 4781
allow user_t node_lo_t:node enforce_dest;
#line 4781

#line 4781
# Communicate within the domain.
#line 4781

#line 4781
allow user_t user_t:udp_socket { sendto };
#line 4781
allow user_t user_t:udp_socket { recvfrom };
#line 4781

#line 4781

#line 4781
allow user_t user_t:tcp_socket { connectto recvfrom };
#line 4781
allow user_t user_t:tcp_socket { acceptfrom recvfrom };
#line 4781
allow user_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781
allow user_t tcp_socket_t:tcp_socket { recvfrom };
#line 4781

#line 4781

#line 4781
# Connect to inetd.
#line 4781

#line 4781

#line 4781

#line 4781

#line 4781
# Connect data port to ftpd.
#line 4781

#line 4781

#line 4781
# Connect to portmap.
#line 4781

#line 4781

#line 4781
# Inherit and use sockets from inetd
#line 4781

#line 4781

#line 4781
# Allow system log read
#line 4781
#allow user_t kernel_t:system syslog_read;
#line 4781
# else do not log it
#line 4781
dontaudit user_t kernel_t:system syslog_read;
#line 4781

#line 4781
# Very permissive allowing every domain to see every type.
#line 4781
allow user_t kernel_t:system { ipc_info };
#line 4781

#line 4781
# When the user domain runs ps, there will be a number of access
#line 4781
# denials when ps tries to search /proc.  Do not audit these denials.
#line 4781
dontaudit user_t domain:dir { read getattr lock search ioctl };
#line 4781
dontaudit user_t domain:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4781

#line 4781
# Some shells ask for w access to utmp, but will operate
#line 4781
# correctly without it.  Do not audit write denials to utmp.
#line 4781
dontaudit user_t initrc_var_run_t:file { getattr read write };
#line 4781

#line 4781
# do not audit getattr on tmpfile, otherwise ls -l /tmp fills the logs
#line 4781
dontaudit user_t tmpfile:{ dir file lnk_file sock_file fifo_file chr_file blk_file } getattr;
#line 4781

#line 4781
# do not audit getattr on disk devices, otherwise KDE fills the logs
#line 4781
dontaudit user_t { removable_device_t fixed_disk_device_t }:blk_file getattr;
#line 4781

#line 4781

#line 4781
# Access the sound device.
#line 4781
allow user_t sound_device_t:chr_file { getattr read write ioctl };
#line 4781

#line 4781

#line 4781

#line 4781
# Get attributes of file systems.
#line 4781
allow user_t fs_type:filesystem getattr;
#line 4781

#line 4781
# Read and write /dev/tty and /dev/null.
#line 4781
allow user_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4781
allow user_t zero_device_t:chr_file { { ioctl read getattr lock write append } execute };
#line 4781
allow user_t random_device_t:chr_file { getattr read };
#line 4781

#line 4781
# Read the devpts root directory.
#line 4781
allow user_t devpts_t:dir { read getattr lock search ioctl };
#line 4781

#line 4781

#line 4781
# When an ordinary user domain runs su, su may try to
#line 4781
# update the /root/.Xauthority file, and the user shell may
#line 4781
# try to update the shell history. This isnt allowed, but 
#line 4781
# we dont need to audit it.
#line 4781

#line 4781
dontaudit user_su_t sysadm_home_t:dir  { read getattr search write add_name remove_name };
#line 4781
dontaudit user_su_t sysadm_home_t:file { read getattr create write link unlink };
#line 4781
dontaudit user_t    sysadm_home_t:dir { read search getattr };
#line 4781
dontaudit user_t    sysadm_home_t:file { read getattr append };
#line 4781

#line 4781

#line 4781
# Some programs that are left in user_t will try to connect
#line 4781
# to syslogd, but we do not want to let them generate log messages.
#line 4781
# Do not audit.
#line 4781
dontaudit user_t devlog_t:sock_file { read write };
#line 4781
dontaudit user_t syslogd_t:unix_dgram_socket sendto;
#line 4781

#line 4781

#line 4781
# stop warnings about "ls -l" on directories with unlabelled files
#line 4781
dontaudit user_t file_t:{ dir file lnk_file } getattr;
#line 4781


#  Do not add any rules referring to user_t to this file!  That will break
#  support for multiple user roles.

# a role for staff that allows seeing all domains and control over the user_t
# domain

#line 4788

#line 4788
# user_t/staff_t is an unprivileged users domain.
#line 4788
type staff_t, domain, userdomain, unpriv_userdomain, web_client_domain;
#line 4788

#line 4788
# staff_r is authorized for staff_t for the initial login domain.
#line 4788
role staff_r types staff_t;
#line 4788
allow system_r staff_r;
#line 4788

#line 4788
# Grant permissions within the domain.
#line 4788

#line 4788
# Access other processes in the same domain.
#line 4788
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4788
# separately if desired.
#line 4788
allow staff_t self:process ~{ptrace setexec setfscreate};
#line 4788

#line 4788
# Access /proc/PID files for processes in the same domain.
#line 4788
allow staff_t self:dir { read getattr lock search ioctl };
#line 4788
allow staff_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Access file descriptions, pipes, and sockets
#line 4788
# created by processes in the same domain.
#line 4788
allow staff_t self:fd *;
#line 4788
allow staff_t self:fifo_file { ioctl read getattr lock write append };
#line 4788
allow staff_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788

#line 4788
# Allow the domain to communicate with other processes in the same domain.
#line 4788
allow staff_t self:unix_dgram_socket sendto;
#line 4788
allow staff_t self:unix_stream_socket connectto;
#line 4788

#line 4788
# Access System V IPC objects created by processes in the same domain.
#line 4788
allow staff_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4788
allow staff_t self:msg  { send receive };
#line 4788
allow staff_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4788
allow staff_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4788
;
#line 4788

#line 4788
# Grant read/search permissions to some of /proc.
#line 4788
allow staff_t proc_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t proc_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
# Read /.
#line 4788
allow staff_t root_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /home.
#line 4788
allow staff_t home_root_t:dir { read getattr lock search ioctl };
#line 4788

#line 4788
# Read /usr.
#line 4788
allow staff_t usr_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read bin and sbin directories.
#line 4788
allow staff_t bin_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788
allow staff_t sbin_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
# Read directories and files with the readable_t type.
#line 4788
# This type is a general type for "world"-readable files.
#line 4788
allow staff_t readable_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t readable_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Stat /...security and lost+found.
#line 4788
allow staff_t file_labels_t:dir getattr;
#line 4788
allow staff_t lost_found_t:dir getattr;
#line 4788

#line 4788

#line 4788

#line 4788
# Read the /tmp directory and any /tmp files with the base type.
#line 4788
# Temporary files created at runtime will typically use derived types.
#line 4788
allow staff_t tmp_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t tmp_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /var, /var/spool, /var/run.
#line 4788
allow staff_t var_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t var_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788
allow staff_t var_spool_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t var_spool_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788
allow staff_t var_run_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t var_run_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
allow staff_t sysctl_kernel_t:dir search;
#line 4788
allow staff_t sysctl_kernel_t:file { getattr read };
#line 4788

#line 4788
# Read /etc.
#line 4788
allow staff_t etc_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788
allow staff_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
# for running depmod as part of the kernel packaging process
#line 4788
allow staff_t modules_conf_t:file { getattr read };
#line 4788

#line 4788
# Read man directories and files.
#line 4788
allow staff_t man_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t man_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /dev directories and any symbolic links.
#line 4788
allow staff_t device_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t device_t:lnk_file { read getattr lock ioctl };
#line 4788

#line 4788
# Do not audit write denials to /etc/ld.so.cache.
#line 4788
dontaudit staff_t ld_so_cache_t:file write;
#line 4788

#line 4788
# Execute from the system shared libraries.
#line 4788

#line 4788
allow staff_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_t device_t:dir search;
#line 4788
allow staff_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788
;
#line 4788

#line 4788
# staff_t is also granted permissions specific to user domains.
#line 4788

#line 4788
# Use capabilities
#line 4788
allow staff_t self:capability { setgid chown fowner };
#line 4788
dontaudit staff_t self:capability { sys_nice fsetid };
#line 4788

#line 4788
# Type for home directory.
#line 4788

#line 4788
type staff_home_dir_t, file_type, sysadmfile, home_dir_type, user_home_dir_type, home_type, user_home_type;
#line 4788
type staff_home_t, file_type, sysadmfile, home_type, user_home_type;
#line 4788
# do not allow privhome access to sysadm_home_dir_t
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow privhome staff_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow privhome staff_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow privhome staff_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition privhome staff_home_dir_t:dir staff_home_t;
#line 4788
type_transition privhome staff_home_dir_t:{ file lnk_file sock_file fifo_file } staff_home_t;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
type staff_tmp_t, file_type, sysadmfile, tmpfile , user_tmpfile;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_t staff_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_t staff_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_t tmp_t:dir staff_tmp_t;
#line 4788
type_transition staff_t tmp_t:{ file lnk_file sock_file fifo_file } staff_tmp_t;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# allow ptrace
#line 4788
#can_ptrace(staff_t, staff_t)
#line 4788

#line 4788
# Create, access, and remove files in home directory.
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_t staff_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_t staff_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_t staff_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_t staff_home_dir_t:dir staff_home_t;
#line 4788
type_transition staff_t staff_home_dir_t:{ file lnk_file sock_file fifo_file } staff_home_t;
#line 4788

#line 4788

#line 4788

#line 4788
allow staff_t staff_home_t:{ dir file lnk_file sock_file fifo_file chr_file blk_file } { relabelfrom relabelto };
#line 4788

#line 4788
# Bind to a Unix domain socket in /tmp.
#line 4788
allow staff_t staff_tmp_t:unix_stream_socket name_bind;
#line 4788

#line 4788
# Type for tty devices.
#line 4788
type staff_tty_device_t, file_type, sysadmfile, ttyfile;
#line 4788
# Access ttys.
#line 4788
allow staff_t privfd:fd use;
#line 4788
allow staff_t staff_tty_device_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 4788
# Use the type when relabeling terminal devices.
#line 4788
type_change staff_t tty_device_t:chr_file staff_tty_device_t;
#line 4788

#line 4788
# Debian login is from shadow utils and does not allow resetting the perms.
#line 4788
# have to fix this!
#line 4788
type_change staff_t ttyfile:chr_file staff_tty_device_t;
#line 4788

#line 4788
# allow reading /var/lib/dpkg etc
#line 4788
allow staff_t var_lib_t:dir { read getattr lock search ioctl };
#line 4788

#line 4788
# Allow reading dpkg origins file
#line 4788

#line 4788
allow staff_t etc_dpkg_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t etc_dpkg_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788

#line 4788
# read localization information
#line 4788
allow staff_t locale_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t locale_t:{file lnk_file} { read getattr lock ioctl };
#line 4788

#line 4788
# for running TeX programs
#line 4788

#line 4788
allow staff_t tetex_data_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t tetex_data_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
allow staff_t tetex_data_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
# Type and access for pty devices.
#line 4788

#line 4788

#line 4788

#line 4788
# Access the pty master multiplexer.
#line 4788
allow staff_t ptmx_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788
allow staff_t devpts_t:filesystem getattr;
#line 4788

#line 4788
# allow searching /dev/pts
#line 4788
allow staff_t devpts_t:dir { getattr read search };
#line 4788

#line 4788
type staff_devpts_t, file_type, sysadmfile, ptyfile , userpty_type;
#line 4788

#line 4788
# Allow the pty to be associated with the file system.
#line 4788
allow staff_devpts_t devpts_t:filesystem associate;
#line 4788

#line 4788
# Label pty files with a derived type.
#line 4788
type_transition staff_t devpts_t:chr_file staff_devpts_t;
#line 4788

#line 4788
# Read and write my pty files.
#line 4788
allow staff_t staff_devpts_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 4788

#line 4788

#line 4788

#line 4788
# Use the type when relabeling pty devices.
#line 4788

#line 4788

#line 4788
type_change staff_t sshd_devpts_t:chr_file staff_devpts_t;
#line 4788

#line 4788
# Connect to sshd.
#line 4788

#line 4788
allow staff_t sshd_t:tcp_socket { connectto recvfrom };
#line 4788
allow sshd_t staff_t:tcp_socket { acceptfrom recvfrom };
#line 4788
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788
allow staff_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788

#line 4788

#line 4788
# Connect to ssh proxy.
#line 4788

#line 4788
allow staff_t staff_ssh_t:tcp_socket { connectto recvfrom };
#line 4788
allow staff_ssh_t staff_t:tcp_socket { acceptfrom recvfrom };
#line 4788
allow staff_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788
allow staff_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788

#line 4788

#line 4788
allow staff_t sshd_t:fd use;
#line 4788
allow staff_t sshd_t:tcp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788
# Use a Unix stream socket inherited from sshd.
#line 4788
allow staff_t sshd_t:unix_stream_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788

#line 4788
allow staff_t boot_t:dir { getattr search };
#line 4788
dontaudit staff_t boot_t:dir read;
#line 4788
dontaudit staff_t boot_t:lnk_file getattr;
#line 4788
dontaudit staff_t boot_t:file { getattr read };
#line 4788
allow staff_t system_map_t:file { getattr read };
#line 4788

#line 4788

#line 4788
type staff_tmpfs_t, file_type, sysadmfile, tmpfsfile;
#line 4788
# Use this type when creating tmpfs/shm objects.
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_t tmpfs_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_t staff_tmpfs_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_t staff_tmpfs_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_t tmpfs_t:dir staff_tmpfs_t;
#line 4788
type_transition staff_t tmpfs_t:{ file lnk_file sock_file fifo_file } staff_tmpfs_t;
#line 4788

#line 4788

#line 4788

#line 4788
allow staff_tmpfs_t tmpfs_t:filesystem associate;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Read and write /var/catman.
#line 4788
allow staff_t catman_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788
allow staff_t catman_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788

#line 4788
# Modify mail spool file.
#line 4788
allow staff_t mail_spool_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_t mail_spool_t:file { ioctl read getattr lock write append };
#line 4788
allow staff_t mail_spool_t:lnk_file read;
#line 4788

#line 4788
#
#line 4788
# Allow the query of filesystem quotas
#line 4788
#
#line 4788
allow staff_t fs_type:filesystem quotaget;
#line 4788

#line 4788
# Run helper programs.
#line 4788

#line 4788
allow staff_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4788

#line 4788
allow staff_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_t device_t:dir search;
#line 4788
allow staff_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788
allow staff_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
# Run programs developed by other users in the same domain.
#line 4788

#line 4788
allow staff_t staff_home_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_t staff_tmp_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
# Run user programs that require different permissions in their own domain.
#line 4788
# These rules were moved into the individual program domains.
#line 4788

#line 4788
# Instantiate derived domains for a number of programs.
#line 4788
# These derived domains encode both information about the calling
#line 4788
# user domain and the program, and allow us to maintain separation
#line 4788
# between different instances of the program being run by different
#line 4788
# user domains.
#line 4788

#line 4788

#line 4788

#line 4788
# Derived domain based on the calling user domain and the program.
#line 4788
type staff_su_t, domain, privlog, auth;
#line 4788

#line 4788
# Transition from the user domain to this domain.
#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow staff_t staff_su_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit staff_t staff_su_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow staff_t su_exec_t:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_su_t staff_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_su_t staff_t:fd use;
#line 4788
allow staff_t staff_su_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_su_t staff_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_su_t su_exec_t:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_su_t su_exec_t:file entrypoint;
#line 4788

#line 4788
type_transition staff_t su_exec_t:process staff_su_t;
#line 4788

#line 4788

#line 4788

#line 4788
allow staff_su_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_su_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_su_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_su_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_su_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_su_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_su_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_su_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_su_t device_t:dir search;
#line 4788
allow staff_su_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
allow staff_su_t etc_t:file { getattr read };
#line 4788

#line 4788
allow staff_su_t etc_t:lnk_file read;
#line 4788

#line 4788
allow staff_su_t locale_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_su_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
allow staff_su_t self:unix_dgram_socket { connect create write };
#line 4788
allow staff_su_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788
allow staff_su_t proc_t:dir search;
#line 4788
allow staff_su_t proc_t:lnk_file read;
#line 4788

#line 4788
allow staff_su_t self:dir { read getattr lock search ioctl };
#line 4788
allow staff_su_t self:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
allow staff_su_t self:process setsched;
#line 4788
allow staff_su_t device_t:dir search;
#line 4788
allow staff_su_t self:process { fork sigchld };
#line 4788

#line 4788
# Use capabilities.
#line 4788
allow staff_su_t self:capability { setuid setgid net_bind_service chown dac_override sys_nice sys_resource };
#line 4788

#line 4788
# Revert to the user domain when a shell is executed.
#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow staff_su_t staff_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit staff_su_t staff_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow staff_su_t shell_exec_t:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_t staff_su_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_t staff_su_t:fd use;
#line 4788
allow staff_su_t staff_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_t staff_su_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_t shell_exec_t:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_t shell_exec_t:file entrypoint;
#line 4788

#line 4788
type_transition staff_su_t shell_exec_t:process staff_t;
#line 4788

#line 4788
allow staff_su_t bin_t:dir search;
#line 4788
allow staff_su_t bin_t:lnk_file read;
#line 4788

#line 4788
allow staff_su_t privfd:fd use;
#line 4788

#line 4788
# Write to utmp.
#line 4788
allow staff_su_t { var_t var_run_t }:dir search;
#line 4788
allow staff_su_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788

#line 4788
# Run chkpwd.
#line 4788

#line 4788
allow staff_su_t chkpwd_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788

#line 4788
# Inherit and use descriptors from gnome-pty-helper.
#line 4788

#line 4788

#line 4788
# The user role is authorized for this domain.
#line 4788
role staff_r types staff_su_t;
#line 4788

#line 4788
# Write to the user domain tty.
#line 4788
allow staff_su_t staff_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_su_t staff_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
allow staff_su_t { home_root_t staff_home_dir_t }:dir search;
#line 4788

#line 4788
# Modify .Xauthority file (via xauth program).
#line 4788

#line 4788

#line 4788
# Access sshd cookie files.
#line 4788
allow staff_su_t sshd_tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788
allow staff_su_t sshd_tmp_t:file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_su_t sshd_tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_su_t staff_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_su_t staff_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_su_t sshd_tmp_t:dir staff_tmp_t;
#line 4788
type_transition staff_su_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } staff_tmp_t;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# stop su complaining if you run it from a directory with restrictive perms
#line 4788
dontaudit staff_su_t file_type:dir search;
#line 4788

#line 4788

#line 4788
# Derived domain based on the calling user domain and the program.
#line 4788
type staff_chkpwd_t, domain, privlog, auth;
#line 4788

#line 4788
# Transition from the user domain to this domain.
#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow staff_t staff_chkpwd_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit staff_t staff_chkpwd_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow staff_t chkpwd_exec_t:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_chkpwd_t staff_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_chkpwd_t staff_t:fd use;
#line 4788
allow staff_t staff_chkpwd_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_chkpwd_t staff_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_chkpwd_t chkpwd_exec_t:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_chkpwd_t chkpwd_exec_t:file entrypoint;
#line 4788

#line 4788
type_transition staff_t chkpwd_exec_t:process staff_chkpwd_t;
#line 4788

#line 4788

#line 4788
# The user role is authorized for this domain.
#line 4788
role staff_r types staff_chkpwd_t;
#line 4788

#line 4788

#line 4788
allow staff_chkpwd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_chkpwd_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_chkpwd_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_chkpwd_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_chkpwd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_chkpwd_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_chkpwd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_chkpwd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_chkpwd_t device_t:dir search;
#line 4788
allow staff_chkpwd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
allow staff_chkpwd_t etc_t:file { getattr read };
#line 4788
allow staff_chkpwd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_chkpwd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788

#line 4788
allow staff_chkpwd_t etc_t:lnk_file read;
#line 4788

#line 4788
allow staff_chkpwd_t locale_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_chkpwd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788

#line 4788
# Use capabilities.
#line 4788
allow staff_chkpwd_t self:capability setuid;
#line 4788

#line 4788
# Inherit and use descriptors from gnome-pty-helper.
#line 4788

#line 4788

#line 4788
# Inherit and use descriptors from newrole.
#line 4788
allow staff_chkpwd_t newrole_t:fd use;
#line 4788

#line 4788
# Write to the user domain tty.
#line 4788
allow staff_chkpwd_t staff_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_chkpwd_t staff_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Derived domain based on the calling user domain and the program.
#line 4788
type staff_crontab_t, domain, privlog;
#line 4788

#line 4788
# Transition from the user domain to the derived domain.
#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow staff_t staff_crontab_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit staff_t staff_crontab_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow staff_t crontab_exec_t:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_crontab_t staff_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_crontab_t staff_t:fd use;
#line 4788
allow staff_t staff_crontab_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_crontab_t staff_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_crontab_t crontab_exec_t:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_crontab_t crontab_exec_t:file entrypoint;
#line 4788

#line 4788
type_transition staff_t crontab_exec_t:process staff_crontab_t;
#line 4788

#line 4788

#line 4788
# The user role is authorized for this domain.
#line 4788
role staff_r types staff_crontab_t;
#line 4788

#line 4788

#line 4788
allow staff_crontab_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_crontab_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crontab_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_crontab_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_crontab_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crontab_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_crontab_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crontab_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_crontab_t device_t:dir search;
#line 4788
allow staff_crontab_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
allow staff_crontab_t etc_t:file { getattr read };
#line 4788
allow staff_crontab_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_crontab_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788

#line 4788
allow staff_crontab_t etc_t:lnk_file read;
#line 4788

#line 4788
allow staff_crontab_t locale_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_crontab_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788

#line 4788
# Use capabilities
#line 4788
allow staff_crontab_t staff_crontab_t:capability { setuid setgid chown };
#line 4788

#line 4788
# Type for temporary files.
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_crontab_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_crontab_t staff_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_crontab_t staff_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_crontab_t tmp_t:dir staff_tmp_t;
#line 4788
type_transition staff_crontab_t tmp_t:{ file lnk_file sock_file fifo_file } staff_tmp_t;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Type of user crontabs once moved to cron spool.
#line 4788
type staff_cron_spool_t, file_type, sysadmfile;
#line 4788
# Use the type when creating files in /var/spool/cron.
#line 4788
allow sysadm_crontab_t staff_cron_spool_t:file { getattr read };
#line 4788
allow staff_crontab_t { var_t var_spool_t }:dir search;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_crontab_t cron_spool_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_crontab_t staff_cron_spool_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_crontab_t staff_cron_spool_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_crontab_t cron_spool_t:dir staff_cron_spool_t;
#line 4788
type_transition staff_crontab_t cron_spool_t:{ file lnk_file sock_file fifo_file } staff_cron_spool_t;
#line 4788

#line 4788

#line 4788

#line 4788
allow staff_crontab_t self:process { fork { sigchld sigkill sigstop signull signal } };
#line 4788

#line 4788
# crontab signals crond by updating the mtime on the spooldir
#line 4788
allow staff_crontab_t cron_spool_t:dir setattr;
#line 4788
# Allow crond to read those crontabs in cron spool.
#line 4788
allow crond_t staff_cron_spool_t:file { read getattr lock ioctl };
#line 4788

#line 4788
# Run helper programs as staff_t
#line 4788
allow staff_crontab_t { bin_t sbin_t }:dir search;
#line 4788
allow staff_crontab_t bin_t:lnk_file read;
#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow staff_crontab_t staff_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit staff_crontab_t staff_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow staff_crontab_t { bin_t sbin_t shell_exec_t }:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_t staff_crontab_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_t staff_crontab_t:fd use;
#line 4788
allow staff_crontab_t staff_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_t staff_crontab_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_t { bin_t sbin_t shell_exec_t }:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_t { bin_t sbin_t shell_exec_t }:file entrypoint;
#line 4788

#line 4788
type_transition staff_crontab_t { bin_t sbin_t shell_exec_t }:process staff_t;
#line 4788

#line 4788

#line 4788
# Read user crontabs 
#line 4788
allow staff_crontab_t { staff_home_t staff_home_dir_t }:dir { read getattr lock search ioctl };  
#line 4788
allow staff_crontab_t staff_home_t:file { read getattr lock ioctl };  
#line 4788
dontaudit staff_crontab_t staff_home_dir_t:dir write;
#line 4788

#line 4788
# Access the cron log file.
#line 4788
allow staff_crontab_t cron_log_t:file { read getattr lock ioctl };
#line 4788
allow staff_crontab_t cron_log_t:file { append };
#line 4788

#line 4788
# Access terminals.
#line 4788
allow staff_crontab_t device_t:dir search;
#line 4788
allow staff_crontab_t staff_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_crontab_t staff_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
# Inherit and use descriptors from gnome-pty-helper.
#line 4788

#line 4788
allow staff_crontab_t privfd:fd use;
#line 4788

#line 4788

#line 4788
# Derived domain based on the calling user domain and the program.
#line 4788
type staff_ssh_t, domain, privlog;
#line 4788

#line 4788
# Transition from the user domain to the derived domain.
#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow staff_t staff_ssh_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit staff_t staff_ssh_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow staff_t ssh_exec_t:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_ssh_t staff_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_ssh_t staff_t:fd use;
#line 4788
allow staff_t staff_ssh_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_ssh_t staff_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_ssh_t ssh_exec_t:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_ssh_t ssh_exec_t:file entrypoint;
#line 4788

#line 4788
type_transition staff_t ssh_exec_t:process staff_ssh_t;
#line 4788

#line 4788

#line 4788
# The user role is authorized for this domain.
#line 4788
role staff_r types staff_ssh_t;
#line 4788

#line 4788
# Grant permissions within the domain.
#line 4788

#line 4788
# Access other processes in the same domain.
#line 4788
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4788
# separately if desired.
#line 4788
allow staff_ssh_t self:process ~{ptrace setexec setfscreate};
#line 4788

#line 4788
# Access /proc/PID files for processes in the same domain.
#line 4788
allow staff_ssh_t self:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Access file descriptions, pipes, and sockets
#line 4788
# created by processes in the same domain.
#line 4788
allow staff_ssh_t self:fd *;
#line 4788
allow staff_ssh_t self:fifo_file { ioctl read getattr lock write append };
#line 4788
allow staff_ssh_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_ssh_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788

#line 4788
# Allow the domain to communicate with other processes in the same domain.
#line 4788
allow staff_ssh_t self:unix_dgram_socket sendto;
#line 4788
allow staff_ssh_t self:unix_stream_socket connectto;
#line 4788

#line 4788
# Access System V IPC objects created by processes in the same domain.
#line 4788
allow staff_ssh_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4788
allow staff_ssh_t self:msg  { send receive };
#line 4788
allow staff_ssh_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4788
allow staff_ssh_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4788

#line 4788

#line 4788
# Use descriptors created by sshd
#line 4788
allow staff_ssh_t privfd:fd use;
#line 4788

#line 4788

#line 4788
allow staff_ssh_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_ssh_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_ssh_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_ssh_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t device_t:dir search;
#line 4788
allow staff_ssh_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788
allow staff_ssh_t etc_t:lnk_file read;
#line 4788

#line 4788
allow staff_ssh_t locale_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788

#line 4788
# Get attributes of file systems.
#line 4788
allow staff_ssh_t fs_type:filesystem getattr;
#line 4788

#line 4788

#line 4788
# Read /.
#line 4788
allow staff_ssh_t root_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /home.
#line 4788
allow staff_ssh_t home_root_t:dir { read getattr lock search ioctl };
#line 4788

#line 4788
# Read /usr.
#line 4788
allow staff_ssh_t usr_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read bin and sbin directories.
#line 4788
allow staff_ssh_t bin_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788
allow staff_ssh_t sbin_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
# Read the devpts root directory.
#line 4788
allow staff_ssh_t devpts_t:dir { read getattr lock search ioctl };
#line 4788

#line 4788
# Read /var.
#line 4788
allow staff_ssh_t var_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t var_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /var/run, /var/log.
#line 4788
allow staff_ssh_t var_run_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t var_run_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788
allow staff_ssh_t var_log_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t var_log_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /etc.
#line 4788
allow staff_ssh_t etc_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788
allow staff_ssh_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788
allow staff_ssh_t resolv_conf_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read the linker, shared library, and executable types.
#line 4788
allow staff_ssh_t ld_so_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788
allow staff_ssh_t shlib_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788
allow staff_ssh_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Read /dev directories and any symbolic links.
#line 4788
allow staff_ssh_t device_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t device_t:lnk_file { read getattr lock ioctl };
#line 4788

#line 4788
# Read /dev/random.
#line 4788
allow staff_ssh_t random_device_t:chr_file { read getattr lock ioctl };
#line 4788

#line 4788
# Read and write /dev/tty and /dev/null.
#line 4788
allow staff_ssh_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_ssh_t { null_device_t zero_device_t }:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
# Grant permissions needed to create TCP and UDP sockets and
#line 4788
# to access the network.
#line 4788

#line 4788
#
#line 4788
# Allow the domain to create and use UDP and TCP sockets.
#line 4788
# Other kinds of sockets must be separately authorized for use.
#line 4788
allow staff_ssh_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_ssh_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send UDP packets.
#line 4788
# Since the destination sockets type is unknown, the generic
#line 4788
# any_socket_t type is used as a placeholder.
#line 4788
#
#line 4788
allow staff_ssh_t any_socket_t:udp_socket sendto;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send using any network interface.
#line 4788
# netif_type is a type attribute for all network interface types.
#line 4788
#
#line 4788
allow staff_ssh_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4788

#line 4788
#
#line 4788
# Allow packets sent by the domain to be received on any network interface.
#line 4788
#
#line 4788
allow staff_ssh_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to receive packets from any network interface.
#line 4788
# netmsg_type is a type attribute for all default message types.
#line 4788
#
#line 4788
allow staff_ssh_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to initiate or accept TCP connections 
#line 4788
# on any network interface.
#line 4788
#
#line 4788
allow staff_ssh_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4788

#line 4788
#
#line 4788
# Receive resets from the TCP reset socket.
#line 4788
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4788
#
#line 4788
allow staff_ssh_t tcp_socket_t:tcp_socket recvfrom;
#line 4788

#line 4788
dontaudit staff_ssh_t tcp_socket_t:tcp_socket connectto;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send to any node.
#line 4788
# node_type is a type attribute for all node types.
#line 4788
#
#line 4788
allow staff_ssh_t node_type:node { tcp_send udp_send rawip_send };
#line 4788

#line 4788
#
#line 4788
# Allow packets sent by the domain to be received from any node.
#line 4788
#
#line 4788
allow staff_ssh_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send NFS client requests via the socket
#line 4788
# created by mount.
#line 4788
#
#line 4788
allow staff_ssh_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788

#line 4788
#
#line 4788
# Bind to the default port type.
#line 4788
# Other port types must be separately authorized.
#line 4788
#
#line 4788
allow staff_ssh_t port_t:udp_socket name_bind;
#line 4788
allow staff_ssh_t port_t:tcp_socket name_bind;
#line 4788

#line 4788

#line 4788
# for sshing to a ssh tunnel
#line 4788

#line 4788
allow staff_ssh_t staff_ssh_t:tcp_socket { connectto recvfrom };
#line 4788
allow staff_ssh_t staff_ssh_t:tcp_socket { acceptfrom recvfrom };
#line 4788
allow staff_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788
allow staff_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788

#line 4788

#line 4788
# Use capabilities.
#line 4788
allow staff_ssh_t self:capability { setuid setgid dac_override dac_read_search };
#line 4788

#line 4788
# Run helpers.
#line 4788

#line 4788
allow staff_ssh_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4788

#line 4788
allow staff_ssh_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_ssh_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_ssh_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_ssh_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_ssh_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_ssh_t device_t:dir search;
#line 4788
allow staff_ssh_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788
allow staff_ssh_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_ssh_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_ssh_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_ssh_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_ssh_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788

#line 4788
# Read the ssh key file.
#line 4788
allow staff_ssh_t sshd_key_t:file { read getattr lock ioctl };
#line 4788

#line 4788
# Access the ssh temporary files.
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_ssh_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_ssh_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_ssh_t sshd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_ssh_t tmp_t:dir sshd_tmp_t;
#line 4788
type_transition staff_ssh_t tmp_t:{ file lnk_file sock_file fifo_file } sshd_tmp_t;
#line 4788

#line 4788

#line 4788

#line 4788
allow staff_ssh_t staff_tmp_t:dir { read getattr lock search ioctl };
#line 4788

#line 4788
# for rsync
#line 4788
allow staff_ssh_t staff_t:unix_stream_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788

#line 4788
# Access the users .ssh directory.
#line 4788
type staff_home_ssh_t, file_type, sysadmfile;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow { sysadm_ssh_t staff_ssh_t } staff_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow { sysadm_ssh_t staff_ssh_t } staff_home_ssh_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow { sysadm_ssh_t staff_ssh_t } staff_home_ssh_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition { sysadm_ssh_t staff_ssh_t } staff_home_dir_t:dir staff_home_ssh_t;
#line 4788
type_transition { sysadm_ssh_t staff_ssh_t } staff_home_dir_t:{ file lnk_file sock_file fifo_file } staff_home_ssh_t;
#line 4788

#line 4788

#line 4788

#line 4788
allow { sysadm_ssh_t staff_ssh_t } staff_home_ssh_t:lnk_file { getattr read };
#line 4788
dontaudit staff_ssh_t staff_home_t:dir search;
#line 4788

#line 4788
allow sshd_t staff_home_ssh_t:dir { read getattr lock search ioctl };
#line 4788
allow sshd_t staff_home_ssh_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
allow staff_t staff_home_ssh_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788
allow staff_t staff_home_ssh_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788

#line 4788

#line 4788
# for /bin/sh used to execute xauth
#line 4788
dontaudit staff_ssh_t proc_t:dir search;
#line 4788

#line 4788
# Inherit and use descriptors from gnome-pty-helper.
#line 4788

#line 4788

#line 4788
# Connect to sshd.
#line 4788

#line 4788

#line 4788
allow staff_ssh_t sshd_t:tcp_socket { connectto recvfrom };
#line 4788
allow sshd_t staff_ssh_t:tcp_socket { acceptfrom recvfrom };
#line 4788
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788
allow staff_ssh_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788

#line 4788

#line 4788

#line 4788
# Write to the user domain tty.
#line 4788
allow staff_ssh_t staff_tty_device_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_ssh_t staff_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788
# Allow the user shell to signal the ssh program.
#line 4788
allow staff_t staff_ssh_t:process signal;
#line 4788
# allow ps to show ssh
#line 4788
allow staff_t staff_ssh_t:dir { search getattr read };
#line 4788
allow staff_t staff_ssh_t:{ file lnk_file } { read getattr };
#line 4788

#line 4788
# Allow the ssh program to communicate with ssh-agent.
#line 4788
allow staff_ssh_t staff_tmp_t:sock_file write;
#line 4788
allow staff_ssh_t staff_t:unix_stream_socket connectto;
#line 4788
allow staff_ssh_t sshd_t:unix_stream_socket connectto;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Instantiate a derived domain for user cron jobs.
#line 4788

#line 4788
# Derived domain for user cron jobs, user user_crond_domain if not system
#line 4788

#line 4788
type staff_crond_t, domain, user_crond_domain;
#line 4788

#line 4788

#line 4788
# Permit a transition from the crond_t domain to this domain.
#line 4788
# The transition is requested explicitly by the modified crond 
#line 4788
# via execve_secure.  There is no way to set up an automatic
#line 4788
# transition, since crontabs are configuration files, not executables.
#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to transition to the new domain.
#line 4788
#
#line 4788
allow crond_t staff_crond_t:process transition;
#line 4788

#line 4788
#
#line 4788
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4788
#
#line 4788
dontaudit crond_t staff_crond_t:process noatsecure;
#line 4788

#line 4788
#
#line 4788
# Allow the process to execute the program.
#line 4788
# 
#line 4788
allow crond_t shell_exec_t:file { read { getattr execute } };
#line 4788

#line 4788
#
#line 4788
# Allow the process to reap the new domain.
#line 4788
#
#line 4788
allow staff_crond_t crond_t:process sigchld;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to inherit and use file 
#line 4788
# descriptions from the creating process and vice versa.
#line 4788
#
#line 4788
allow staff_crond_t crond_t:fd use;
#line 4788
allow crond_t staff_crond_t:fd use;
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to write back to the old domain via a pipe.
#line 4788
#
#line 4788
allow staff_crond_t crond_t:fifo_file { ioctl read getattr lock write append };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to read and execute the program.
#line 4788
#
#line 4788
allow staff_crond_t shell_exec_t:file { read getattr lock execute ioctl };
#line 4788

#line 4788
#
#line 4788
# Allow the new domain to be entered via the program.
#line 4788
#
#line 4788
allow staff_crond_t shell_exec_t:file entrypoint;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# The user role is authorized for this domain.
#line 4788
role staff_r types staff_crond_t;
#line 4788

#line 4788
# This domain is granted permissions common to most domains.
#line 4788

#line 4788
#
#line 4788
# Allow the domain to create and use UDP and TCP sockets.
#line 4788
# Other kinds of sockets must be separately authorized for use.
#line 4788
allow staff_crond_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_crond_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send UDP packets.
#line 4788
# Since the destination sockets type is unknown, the generic
#line 4788
# any_socket_t type is used as a placeholder.
#line 4788
#
#line 4788
allow staff_crond_t any_socket_t:udp_socket sendto;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send using any network interface.
#line 4788
# netif_type is a type attribute for all network interface types.
#line 4788
#
#line 4788
allow staff_crond_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4788

#line 4788
#
#line 4788
# Allow packets sent by the domain to be received on any network interface.
#line 4788
#
#line 4788
allow staff_crond_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to receive packets from any network interface.
#line 4788
# netmsg_type is a type attribute for all default message types.
#line 4788
#
#line 4788
allow staff_crond_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to initiate or accept TCP connections 
#line 4788
# on any network interface.
#line 4788
#
#line 4788
allow staff_crond_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4788

#line 4788
#
#line 4788
# Receive resets from the TCP reset socket.
#line 4788
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4788
#
#line 4788
allow staff_crond_t tcp_socket_t:tcp_socket recvfrom;
#line 4788

#line 4788
dontaudit staff_crond_t tcp_socket_t:tcp_socket connectto;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send to any node.
#line 4788
# node_type is a type attribute for all node types.
#line 4788
#
#line 4788
allow staff_crond_t node_type:node { tcp_send udp_send rawip_send };
#line 4788

#line 4788
#
#line 4788
# Allow packets sent by the domain to be received from any node.
#line 4788
#
#line 4788
allow staff_crond_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send NFS client requests via the socket
#line 4788
# created by mount.
#line 4788
#
#line 4788
allow staff_crond_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788

#line 4788
#
#line 4788
# Bind to the default port type.
#line 4788
# Other port types must be separately authorized.
#line 4788
#
#line 4788
allow staff_crond_t port_t:udp_socket name_bind;
#line 4788
allow staff_crond_t port_t:tcp_socket name_bind;
#line 4788

#line 4788
allow staff_crond_t resolv_conf_t:file { getattr read };
#line 4788

#line 4788
allow staff_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_crond_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_crond_t device_t:dir search;
#line 4788
allow staff_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788
allow staff_crond_t self:dir { read getattr lock search ioctl };
#line 4788
allow staff_crond_t self:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788
allow staff_crond_t self:fifo_file { ioctl read getattr lock write append };
#line 4788
allow staff_crond_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_crond_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_crond_t etc_runtime_t:file { getattr read };
#line 4788
allow staff_crond_t self:process { fork { sigchld sigkill sigstop signull signal } };
#line 4788
allow staff_crond_t proc_t:dir { getattr search read };
#line 4788
allow staff_crond_t proc_t:file { getattr read };
#line 4788

#line 4788
allow staff_crond_t etc_t:lnk_file read;
#line 4788

#line 4788
allow staff_crond_t locale_t:dir { read getattr lock search ioctl };
#line 4788
allow staff_crond_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4788

#line 4788

#line 4788
allow staff_crond_t sysctl_kernel_t:dir search;
#line 4788
allow staff_crond_t sysctl_kernel_t:file { getattr read };
#line 4788
allow staff_crond_t var_spool_t:dir search;
#line 4788
allow staff_crond_t fs_t:filesystem getattr;
#line 4788

#line 4788
allow staff_crond_t devtty_t:chr_file { read write };
#line 4788

#line 4788
# Use capabilities.
#line 4788
allow staff_crond_t staff_crond_t:capability dac_override;
#line 4788

#line 4788
# Inherit and use descriptors from initrc.
#line 4788
allow staff_crond_t initrc_t:fd use;
#line 4788

#line 4788
# 
#line 4788
# Since crontab files are not directly executed,
#line 4788
# crond must ensure that the crontab file has
#line 4788
# a type that is appropriate for the domain of
#line 4788
# the user cron job.  It performs an entrypoint
#line 4788
# permission check for this purpose.
#line 4788
#
#line 4788
allow staff_crond_t staff_cron_spool_t:file entrypoint;
#line 4788

#line 4788
# Access user files and dirs.
#line 4788
allow staff_crond_t home_root_t:dir search;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_crond_t staff_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_crond_t staff_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_crond_t staff_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_crond_t staff_home_dir_t:dir staff_home_t;
#line 4788
type_transition staff_crond_t staff_home_dir_t:{ file lnk_file sock_file fifo_file } staff_home_t;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
#
#line 4788
# Allow the process to modify the directory.
#line 4788
#
#line 4788
allow staff_crond_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4788

#line 4788
#
#line 4788
# Allow the process to create the file.
#line 4788
#
#line 4788

#line 4788
allow staff_crond_t staff_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4788
allow staff_crond_t staff_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4788

#line 4788

#line 4788
type_transition staff_crond_t tmp_t:dir staff_tmp_t;
#line 4788
type_transition staff_crond_t tmp_t:{ file lnk_file sock_file fifo_file } staff_tmp_t;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Run helper programs.
#line 4788

#line 4788
allow staff_crond_t { bin_t sbin_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_crond_t { bin_t sbin_t etc_t }:lnk_file { getattr read };
#line 4788

#line 4788
allow staff_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4788
allow staff_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_crond_t ld_so_t:file execute_no_trans;
#line 4788
allow staff_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4788
allow staff_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4788
allow staff_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4788
allow staff_crond_t device_t:dir search;
#line 4788
allow staff_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788

#line 4788

#line 4788
allow staff_crond_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_crond_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_crond_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_crond_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
allow staff_crond_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788

#line 4788
# Run scripts in user home directory.
#line 4788

#line 4788
allow staff_crond_t staff_home_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4788

#line 4788

#line 4788
# ps does not need to access /boot when run from cron
#line 4788
dontaudit staff_crond_t boot_t:dir search;
#line 4788
# quiet other ps operations
#line 4788
dontaudit staff_crond_t domain:dir getattr;
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Read the mouse.
#line 4788
allow staff_t mouse_device_t:chr_file { read getattr lock ioctl };
#line 4788
# Access other miscellaneous devices.
#line 4788
allow staff_t misc_device_t:{ file lnk_file sock_file fifo_file chr_file blk_file } { ioctl read getattr lock write append };
#line 4788

#line 4788
# Use the network.
#line 4788

#line 4788
#
#line 4788
# Allow the domain to create and use UDP and TCP sockets.
#line 4788
# Other kinds of sockets must be separately authorized for use.
#line 4788
allow staff_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788
allow staff_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send UDP packets.
#line 4788
# Since the destination sockets type is unknown, the generic
#line 4788
# any_socket_t type is used as a placeholder.
#line 4788
#
#line 4788
allow staff_t any_socket_t:udp_socket sendto;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send using any network interface.
#line 4788
# netif_type is a type attribute for all network interface types.
#line 4788
#
#line 4788
allow staff_t netif_type:netif { tcp_send udp_send rawip_send };
#line 4788

#line 4788
#
#line 4788
# Allow packets sent by the domain to be received on any network interface.
#line 4788
#
#line 4788
allow staff_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to receive packets from any network interface.
#line 4788
# netmsg_type is a type attribute for all default message types.
#line 4788
#
#line 4788
allow staff_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to initiate or accept TCP connections 
#line 4788
# on any network interface.
#line 4788
#
#line 4788
allow staff_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 4788

#line 4788
#
#line 4788
# Receive resets from the TCP reset socket.
#line 4788
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 4788
#
#line 4788
allow staff_t tcp_socket_t:tcp_socket recvfrom;
#line 4788

#line 4788
dontaudit staff_t tcp_socket_t:tcp_socket connectto;
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send to any node.
#line 4788
# node_type is a type attribute for all node types.
#line 4788
#
#line 4788
allow staff_t node_type:node { tcp_send udp_send rawip_send };
#line 4788

#line 4788
#
#line 4788
# Allow packets sent by the domain to be received from any node.
#line 4788
#
#line 4788
allow staff_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 4788

#line 4788
#
#line 4788
# Allow the domain to send NFS client requests via the socket
#line 4788
# created by mount.
#line 4788
#
#line 4788
allow staff_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4788

#line 4788
#
#line 4788
# Bind to the default port type.
#line 4788
# Other port types must be separately authorized.
#line 4788
#
#line 4788
allow staff_t port_t:udp_socket name_bind;
#line 4788
allow staff_t port_t:tcp_socket name_bind;
#line 4788

#line 4788
allow staff_t resolv_conf_t:file { getattr read };
#line 4788
# for perl
#line 4788
dontaudit staff_t resolv_conf_t:file ioctl;
#line 4788

#line 4788
#
#line 4788
# connect_secure and sendmsg_secure calls with a 
#line 4788
# peer or destination socket SID can be enforced
#line 4788
# when using the loopback interface.  Enforcement
#line 4788
# for real network interfaces will be possible when
#line 4788
# a packet labeling mechanism is integrated.
#line 4788
#
#line 4788
allow staff_t node_lo_t:node enforce_dest;
#line 4788

#line 4788
# Communicate within the domain.
#line 4788

#line 4788
allow staff_t staff_t:udp_socket { sendto };
#line 4788
allow staff_t staff_t:udp_socket { recvfrom };
#line 4788

#line 4788

#line 4788
allow staff_t staff_t:tcp_socket { connectto recvfrom };
#line 4788
allow staff_t staff_t:tcp_socket { acceptfrom recvfrom };
#line 4788
allow staff_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788
allow staff_t tcp_socket_t:tcp_socket { recvfrom };
#line 4788

#line 4788

#line 4788
# Connect to inetd.
#line 4788

#line 4788

#line 4788

#line 4788

#line 4788
# Connect data port to ftpd.
#line 4788

#line 4788

#line 4788
# Connect to portmap.
#line 4788

#line 4788

#line 4788
# Inherit and use sockets from inetd
#line 4788

#line 4788

#line 4788
# Allow system log read
#line 4788
#allow staff_t kernel_t:system syslog_read;
#line 4788
# else do not log it
#line 4788
dontaudit staff_t kernel_t:system syslog_read;
#line 4788

#line 4788
# Very permissive allowing every domain to see every type.
#line 4788
allow staff_t kernel_t:system { ipc_info };
#line 4788

#line 4788
# When the user domain runs ps, there will be a number of access
#line 4788
# denials when ps tries to search /proc.  Do not audit these denials.
#line 4788
dontaudit staff_t domain:dir { read getattr lock search ioctl };
#line 4788
dontaudit staff_t domain:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4788

#line 4788
# Some shells ask for w access to utmp, but will operate
#line 4788
# correctly without it.  Do not audit write denials to utmp.
#line 4788
dontaudit staff_t initrc_var_run_t:file { getattr read write };
#line 4788

#line 4788
# do not audit getattr on tmpfile, otherwise ls -l /tmp fills the logs
#line 4788
dontaudit staff_t tmpfile:{ dir file lnk_file sock_file fifo_file chr_file blk_file } getattr;
#line 4788

#line 4788
# do not audit getattr on disk devices, otherwise KDE fills the logs
#line 4788
dontaudit staff_t { removable_device_t fixed_disk_device_t }:blk_file getattr;
#line 4788

#line 4788

#line 4788
# Access the sound device.
#line 4788
allow staff_t sound_device_t:chr_file { getattr read write ioctl };
#line 4788

#line 4788

#line 4788

#line 4788
# Get attributes of file systems.
#line 4788
allow staff_t fs_type:filesystem getattr;
#line 4788

#line 4788
# Read and write /dev/tty and /dev/null.
#line 4788
allow staff_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4788
allow staff_t zero_device_t:chr_file { { ioctl read getattr lock write append } execute };
#line 4788
allow staff_t random_device_t:chr_file { getattr read };
#line 4788

#line 4788
# Read the devpts root directory.
#line 4788
allow staff_t devpts_t:dir { read getattr lock search ioctl };
#line 4788

#line 4788

#line 4788
# When an ordinary user domain runs su, su may try to
#line 4788
# update the /root/.Xauthority file, and the user shell may
#line 4788
# try to update the shell history. This isnt allowed, but 
#line 4788
# we dont need to audit it.
#line 4788

#line 4788
dontaudit staff_su_t sysadm_home_t:dir  { read getattr search write add_name remove_name };
#line 4788
dontaudit staff_su_t sysadm_home_t:file { read getattr create write link unlink };
#line 4788
dontaudit staff_t    sysadm_home_t:dir { read search getattr };
#line 4788
dontaudit staff_t    sysadm_home_t:file { read getattr append };
#line 4788

#line 4788

#line 4788
# Some programs that are left in staff_t will try to connect
#line 4788
# to syslogd, but we do not want to let them generate log messages.
#line 4788
# Do not audit.
#line 4788
dontaudit staff_t devlog_t:sock_file { read write };
#line 4788
dontaudit staff_t syslogd_t:unix_dgram_socket sendto;
#line 4788

#line 4788

#line 4788
# stop warnings about "ls -l" on directories with unlabelled files
#line 4788
dontaudit staff_t file_t:{ dir file lnk_file } getattr;
#line 4788

allow staff_t unpriv_userdomain:process { sigchld sigkill sigstop signull signal };

#line 4790
allow staff_t domain:dir { read getattr lock search ioctl };
#line 4790
allow staff_t domain:{ file lnk_file } { read getattr lock ioctl };
#line 4790

allow staff_t { ttyfile ptyfile tty_device_t }:chr_file getattr;

# if adding new user roles make sure you edit the in_user_role macro in
# macros/user_macros.te to match

# lots of user programs accidentally search /root, and also the admin often
# logs in as UID=0 domain=user_t...
dontaudit unpriv_userdomain sysadm_home_dir_t:dir { getattr search };

# system_r is authorized for user_t as a fallback for unmodified daemons.
role system_r types user_t;

# change from role $1_r to $2_r and relabel tty appropriately
#line 4808


#line 4817

#line 4817
#
#line 4817
# Allow the user roles to transition
#line 4817
# into each other.
#line 4817

#line 4817
allow sysadm_r user_r;
#line 4817
type_change user_t sysadm_devpts_t:chr_file user_devpts_t;
#line 4817
type_change user_t sysadm_tty_device_t:chr_file user_tty_device_t;
#line 4817

#line 4817

#line 4817
allow staff_r sysadm_r;
#line 4817
type_change sysadm_t staff_devpts_t:chr_file sysadm_devpts_t;
#line 4817
type_change sysadm_t staff_tty_device_t:chr_file sysadm_tty_device_t;
#line 4817

#line 4817

#line 4817
allow sysadm_r staff_r;
#line 4817
type_change staff_t sysadm_devpts_t:chr_file staff_devpts_t;
#line 4817
type_change staff_t sysadm_tty_device_t:chr_file staff_tty_device_t;
#line 4817

#line 4817


# "ps aux" and "ls -l /dev/pts" make too much noise without this
dontaudit unpriv_userdomain ptyfile:chr_file getattr;
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

#################################
#
# Rules for the kernel_t domain.
#

#
# kernel_t is the domain of kernel threads.
# It is also the target type when checking permissions in the system class.
# 
type kernel_t, domain;
role system_r types kernel_t;

#line 4836
# Access other processes in the same domain.
#line 4836
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4836
# separately if desired.
#line 4836
allow kernel_t self:process ~{ptrace setexec setfscreate};
#line 4836

#line 4836
# Access /proc/PID files for processes in the same domain.
#line 4836
allow kernel_t self:dir { read getattr lock search ioctl };
#line 4836
allow kernel_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4836

#line 4836
# Access file descriptions, pipes, and sockets
#line 4836
# created by processes in the same domain.
#line 4836
allow kernel_t self:fd *;
#line 4836
allow kernel_t self:fifo_file { ioctl read getattr lock write append };
#line 4836
allow kernel_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4836
allow kernel_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4836

#line 4836
# Allow the domain to communicate with other processes in the same domain.
#line 4836
allow kernel_t self:unix_dgram_socket sendto;
#line 4836
allow kernel_t self:unix_stream_socket connectto;
#line 4836

#line 4836
# Access System V IPC objects created by processes in the same domain.
#line 4836
allow kernel_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4836
allow kernel_t self:msg  { send receive };
#line 4836
allow kernel_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4836
allow kernel_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4836


#line 4837
# Read system information files in /proc.
#line 4837
allow kernel_t proc_t:dir { read getattr lock search ioctl };
#line 4837
allow kernel_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4837

#line 4837
# Stat /proc/kmsg and /proc/kcore.
#line 4837
allow kernel_t proc_kmsg_t:file { getattr };
#line 4837
allow kernel_t proc_kcore_t:file { getattr };
#line 4837

#line 4837
# Read system variables in /proc/sys.
#line 4837
allow kernel_t sysctl_modprobe_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_t:dir { read getattr lock search ioctl };
#line 4837
allow kernel_t sysctl_fs_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_fs_t:dir { read getattr lock search ioctl };
#line 4837
allow kernel_t sysctl_kernel_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_kernel_t:dir { read getattr lock search ioctl };
#line 4837
allow kernel_t sysctl_net_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_net_t:dir { read getattr lock search ioctl };
#line 4837
allow kernel_t sysctl_vm_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_vm_t:dir { read getattr lock search ioctl };
#line 4837
allow kernel_t sysctl_dev_t:file { read getattr lock ioctl };
#line 4837
allow kernel_t sysctl_dev_t:dir { read getattr lock search ioctl };
#line 4837


#line 4838
# Read /.
#line 4838
allow kernel_t root_t:dir { read getattr lock search ioctl };
#line 4838
allow kernel_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4838

#line 4838
# Read /home.
#line 4838
allow kernel_t home_root_t:dir { read getattr lock search ioctl };
#line 4838

#line 4838
# Read /usr.
#line 4838
allow kernel_t usr_t:dir { read getattr lock search ioctl };
#line 4838
allow kernel_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4838

#line 4838
# Read bin and sbin directories.
#line 4838
allow kernel_t bin_t:dir { read getattr lock search ioctl };
#line 4838
allow kernel_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4838
allow kernel_t sbin_t:dir { read getattr lock search ioctl };
#line 4838
allow kernel_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4838


#line 4839
allow kernel_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4839
allow kernel_t lib_t:lnk_file { read getattr lock ioctl };
#line 4839
allow kernel_t ld_so_t:file { read getattr lock execute ioctl };
#line 4839
allow kernel_t ld_so_t:file execute_no_trans;
#line 4839
allow kernel_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4839
allow kernel_t shlib_t:file { read getattr lock execute ioctl };
#line 4839
allow kernel_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4839
allow kernel_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4839
allow kernel_t device_t:dir search;
#line 4839
allow kernel_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4839


# Use capabilities.
allow kernel_t self:capability *;

# Run init in the init_t domain.

#line 4845

#line 4845

#line 4845
#
#line 4845
# Allow the process to transition to the new domain.
#line 4845
#
#line 4845
allow kernel_t init_t:process transition;
#line 4845

#line 4845
#
#line 4845
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4845
#
#line 4845
dontaudit kernel_t init_t:process noatsecure;
#line 4845

#line 4845
#
#line 4845
# Allow the process to execute the program.
#line 4845
# 
#line 4845
allow kernel_t init_exec_t:file { read { getattr execute } };
#line 4845

#line 4845
#
#line 4845
# Allow the process to reap the new domain.
#line 4845
#
#line 4845
allow init_t kernel_t:process sigchld;
#line 4845

#line 4845
#
#line 4845
# Allow the new domain to inherit and use file 
#line 4845
# descriptions from the creating process and vice versa.
#line 4845
#
#line 4845
allow init_t kernel_t:fd use;
#line 4845
allow kernel_t init_t:fd use;
#line 4845

#line 4845
#
#line 4845
# Allow the new domain to write back to the old domain via a pipe.
#line 4845
#
#line 4845
allow init_t kernel_t:fifo_file { ioctl read getattr lock write append };
#line 4845

#line 4845
#
#line 4845
# Allow the new domain to read and execute the program.
#line 4845
#
#line 4845
allow init_t init_exec_t:file { read getattr lock execute ioctl };
#line 4845

#line 4845
#
#line 4845
# Allow the new domain to be entered via the program.
#line 4845
#
#line 4845
allow init_t init_exec_t:file entrypoint;
#line 4845

#line 4845
type_transition kernel_t init_exec_t:process init_t;
#line 4845


# Share state with the init process.
allow kernel_t init_t:process share;

# Mount and unmount file systems.
allow kernel_t fs_type:filesystem { mount remount unmount getattr };

# Search the persistent label mapping.
allow kernel_t file_labels_t:dir search;

# Send signal to any process.
allow kernel_t domain:process signal;

# Access the console.
allow kernel_t device_t:dir search;
allow kernel_t console_device_t:chr_file { ioctl read getattr lock write append };

# Access the initrd filesystem.
allow kernel_t file_t:chr_file { ioctl read getattr lock write append };

#line 4865
allow kernel_t file_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4865

#line 4868

allow kernel_t self:capability sys_chroot;

allow kernel_t { unlabeled_t root_t file_t }:dir mounton;
allow kernel_t file_t:dir { read getattr lock search ioctl add_name remove_name write };
allow kernel_t file_t:blk_file { create ioctl read getattr lock write setattr append link unlink rename };
allow kernel_t { sysctl_t sysctl_kernel_t }:file { setattr { ioctl read getattr lock write append } };

# Lookup the policy.
allow kernel_t policy_config_t:dir { read getattr lock search ioctl };

# Load the policy configuration.

#line 4880
allow kernel_t security_t:dir { read search getattr };
#line 4880
allow kernel_t security_t:file { getattr read write };
#line 4880
allow kernel_t security_t:security load_policy;
#line 4880
auditallow kernel_t security_t:security load_policy;
#line 4880


#line 4886

#line 4886
# Run insmod (for kernel module loader).
#line 4886

#line 4886

#line 4886

#line 4886
#
#line 4886
# Allow the process to transition to the new domain.
#line 4886
#
#line 4886
allow kernel_t insmod_t:process transition;
#line 4886

#line 4886
#
#line 4886
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4886
#
#line 4886
dontaudit kernel_t insmod_t:process noatsecure;
#line 4886

#line 4886
#
#line 4886
# Allow the process to execute the program.
#line 4886
# 
#line 4886
allow kernel_t insmod_exec_t:file { read { getattr execute } };
#line 4886

#line 4886
#
#line 4886
# Allow the process to reap the new domain.
#line 4886
#
#line 4886
allow insmod_t kernel_t:process sigchld;
#line 4886

#line 4886
#
#line 4886
# Allow the new domain to inherit and use file 
#line 4886
# descriptions from the creating process and vice versa.
#line 4886
#
#line 4886
allow insmod_t kernel_t:fd use;
#line 4886
allow kernel_t insmod_t:fd use;
#line 4886

#line 4886
#
#line 4886
# Allow the new domain to write back to the old domain via a pipe.
#line 4886
#
#line 4886
allow insmod_t kernel_t:fifo_file { ioctl read getattr lock write append };
#line 4886

#line 4886
#
#line 4886
# Allow the new domain to read and execute the program.
#line 4886
#
#line 4886
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 4886

#line 4886
#
#line 4886
# Allow the new domain to be entered via the program.
#line 4886
#
#line 4886
allow insmod_t insmod_exec_t:file entrypoint;
#line 4886

#line 4886
type_transition kernel_t insmod_exec_t:process insmod_t;
#line 4886

#line 4886
allow kernel_t insmod_exec_t:lnk_file read;
#line 4886


# /proc/sys/kernel/modprobe is set to /bin/true if not using modules.

#line 4889
allow kernel_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4889



#DESC Checkpolicy - SELinux policy compliler
#
# Authors:  Frank Mayer, mayerf@tresys.com
# X-Debian-Packages: checkpolicy
#

###########################
# 
# checkpolicy_t is the domain type for checkpolicy
# checkpolicy_exec_t if file type for the executable

type checkpolicy_t, domain;
role sysadm_r types checkpolicy_t;

type checkpolicy_exec_t, file_type, exec_type, sysadmfile;

##########################
# 
# Rules


#line 4912

#line 4912

#line 4912
#
#line 4912
# Allow the process to transition to the new domain.
#line 4912
#
#line 4912
allow sysadm_t checkpolicy_t:process transition;
#line 4912

#line 4912
#
#line 4912
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4912
#
#line 4912
dontaudit sysadm_t checkpolicy_t:process noatsecure;
#line 4912

#line 4912
#
#line 4912
# Allow the process to execute the program.
#line 4912
# 
#line 4912
allow sysadm_t checkpolicy_exec_t:file { read { getattr execute } };
#line 4912

#line 4912
#
#line 4912
# Allow the process to reap the new domain.
#line 4912
#
#line 4912
allow checkpolicy_t sysadm_t:process sigchld;
#line 4912

#line 4912
#
#line 4912
# Allow the new domain to inherit and use file 
#line 4912
# descriptions from the creating process and vice versa.
#line 4912
#
#line 4912
allow checkpolicy_t sysadm_t:fd use;
#line 4912
allow sysadm_t checkpolicy_t:fd use;
#line 4912

#line 4912
#
#line 4912
# Allow the new domain to write back to the old domain via a pipe.
#line 4912
#
#line 4912
allow checkpolicy_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 4912

#line 4912
#
#line 4912
# Allow the new domain to read and execute the program.
#line 4912
#
#line 4912
allow checkpolicy_t checkpolicy_exec_t:file { read getattr lock execute ioctl };
#line 4912

#line 4912
#
#line 4912
# Allow the new domain to be entered via the program.
#line 4912
#
#line 4912
allow checkpolicy_t checkpolicy_exec_t:file entrypoint;
#line 4912

#line 4912
type_transition sysadm_t checkpolicy_exec_t:process checkpolicy_t;
#line 4912


# able to create and modify binary policy files
allow checkpolicy_t policy_config_t:dir { read getattr lock search ioctl add_name remove_name write };
allow checkpolicy_t policy_config_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };

###########################
# constrain what checkpolicy can use as source files
#

# only allow read of policy source files
allow checkpolicy_t policy_src_t:dir { read getattr lock search ioctl };
allow checkpolicy_t policy_src_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };

# allow test policies to be created in src directories

#line 4927

#line 4927

#line 4927

#line 4927
#
#line 4927
# Allow the process to modify the directory.
#line 4927
#
#line 4927
allow checkpolicy_t policy_src_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4927

#line 4927
#
#line 4927
# Allow the process to create the file.
#line 4927
#
#line 4927

#line 4927
allow checkpolicy_t policy_config_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4927
allow checkpolicy_t policy_config_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4927

#line 4927

#line 4927
type_transition checkpolicy_t policy_src_t:dir policy_config_t;
#line 4927
type_transition checkpolicy_t policy_src_t:{ file lnk_file sock_file fifo_file } policy_config_t;
#line 4927

#line 4927

#line 4927


# directory search permissions for path to source and binary policy files
allow checkpolicy_t root_t:dir search;
allow checkpolicy_t etc_t:dir search;

# Read the devpts root directory.  
allow checkpolicy_t devpts_t:dir { read getattr lock search ioctl };
#line 4936


# Other access
allow checkpolicy_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { read write ioctl getattr };

#line 4940
allow checkpolicy_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4940
allow checkpolicy_t lib_t:lnk_file { read getattr lock ioctl };
#line 4940
allow checkpolicy_t ld_so_t:file { read getattr lock execute ioctl };
#line 4940
allow checkpolicy_t ld_so_t:file execute_no_trans;
#line 4940
allow checkpolicy_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4940
allow checkpolicy_t shlib_t:file { read getattr lock execute ioctl };
#line 4940
allow checkpolicy_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4940
allow checkpolicy_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4940
allow checkpolicy_t device_t:dir search;
#line 4940
allow checkpolicy_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4940

allow checkpolicy_t self:capability dac_override;

allow checkpolicy_t sysadm_tmp_t:file { getattr write } ;

##########################
# Allow users to execute checkpolicy without a domain transition
# so it can be used without privilege to write real binary policy file

#line 4948
allow user_t checkpolicy_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 4948


allow checkpolicy_t privfd:fd use;

#DESC Chkpwd - PAM password checking programs
# X-Debian-Packages: libpam-modules
#
# Domains for the /sbin/.*_chkpwd utilities.
#

#
# chkpwd_exec_t is the type of the /sbin/.*_chkpwd executables.
#
type chkpwd_exec_t, file_type, sysadmfile, exec_type;

# Everything else is in the chkpwd_domain macro in
# macros/program/chkpwd_macros.te.
#DESC Crond - Crond daemon
#
# Domains for the top-level crond daemon process and
# for system cron jobs.  The domains for user cron jobs
# are in macros/program/crond_macros.te.
#
# X-Debian-Packages: cron
# Authors:  Jonathan Crowley (MITRE) <jonathan@mitre.org>,
#	    Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser
#

# Domain for crond.  It needs auth to check for locked accounts.

#line 4977

#line 4977
type crond_t, domain, privlog , privrole, privmail, auth, privfd;
#line 4977
type crond_exec_t, file_type, sysadmfile, exec_type;
#line 4977

#line 4977
role system_r types crond_t;
#line 4977

#line 4977

#line 4977

#line 4977

#line 4977
#
#line 4977
# Allow the process to transition to the new domain.
#line 4977
#
#line 4977
allow initrc_t crond_t:process transition;
#line 4977

#line 4977
#
#line 4977
# Do not audit when glibc secure mode is enabled upon the transition.
#line 4977
#
#line 4977
dontaudit initrc_t crond_t:process noatsecure;
#line 4977

#line 4977
#
#line 4977
# Allow the process to execute the program.
#line 4977
# 
#line 4977
allow initrc_t crond_exec_t:file { read { getattr execute } };
#line 4977

#line 4977
#
#line 4977
# Allow the process to reap the new domain.
#line 4977
#
#line 4977
allow crond_t initrc_t:process sigchld;
#line 4977

#line 4977
#
#line 4977
# Allow the new domain to inherit and use file 
#line 4977
# descriptions from the creating process and vice versa.
#line 4977
#
#line 4977
allow crond_t initrc_t:fd use;
#line 4977
allow initrc_t crond_t:fd use;
#line 4977

#line 4977
#
#line 4977
# Allow the new domain to write back to the old domain via a pipe.
#line 4977
#
#line 4977
allow crond_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 4977

#line 4977
#
#line 4977
# Allow the new domain to read and execute the program.
#line 4977
#
#line 4977
allow crond_t crond_exec_t:file { read getattr lock execute ioctl };
#line 4977

#line 4977
#
#line 4977
# Allow the new domain to be entered via the program.
#line 4977
#
#line 4977
allow crond_t crond_exec_t:file entrypoint;
#line 4977

#line 4977
type_transition initrc_t crond_exec_t:process crond_t;
#line 4977

#line 4977

#line 4977
# Inherit and use descriptors from init.
#line 4977
allow crond_t init_t:fd use;
#line 4977
allow crond_t init_t:process sigchld;
#line 4977
allow crond_t privfd:fd use;
#line 4977
allow crond_t newrole_t:process sigchld;
#line 4977
allow crond_t self:process { { sigchld sigkill sigstop signull signal } fork };
#line 4977

#line 4977

#line 4977
allow crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 4977
allow crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 4977
allow crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 4977
allow crond_t ld_so_t:file execute_no_trans;
#line 4977
allow crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 4977
allow crond_t shlib_t:file { read getattr lock execute ioctl };
#line 4977
allow crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 4977
allow crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 4977
allow crond_t device_t:dir search;
#line 4977
allow crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4977

#line 4977

#line 4977
allow crond_t { self proc_t }:dir { read getattr lock search ioctl };
#line 4977
allow crond_t { self proc_t }:lnk_file read;
#line 4977

#line 4977
allow crond_t device_t:dir { getattr search };
#line 4977
allow crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 4977
allow crond_t console_device_t:chr_file { ioctl read getattr lock write append };
#line 4977
allow crond_t initrc_devpts_t:chr_file { ioctl read getattr lock write append };
#line 4977

#line 4977
# Create pid file.
#line 4977
allow crond_t var_t:dir { getattr search };
#line 4977

#line 4977
type crond_var_run_t alias var_run_crond_t, file_type, sysadmfile, pidfile;
#line 4977

#line 4977

#line 4977

#line 4977

#line 4977
#
#line 4977
# Allow the process to modify the directory.
#line 4977
#
#line 4977
allow crond_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4977

#line 4977
#
#line 4977
# Allow the process to create the file.
#line 4977
#
#line 4977

#line 4977

#line 4977
allow crond_t crond_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 4977

#line 4977

#line 4977
type_transition crond_t var_run_t:file crond_var_run_t;
#line 4977

#line 4977

#line 4977

#line 4977
allow crond_t crond_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4977

#line 4977

#line 4977
allow crond_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 4977

#line 4977
# for daemons that look at /root on startup
#line 4977
dontaudit crond_t sysadm_home_dir_t:dir search;
#line 4977

#line 4977
# for df
#line 4977
allow crond_t fs_type:filesystem getattr;
#line 4977

#line 4977

#line 4977
allow crond_t etc_t:lnk_file read;
#line 4977

#line 4977
allow crond_t locale_t:dir { read getattr lock search ioctl };
#line 4977
allow crond_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 4977

#line 4977

#line 4977


# This domain is granted permissions common to most domains (including can_net)

#line 4980
# Access other processes in the same domain.
#line 4980
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 4980
# separately if desired.
#line 4980
allow crond_t self:process ~{ptrace setexec setfscreate};
#line 4980

#line 4980
# Access /proc/PID files for processes in the same domain.
#line 4980
allow crond_t self:dir { read getattr lock search ioctl };
#line 4980
allow crond_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 4980

#line 4980
# Access file descriptions, pipes, and sockets
#line 4980
# created by processes in the same domain.
#line 4980
allow crond_t self:fd *;
#line 4980
allow crond_t self:fifo_file { ioctl read getattr lock write append };
#line 4980
allow crond_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 4980
allow crond_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 4980

#line 4980
# Allow the domain to communicate with other processes in the same domain.
#line 4980
allow crond_t self:unix_dgram_socket sendto;
#line 4980
allow crond_t self:unix_stream_socket connectto;
#line 4980

#line 4980
# Access System V IPC objects created by processes in the same domain.
#line 4980
allow crond_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 4980
allow crond_t self:msg  { send receive };
#line 4980
allow crond_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 4980
allow crond_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 4980


# Type for the anacron executable.
type anacron_exec_t, file_type, sysadmfile, exec_type;

# Type for temporary files.

#line 4986
type crond_tmp_t, file_type, sysadmfile, tmpfile ;
#line 4986

#line 4986

#line 4986

#line 4986

#line 4986
#
#line 4986
# Allow the process to modify the directory.
#line 4986
#
#line 4986
allow crond_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 4986

#line 4986
#
#line 4986
# Allow the process to create the file.
#line 4986
#
#line 4986

#line 4986
allow crond_t crond_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 4986
allow crond_t crond_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 4986

#line 4986

#line 4986
type_transition crond_t tmp_t:dir crond_tmp_t;
#line 4986
type_transition crond_t tmp_t:{ file lnk_file sock_file fifo_file } crond_tmp_t;
#line 4986

#line 4986

#line 4986

#line 4986


# Domain for system cron jobs.
type system_crond_t, domain, privlog, privmail;

# Read and write console and ttys.
allow system_crond_t devtty_t:chr_file { ioctl read getattr lock write append };

# Read system information files in /proc.
allow system_crond_t proc_t:dir { read getattr lock search ioctl };
allow system_crond_t proc_t:file { read getattr lock ioctl };
allow system_crond_t proc_t:lnk_file read;

#line 5001


# to stop killall type operations from filling our logs
dontaudit system_crond_t domain:dir search;

# read files in /etc
allow system_crond_t etc_t:{ file lnk_file } { read getattr lock ioctl };
allow system_crond_t etc_runtime_t:file read;


#line 5010
allow { crond_t system_crond_t } etc_t:lnk_file read;
#line 5010

#line 5010
allow { crond_t system_crond_t } locale_t:dir { read getattr lock search ioctl };
#line 5010
allow { crond_t system_crond_t } locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 5010

#line 5010


# Type for log files.
type cron_log_t, file_type, sysadmfile, logfile;
# Use this type when creating files in /var/log.

#line 5015

#line 5015

#line 5015

#line 5015
#
#line 5015
# Allow the process to modify the directory.
#line 5015
#
#line 5015
allow crond_t var_log_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5015

#line 5015
#
#line 5015
# Allow the process to create the file.
#line 5015
#
#line 5015

#line 5015
allow crond_t cron_log_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5015
allow crond_t cron_log_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5015

#line 5015

#line 5015
type_transition crond_t var_log_t:dir cron_log_t;
#line 5015
type_transition crond_t var_log_t:{ file lnk_file sock_file fifo_file } cron_log_t;
#line 5015

#line 5015

#line 5015


# Use capabilities.
allow crond_t crond_t:capability { setgid setuid net_bind_service };

# Get security policy decisions.

#line 5021
allow crond_t security_t:dir { read search getattr };
#line 5021
allow crond_t security_t:file { getattr read write };
#line 5021
allow crond_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 5021


# for finding binaries and /bin/sh
allow crond_t { bin_t sbin_t }:dir search;
allow crond_t bin_t:lnk_file read;

# Read from /var/spool/cron.
allow crond_t var_lib_t:dir search;
allow crond_t var_spool_t:dir { read getattr lock search ioctl };
allow crond_t cron_spool_t:dir { read getattr lock search ioctl };
allow crond_t cron_spool_t:file { read getattr lock ioctl };

# Read system crontabs
allow crond_t system_crond_script_t:file { read getattr lock ioctl };
allow crond_t system_crond_script_t:dir { read getattr lock search ioctl };

# Read /etc/security/default_contexts.
allow crond_t default_context_t:file { read getattr lock ioctl };

allow crond_t etc_t:file { getattr read };
allow crond_t etc_t:lnk_file read;

# crond tries to search /root.  Not sure why.
allow crond_t sysadm_home_dir_t:dir { read getattr lock search ioctl };

# to search /home
allow crond_t home_root_t:dir { getattr search };
allow crond_t user_home_dir_type:dir { read getattr lock search ioctl };

# Run a shell.

#line 5051
allow crond_t shell_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5051


#line 5059


# This system_r role is authorized for this domain.
role system_r types system_crond_t;


#line 5064
allow system_crond_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5064
allow system_crond_t lib_t:lnk_file { read getattr lock ioctl };
#line 5064
allow system_crond_t ld_so_t:file { read getattr lock execute ioctl };
#line 5064
allow system_crond_t ld_so_t:file execute_no_trans;
#line 5064
allow system_crond_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 5064
allow system_crond_t shlib_t:file { read getattr lock execute ioctl };
#line 5064
allow system_crond_t shlib_t:lnk_file { read getattr lock ioctl };
#line 5064
allow system_crond_t ld_so_cache_t:file { read getattr lock ioctl };
#line 5064
allow system_crond_t device_t:dir search;
#line 5064
allow system_crond_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5064
;

#line 5065
# Access other processes in the same domain.
#line 5065
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 5065
# separately if desired.
#line 5065
allow system_crond_t self:process ~{ptrace setexec setfscreate};
#line 5065

#line 5065
# Access /proc/PID files for processes in the same domain.
#line 5065
allow system_crond_t self:dir { read getattr lock search ioctl };
#line 5065
allow system_crond_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5065

#line 5065
# Access file descriptions, pipes, and sockets
#line 5065
# created by processes in the same domain.
#line 5065
allow system_crond_t self:fd *;
#line 5065
allow system_crond_t self:fifo_file { ioctl read getattr lock write append };
#line 5065
allow system_crond_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 5065
allow system_crond_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 5065

#line 5065
# Allow the domain to communicate with other processes in the same domain.
#line 5065
allow system_crond_t self:unix_dgram_socket sendto;
#line 5065
allow system_crond_t self:unix_stream_socket connectto;
#line 5065

#line 5065
# Access System V IPC objects created by processes in the same domain.
#line 5065
allow system_crond_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 5065
allow system_crond_t self:msg  { send receive };
#line 5065
allow system_crond_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 5065
allow system_crond_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 5065
;
allow system_crond_t var_log_t:file { read getattr lock ioctl };

# Type for system crontab files.
type system_crond_script_t, file_type, sysadmfile;

# Permit crond_t to transition to this domain.
# The transition is requested explicitly by the modified crond 
# via execve_secure.  There is no way to set up an automatic
# transition, since crontabs are configuration files, not executables.

#line 5075

#line 5075
#
#line 5075
# Allow the process to transition to the new domain.
#line 5075
#
#line 5075
allow crond_t system_crond_t:process transition;
#line 5075

#line 5075
#
#line 5075
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5075
#
#line 5075
dontaudit crond_t system_crond_t:process noatsecure;
#line 5075

#line 5075
#
#line 5075
# Allow the process to execute the program.
#line 5075
# 
#line 5075
allow crond_t shell_exec_t:file { read { getattr execute } };
#line 5075

#line 5075
#
#line 5075
# Allow the process to reap the new domain.
#line 5075
#
#line 5075
allow system_crond_t crond_t:process sigchld;
#line 5075

#line 5075
#
#line 5075
# Allow the new domain to inherit and use file 
#line 5075
# descriptions from the creating process and vice versa.
#line 5075
#
#line 5075
allow system_crond_t crond_t:fd use;
#line 5075
allow crond_t system_crond_t:fd use;
#line 5075

#line 5075
#
#line 5075
# Allow the new domain to write back to the old domain via a pipe.
#line 5075
#
#line 5075
allow system_crond_t crond_t:fifo_file { ioctl read getattr lock write append };
#line 5075

#line 5075
#
#line 5075
# Allow the new domain to read and execute the program.
#line 5075
#
#line 5075
allow system_crond_t shell_exec_t:file { read getattr lock execute ioctl };
#line 5075

#line 5075
#
#line 5075
# Allow the new domain to be entered via the program.
#line 5075
#
#line 5075
allow system_crond_t shell_exec_t:file entrypoint;
#line 5075


# Set exec context.

#line 5078
allow crond_t self:process setexec;
#line 5078
allow crond_t proc_t:dir search;
#line 5078
allow crond_t proc_t:lnk_file read;
#line 5078
allow crond_t self:dir search;
#line 5078
allow crond_t self:file { read write };
#line 5078


# Transition to this domain for anacron as well.
# Still need to study anacron.

#line 5082

#line 5082

#line 5082
#
#line 5082
# Allow the process to transition to the new domain.
#line 5082
#
#line 5082
allow initrc_t system_crond_t:process transition;
#line 5082

#line 5082
#
#line 5082
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5082
#
#line 5082
dontaudit initrc_t system_crond_t:process noatsecure;
#line 5082

#line 5082
#
#line 5082
# Allow the process to execute the program.
#line 5082
# 
#line 5082
allow initrc_t anacron_exec_t:file { read { getattr execute } };
#line 5082

#line 5082
#
#line 5082
# Allow the process to reap the new domain.
#line 5082
#
#line 5082
allow system_crond_t initrc_t:process sigchld;
#line 5082

#line 5082
#
#line 5082
# Allow the new domain to inherit and use file 
#line 5082
# descriptions from the creating process and vice versa.
#line 5082
#
#line 5082
allow system_crond_t initrc_t:fd use;
#line 5082
allow initrc_t system_crond_t:fd use;
#line 5082

#line 5082
#
#line 5082
# Allow the new domain to write back to the old domain via a pipe.
#line 5082
#
#line 5082
allow system_crond_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 5082

#line 5082
#
#line 5082
# Allow the new domain to read and execute the program.
#line 5082
#
#line 5082
allow system_crond_t anacron_exec_t:file { read getattr lock execute ioctl };
#line 5082

#line 5082
#
#line 5082
# Allow the new domain to be entered via the program.
#line 5082
#
#line 5082
allow system_crond_t anacron_exec_t:file entrypoint;
#line 5082

#line 5082
type_transition initrc_t anacron_exec_t:process system_crond_t;
#line 5082


# Access log files

#line 5085

#line 5085

#line 5085

#line 5085
#
#line 5085
# Allow the process to modify the directory.
#line 5085
#
#line 5085
allow system_crond_t var_log_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5085

#line 5085
#
#line 5085
# Allow the process to create the file.
#line 5085
#
#line 5085

#line 5085
allow system_crond_t cron_log_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5085
allow system_crond_t cron_log_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5085

#line 5085

#line 5085
type_transition system_crond_t var_log_t:dir cron_log_t;
#line 5085
type_transition system_crond_t var_log_t:{ file lnk_file sock_file fifo_file } cron_log_t;
#line 5085

#line 5085

#line 5085


# Inherit and use descriptors from init.
allow system_crond_t init_t:fd use;

# Inherit and use descriptors from initrc.
allow system_crond_t initrc_t:fd use;

# Write to a socket from initrc.
allow system_crond_t initrc_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Use capabilities.
allow system_crond_t system_crond_t:capability { setgid setuid dac_override fowner net_bind_service fsetid };

# Read the system crontabs.
allow system_crond_t system_crond_script_t:file { read getattr lock ioctl };

# 
# Since crontab files are not directly executed,
# crond must ensure that the crontab file has
# a type that is appropriate for the domain of
# the system cron job.  It performs an entrypoint
# permission check for this purpose.
#
allow system_crond_t system_crond_script_t:file entrypoint;
#line 5112


# Run helper programs in the system_crond_t domain.
allow system_crond_t { bin_t sbin_t }:dir { read getattr lock search ioctl };
allow system_crond_t { bin_t sbin_t }:lnk_file read;

#line 5117
allow system_crond_t { bin_t sbin_t exec_type etc_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5117


# Read from /var/spool/cron.
allow system_crond_t cron_spool_t:dir { read getattr lock search ioctl };
allow system_crond_t cron_spool_t:file { read getattr lock ioctl };

# Write to /var/lib/slocate.db.
allow system_crond_t var_lib_t:dir { read getattr lock search ioctl add_name remove_name write };
allow system_crond_t var_lib_t:file { create ioctl read getattr lock write setattr append link unlink rename };

# Update whatis files.
allow system_crond_t catman_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
allow system_crond_t catman_t:file { create ioctl read getattr lock write setattr append link unlink rename };
allow system_crond_t man_t:file { read getattr lock ioctl };
allow system_crond_t man_t:lnk_file read;

# Write /var/lock/makewhatis.lock.

#line 5134
type system_crond_lock_t, file_type, sysadmfile, lockfile;
#line 5134

#line 5134

#line 5134

#line 5134

#line 5134
#
#line 5134
# Allow the process to modify the directory.
#line 5134
#
#line 5134
allow system_crond_t var_lock_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5134

#line 5134
#
#line 5134
# Allow the process to create the file.
#line 5134
#
#line 5134

#line 5134

#line 5134
allow system_crond_t system_crond_lock_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 5134

#line 5134

#line 5134
type_transition system_crond_t var_lock_t:file system_crond_lock_t;
#line 5134

#line 5134

#line 5134

#line 5134


# Modutils are now combined, so we can no longer distinguish them.
# Let crond run the insmod executable in the insmod_t domain.
#line 5140

#line 5140

#line 5140

#line 5140

#line 5140

#line 5140
#
#line 5140
# Allow the process to transition to the new domain.
#line 5140
#
#line 5140
allow system_crond_t insmod_t:process transition;
#line 5140

#line 5140
#
#line 5140
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5140
#
#line 5140
dontaudit system_crond_t insmod_t:process noatsecure;
#line 5140

#line 5140
#
#line 5140
# Allow the process to execute the program.
#line 5140
# 
#line 5140
allow system_crond_t insmod_exec_t:file { read { getattr execute } };
#line 5140

#line 5140
#
#line 5140
# Allow the process to reap the new domain.
#line 5140
#
#line 5140
allow insmod_t system_crond_t:process sigchld;
#line 5140

#line 5140
#
#line 5140
# Allow the new domain to inherit and use file 
#line 5140
# descriptions from the creating process and vice versa.
#line 5140
#
#line 5140
allow insmod_t system_crond_t:fd use;
#line 5140
allow system_crond_t insmod_t:fd use;
#line 5140

#line 5140
#
#line 5140
# Allow the new domain to write back to the old domain via a pipe.
#line 5140
#
#line 5140
allow insmod_t system_crond_t:fifo_file { ioctl read getattr lock write append };
#line 5140

#line 5140
#
#line 5140
# Allow the new domain to read and execute the program.
#line 5140
#
#line 5140
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 5140

#line 5140
#
#line 5140
# Allow the new domain to be entered via the program.
#line 5140
#
#line 5140
allow insmod_t insmod_exec_t:file entrypoint;
#line 5140

#line 5140
type_transition system_crond_t insmod_exec_t:process insmod_t;
#line 5140

#line 5140
allow insmod_t crond_t:fifo_file { getattr read write ioctl };
#line 5140
# a rule for privfd may make this obsolete
#line 5140
allow insmod_t crond_t:fd use;
#line 5140
allow insmod_t crond_t:process sigchld;
#line 5140

#line 5140


# for if /var/mail is a symlink
allow crond_t mail_spool_t:lnk_file read;

# Run logrotate in the logrotate_t domain.
#line 5148

#line 5148

#line 5148

#line 5148

#line 5148

#line 5148
#
#line 5148
# Allow the process to transition to the new domain.
#line 5148
#
#line 5148
allow system_crond_t logrotate_t:process transition;
#line 5148

#line 5148
#
#line 5148
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5148
#
#line 5148
dontaudit system_crond_t logrotate_t:process noatsecure;
#line 5148

#line 5148
#
#line 5148
# Allow the process to execute the program.
#line 5148
# 
#line 5148
allow system_crond_t logrotate_exec_t:file { read { getattr execute } };
#line 5148

#line 5148
#
#line 5148
# Allow the process to reap the new domain.
#line 5148
#
#line 5148
allow logrotate_t system_crond_t:process sigchld;
#line 5148

#line 5148
#
#line 5148
# Allow the new domain to inherit and use file 
#line 5148
# descriptions from the creating process and vice versa.
#line 5148
#
#line 5148
allow logrotate_t system_crond_t:fd use;
#line 5148
allow system_crond_t logrotate_t:fd use;
#line 5148

#line 5148
#
#line 5148
# Allow the new domain to write back to the old domain via a pipe.
#line 5148
#
#line 5148
allow logrotate_t system_crond_t:fifo_file { ioctl read getattr lock write append };
#line 5148

#line 5148
#
#line 5148
# Allow the new domain to read and execute the program.
#line 5148
#
#line 5148
allow logrotate_t logrotate_exec_t:file { read getattr lock execute ioctl };
#line 5148

#line 5148
#
#line 5148
# Allow the new domain to be entered via the program.
#line 5148
#
#line 5148
allow logrotate_t logrotate_exec_t:file entrypoint;
#line 5148

#line 5148
type_transition system_crond_t logrotate_exec_t:process logrotate_t;
#line 5148

#line 5148
allow logrotate_t crond_t:fifo_file { getattr read write ioctl };
#line 5148
# a rule for privfd may make this obsolete
#line 5148
allow logrotate_t crond_t:fd use;
#line 5148
allow logrotate_t crond_t:process sigchld;
#line 5148

#line 5148


#line 5152


# Stat any file and search any directory for find.
allow system_crond_t file_type:{ file lnk_file sock_file fifo_file chr_file blk_file } getattr;
allow system_crond_t file_type:dir { read search getattr };

# Create temporary files.
type system_crond_tmp_t, file_type, sysadmfile, tmpfile;

#line 5160

#line 5160

#line 5160

#line 5160
#
#line 5160
# Allow the process to modify the directory.
#line 5160
#
#line 5160
allow system_crond_t { tmp_t crond_tmp_t }:dir { read getattr lock search ioctl add_name remove_name write };
#line 5160

#line 5160
#
#line 5160
# Allow the process to create the file.
#line 5160
#
#line 5160

#line 5160
allow system_crond_t system_crond_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5160
allow system_crond_t system_crond_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5160

#line 5160

#line 5160
type_transition system_crond_t { tmp_t crond_tmp_t }:dir system_crond_tmp_t;
#line 5160
type_transition system_crond_t { tmp_t crond_tmp_t }:{ file lnk_file sock_file fifo_file } system_crond_tmp_t;
#line 5160

#line 5160

#line 5160


# /sbin/runlevel ask for w access to utmp, but will operate
# correctly without it.  Do not audit write denials to utmp.
dontaudit system_crond_t initrc_var_run_t:file { read write };

# Access accounting summary files.
allow system_crond_t var_log_sa_t:file { create ioctl read getattr lock write setattr append link unlink rename };
allow system_crond_t var_log_sa_t:dir { read getattr lock search ioctl add_name remove_name write };

allow system_crond_t fs_t:filesystem getattr;

allow system_crond_t { sysctl_t sysctl_kernel_t }:dir search;
allow system_crond_t sysctl_kernel_t:file { getattr read };

# Access other spool directories like
# /var/spool/anacron and /var/spool/slrnpull.
allow system_crond_t var_spool_t:file { create ioctl read getattr lock write setattr append link unlink rename };
allow system_crond_t var_spool_t:dir { read getattr lock search ioctl add_name remove_name write };

# Do not audit attempts to search unlabeled directories (e.g. slocate).
dontaudit system_crond_t unlabeled_t:dir { read getattr lock search ioctl };
dontaudit system_crond_t unlabeled_t:file { read getattr lock ioctl };

#DESC Crontab - Crontab manipulation programs
#
# Domains for the crontab program.
#
# X-Debian-Packages: cron
#

# Type for the crontab executable.
type crontab_exec_t, file_type, sysadmfile, exec_type;

# Everything else is in the crontab_domain macro in
# macros/program/crontab_macros.te.
#DESC Dpkg - Debian package manager
#
# Author:  Russell Coker <russell@coker.com.au>
# X-Debian-Packages: dpkg
#

#################################
#
# Rules for the dpkg_t domain.
#
type dpkg_t, domain, admin, privlog, privmail;
type dpkg_exec_t, file_type, sysadmfile, exec_type;
type dpkg_var_lib_t alias var_lib_dpkg_t, file_type, sysadmfile;
type etc_dpkg_t, file_type, sysadmfile;
type dpkg_lock_t, file_type, sysadmfile;
type debconf_cache_t, file_type, sysadmfile;


#line 5213
type dpkg_tmp_t, file_type, sysadmfile, tmpfile ;
#line 5213

#line 5213

#line 5213

#line 5213

#line 5213
#
#line 5213
# Allow the process to modify the directory.
#line 5213
#
#line 5213
allow dpkg_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5213

#line 5213
#
#line 5213
# Allow the process to create the file.
#line 5213
#
#line 5213

#line 5213
allow dpkg_t dpkg_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5213
allow dpkg_t dpkg_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5213

#line 5213

#line 5213
type_transition dpkg_t tmp_t:dir dpkg_tmp_t;
#line 5213
type_transition dpkg_t tmp_t:{ file lnk_file sock_file fifo_file } dpkg_tmp_t;
#line 5213

#line 5213

#line 5213

#line 5213


#line 5214
allow dpkg_t { dpkg_exec_t bin_t shell_exec_t dpkg_tmp_t ls_exec_t dpkg_var_lib_t etc_dpkg_t sbin_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5214

#line 5218


#line 5219
allow dpkg_t init_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5219

#line 5222


allow dpkg_t random_device_t:chr_file read;

#line 5227

#line 5227

#line 5227

#line 5227
#
#line 5227
# Allow the process to transition to the new domain.
#line 5227
#
#line 5227
allow dpkg_t setfiles_t:process transition;
#line 5227

#line 5227
#
#line 5227
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5227
#
#line 5227
dontaudit dpkg_t setfiles_t:process noatsecure;
#line 5227

#line 5227
#
#line 5227
# Allow the process to execute the program.
#line 5227
# 
#line 5227
allow dpkg_t setfiles_exec_t:file { read { getattr execute } };
#line 5227

#line 5227
#
#line 5227
# Allow the process to reap the new domain.
#line 5227
#
#line 5227
allow setfiles_t dpkg_t:process sigchld;
#line 5227

#line 5227
#
#line 5227
# Allow the new domain to inherit and use file 
#line 5227
# descriptions from the creating process and vice versa.
#line 5227
#
#line 5227
allow setfiles_t dpkg_t:fd use;
#line 5227
allow dpkg_t setfiles_t:fd use;
#line 5227

#line 5227
#
#line 5227
# Allow the new domain to write back to the old domain via a pipe.
#line 5227
#
#line 5227
allow setfiles_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5227

#line 5227
#
#line 5227
# Allow the new domain to read and execute the program.
#line 5227
#
#line 5227
allow setfiles_t setfiles_exec_t:file { read getattr lock execute ioctl };
#line 5227

#line 5227
#
#line 5227
# Allow the new domain to be entered via the program.
#line 5227
#
#line 5227
allow setfiles_t setfiles_exec_t:file entrypoint;
#line 5227

#line 5227
type_transition dpkg_t setfiles_exec_t:process setfiles_t;
#line 5227


#line 5233

#line 5233

#line 5233

#line 5233

#line 5233
#
#line 5233
# Allow the process to transition to the new domain.
#line 5233
#
#line 5233
allow dpkg_t update_modules_t:process transition;
#line 5233

#line 5233
#
#line 5233
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5233
#
#line 5233
dontaudit dpkg_t update_modules_t:process noatsecure;
#line 5233

#line 5233
#
#line 5233
# Allow the process to execute the program.
#line 5233
# 
#line 5233
allow dpkg_t update_modules_exec_t:file { read { getattr execute } };
#line 5233

#line 5233
#
#line 5233
# Allow the process to reap the new domain.
#line 5233
#
#line 5233
allow update_modules_t dpkg_t:process sigchld;
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to inherit and use file 
#line 5233
# descriptions from the creating process and vice versa.
#line 5233
#
#line 5233
allow update_modules_t dpkg_t:fd use;
#line 5233
allow dpkg_t update_modules_t:fd use;
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to write back to the old domain via a pipe.
#line 5233
#
#line 5233
allow update_modules_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to read and execute the program.
#line 5233
#
#line 5233
allow update_modules_t update_modules_exec_t:file { read getattr lock execute ioctl };
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to be entered via the program.
#line 5233
#
#line 5233
allow update_modules_t update_modules_exec_t:file entrypoint;
#line 5233

#line 5233
type_transition dpkg_t update_modules_exec_t:process update_modules_t;
#line 5233

#line 5233

#line 5233

#line 5233

#line 5233
#
#line 5233
# Allow the process to transition to the new domain.
#line 5233
#
#line 5233
allow dpkg_t insmod_t:process transition;
#line 5233

#line 5233
#
#line 5233
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5233
#
#line 5233
dontaudit dpkg_t insmod_t:process noatsecure;
#line 5233

#line 5233
#
#line 5233
# Allow the process to execute the program.
#line 5233
# 
#line 5233
allow dpkg_t insmod_exec_t:file { read { getattr execute } };
#line 5233

#line 5233
#
#line 5233
# Allow the process to reap the new domain.
#line 5233
#
#line 5233
allow insmod_t dpkg_t:process sigchld;
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to inherit and use file 
#line 5233
# descriptions from the creating process and vice versa.
#line 5233
#
#line 5233
allow insmod_t dpkg_t:fd use;
#line 5233
allow dpkg_t insmod_t:fd use;
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to write back to the old domain via a pipe.
#line 5233
#
#line 5233
allow insmod_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to read and execute the program.
#line 5233
#
#line 5233
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to be entered via the program.
#line 5233
#
#line 5233
allow insmod_t insmod_exec_t:file entrypoint;
#line 5233

#line 5233
type_transition dpkg_t insmod_exec_t:process insmod_t;
#line 5233

#line 5233

#line 5233

#line 5233

#line 5233
#
#line 5233
# Allow the process to transition to the new domain.
#line 5233
#
#line 5233
allow dpkg_t depmod_t:process transition;
#line 5233

#line 5233
#
#line 5233
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5233
#
#line 5233
dontaudit dpkg_t depmod_t:process noatsecure;
#line 5233

#line 5233
#
#line 5233
# Allow the process to execute the program.
#line 5233
# 
#line 5233
allow dpkg_t depmod_exec_t:file { read { getattr execute } };
#line 5233

#line 5233
#
#line 5233
# Allow the process to reap the new domain.
#line 5233
#
#line 5233
allow depmod_t dpkg_t:process sigchld;
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to inherit and use file 
#line 5233
# descriptions from the creating process and vice versa.
#line 5233
#
#line 5233
allow depmod_t dpkg_t:fd use;
#line 5233
allow dpkg_t depmod_t:fd use;
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to write back to the old domain via a pipe.
#line 5233
#
#line 5233
allow depmod_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to read and execute the program.
#line 5233
#
#line 5233
allow depmod_t depmod_exec_t:file { read getattr lock execute ioctl };
#line 5233

#line 5233
#
#line 5233
# Allow the new domain to be entered via the program.
#line 5233
#
#line 5233
allow depmod_t depmod_exec_t:file entrypoint;
#line 5233

#line 5233
type_transition dpkg_t depmod_exec_t:process depmod_t;
#line 5233

#line 5233

#line 5240

#line 5247

#line 5251

#line 5254

#line 5257

#line 5260

# because the syslogd package is broken and does not use the start scripts
#line 5268

#line 5268

#line 5268

#line 5268

#line 5268
#
#line 5268
# Allow the process to transition to the new domain.
#line 5268
#
#line 5268
allow dpkg_t klogd_t:process transition;
#line 5268

#line 5268
#
#line 5268
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5268
#
#line 5268
dontaudit dpkg_t klogd_t:process noatsecure;
#line 5268

#line 5268
#
#line 5268
# Allow the process to execute the program.
#line 5268
# 
#line 5268
allow dpkg_t klogd_exec_t:file { read { getattr execute } };
#line 5268

#line 5268
#
#line 5268
# Allow the process to reap the new domain.
#line 5268
#
#line 5268
allow klogd_t dpkg_t:process sigchld;
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to inherit and use file 
#line 5268
# descriptions from the creating process and vice versa.
#line 5268
#
#line 5268
allow klogd_t dpkg_t:fd use;
#line 5268
allow dpkg_t klogd_t:fd use;
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to write back to the old domain via a pipe.
#line 5268
#
#line 5268
allow klogd_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to read and execute the program.
#line 5268
#
#line 5268
allow klogd_t klogd_exec_t:file { read getattr lock execute ioctl };
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to be entered via the program.
#line 5268
#
#line 5268
allow klogd_t klogd_exec_t:file entrypoint;
#line 5268

#line 5268
type_transition dpkg_t klogd_exec_t:process klogd_t;
#line 5268

#line 5268

#line 5268

#line 5268

#line 5268
#
#line 5268
# Allow the process to transition to the new domain.
#line 5268
#
#line 5268
allow dpkg_t syslogd_t:process transition;
#line 5268

#line 5268
#
#line 5268
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5268
#
#line 5268
dontaudit dpkg_t syslogd_t:process noatsecure;
#line 5268

#line 5268
#
#line 5268
# Allow the process to execute the program.
#line 5268
# 
#line 5268
allow dpkg_t syslogd_exec_t:file { read { getattr execute } };
#line 5268

#line 5268
#
#line 5268
# Allow the process to reap the new domain.
#line 5268
#
#line 5268
allow syslogd_t dpkg_t:process sigchld;
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to inherit and use file 
#line 5268
# descriptions from the creating process and vice versa.
#line 5268
#
#line 5268
allow syslogd_t dpkg_t:fd use;
#line 5268
allow dpkg_t syslogd_t:fd use;
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to write back to the old domain via a pipe.
#line 5268
#
#line 5268
allow syslogd_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to read and execute the program.
#line 5268
#
#line 5268
allow syslogd_t syslogd_exec_t:file { read getattr lock execute ioctl };
#line 5268

#line 5268
#
#line 5268
# Allow the new domain to be entered via the program.
#line 5268
#
#line 5268
allow syslogd_t syslogd_exec_t:file entrypoint;
#line 5268

#line 5268
type_transition dpkg_t syslogd_exec_t:process syslogd_t;
#line 5268

#line 5268
allow system_crond_t syslogd_t:dir search;
#line 5268
allow system_crond_t syslogd_t:file { getattr read };
#line 5268
allow system_crond_t syslogd_t:process signal;
#line 5268

# mysqld is broken too
#line 5274

#line 5284

#line 5287

#line 5291

#line 5294

#line 5294
allow load_policy_t initrc_t:fifo_file { read write };
#line 5294

#line 5300

#line 5300

#line 5300

#line 5300

#line 5300
#
#line 5300
# Allow the process to transition to the new domain.
#line 5300
#
#line 5300
allow dpkg_t checkpolicy_t:process transition;
#line 5300

#line 5300
#
#line 5300
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5300
#
#line 5300
dontaudit dpkg_t checkpolicy_t:process noatsecure;
#line 5300

#line 5300
#
#line 5300
# Allow the process to execute the program.
#line 5300
# 
#line 5300
allow dpkg_t checkpolicy_exec_t:file { read { getattr execute } };
#line 5300

#line 5300
#
#line 5300
# Allow the process to reap the new domain.
#line 5300
#
#line 5300
allow checkpolicy_t dpkg_t:process sigchld;
#line 5300

#line 5300
#
#line 5300
# Allow the new domain to inherit and use file 
#line 5300
# descriptions from the creating process and vice versa.
#line 5300
#
#line 5300
allow checkpolicy_t dpkg_t:fd use;
#line 5300
allow dpkg_t checkpolicy_t:fd use;
#line 5300

#line 5300
#
#line 5300
# Allow the new domain to write back to the old domain via a pipe.
#line 5300
#
#line 5300
allow checkpolicy_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5300

#line 5300
#
#line 5300
# Allow the new domain to read and execute the program.
#line 5300
#
#line 5300
allow checkpolicy_t checkpolicy_exec_t:file { read getattr lock execute ioctl };
#line 5300

#line 5300
#
#line 5300
# Allow the new domain to be entered via the program.
#line 5300
#
#line 5300
allow checkpolicy_t checkpolicy_exec_t:file entrypoint;
#line 5300

#line 5300
type_transition dpkg_t checkpolicy_exec_t:process checkpolicy_t;
#line 5300

#line 5300
role system_r types checkpolicy_t;
#line 5300
allow checkpolicy_t initrc_t:fd use;
#line 5300
allow checkpolicy_t initrc_devpts_t:chr_file { read write };
#line 5300

#line 5303

#line 5306

#line 5310


# for apt
type apt_t, domain, admin, privmail, web_client_domain;
type apt_exec_t, file_type, sysadmfile, exec_type;
type apt_var_lib_t alias var_lib_apt_t, file_type, sysadmfile;
type var_cache_apt_t, file_type, sysadmfile;
type etc_apt_t, file_type, sysadmfile;

#line 5318
type apt_tmp_t, file_type, sysadmfile, tmpfile ;
#line 5318

#line 5318

#line 5318

#line 5318

#line 5318
#
#line 5318
# Allow the process to modify the directory.
#line 5318
#
#line 5318
allow apt_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5318

#line 5318
#
#line 5318
# Allow the process to create the file.
#line 5318
#
#line 5318

#line 5318
allow apt_t apt_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5318
allow apt_t apt_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5318

#line 5318

#line 5318
type_transition apt_t tmp_t:dir apt_tmp_t;
#line 5318
type_transition apt_t tmp_t:{ file lnk_file sock_file fifo_file } apt_tmp_t;
#line 5318

#line 5318

#line 5318

#line 5318


#line 5319
allow apt_t apt_tmp_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5319


allow { apt_t dpkg_t install_menu_t } device_t:dir search;

dontaudit apt_t var_log_t:dir getattr;
dontaudit apt_t var_run_t:dir search;

# for rc files such as ~/.less

#line 5327
allow apt_t sysadm_home_t:dir { read getattr lock search ioctl };
#line 5327
allow apt_t sysadm_home_t:{ file lnk_file } { read getattr lock ioctl };
#line 5327

allow apt_t sysadm_home_dir_t:dir { search getattr };

allow apt_t bin_t:lnk_file { read getattr lock ioctl };


#line 5332
allow apt_t debconf_cache_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5332
allow apt_t debconf_cache_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5332


#line 5333
allow userdomain debconf_cache_t:dir { read getattr lock search ioctl };
#line 5333
allow userdomain debconf_cache_t:{ file lnk_file } { read getattr lock ioctl };
#line 5333


# for python
allow { apt_t dpkg_t } sysctl_kernel_t:dir search;
allow { apt_t dpkg_t } sysctl_kernel_t:file { read getattr lock ioctl };

allow dpkg_t sysctl_t:dir { search };

allow dpkg_t console_device_t:chr_file { ioctl read getattr lock write append };

allow apt_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

allow dpkg_t domain:dir { read getattr lock search ioctl };
allow dpkg_t domain:{ file lnk_file } { read getattr lock ioctl };

# for shared objects that are not yet labelled (upgrades)
allow { apt_t dpkg_t } lib_t:file execute;

# when dpkg runs postinst scripts run them in initrc_t domain so that the
# daemons are started in the correct context

#line 5353

#line 5353

#line 5353
#
#line 5353
# Allow the process to transition to the new domain.
#line 5353
#
#line 5353
allow dpkg_t initrc_t:process transition;
#line 5353

#line 5353
#
#line 5353
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5353
#
#line 5353
dontaudit dpkg_t initrc_t:process noatsecure;
#line 5353

#line 5353
#
#line 5353
# Allow the process to execute the program.
#line 5353
# 
#line 5353
allow dpkg_t etc_t:file { read { getattr execute } };
#line 5353

#line 5353
#
#line 5353
# Allow the process to reap the new domain.
#line 5353
#
#line 5353
allow initrc_t dpkg_t:process sigchld;
#line 5353

#line 5353
#
#line 5353
# Allow the new domain to inherit and use file 
#line 5353
# descriptions from the creating process and vice versa.
#line 5353
#
#line 5353
allow initrc_t dpkg_t:fd use;
#line 5353
allow dpkg_t initrc_t:fd use;
#line 5353

#line 5353
#
#line 5353
# Allow the new domain to write back to the old domain via a pipe.
#line 5353
#
#line 5353
allow initrc_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5353

#line 5353
#
#line 5353
# Allow the new domain to read and execute the program.
#line 5353
#
#line 5353
allow initrc_t etc_t:file { read getattr lock execute ioctl };
#line 5353

#line 5353
#
#line 5353
# Allow the new domain to be entered via the program.
#line 5353
#
#line 5353
allow initrc_t etc_t:file entrypoint;
#line 5353

#line 5353
type_transition dpkg_t etc_t:process initrc_t;
#line 5353


#line 5361


# for kernel-image postinst
dontaudit dpkg_t fixed_disk_device_t:blk_file read;

# allow user domains to execute dpkg
allow userdomain dpkg_exec_t:dir { read getattr lock search ioctl };

#line 5368
allow userdomain { dpkg_exec_t apt_exec_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5368


# allow everyone to read dpkg database

#line 5371
allow { apt_t userdomain } { dpkg_var_lib_t apt_var_lib_t var_cache_apt_t }:dir { read getattr lock search ioctl };
#line 5371
allow { apt_t userdomain } { dpkg_var_lib_t apt_var_lib_t var_cache_apt_t }:{ file lnk_file } { read getattr lock ioctl };
#line 5371


# for /var/lib/dpkg/lock

#line 5374
allow apt_t dpkg_var_lib_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5374
allow apt_t dpkg_var_lib_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5374


#line 5384

#line 5384

#line 5384
allow system_crond_t dpkg_var_lib_t:dir { read getattr lock search ioctl };
#line 5384
allow system_crond_t dpkg_var_lib_t:{ file lnk_file } { read getattr lock ioctl };
#line 5384

#line 5384
allow system_crond_t etc_dpkg_t:file { read getattr lock ioctl };
#line 5384

#line 5384
# for Debian cron job
#line 5384
allow system_crond_t shadow_t:file { read getattr };
#line 5384

#line 5384
allow system_crond_t tetex_data_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5384
allow system_crond_t tetex_data_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5384

#line 5384

#line 5384
allow dpkg_t tetex_data_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5384

#line 5384



#line 5386
allow install_menu_t { var_lib_t dpkg_var_lib_t }:dir { read getattr lock search ioctl };
#line 5386
allow install_menu_t { var_lib_t dpkg_var_lib_t }:{ file lnk_file } { read getattr lock ioctl };
#line 5386

allow install_menu_t initrc_t:fifo_file { read write };
allow { apt_t install_menu_t userdomain } etc_dpkg_t:file { read getattr lock ioctl };

#line 5389
allow sysadm_t etc_dpkg_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5389


# Inherit and use descriptors from open_init_pty
allow { apt_t dpkg_t install_menu_t } initrc_t:fd use;
dontaudit dpkg_t privfd:fd use;
allow { apt_t dpkg_t install_menu_t } devpts_t:dir search;
allow { apt_t dpkg_t install_menu_t } initrc_devpts_t:chr_file { ioctl read getattr lock write append };

allow ifconfig_t dpkg_t:fd use;
allow ifconfig_t dpkg_t:fifo_file { read write };


#line 5400
allow { dpkg_t apt_t } { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5400
allow { dpkg_t apt_t } lib_t:lnk_file { read getattr lock ioctl };
#line 5400
allow { dpkg_t apt_t } ld_so_t:file { read getattr lock execute ioctl };
#line 5400
allow { dpkg_t apt_t } ld_so_t:file execute_no_trans;
#line 5400
allow { dpkg_t apt_t } ld_so_t:lnk_file { read getattr lock ioctl };
#line 5400
allow { dpkg_t apt_t } shlib_t:file { read getattr lock execute ioctl };
#line 5400
allow { dpkg_t apt_t } shlib_t:lnk_file { read getattr lock ioctl };
#line 5400
allow { dpkg_t apt_t } ld_so_cache_t:file { read getattr lock ioctl };
#line 5400
allow { dpkg_t apt_t } device_t:dir search;
#line 5400
allow { dpkg_t apt_t } null_device_t:chr_file { ioctl read getattr lock write append };
#line 5400

allow dpkg_t proc_t:dir { read getattr lock search ioctl };
allow dpkg_t proc_t:{ file lnk_file } { read getattr lock ioctl };
allow dpkg_t fs_t:filesystem getattr;

allow dpkg_t self:capability { chown dac_override fowner fsetid kill setgid setuid sys_resource mknod };

allow dpkg_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow dpkg_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
allow dpkg_t self:unix_dgram_socket sendto;
allow dpkg_t self:unix_stream_socket connect;

allow { dpkg_t apt_t } devtty_t:chr_file { ioctl read getattr lock write append };
allow { dpkg_t apt_t } sysadm_tty_device_t:chr_file { ioctl read getattr lock write append };

# dpkg really needs to be able to kill any process, unfortunate but true
allow dpkg_t domain:process signal;
allow dpkg_t sysadm_t:process sigchld;
allow dpkg_t self:process { setpgid { sigchld sigkill sigstop signull signal } fork getsched };

# read/write/create any files in the system
allow dpkg_t file_type:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
allow dpkg_t file_type:{ file lnk_file fifo_file sock_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5425
allow dpkg_t file_type:{ chr_file blk_file } { create getattr setattr rename };
dontaudit dpkg_t domain:{ socket unix_dgram_socket udp_socket unix_stream_socket tcp_socket fifo_file rawip_socket packet_socket } getattr;
allow dpkg_t proc_kmsg_t:file getattr;
allow dpkg_t root_dir_type:dir getattr;

# change to the apt_t domain on exec from dpkg_t (dselect)

#line 5431

#line 5431

#line 5431
#
#line 5431
# Allow the process to transition to the new domain.
#line 5431
#
#line 5431
allow dpkg_t apt_t:process transition;
#line 5431

#line 5431
#
#line 5431
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5431
#
#line 5431
dontaudit dpkg_t apt_t:process noatsecure;
#line 5431

#line 5431
#
#line 5431
# Allow the process to execute the program.
#line 5431
# 
#line 5431
allow dpkg_t apt_exec_t:file { read { getattr execute } };
#line 5431

#line 5431
#
#line 5431
# Allow the process to reap the new domain.
#line 5431
#
#line 5431
allow apt_t dpkg_t:process sigchld;
#line 5431

#line 5431
#
#line 5431
# Allow the new domain to inherit and use file 
#line 5431
# descriptions from the creating process and vice versa.
#line 5431
#
#line 5431
allow apt_t dpkg_t:fd use;
#line 5431
allow dpkg_t apt_t:fd use;
#line 5431

#line 5431
#
#line 5431
# Allow the new domain to write back to the old domain via a pipe.
#line 5431
#
#line 5431
allow apt_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5431

#line 5431
#
#line 5431
# Allow the new domain to read and execute the program.
#line 5431
#
#line 5431
allow apt_t apt_exec_t:file { read getattr lock execute ioctl };
#line 5431

#line 5431
#
#line 5431
# Allow the new domain to be entered via the program.
#line 5431
#
#line 5431
allow apt_t apt_exec_t:file entrypoint;
#line 5431

#line 5431
type_transition dpkg_t apt_exec_t:process apt_t;
#line 5431


# allow apt to change /var/lib/apt files
allow apt_t { apt_var_lib_t var_cache_apt_t }:dir { read getattr lock search ioctl add_name remove_name write };
allow apt_t { apt_var_lib_t var_cache_apt_t }:file { create ioctl read getattr lock write setattr append link unlink rename };

# allow apt to create /usr/lib/site-python/DebianControlParser.pyc

#line 5438
allow apt_t lib_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5438
allow apt_t lib_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5438


# allow /var/cache/apt/archives to be owned by non-root
allow apt_t self:capability dac_override;


#line 5443
allow apt_t { apt_exec_t bin_t sbin_t shell_exec_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5443

allow apt_t { bin_t sbin_t }:dir search;
allow apt_t self:process { signal sigchld fork };
allow apt_t sysadm_t:process sigchld;

#line 5447
#
#line 5447
# Allow the domain to create and use UDP and TCP sockets.
#line 5447
# Other kinds of sockets must be separately authorized for use.
#line 5447
allow { apt_t dpkg_t } self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 5447
allow { apt_t dpkg_t } self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 5447

#line 5447
#
#line 5447
# Allow the domain to send UDP packets.
#line 5447
# Since the destination sockets type is unknown, the generic
#line 5447
# any_socket_t type is used as a placeholder.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } any_socket_t:udp_socket sendto;
#line 5447

#line 5447
#
#line 5447
# Allow the domain to send using any network interface.
#line 5447
# netif_type is a type attribute for all network interface types.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } netif_type:netif { tcp_send udp_send rawip_send };
#line 5447

#line 5447
#
#line 5447
# Allow packets sent by the domain to be received on any network interface.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 5447

#line 5447
#
#line 5447
# Allow the domain to receive packets from any network interface.
#line 5447
# netmsg_type is a type attribute for all default message types.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 5447

#line 5447
#
#line 5447
# Allow the domain to initiate or accept TCP connections 
#line 5447
# on any network interface.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } netmsg_type:tcp_socket { connectto acceptfrom };
#line 5447

#line 5447
#
#line 5447
# Receive resets from the TCP reset socket.
#line 5447
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } tcp_socket_t:tcp_socket recvfrom;
#line 5447

#line 5447
dontaudit { apt_t dpkg_t } tcp_socket_t:tcp_socket connectto;
#line 5447

#line 5447
#
#line 5447
# Allow the domain to send to any node.
#line 5447
# node_type is a type attribute for all node types.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } node_type:node { tcp_send udp_send rawip_send };
#line 5447

#line 5447
#
#line 5447
# Allow packets sent by the domain to be received from any node.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } node_type:node { tcp_recv udp_recv rawip_recv };
#line 5447

#line 5447
#
#line 5447
# Allow the domain to send NFS client requests via the socket
#line 5447
# created by mount.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 5447

#line 5447
#
#line 5447
# Bind to the default port type.
#line 5447
# Other port types must be separately authorized.
#line 5447
#
#line 5447
allow { apt_t dpkg_t } port_t:udp_socket name_bind;
#line 5447
allow { apt_t dpkg_t } port_t:tcp_socket name_bind;
#line 5447

allow apt_t tcp_socket_t:tcp_socket connectto;

allow { apt_t dpkg_t } var_t:dir { search getattr };
dontaudit apt_t { fs_type file_type }:dir getattr;
allow { apt_t dpkg_t } { var_lib_t bin_t }:dir { read getattr lock search ioctl };

allow { apt_t dpkg_t } dpkg_lock_t:file { setattr { ioctl read getattr lock write append } };

# for /proc/meminfo and for "ps"
allow apt_t { proc_t apt_t }:dir { read getattr lock search ioctl };
allow apt_t { proc_t apt_t }:{ file lnk_file } { read getattr lock ioctl };
allow apt_t self:fifo_file { ioctl read getattr lock write append };
allow dpkg_t self:fifo_file { ioctl read getattr lock write append };

allow apt_t etc_t:dir { read getattr lock search ioctl };
allow apt_t { etc_t resolv_conf_t }:file { read getattr lock ioctl };
allow apt_t etc_t:lnk_file read;

#line 5465
allow apt_t etc_t:lnk_file read;
#line 5465

#line 5465
allow apt_t locale_t:dir { read getattr lock search ioctl };
#line 5465
allow apt_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 5465

#line 5465


#line 5466
allow { userdomain apt_t } etc_apt_t:dir { read getattr lock search ioctl };
#line 5466
allow { userdomain apt_t } etc_apt_t:{ file lnk_file } { read getattr lock ioctl };
#line 5466


# apt wants to check available disk space
allow apt_t fs_t:filesystem getattr;
allow apt_t etc_runtime_t:file { read getattr lock ioctl };

# auto transition from apt_t to dpkg_t because for 99% of Debian upgrades you
# have apt run dpkg.
# This means that getting apt_t access is almost as good as dpkg_t which has
# as much power as sysadm_t...

#line 5476

#line 5476

#line 5476
#
#line 5476
# Allow the process to transition to the new domain.
#line 5476
#
#line 5476
allow apt_t dpkg_t:process transition;
#line 5476

#line 5476
#
#line 5476
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5476
#
#line 5476
dontaudit apt_t dpkg_t:process noatsecure;
#line 5476

#line 5476
#
#line 5476
# Allow the process to execute the program.
#line 5476
# 
#line 5476
allow apt_t dpkg_exec_t:file { read { getattr execute } };
#line 5476

#line 5476
#
#line 5476
# Allow the process to reap the new domain.
#line 5476
#
#line 5476
allow dpkg_t apt_t:process sigchld;
#line 5476

#line 5476
#
#line 5476
# Allow the new domain to inherit and use file 
#line 5476
# descriptions from the creating process and vice versa.
#line 5476
#
#line 5476
allow dpkg_t apt_t:fd use;
#line 5476
allow apt_t dpkg_t:fd use;
#line 5476

#line 5476
#
#line 5476
# Allow the new domain to write back to the old domain via a pipe.
#line 5476
#
#line 5476
allow dpkg_t apt_t:fifo_file { ioctl read getattr lock write append };
#line 5476

#line 5476
#
#line 5476
# Allow the new domain to read and execute the program.
#line 5476
#
#line 5476
allow dpkg_t dpkg_exec_t:file { read getattr lock execute ioctl };
#line 5476

#line 5476
#
#line 5476
# Allow the new domain to be entered via the program.
#line 5476
#
#line 5476
allow dpkg_t dpkg_exec_t:file entrypoint;
#line 5476

#line 5476
type_transition apt_t dpkg_exec_t:process dpkg_t;
#line 5476


# hack to allow update-menus/install-menu to manage menus
type install_menu_t, domain, admin;
type install_menu_exec_t, file_type, sysadmfile, exec_type;

#line 5481
type install_menu_var_run_t alias var_run_install_menu_t, file_type, sysadmfile, pidfile;
#line 5481

#line 5481

#line 5481

#line 5481

#line 5481
#
#line 5481
# Allow the process to modify the directory.
#line 5481
#
#line 5481
allow install_menu_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5481

#line 5481
#
#line 5481
# Allow the process to create the file.
#line 5481
#
#line 5481

#line 5481

#line 5481
allow install_menu_t install_menu_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 5481

#line 5481

#line 5481
type_transition install_menu_t var_run_t:file install_menu_var_run_t;
#line 5481

#line 5481

#line 5481

#line 5481
allow install_menu_t install_menu_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5481


allow install_menu_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

type debian_menu_t, file_type, sysadmfile;


#line 5487
allow userdomain debian_menu_t:dir { read getattr lock search ioctl };
#line 5487
allow userdomain debian_menu_t:{ file lnk_file } { read getattr lock ioctl };
#line 5487

dontaudit install_menu_t sysadm_home_dir_t:dir search;
allow install_menu_t debian_menu_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
allow install_menu_t debian_menu_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
allow install_menu_t dpkg_lock_t:file { setattr { ioctl read getattr lock write append } };
allow install_menu_t self:process signal;
allow install_menu_t proc_t:dir search;
allow install_menu_t proc_t:file { read getattr lock ioctl };

#line 5495
allow install_menu_t { bin_t sbin_t shell_exec_t install_menu_exec_t dpkg_exec_t install_menu_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5495

allow install_menu_t { bin_t sbin_t }:dir search;
allow install_menu_t bin_t:lnk_file read;

# for /etc/kde3/debian/kde-update-menu.sh

#line 5500
allow install_menu_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5500


allow install_menu_t var_t:dir search;

#line 5503
type install_menu_tmp_t, file_type, sysadmfile, tmpfile ;
#line 5503

#line 5503

#line 5503

#line 5503

#line 5503
#
#line 5503
# Allow the process to modify the directory.
#line 5503
#
#line 5503
allow install_menu_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5503

#line 5503
#
#line 5503
# Allow the process to create the file.
#line 5503
#
#line 5503

#line 5503
allow install_menu_t install_menu_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5503
allow install_menu_t install_menu_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5503

#line 5503

#line 5503
type_transition install_menu_t tmp_t:dir install_menu_tmp_t;
#line 5503
type_transition install_menu_t tmp_t:{ file lnk_file sock_file fifo_file } install_menu_tmp_t;
#line 5503

#line 5503

#line 5503

#line 5503



#line 5505
allow install_menu_t var_lib_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5505
allow install_menu_t var_lib_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5505

#line 5508

allow install_menu_t { var_spool_t etc_t }:dir { read getattr lock search ioctl add_name remove_name write };
allow install_menu_t { var_spool_t etc_t }:file { create ioctl read getattr lock write setattr append link unlink rename };
allow install_menu_t self:fifo_file { ioctl read getattr lock write append };
allow install_menu_t etc_runtime_t:file { read getattr lock ioctl };
allow install_menu_t devtty_t:chr_file { ioctl read getattr lock write append };
allow install_menu_t fs_t:filesystem getattr;


#line 5516

#line 5516

#line 5516
#
#line 5516
# Allow the process to transition to the new domain.
#line 5516
#
#line 5516
allow dpkg_t install_menu_t:process transition;
#line 5516

#line 5516
#
#line 5516
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5516
#
#line 5516
dontaudit dpkg_t install_menu_t:process noatsecure;
#line 5516

#line 5516
#
#line 5516
# Allow the process to execute the program.
#line 5516
# 
#line 5516
allow dpkg_t install_menu_exec_t:file { read { getattr execute } };
#line 5516

#line 5516
#
#line 5516
# Allow the process to reap the new domain.
#line 5516
#
#line 5516
allow install_menu_t dpkg_t:process sigchld;
#line 5516

#line 5516
#
#line 5516
# Allow the new domain to inherit and use file 
#line 5516
# descriptions from the creating process and vice versa.
#line 5516
#
#line 5516
allow install_menu_t dpkg_t:fd use;
#line 5516
allow dpkg_t install_menu_t:fd use;
#line 5516

#line 5516
#
#line 5516
# Allow the new domain to write back to the old domain via a pipe.
#line 5516
#
#line 5516
allow install_menu_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 5516

#line 5516
#
#line 5516
# Allow the new domain to read and execute the program.
#line 5516
#
#line 5516
allow install_menu_t install_menu_exec_t:file { read getattr lock execute ioctl };
#line 5516

#line 5516
#
#line 5516
# Allow the new domain to be entered via the program.
#line 5516
#
#line 5516
allow install_menu_t install_menu_exec_t:file entrypoint;
#line 5516

#line 5516
type_transition dpkg_t install_menu_exec_t:process install_menu_t;
#line 5516

allow dpkg_t install_menu_t:process { sigchld sigkill sigstop signull signal };

allow install_menu_t privfd:fd use;

#line 5520
allow install_menu_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5520
allow install_menu_t lib_t:lnk_file { read getattr lock ioctl };
#line 5520
allow install_menu_t ld_so_t:file { read getattr lock execute ioctl };
#line 5520
allow install_menu_t ld_so_t:file execute_no_trans;
#line 5520
allow install_menu_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 5520
allow install_menu_t shlib_t:file { read getattr lock execute ioctl };
#line 5520
allow install_menu_t shlib_t:lnk_file { read getattr lock ioctl };
#line 5520
allow install_menu_t ld_so_cache_t:file { read getattr lock ioctl };
#line 5520
allow install_menu_t device_t:dir search;
#line 5520
allow install_menu_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5520


allow install_menu_t self:process { fork sigchld };

role system_r types { dpkg_t apt_t install_menu_t };

#################################
#
# Rules for the run_deb_t domain.
#
#run_program(sysadm_t, sysadm_r, deb, dpkg_exec_t, dpkg_t)
#domain_trans(run_deb_t, apt_exec_t, apt_t)

#line 5532

#line 5532

#line 5532
#
#line 5532
# Allow the process to transition to the new domain.
#line 5532
#
#line 5532
allow initrc_t dpkg_t:process transition;
#line 5532

#line 5532
#
#line 5532
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5532
#
#line 5532
dontaudit initrc_t dpkg_t:process noatsecure;
#line 5532

#line 5532
#
#line 5532
# Allow the process to execute the program.
#line 5532
# 
#line 5532
allow initrc_t dpkg_exec_t:file { read { getattr execute } };
#line 5532

#line 5532
#
#line 5532
# Allow the process to reap the new domain.
#line 5532
#
#line 5532
allow dpkg_t initrc_t:process sigchld;
#line 5532

#line 5532
#
#line 5532
# Allow the new domain to inherit and use file 
#line 5532
# descriptions from the creating process and vice versa.
#line 5532
#
#line 5532
allow dpkg_t initrc_t:fd use;
#line 5532
allow initrc_t dpkg_t:fd use;
#line 5532

#line 5532
#
#line 5532
# Allow the new domain to write back to the old domain via a pipe.
#line 5532
#
#line 5532
allow dpkg_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 5532

#line 5532
#
#line 5532
# Allow the new domain to read and execute the program.
#line 5532
#
#line 5532
allow dpkg_t dpkg_exec_t:file { read getattr lock execute ioctl };
#line 5532

#line 5532
#
#line 5532
# Allow the new domain to be entered via the program.
#line 5532
#
#line 5532
allow dpkg_t dpkg_exec_t:file entrypoint;
#line 5532

#line 5532
type_transition initrc_t dpkg_exec_t:process dpkg_t;
#line 5532


#line 5533

#line 5533

#line 5533
#
#line 5533
# Allow the process to transition to the new domain.
#line 5533
#
#line 5533
allow initrc_t apt_t:process transition;
#line 5533

#line 5533
#
#line 5533
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5533
#
#line 5533
dontaudit initrc_t apt_t:process noatsecure;
#line 5533

#line 5533
#
#line 5533
# Allow the process to execute the program.
#line 5533
# 
#line 5533
allow initrc_t apt_exec_t:file { read { getattr execute } };
#line 5533

#line 5533
#
#line 5533
# Allow the process to reap the new domain.
#line 5533
#
#line 5533
allow apt_t initrc_t:process sigchld;
#line 5533

#line 5533
#
#line 5533
# Allow the new domain to inherit and use file 
#line 5533
# descriptions from the creating process and vice versa.
#line 5533
#
#line 5533
allow apt_t initrc_t:fd use;
#line 5533
allow initrc_t apt_t:fd use;
#line 5533

#line 5533
#
#line 5533
# Allow the new domain to write back to the old domain via a pipe.
#line 5533
#
#line 5533
allow apt_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 5533

#line 5533
#
#line 5533
# Allow the new domain to read and execute the program.
#line 5533
#
#line 5533
allow apt_t apt_exec_t:file { read getattr lock execute ioctl };
#line 5533

#line 5533
#
#line 5533
# Allow the new domain to be entered via the program.
#line 5533
#
#line 5533
allow apt_t apt_exec_t:file entrypoint;
#line 5533

#line 5533
type_transition initrc_t apt_exec_t:process apt_t;
#line 5533

#DESC Fsadm - Disk and file system administration
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: util-linux e2fsprogs xfsprogs reiserfsprogs parted raidtools2 mount
#

#################################
#
# Rules for the fsadm_t domain.
#
# fsadm_t is the domain for disk and file system
# administration.
# fsadm_exec_t is the type of the corresponding programs.
#
type fsadm_t, domain, privlog;
role system_r types fsadm_t;
role sysadm_r types fsadm_t;


#line 5552
# Access other processes in the same domain.
#line 5552
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 5552
# separately if desired.
#line 5552
allow fsadm_t self:process ~{ptrace setexec setfscreate};
#line 5552

#line 5552
# Access /proc/PID files for processes in the same domain.
#line 5552
allow fsadm_t self:dir { read getattr lock search ioctl };
#line 5552
allow fsadm_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5552

#line 5552
# Access file descriptions, pipes, and sockets
#line 5552
# created by processes in the same domain.
#line 5552
allow fsadm_t self:fd *;
#line 5552
allow fsadm_t self:fifo_file { ioctl read getattr lock write append };
#line 5552
allow fsadm_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 5552
allow fsadm_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 5552

#line 5552
# Allow the domain to communicate with other processes in the same domain.
#line 5552
allow fsadm_t self:unix_dgram_socket sendto;
#line 5552
allow fsadm_t self:unix_stream_socket connectto;
#line 5552

#line 5552
# Access System V IPC objects created by processes in the same domain.
#line 5552
allow fsadm_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 5552
allow fsadm_t self:msg  { send receive };
#line 5552
allow fsadm_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 5552
allow fsadm_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 5552


# Read system information files in /proc.
allow fsadm_t proc_t:dir { read getattr lock search ioctl };
allow fsadm_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };

# Read system variables in /proc/sys
allow fsadm_t sysctl_kernel_t:file { read getattr lock ioctl };
allow fsadm_t sysctl_kernel_t:dir { read getattr lock search ioctl };


#line 5562
# Read /.
#line 5562
allow fsadm_t root_t:dir { read getattr lock search ioctl };
#line 5562
allow fsadm_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5562

#line 5562
# Read /home.
#line 5562
allow fsadm_t home_root_t:dir { read getattr lock search ioctl };
#line 5562

#line 5562
# Read /usr.
#line 5562
allow fsadm_t usr_t:dir { read getattr lock search ioctl };
#line 5562
allow fsadm_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5562

#line 5562
# Read bin and sbin directories.
#line 5562
allow fsadm_t bin_t:dir { read getattr lock search ioctl };
#line 5562
allow fsadm_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5562
allow fsadm_t sbin_t:dir { read getattr lock search ioctl };
#line 5562
allow fsadm_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5562


# Read /etc.
allow fsadm_t etc_t:dir { read getattr lock search ioctl };
allow fsadm_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };

# Read module-related files.
allow fsadm_t modules_conf_t:{ file lnk_file } { read getattr lock ioctl };

# Read /dev directories and any symbolic links.
allow fsadm_t device_t:dir { read getattr lock search ioctl };
allow fsadm_t device_t:lnk_file { read getattr lock ioctl };


#line 5575
allow fsadm_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5575
allow fsadm_t lib_t:lnk_file { read getattr lock ioctl };
#line 5575
allow fsadm_t ld_so_t:file { read getattr lock execute ioctl };
#line 5575
allow fsadm_t ld_so_t:file execute_no_trans;
#line 5575
allow fsadm_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 5575
allow fsadm_t shlib_t:file { read getattr lock execute ioctl };
#line 5575
allow fsadm_t shlib_t:lnk_file { read getattr lock ioctl };
#line 5575
allow fsadm_t ld_so_cache_t:file { read getattr lock ioctl };
#line 5575
allow fsadm_t device_t:dir search;
#line 5575
allow fsadm_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5575


type fsadm_exec_t, file_type, sysadmfile, exec_type;

#line 5578

#line 5578

#line 5578
#
#line 5578
# Allow the process to transition to the new domain.
#line 5578
#
#line 5578
allow initrc_t fsadm_t:process transition;
#line 5578

#line 5578
#
#line 5578
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5578
#
#line 5578
dontaudit initrc_t fsadm_t:process noatsecure;
#line 5578

#line 5578
#
#line 5578
# Allow the process to execute the program.
#line 5578
# 
#line 5578
allow initrc_t fsadm_exec_t:file { read { getattr execute } };
#line 5578

#line 5578
#
#line 5578
# Allow the process to reap the new domain.
#line 5578
#
#line 5578
allow fsadm_t initrc_t:process sigchld;
#line 5578

#line 5578
#
#line 5578
# Allow the new domain to inherit and use file 
#line 5578
# descriptions from the creating process and vice versa.
#line 5578
#
#line 5578
allow fsadm_t initrc_t:fd use;
#line 5578
allow initrc_t fsadm_t:fd use;
#line 5578

#line 5578
#
#line 5578
# Allow the new domain to write back to the old domain via a pipe.
#line 5578
#
#line 5578
allow fsadm_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 5578

#line 5578
#
#line 5578
# Allow the new domain to read and execute the program.
#line 5578
#
#line 5578
allow fsadm_t fsadm_exec_t:file { read getattr lock execute ioctl };
#line 5578

#line 5578
#
#line 5578
# Allow the new domain to be entered via the program.
#line 5578
#
#line 5578
allow fsadm_t fsadm_exec_t:file entrypoint;
#line 5578

#line 5578
type_transition initrc_t fsadm_exec_t:process fsadm_t;
#line 5578


#line 5579

#line 5579

#line 5579
#
#line 5579
# Allow the process to transition to the new domain.
#line 5579
#
#line 5579
allow sysadm_t fsadm_t:process transition;
#line 5579

#line 5579
#
#line 5579
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5579
#
#line 5579
dontaudit sysadm_t fsadm_t:process noatsecure;
#line 5579

#line 5579
#
#line 5579
# Allow the process to execute the program.
#line 5579
# 
#line 5579
allow sysadm_t fsadm_exec_t:file { read { getattr execute } };
#line 5579

#line 5579
#
#line 5579
# Allow the process to reap the new domain.
#line 5579
#
#line 5579
allow fsadm_t sysadm_t:process sigchld;
#line 5579

#line 5579
#
#line 5579
# Allow the new domain to inherit and use file 
#line 5579
# descriptions from the creating process and vice versa.
#line 5579
#
#line 5579
allow fsadm_t sysadm_t:fd use;
#line 5579
allow sysadm_t fsadm_t:fd use;
#line 5579

#line 5579
#
#line 5579
# Allow the new domain to write back to the old domain via a pipe.
#line 5579
#
#line 5579
allow fsadm_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 5579

#line 5579
#
#line 5579
# Allow the new domain to read and execute the program.
#line 5579
#
#line 5579
allow fsadm_t fsadm_exec_t:file { read getattr lock execute ioctl };
#line 5579

#line 5579
#
#line 5579
# Allow the new domain to be entered via the program.
#line 5579
#
#line 5579
allow fsadm_t fsadm_exec_t:file entrypoint;
#line 5579

#line 5579
type_transition sysadm_t fsadm_exec_t:process fsadm_t;
#line 5579


type fsadm_tmp_t, file_type, sysadmfile, tmpfile;

#line 5582

#line 5582

#line 5582

#line 5582
#
#line 5582
# Allow the process to modify the directory.
#line 5582
#
#line 5582
allow fsadm_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5582

#line 5582
#
#line 5582
# Allow the process to create the file.
#line 5582
#
#line 5582

#line 5582
allow fsadm_t fsadm_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5582
allow fsadm_t fsadm_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5582

#line 5582

#line 5582
type_transition fsadm_t tmp_t:dir fsadm_tmp_t;
#line 5582
type_transition fsadm_t tmp_t:{ file lnk_file sock_file fifo_file } fsadm_tmp_t;
#line 5582

#line 5582

#line 5582


# remount file system to apply changes
allow fsadm_t fs_t:filesystem remount;

# mkreiserfs needs this
allow fsadm_t proc_t:filesystem getattr;

# mkreiserfs and other programs need this for UUID
allow fsadm_t random_device_t:chr_file { getattr read };

# Use capabilities.  ipc_lock is for losetup
allow fsadm_t self:capability { ipc_lock sys_rawio sys_admin };

# Write to /etc/mtab.

#line 5597

#line 5597

#line 5597

#line 5597
#
#line 5597
# Allow the process to modify the directory.
#line 5597
#
#line 5597
allow fsadm_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5597

#line 5597
#
#line 5597
# Allow the process to create the file.
#line 5597
#
#line 5597

#line 5597

#line 5597
allow fsadm_t etc_runtime_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 5597

#line 5597

#line 5597
type_transition fsadm_t etc_t:file etc_runtime_t;
#line 5597

#line 5597

#line 5597


# Inherit and use descriptors from init.
allow fsadm_t init_t:fd use;

# Run other fs admin programs in the fsadm_t domain.

#line 5603
allow fsadm_t fsadm_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5603


# Access disk devices.
allow fsadm_t fixed_disk_device_t:{ chr_file blk_file } { ioctl read getattr lock write append };
allow fsadm_t removable_device_t:{ chr_file blk_file } { ioctl read getattr lock write append };

# Access lost+found.
allow fsadm_t lost_found_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
allow fsadm_t lost_found_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };

# Recreate /mnt/cdrom. 
allow fsadm_t file_t:dir { search read getattr rmdir create };

# Recreate /dev/cdrom.
allow fsadm_t device_t:dir { read getattr lock search ioctl add_name remove_name write };
allow fsadm_t device_t:lnk_file { unlink create };

# Enable swapping to devices and files
allow fsadm_t swapfile_t:file { getattr swapon };
allow fsadm_t fixed_disk_device_t:blk_file { getattr swapon };

# XXX Why does updfstab run insmod?

#line 5625

#line 5625

#line 5625
#
#line 5625
# Allow the process to transition to the new domain.
#line 5625
#
#line 5625
allow fsadm_t insmod_t:process transition;
#line 5625

#line 5625
#
#line 5625
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5625
#
#line 5625
dontaudit fsadm_t insmod_t:process noatsecure;
#line 5625

#line 5625
#
#line 5625
# Allow the process to execute the program.
#line 5625
# 
#line 5625
allow fsadm_t insmod_exec_t:file { read { getattr execute } };
#line 5625

#line 5625
#
#line 5625
# Allow the process to reap the new domain.
#line 5625
#
#line 5625
allow insmod_t fsadm_t:process sigchld;
#line 5625

#line 5625
#
#line 5625
# Allow the new domain to inherit and use file 
#line 5625
# descriptions from the creating process and vice versa.
#line 5625
#
#line 5625
allow insmod_t fsadm_t:fd use;
#line 5625
allow fsadm_t insmod_t:fd use;
#line 5625

#line 5625
#
#line 5625
# Allow the new domain to write back to the old domain via a pipe.
#line 5625
#
#line 5625
allow insmod_t fsadm_t:fifo_file { ioctl read getattr lock write append };
#line 5625

#line 5625
#
#line 5625
# Allow the new domain to read and execute the program.
#line 5625
#
#line 5625
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 5625

#line 5625
#
#line 5625
# Allow the new domain to be entered via the program.
#line 5625
#
#line 5625
allow insmod_t insmod_exec_t:file entrypoint;
#line 5625

#line 5625
type_transition fsadm_t insmod_exec_t:process insmod_t;
#line 5625


# Allow console log change (updfstab)
allow fsadm_t kernel_t:system syslog_console;

# Access terminals.
allow fsadm_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { ioctl read getattr lock write append };

allow fsadm_t privfd:fd use;


#line 5635
allow fsadm_t etc_t:lnk_file read;
#line 5635

#line 5635
allow fsadm_t locale_t:dir { read getattr lock search ioctl };
#line 5635
allow fsadm_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 5635

#line 5635

#DESC Getty - Manage ttys
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: util-linux fbgetty mingetty mgetty rungetty
#

#################################
#
# Rules for the getty_t domain.
#

#line 5646

#line 5646
type getty_t, domain, privlog , mlstrustedwriter, privfd;
#line 5646
type getty_exec_t, file_type, sysadmfile, exec_type;
#line 5646

#line 5646
role system_r types getty_t;
#line 5646

#line 5646

#line 5646

#line 5646

#line 5646
#
#line 5646
# Allow the process to transition to the new domain.
#line 5646
#
#line 5646
allow initrc_t getty_t:process transition;
#line 5646

#line 5646
#
#line 5646
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5646
#
#line 5646
dontaudit initrc_t getty_t:process noatsecure;
#line 5646

#line 5646
#
#line 5646
# Allow the process to execute the program.
#line 5646
# 
#line 5646
allow initrc_t getty_exec_t:file { read { getattr execute } };
#line 5646

#line 5646
#
#line 5646
# Allow the process to reap the new domain.
#line 5646
#
#line 5646
allow getty_t initrc_t:process sigchld;
#line 5646

#line 5646
#
#line 5646
# Allow the new domain to inherit and use file 
#line 5646
# descriptions from the creating process and vice versa.
#line 5646
#
#line 5646
allow getty_t initrc_t:fd use;
#line 5646
allow initrc_t getty_t:fd use;
#line 5646

#line 5646
#
#line 5646
# Allow the new domain to write back to the old domain via a pipe.
#line 5646
#
#line 5646
allow getty_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 5646

#line 5646
#
#line 5646
# Allow the new domain to read and execute the program.
#line 5646
#
#line 5646
allow getty_t getty_exec_t:file { read getattr lock execute ioctl };
#line 5646

#line 5646
#
#line 5646
# Allow the new domain to be entered via the program.
#line 5646
#
#line 5646
allow getty_t getty_exec_t:file entrypoint;
#line 5646

#line 5646
type_transition initrc_t getty_exec_t:process getty_t;
#line 5646

#line 5646

#line 5646
# Inherit and use descriptors from init.
#line 5646
allow getty_t init_t:fd use;
#line 5646
allow getty_t init_t:process sigchld;
#line 5646
allow getty_t privfd:fd use;
#line 5646
allow getty_t newrole_t:process sigchld;
#line 5646
allow getty_t self:process { { sigchld sigkill sigstop signull signal } fork };
#line 5646

#line 5646

#line 5646
allow getty_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5646
allow getty_t lib_t:lnk_file { read getattr lock ioctl };
#line 5646
allow getty_t ld_so_t:file { read getattr lock execute ioctl };
#line 5646
allow getty_t ld_so_t:file execute_no_trans;
#line 5646
allow getty_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 5646
allow getty_t shlib_t:file { read getattr lock execute ioctl };
#line 5646
allow getty_t shlib_t:lnk_file { read getattr lock ioctl };
#line 5646
allow getty_t ld_so_cache_t:file { read getattr lock ioctl };
#line 5646
allow getty_t device_t:dir search;
#line 5646
allow getty_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5646

#line 5646

#line 5646
allow getty_t { self proc_t }:dir { read getattr lock search ioctl };
#line 5646
allow getty_t { self proc_t }:lnk_file read;
#line 5646

#line 5646
allow getty_t device_t:dir { getattr search };
#line 5646
allow getty_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5646
allow getty_t console_device_t:chr_file { ioctl read getattr lock write append };
#line 5646
allow getty_t initrc_devpts_t:chr_file { ioctl read getattr lock write append };
#line 5646

#line 5646
# Create pid file.
#line 5646
allow getty_t var_t:dir { getattr search };
#line 5646

#line 5646
type getty_var_run_t alias var_run_getty_t, file_type, sysadmfile, pidfile;
#line 5646

#line 5646

#line 5646

#line 5646

#line 5646
#
#line 5646
# Allow the process to modify the directory.
#line 5646
#
#line 5646
allow getty_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5646

#line 5646
#
#line 5646
# Allow the process to create the file.
#line 5646
#
#line 5646

#line 5646

#line 5646
allow getty_t getty_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 5646

#line 5646

#line 5646
type_transition getty_t var_run_t:file getty_var_run_t;
#line 5646

#line 5646

#line 5646

#line 5646
allow getty_t getty_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5646

#line 5646

#line 5646
allow getty_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 5646

#line 5646
# for daemons that look at /root on startup
#line 5646
dontaudit getty_t sysadm_home_dir_t:dir search;
#line 5646

#line 5646
# for df
#line 5646
allow getty_t fs_type:filesystem getattr;
#line 5646

#line 5646

#line 5646
allow getty_t etc_t:lnk_file read;
#line 5646

#line 5646
allow getty_t locale_t:dir { read getattr lock search ioctl };
#line 5646
allow getty_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 5646

#line 5646

#line 5646


type etc_getty_t, file_type, sysadmfile;

#line 5649
allow getty_t etc_getty_t:dir { read getattr lock search ioctl };
#line 5649
allow getty_t etc_getty_t:{ file lnk_file } { read getattr lock ioctl };
#line 5649


allow getty_t console_device_t:chr_file setattr;


#line 5653
type getty_tmp_t, file_type, sysadmfile, tmpfile ;
#line 5653

#line 5653

#line 5653

#line 5653

#line 5653
#
#line 5653
# Allow the process to modify the directory.
#line 5653
#
#line 5653
allow getty_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5653

#line 5653
#
#line 5653
# Allow the process to create the file.
#line 5653
#
#line 5653

#line 5653
allow getty_t getty_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5653
allow getty_t getty_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5653

#line 5653

#line 5653
type_transition getty_t tmp_t:dir getty_tmp_t;
#line 5653
type_transition getty_t tmp_t:{ file lnk_file sock_file fifo_file } getty_tmp_t;
#line 5653

#line 5653

#line 5653

#line 5653


#line 5654
type getty_log_t, file_type, sysadmfile, logfile;
#line 5654

#line 5654

#line 5654

#line 5654

#line 5654
#
#line 5654
# Allow the process to modify the directory.
#line 5654
#
#line 5654
allow getty_t var_log_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5654

#line 5654
#
#line 5654
# Allow the process to create the file.
#line 5654
#
#line 5654

#line 5654
allow getty_t getty_log_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5654
allow getty_t getty_log_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5654

#line 5654

#line 5654
type_transition getty_t var_log_t:dir getty_log_t;
#line 5654
type_transition getty_t var_log_t:{ file lnk_file sock_file fifo_file } getty_log_t;
#line 5654

#line 5654

#line 5654

#line 5654


allow getty_t { etc_t etc_runtime_t }:file { getattr read };
allow getty_t etc_t:lnk_file read;
allow getty_t self:process { getpgid getsession };
allow getty_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow getty_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# for ldap and other authentication services
allow getty_t resolv_conf_t:file { getattr read };

# to allow w to display everyone...
#allow userdomain ttyfile:chr_file getattr;

# Use capabilities.
allow getty_t getty_t:capability { dac_override chown sys_tty_config };

# fbgetty needs fsetid for some reason
allow getty_t getty_t:capability fsetid;

# Run getty in its own domain.

#line 5675

#line 5675

#line 5675
#
#line 5675
# Allow the process to transition to the new domain.
#line 5675
#
#line 5675
allow init_t getty_t:process transition;
#line 5675

#line 5675
#
#line 5675
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5675
#
#line 5675
dontaudit init_t getty_t:process noatsecure;
#line 5675

#line 5675
#
#line 5675
# Allow the process to execute the program.
#line 5675
# 
#line 5675
allow init_t getty_exec_t:file { read { getattr execute } };
#line 5675

#line 5675
#
#line 5675
# Allow the process to reap the new domain.
#line 5675
#
#line 5675
allow getty_t init_t:process sigchld;
#line 5675

#line 5675
#
#line 5675
# Allow the new domain to inherit and use file 
#line 5675
# descriptions from the creating process and vice versa.
#line 5675
#
#line 5675
allow getty_t init_t:fd use;
#line 5675
allow init_t getty_t:fd use;
#line 5675

#line 5675
#
#line 5675
# Allow the new domain to write back to the old domain via a pipe.
#line 5675
#
#line 5675
allow getty_t init_t:fifo_file { ioctl read getattr lock write append };
#line 5675

#line 5675
#
#line 5675
# Allow the new domain to read and execute the program.
#line 5675
#
#line 5675
allow getty_t getty_exec_t:file { read getattr lock execute ioctl };
#line 5675

#line 5675
#
#line 5675
# Allow the new domain to be entered via the program.
#line 5675
#
#line 5675
allow getty_t getty_exec_t:file entrypoint;
#line 5675

#line 5675
type_transition init_t getty_exec_t:process getty_t;
#line 5675


# Inherit and use descriptors from init.
allow getty_t init_t:fd use;

# Run login in local_login_t domain.
allow getty_t bin_t:dir search;

#line 5682

#line 5682

#line 5682
#
#line 5682
# Allow the process to transition to the new domain.
#line 5682
#
#line 5682
allow getty_t local_login_t:process transition;
#line 5682

#line 5682
#
#line 5682
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5682
#
#line 5682
dontaudit getty_t local_login_t:process noatsecure;
#line 5682

#line 5682
#
#line 5682
# Allow the process to execute the program.
#line 5682
# 
#line 5682
allow getty_t login_exec_t:file { read { getattr execute } };
#line 5682

#line 5682
#
#line 5682
# Allow the process to reap the new domain.
#line 5682
#
#line 5682
allow local_login_t getty_t:process sigchld;
#line 5682

#line 5682
#
#line 5682
# Allow the new domain to inherit and use file 
#line 5682
# descriptions from the creating process and vice versa.
#line 5682
#
#line 5682
allow local_login_t getty_t:fd use;
#line 5682
allow getty_t local_login_t:fd use;
#line 5682

#line 5682
#
#line 5682
# Allow the new domain to write back to the old domain via a pipe.
#line 5682
#
#line 5682
allow local_login_t getty_t:fifo_file { ioctl read getattr lock write append };
#line 5682

#line 5682
#
#line 5682
# Allow the new domain to read and execute the program.
#line 5682
#
#line 5682
allow local_login_t login_exec_t:file { read getattr lock execute ioctl };
#line 5682

#line 5682
#
#line 5682
# Allow the new domain to be entered via the program.
#line 5682
#
#line 5682
allow local_login_t login_exec_t:file entrypoint;
#line 5682

#line 5682
type_transition getty_t login_exec_t:process local_login_t;
#line 5682


# Write to /var/run/utmp.
allow getty_t initrc_var_run_t:file { ioctl read getattr lock write append };

# Write to /var/log/wtmp.
allow getty_t wtmp_t:file { ioctl read getattr lock write append };

# Chown, chmod, read and write ttys.
allow getty_t tty_device_t:chr_file { setattr { ioctl read getattr lock write append } };
allow getty_t ttyfile:chr_file { setattr { ioctl read getattr lock write append } };



#line 5695
allow getty_t var_lock_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5695
allow getty_t var_lock_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5695

#DESC Ifconfig - Configure network interfaces
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: net-tools
#

#################################
#
# Rules for the ifconfig_t domain.
#
# ifconfig_t is the domain for the ifconfig program.
# ifconfig_exec_t is the type of the corresponding program.
#
type ifconfig_t, domain, privlog;
type ifconfig_exec_t, file_type, sysadmfile, exec_type;

role system_r types ifconfig_t;
role sysadm_r types ifconfig_t;


#line 5715
allow ifconfig_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5715
allow ifconfig_t lib_t:lnk_file { read getattr lock ioctl };
#line 5715
allow ifconfig_t ld_so_t:file { read getattr lock execute ioctl };
#line 5715
allow ifconfig_t ld_so_t:file execute_no_trans;
#line 5715
allow ifconfig_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 5715
allow ifconfig_t shlib_t:file { read getattr lock execute ioctl };
#line 5715
allow ifconfig_t shlib_t:lnk_file { read getattr lock ioctl };
#line 5715
allow ifconfig_t ld_so_cache_t:file { read getattr lock ioctl };
#line 5715
allow ifconfig_t device_t:dir search;
#line 5715
allow ifconfig_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5715


#line 5716
# Access other processes in the same domain.
#line 5716
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 5716
# separately if desired.
#line 5716
allow ifconfig_t self:process ~{ptrace setexec setfscreate};
#line 5716

#line 5716
# Access /proc/PID files for processes in the same domain.
#line 5716
allow ifconfig_t self:dir { read getattr lock search ioctl };
#line 5716
allow ifconfig_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 5716

#line 5716
# Access file descriptions, pipes, and sockets
#line 5716
# created by processes in the same domain.
#line 5716
allow ifconfig_t self:fd *;
#line 5716
allow ifconfig_t self:fifo_file { ioctl read getattr lock write append };
#line 5716
allow ifconfig_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 5716
allow ifconfig_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 5716

#line 5716
# Allow the domain to communicate with other processes in the same domain.
#line 5716
allow ifconfig_t self:unix_dgram_socket sendto;
#line 5716
allow ifconfig_t self:unix_stream_socket connectto;
#line 5716

#line 5716
# Access System V IPC objects created by processes in the same domain.
#line 5716
allow ifconfig_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 5716
allow ifconfig_t self:msg  { send receive };
#line 5716
allow ifconfig_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 5716
allow ifconfig_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 5716



#line 5718

#line 5718

#line 5718
#
#line 5718
# Allow the process to transition to the new domain.
#line 5718
#
#line 5718
allow initrc_t ifconfig_t:process transition;
#line 5718

#line 5718
#
#line 5718
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5718
#
#line 5718
dontaudit initrc_t ifconfig_t:process noatsecure;
#line 5718

#line 5718
#
#line 5718
# Allow the process to execute the program.
#line 5718
# 
#line 5718
allow initrc_t ifconfig_exec_t:file { read { getattr execute } };
#line 5718

#line 5718
#
#line 5718
# Allow the process to reap the new domain.
#line 5718
#
#line 5718
allow ifconfig_t initrc_t:process sigchld;
#line 5718

#line 5718
#
#line 5718
# Allow the new domain to inherit and use file 
#line 5718
# descriptions from the creating process and vice versa.
#line 5718
#
#line 5718
allow ifconfig_t initrc_t:fd use;
#line 5718
allow initrc_t ifconfig_t:fd use;
#line 5718

#line 5718
#
#line 5718
# Allow the new domain to write back to the old domain via a pipe.
#line 5718
#
#line 5718
allow ifconfig_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 5718

#line 5718
#
#line 5718
# Allow the new domain to read and execute the program.
#line 5718
#
#line 5718
allow ifconfig_t ifconfig_exec_t:file { read getattr lock execute ioctl };
#line 5718

#line 5718
#
#line 5718
# Allow the new domain to be entered via the program.
#line 5718
#
#line 5718
allow ifconfig_t ifconfig_exec_t:file entrypoint;
#line 5718

#line 5718
type_transition initrc_t ifconfig_exec_t:process ifconfig_t;
#line 5718


#line 5719

#line 5719

#line 5719
#
#line 5719
# Allow the process to transition to the new domain.
#line 5719
#
#line 5719
allow sysadm_t ifconfig_t:process transition;
#line 5719

#line 5719
#
#line 5719
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5719
#
#line 5719
dontaudit sysadm_t ifconfig_t:process noatsecure;
#line 5719

#line 5719
#
#line 5719
# Allow the process to execute the program.
#line 5719
# 
#line 5719
allow sysadm_t ifconfig_exec_t:file { read { getattr execute } };
#line 5719

#line 5719
#
#line 5719
# Allow the process to reap the new domain.
#line 5719
#
#line 5719
allow ifconfig_t sysadm_t:process sigchld;
#line 5719

#line 5719
#
#line 5719
# Allow the new domain to inherit and use file 
#line 5719
# descriptions from the creating process and vice versa.
#line 5719
#
#line 5719
allow ifconfig_t sysadm_t:fd use;
#line 5719
allow sysadm_t ifconfig_t:fd use;
#line 5719

#line 5719
#
#line 5719
# Allow the new domain to write back to the old domain via a pipe.
#line 5719
#
#line 5719
allow ifconfig_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 5719

#line 5719
#
#line 5719
# Allow the new domain to read and execute the program.
#line 5719
#
#line 5719
allow ifconfig_t ifconfig_exec_t:file { read getattr lock execute ioctl };
#line 5719

#line 5719
#
#line 5719
# Allow the new domain to be entered via the program.
#line 5719
#
#line 5719
allow ifconfig_t ifconfig_exec_t:file entrypoint;
#line 5719

#line 5719
type_transition sysadm_t ifconfig_exec_t:process ifconfig_t;
#line 5719


# Use capabilities.
allow ifconfig_t ifconfig_t:capability { sys_module net_admin };

# Inherit and use descriptors from init.
allow ifconfig_t init_t:fd use;

# Execute insmod.

#line 5728

#line 5728

#line 5728
#
#line 5728
# Allow the process to transition to the new domain.
#line 5728
#
#line 5728
allow ifconfig_t insmod_t:process transition;
#line 5728

#line 5728
#
#line 5728
# Do not audit when glibc secure mode is enabled upon the transition.
#line 5728
#
#line 5728
dontaudit ifconfig_t insmod_t:process noatsecure;
#line 5728

#line 5728
#
#line 5728
# Allow the process to execute the program.
#line 5728
# 
#line 5728
allow ifconfig_t insmod_exec_t:file { read { getattr execute } };
#line 5728

#line 5728
#
#line 5728
# Allow the process to reap the new domain.
#line 5728
#
#line 5728
allow insmod_t ifconfig_t:process sigchld;
#line 5728

#line 5728
#
#line 5728
# Allow the new domain to inherit and use file 
#line 5728
# descriptions from the creating process and vice versa.
#line 5728
#
#line 5728
allow insmod_t ifconfig_t:fd use;
#line 5728
allow ifconfig_t insmod_t:fd use;
#line 5728

#line 5728
#
#line 5728
# Allow the new domain to write back to the old domain via a pipe.
#line 5728
#
#line 5728
allow insmod_t ifconfig_t:fifo_file { ioctl read getattr lock write append };
#line 5728

#line 5728
#
#line 5728
# Allow the new domain to read and execute the program.
#line 5728
#
#line 5728
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 5728

#line 5728
#
#line 5728
# Allow the new domain to be entered via the program.
#line 5728
#
#line 5728
allow insmod_t insmod_exec_t:file entrypoint;
#line 5728

#line 5728
type_transition ifconfig_t insmod_exec_t:process insmod_t;
#line 5728


# Access /proc
allow ifconfig_t proc_t:dir { read getattr lock search ioctl };
allow ifconfig_t proc_t:file { read getattr lock ioctl };

allow ifconfig_t privfd:fd use;

# Create UDP sockets, necessary when called from dhcpc
allow ifconfig_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Access terminals.
allow ifconfig_t { initrc_devpts_t { sysadm_tty_device_t sysadm_devpts_t } }:chr_file { ioctl read getattr lock write append };


allow ifconfig_t tun_tap_device_t:chr_file { read write };

# ifconfig attempts to search some sysctl entries.
# Do not audit those attempts; comment out these rules if it is desired to
# see the denials.
dontaudit ifconfig_t { sysctl_t sysctl_net_t }:dir search;

allow ifconfig_t fs_t:filesystem getattr;
#DESC Initrc - System initialization scripts
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: sysvinit policycoreutils
#

#################################
#
# Rules for the initrc_t domain.
#
# initrc_t is the domain of the init rc scripts.
# initrc_exec_t is the type of the init program.
#
#line 5769

#line 5769
type initrc_t, domain, privlog, privowner, privmail;
#line 5769

role system_r types initrc_t;

#line 5771
allow initrc_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 5771
allow initrc_t lib_t:lnk_file { read getattr lock ioctl };
#line 5771
allow initrc_t ld_so_t:file { read getattr lock execute ioctl };
#line 5771
allow initrc_t ld_so_t:file execute_no_trans;
#line 5771
allow initrc_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 5771
allow initrc_t shlib_t:file { read getattr lock execute ioctl };
#line 5771
allow initrc_t shlib_t:lnk_file { read getattr lock ioctl };
#line 5771
allow initrc_t ld_so_cache_t:file { read getattr lock ioctl };
#line 5771
allow initrc_t device_t:dir search;
#line 5771
allow initrc_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 5771
;
type initrc_exec_t, file_type, sysadmfile, exec_type;

# read files in /etc/init.d
allow initrc_t etc_t:lnk_file { read getattr lock ioctl };


#line 5777
allow initrc_t etc_t:lnk_file read;
#line 5777

#line 5777
allow initrc_t locale_t:dir { read getattr lock search ioctl };
#line 5777
allow initrc_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 5777

#line 5777


# Read system information files in /proc.
allow initrc_t proc_t:dir { read getattr lock search ioctl };
allow initrc_t proc_t:{ file lnk_file } { read getattr lock ioctl };

# Allow IPC with self
allow initrc_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow initrc_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow initrc_t self:fifo_file { ioctl read getattr lock write append };

# Read the root directory of a usbdevfs filesystem, and
# the devices and drivers files.  Permit stating of the
# device nodes, but nothing else.
allow initrc_t usbdevfs_t:dir { read getattr lock search ioctl };
allow initrc_t usbdevfs_t:{ file lnk_file } { read getattr lock ioctl };
allow initrc_t usbdevfs_device_t:file getattr;

# allow initrc to fork and renice itself
allow initrc_t self:process { fork sigchld setsched };

# Can create ptys for open_init_pty

#line 5799

#line 5799
# Access the pty master multiplexer.
#line 5799
allow initrc_t ptmx_t:chr_file { ioctl read getattr lock write append };
#line 5799

#line 5799

#line 5799
allow initrc_t devpts_t:filesystem getattr;
#line 5799

#line 5799
# allow searching /dev/pts
#line 5799
allow initrc_t devpts_t:dir { getattr read search };
#line 5799

#line 5799
type initrc_devpts_t, file_type, sysadmfile, ptyfile ;
#line 5799

#line 5799
# Allow the pty to be associated with the file system.
#line 5799
allow initrc_devpts_t devpts_t:filesystem associate;
#line 5799

#line 5799
# Label pty files with a derived type.
#line 5799
type_transition initrc_t devpts_t:chr_file initrc_devpts_t;
#line 5799

#line 5799
# Read and write my pty files.
#line 5799
allow initrc_t initrc_devpts_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 5799


type initrc_tmp_t, file_type, sysadmfile, tmpfile;

#line 5802

#line 5802

#line 5802

#line 5802
#
#line 5802
# Allow the process to modify the directory.
#line 5802
#
#line 5802
allow initrc_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5802

#line 5802
#
#line 5802
# Allow the process to create the file.
#line 5802
#
#line 5802

#line 5802
allow initrc_t initrc_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5802
allow initrc_t initrc_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5802

#line 5802

#line 5802
type_transition initrc_t tmp_t:dir initrc_tmp_t;
#line 5802
type_transition initrc_t tmp_t:{ file lnk_file sock_file fifo_file } initrc_tmp_t;
#line 5802

#line 5802

#line 5802



#line 5804
type initrc_var_run_t alias var_run_initrc_t, file_type, sysadmfile, pidfile;
#line 5804

#line 5804

#line 5804

#line 5804

#line 5804
#
#line 5804
# Allow the process to modify the directory.
#line 5804
#
#line 5804
allow initrc_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5804

#line 5804
#
#line 5804
# Allow the process to create the file.
#line 5804
#
#line 5804

#line 5804

#line 5804
allow initrc_t initrc_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 5804

#line 5804

#line 5804
type_transition initrc_t var_run_t:file initrc_var_run_t;
#line 5804

#line 5804

#line 5804

#line 5804
allow initrc_t initrc_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5804

allow initrc_t var_run_t:{ file sock_file lnk_file } unlink;

allow initrc_t framebuf_device_t:chr_file { read getattr lock ioctl };

# Use capabilities.
allow initrc_t initrc_t:capability ~sys_module;

# Use system operations.
allow initrc_t kernel_t:system *;

# Allow initrc_t to set the enforcing flag.
# The initrc_t domain is granted this permission for 
# people who want to use a development kernel and toggle
# into enforcing mode from an /etc/rc.d script.

#line 5819
allow initrc_t security_t:dir { read search getattr };
#line 5819
allow initrc_t security_t:file { getattr read write };
#line 5819
allow initrc_t security_t:security setenforce;
#line 5819
auditallow initrc_t security_t:security setenforce;
#line 5819


# Set values in /proc/sys.

#line 5822
allow initrc_t sysctl_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_fs_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_fs_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_kernel_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_kernel_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_net_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_net_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_net_unix_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_net_unix_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_vm_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_vm_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_dev_t:dir { read getattr lock search ioctl };
#line 5822
allow initrc_t sysctl_dev_t:file { setattr { ioctl read getattr lock write append } };
#line 5822
allow initrc_t sysctl_modprobe_t:file { setattr { ioctl read getattr lock write append } };
#line 5822


# Run helper programs in the initrc_t domain.
allow initrc_t {bin_t sbin_t }:dir { read getattr lock search ioctl };
allow initrc_t {bin_t sbin_t }:lnk_file read;

#line 5827
allow initrc_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5827


#line 5828
allow initrc_t lib_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5828


#line 5829
allow initrc_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5829


#line 5830
allow initrc_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5830


#line 5831
allow initrc_t exec_type:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5831



# Read conf.modules.
allow initrc_t modules_conf_t:file { read getattr lock ioctl };

# Run other rc scripts in the initrc_t domain.

#line 5838
allow initrc_t initrc_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5838


# Run init (telinit) in the initrc_t domain.

#line 5841
allow initrc_t init_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 5841


# Communicate with the init process.
allow initrc_t initctl_t:fifo_file { ioctl read getattr lock write append };

# Send messages to portmap and ypbind.



# Search persistent label mappings.
allow initrc_t file_labels_t:dir { read getattr lock search ioctl };
allow initrc_t file_labels_t:file { getattr };

# Read /proc/PID directories for all domains.
allow initrc_t domain:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
allow initrc_t domain:dir { read getattr lock search ioctl };

# Mount and unmount file systems.
allow initrc_t fs_type:filesystem { mount remount unmount getattr };
allow initrc_t file_t:dir { read search getattr mounton };

# Create runtime files in /etc, e.g. /etc/mtab, /etc/HOSTNAME.

#line 5863

#line 5863

#line 5863

#line 5863
#
#line 5863
# Allow the process to modify the directory.
#line 5863
#
#line 5863
allow initrc_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5863

#line 5863
#
#line 5863
# Allow the process to create the file.
#line 5863
#
#line 5863

#line 5863

#line 5863
allow initrc_t etc_runtime_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 5863

#line 5863

#line 5863
type_transition initrc_t etc_t:file etc_runtime_t;
#line 5863

#line 5863

#line 5863


# Update /etc/ld.so.cache.
allow initrc_t ld_so_cache_t:file { ioctl read getattr lock write append };

#line 5871


#line 5877


# Update /var/log/wtmp and /var/log/dmesg.
allow initrc_t wtmp_t:file { setattr { ioctl read getattr lock write append } };
allow initrc_t var_log_t:file { setattr { ioctl read getattr lock write append } };
allow initrc_t lastlog_t:file { setattr { ioctl read getattr lock write append } };

# remove old locks
allow initrc_t lockfile:dir { read getattr lock search ioctl add_name remove_name write };
allow initrc_t lockfile:file { getattr unlink };

# Access /var/lib/random-seed.
allow initrc_t var_lib_t:file { ioctl read getattr lock write append };
allow initrc_t var_lib_t:file unlink;

# Create lock file.
allow initrc_t var_lock_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
allow initrc_t var_lock_t:file { create ioctl read getattr lock write setattr append link unlink rename };

# Set the clock.
allow initrc_t clock_device_t:{ chr_file blk_file } { ioctl read getattr lock write append };

# Kill all processes.
allow initrc_t domain:process { sigchld sigkill sigstop signull signal };

# Read and unlink /var/run/*.pid files.
allow initrc_t pidfile:file { getattr read unlink };

# Write to /dev/urandom.
allow initrc_t random_device_t:chr_file { ioctl read getattr lock write append };

# Access /dev/psaux (for kudzu).
allow initrc_t psaux_t:chr_file { ioctl read getattr lock write append };

# Set device ownerships/modes.
allow initrc_t framebuf_device_t:lnk_file read;
allow initrc_t framebuf_device_t:{ chr_file blk_file } setattr;
allow initrc_t misc_device_t:{ chr_file blk_file } setattr;
allow initrc_t device_t:{ chr_file blk_file } setattr;
allow initrc_t fixed_disk_device_t:{ chr_file blk_file } setattr;
allow initrc_t removable_device_t:{ chr_file blk_file } setattr;

# Stat any file.
allow initrc_t file_type:{ file lnk_file sock_file fifo_file chr_file blk_file } getattr;
allow initrc_t file_type:dir { search getattr };

# Read and write console and ttys.
allow initrc_t devtty_t:chr_file { ioctl read getattr lock write append };
allow initrc_t console_device_t:chr_file { ioctl read getattr lock write append };
allow initrc_t tty_device_t:chr_file { ioctl read getattr lock write append };
allow initrc_t ttyfile:chr_file { ioctl read getattr lock write append };
allow initrc_t ptyfile:chr_file { ioctl read getattr lock write append };

# Reset tty labels.
allow initrc_t ttyfile:chr_file relabelfrom;
allow initrc_t tty_device_t:chr_file relabelto;

#line 5943


allow initrc_t system_map_t:{ file lnk_file } { read getattr lock ioctl };

# Unlink /halt.
allow initrc_t root_t:dir { search write remove_name };
allow initrc_t root_t:file { unlink write };



allow initrc_t var_spool_t:file { ioctl read getattr lock write append };

#
# quota control
#
allow initrc_t fs_type:filesystem { quotamod quotaget };

# Access the mouse (for kudzu).
allow initrc_t mouse_device_t:chr_file { ioctl read getattr lock write append };

# Allow access to the sysadm TTYs. Note that this will give access to the 
# TTYs to any process in the initrc_t domain. Therefore, daemons and such
# started from init should be placed in their own domain.
allow initrc_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { ioctl read getattr lock write append };

# Access sound device and files.
allow initrc_t sound_device_t:chr_file { setattr ioctl read write };


#line 5976


# Update /var/log/ksyms.*.

#line 5979

#line 5979

#line 5979

#line 5979
#
#line 5979
# Allow the process to modify the directory.
#line 5979
#
#line 5979
allow initrc_t var_log_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 5979

#line 5979
#
#line 5979
# Allow the process to create the file.
#line 5979
#
#line 5979

#line 5979
allow initrc_t var_log_ksyms_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 5979
allow initrc_t var_log_ksyms_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 5979

#line 5979

#line 5979
type_transition initrc_t var_log_t:dir var_log_ksyms_t;
#line 5979
type_transition initrc_t var_log_t:{ file lnk_file sock_file fifo_file } var_log_ksyms_t;
#line 5979

#line 5979

#line 5979


#line 5983


# Access /dev/fd0 (for kudzu). Need a separate domain for kudzu?
allow initrc_t removable_device_t:blk_file { ioctl read getattr lock write append };

#line 5991


# Create and delete /.autofsck
allow initrc_t root_t:dir { search write add_name };
allow initrc_t root_t:file { create setattr unlink getattr };
allow initrc_t file_t:file { unlink getattr };

# Read user home directories.
allow initrc_t { home_root_t home_type }:dir { read getattr lock search ioctl };
allow initrc_t home_type:file { read getattr lock ioctl };

# for system start scripts
allow initrc_t pidfile:dir { read getattr lock search ioctl add_name remove_name write };
allow initrc_t pidfile:sock_file unlink;
allow initrc_t tmpfile:sock_file unlink;

#line 6006
allow initrc_t var_lib_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6006
allow initrc_t var_lib_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6006


# allow start scripts to clean /tmp
allow initrc_t tmpfile:dir { { read getattr lock search ioctl add_name remove_name write } rmdir };
allow initrc_t tmpfile:{ file lnk_file sock_file fifo_file } { getattr unlink };

#################################
#
# Rules for the run_init_t domain.
#

#line 6016
type run_init_exec_t, file_type, exec_type, sysadmfile;
#line 6016

#line 6016
# domain for program to run in, needs to change role (privrole), change
#line 6016
# identity to system_u (privuser), log failures to syslog (privlog) and
#line 6016
# authenticate users (auth)
#line 6016
type run_init_t, domain, privrole, privuser, privlog, auth;
#line 6016

#line 6016

#line 6016

#line 6016
#
#line 6016
# Allow the process to transition to the new domain.
#line 6016
#
#line 6016
allow sysadm_t run_init_t:process transition;
#line 6016

#line 6016
#
#line 6016
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6016
#
#line 6016
dontaudit sysadm_t run_init_t:process noatsecure;
#line 6016

#line 6016
#
#line 6016
# Allow the process to execute the program.
#line 6016
# 
#line 6016
allow sysadm_t run_init_exec_t:file { read { getattr execute } };
#line 6016

#line 6016
#
#line 6016
# Allow the process to reap the new domain.
#line 6016
#
#line 6016
allow run_init_t sysadm_t:process sigchld;
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to inherit and use file 
#line 6016
# descriptions from the creating process and vice versa.
#line 6016
#
#line 6016
allow run_init_t sysadm_t:fd use;
#line 6016
allow sysadm_t run_init_t:fd use;
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to write back to the old domain via a pipe.
#line 6016
#
#line 6016
allow run_init_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to read and execute the program.
#line 6016
#
#line 6016
allow run_init_t run_init_exec_t:file { read getattr lock execute ioctl };
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to be entered via the program.
#line 6016
#
#line 6016
allow run_init_t run_init_exec_t:file entrypoint;
#line 6016

#line 6016
type_transition sysadm_t run_init_exec_t:process run_init_t;
#line 6016

#line 6016
role sysadm_r types run_init_t;
#line 6016

#line 6016
# for utmp
#line 6016
allow run_init_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 6016
allow run_init_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { ioctl read getattr lock write append };
#line 6016

#line 6016
dontaudit run_init_t devpts_t:dir { getattr read };
#line 6016
dontaudit run_init_t device_t:dir read;
#line 6016

#line 6016
# often the administrator runs such programs from a directory that is owned
#line 6016
# by a different user or has restrictive SE permissions, do not want to audit
#line 6016
# the failed access to the current directory
#line 6016
dontaudit run_init_t file_type:dir search;
#line 6016
dontaudit run_init_t self:capability { dac_override dac_read_search };
#line 6016

#line 6016

#line 6016
allow run_init_t { bin_t chkpwd_exec_t shell_exec_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6016

#line 6016

#line 6016

#line 6016

#line 6016
#
#line 6016
# Allow the process to transition to the new domain.
#line 6016
#
#line 6016
allow run_init_t initrc_t:process transition;
#line 6016

#line 6016
#
#line 6016
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6016
#
#line 6016
dontaudit run_init_t initrc_t:process noatsecure;
#line 6016

#line 6016
#
#line 6016
# Allow the process to execute the program.
#line 6016
# 
#line 6016
allow run_init_t initrc_exec_t:file { read { getattr execute } };
#line 6016

#line 6016
#
#line 6016
# Allow the process to reap the new domain.
#line 6016
#
#line 6016
allow initrc_t run_init_t:process sigchld;
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to inherit and use file 
#line 6016
# descriptions from the creating process and vice versa.
#line 6016
#
#line 6016
allow initrc_t run_init_t:fd use;
#line 6016
allow run_init_t initrc_t:fd use;
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to write back to the old domain via a pipe.
#line 6016
#
#line 6016
allow initrc_t run_init_t:fifo_file { ioctl read getattr lock write append };
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to read and execute the program.
#line 6016
#
#line 6016
allow initrc_t initrc_exec_t:file { read getattr lock execute ioctl };
#line 6016

#line 6016
#
#line 6016
# Allow the new domain to be entered via the program.
#line 6016
#
#line 6016
allow initrc_t initrc_exec_t:file entrypoint;
#line 6016

#line 6016

#line 6016
allow run_init_t self:process setexec;
#line 6016
allow run_init_t proc_t:dir search;
#line 6016
allow run_init_t proc_t:lnk_file read;
#line 6016
allow run_init_t self:dir search;
#line 6016
allow run_init_t self:file { read write };
#line 6016

#line 6016

#line 6016
allow run_init_t privfd:fd use;
#line 6016

#line 6016
allow run_init_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6016
allow run_init_t lib_t:lnk_file { read getattr lock ioctl };
#line 6016
allow run_init_t ld_so_t:file { read getattr lock execute ioctl };
#line 6016
allow run_init_t ld_so_t:file execute_no_trans;
#line 6016
allow run_init_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6016
allow run_init_t shlib_t:file { read getattr lock execute ioctl };
#line 6016
allow run_init_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6016
allow run_init_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6016
allow run_init_t device_t:dir search;
#line 6016
allow run_init_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6016

#line 6016

#line 6016
allow run_init_t security_t:dir { read search getattr };
#line 6016
allow run_init_t security_t:file { getattr read write };
#line 6016
allow run_init_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 6016

#line 6016
allow run_init_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6016
allow run_init_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6016
allow run_init_t etc_t:file { getattr read };
#line 6016

#line 6016
allow run_init_t etc_t:lnk_file read;
#line 6016

#line 6016
allow run_init_t locale_t:dir { read getattr lock search ioctl };
#line 6016
allow run_init_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6016

#line 6016

#line 6016
allow run_init_t fs_t:filesystem getattr;
#line 6016
allow run_init_t { bin_t sbin_t }:dir search;
#line 6016
dontaudit run_init_t { proc_t device_t }:dir { getattr search };
#line 6016

allow initrc_t privfd:fd use;
#DESC Init - Process initialization
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: sysvinit
#

#################################
#
# Rules for the init_t domain.
#
# init_t is the domain of the init process.
# init_exec_t is the type of the init program.
# initctl_t is the type of the named pipe created 
# by init during initialization.  This pipe is used
# to communicate with init.
# sulogin_exec_t is the type of sulogin.
#
type init_t, domain, privlog, mlstrustedreader, mlstrustedwriter;
role system_r types init_t;

#line 6037
allow init_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6037
allow init_t lib_t:lnk_file { read getattr lock ioctl };
#line 6037
allow init_t ld_so_t:file { read getattr lock execute ioctl };
#line 6037
allow init_t ld_so_t:file execute_no_trans;
#line 6037
allow init_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6037
allow init_t shlib_t:file { read getattr lock execute ioctl };
#line 6037
allow init_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6037
allow init_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6037
allow init_t device_t:dir search;
#line 6037
allow init_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6037
;
type init_exec_t, file_type, sysadmfile, exec_type;
type initctl_t, file_type, sysadmfile;
type sulogin_exec_t, file_type, exec_type, sysadmfile;

# for mount points
allow init_t file_t:dir search;

# Use capabilities.
allow init_t init_t:capability ~sys_module;

# Run /etc/rc.sysinit, /etc/rc, /etc/rc.local in the initrc_t domain.

#line 6049

#line 6049

#line 6049
#
#line 6049
# Allow the process to transition to the new domain.
#line 6049
#
#line 6049
allow init_t initrc_t:process transition;
#line 6049

#line 6049
#
#line 6049
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6049
#
#line 6049
dontaudit init_t initrc_t:process noatsecure;
#line 6049

#line 6049
#
#line 6049
# Allow the process to execute the program.
#line 6049
# 
#line 6049
allow init_t initrc_exec_t:file { read { getattr execute } };
#line 6049

#line 6049
#
#line 6049
# Allow the process to reap the new domain.
#line 6049
#
#line 6049
allow initrc_t init_t:process sigchld;
#line 6049

#line 6049
#
#line 6049
# Allow the new domain to inherit and use file 
#line 6049
# descriptions from the creating process and vice versa.
#line 6049
#
#line 6049
allow initrc_t init_t:fd use;
#line 6049
allow init_t initrc_t:fd use;
#line 6049

#line 6049
#
#line 6049
# Allow the new domain to write back to the old domain via a pipe.
#line 6049
#
#line 6049
allow initrc_t init_t:fifo_file { ioctl read getattr lock write append };
#line 6049

#line 6049
#
#line 6049
# Allow the new domain to read and execute the program.
#line 6049
#
#line 6049
allow initrc_t initrc_exec_t:file { read getattr lock execute ioctl };
#line 6049

#line 6049
#
#line 6049
# Allow the new domain to be entered via the program.
#line 6049
#
#line 6049
allow initrc_t initrc_exec_t:file entrypoint;
#line 6049

#line 6049
type_transition init_t initrc_exec_t:process initrc_t;
#line 6049


# Run the shell or sulogin in the sysadm_t domain for single-user mode.

#line 6052

#line 6052

#line 6052
#
#line 6052
# Allow the process to transition to the new domain.
#line 6052
#
#line 6052
allow init_t sysadm_t:process transition;
#line 6052

#line 6052
#
#line 6052
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6052
#
#line 6052
dontaudit init_t sysadm_t:process noatsecure;
#line 6052

#line 6052
#
#line 6052
# Allow the process to execute the program.
#line 6052
# 
#line 6052
allow init_t shell_exec_t:file { read { getattr execute } };
#line 6052

#line 6052
#
#line 6052
# Allow the process to reap the new domain.
#line 6052
#
#line 6052
allow sysadm_t init_t:process sigchld;
#line 6052

#line 6052
#
#line 6052
# Allow the new domain to inherit and use file 
#line 6052
# descriptions from the creating process and vice versa.
#line 6052
#
#line 6052
allow sysadm_t init_t:fd use;
#line 6052
allow init_t sysadm_t:fd use;
#line 6052

#line 6052
#
#line 6052
# Allow the new domain to write back to the old domain via a pipe.
#line 6052
#
#line 6052
allow sysadm_t init_t:fifo_file { ioctl read getattr lock write append };
#line 6052

#line 6052
#
#line 6052
# Allow the new domain to read and execute the program.
#line 6052
#
#line 6052
allow sysadm_t shell_exec_t:file { read getattr lock execute ioctl };
#line 6052

#line 6052
#
#line 6052
# Allow the new domain to be entered via the program.
#line 6052
#
#line 6052
allow sysadm_t shell_exec_t:file entrypoint;
#line 6052

#line 6052
type_transition init_t shell_exec_t:process sysadm_t;
#line 6052


#line 6053

#line 6053

#line 6053
#
#line 6053
# Allow the process to transition to the new domain.
#line 6053
#
#line 6053
allow init_t sysadm_t:process transition;
#line 6053

#line 6053
#
#line 6053
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6053
#
#line 6053
dontaudit init_t sysadm_t:process noatsecure;
#line 6053

#line 6053
#
#line 6053
# Allow the process to execute the program.
#line 6053
# 
#line 6053
allow init_t sulogin_exec_t:file { read { getattr execute } };
#line 6053

#line 6053
#
#line 6053
# Allow the process to reap the new domain.
#line 6053
#
#line 6053
allow sysadm_t init_t:process sigchld;
#line 6053

#line 6053
#
#line 6053
# Allow the new domain to inherit and use file 
#line 6053
# descriptions from the creating process and vice versa.
#line 6053
#
#line 6053
allow sysadm_t init_t:fd use;
#line 6053
allow init_t sysadm_t:fd use;
#line 6053

#line 6053
#
#line 6053
# Allow the new domain to write back to the old domain via a pipe.
#line 6053
#
#line 6053
allow sysadm_t init_t:fifo_file { ioctl read getattr lock write append };
#line 6053

#line 6053
#
#line 6053
# Allow the new domain to read and execute the program.
#line 6053
#
#line 6053
allow sysadm_t sulogin_exec_t:file { read getattr lock execute ioctl };
#line 6053

#line 6053
#
#line 6053
# Allow the new domain to be entered via the program.
#line 6053
#
#line 6053
allow sysadm_t sulogin_exec_t:file entrypoint;
#line 6053

#line 6053
type_transition init_t sulogin_exec_t:process sysadm_t;
#line 6053


# Run /sbin/update in the init_t domain.

#line 6056
allow init_t sbin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6056


# Run init.

#line 6059
allow init_t init_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6059


# Run chroot from initrd scripts.
#line 6064


# Create /dev/initctl.

#line 6067

#line 6067

#line 6067

#line 6067
#
#line 6067
# Allow the process to modify the directory.
#line 6067
#
#line 6067
allow init_t device_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6067

#line 6067
#
#line 6067
# Allow the process to create the file.
#line 6067
#
#line 6067

#line 6067

#line 6067
allow init_t initctl_t:fifo_file { create ioctl read getattr lock write setattr append link unlink rename };
#line 6067

#line 6067

#line 6067
type_transition init_t device_t:fifo_file initctl_t;
#line 6067

#line 6067

#line 6067


# Create ioctl.save.

#line 6070

#line 6070

#line 6070

#line 6070
#
#line 6070
# Allow the process to modify the directory.
#line 6070
#
#line 6070
allow init_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6070

#line 6070
#
#line 6070
# Allow the process to create the file.
#line 6070
#
#line 6070

#line 6070

#line 6070
allow init_t etc_runtime_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 6070

#line 6070

#line 6070
type_transition init_t etc_t:file etc_runtime_t;
#line 6070

#line 6070

#line 6070


# Update /etc/ld.so.cache
allow init_t ld_so_cache_t:file { ioctl read getattr lock write append };

# Allow access to log files
allow init_t var_t:dir search;
allow init_t var_log_t:dir search;


#line 6079
allow init_t etc_t:lnk_file read;
#line 6079

#line 6079
allow init_t locale_t:dir { read getattr lock search ioctl };
#line 6079
allow init_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6079

#line 6079


# Create unix sockets
allow init_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow init_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow init_t self:fifo_file { ioctl read getattr lock write append };

# Permissions required for system startup
allow init_t bin_t:dir { read getattr lock search ioctl };
allow init_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
allow init_t exec_type:{ file lnk_file } { read getattr lock ioctl };
allow init_t sbin_t:dir { read getattr lock search ioctl };
allow init_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };

# allow init to fork
allow init_t self:process { fork sigchld };

# Modify utmp.
allow init_t var_run_t:file { ioctl read getattr lock write append };
allow init_t initrc_var_run_t:file { setattr { ioctl read getattr lock write append } };

# For /var/run/shutdown.pid.

#line 6101
type init_var_run_t alias var_run_init_t, file_type, sysadmfile, pidfile;
#line 6101

#line 6101

#line 6101

#line 6101

#line 6101
#
#line 6101
# Allow the process to modify the directory.
#line 6101
#
#line 6101
allow init_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6101

#line 6101
#
#line 6101
# Allow the process to create the file.
#line 6101
#
#line 6101

#line 6101

#line 6101
allow init_t init_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 6101

#line 6101

#line 6101
type_transition init_t var_run_t:file init_var_run_t;
#line 6101

#line 6101

#line 6101

#line 6101
allow init_t init_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6101


# Shutdown permissions
allow init_t proc_t:dir { read getattr lock search ioctl };
allow init_t proc_t:lnk_file { read getattr lock ioctl };
allow init_t self:dir { read getattr lock search ioctl };
allow init_t self:lnk_file { read getattr lock ioctl };
allow init_t devpts_t:dir { read getattr lock ioctl };

# Modify wtmp.
allow init_t wtmp_t:file { ioctl read getattr lock write append };

# Kill all processes.
allow init_t domain:process { sigchld sigkill sigstop signull signal };

# Allow all processes to send SIGCHLD to init.
allow domain init_t:process { sigchld signull };

# If you load a new policy that removes active domains, processes can
# get stuck if you do not allow unlabeled processes to signal init
# If you load an incompatible policy, you should probably reboot,
# since you may have compromised system security.
allow unlabeled_t init_t:process sigchld;

# Read and write the console and ttys.
allow init_t console_device_t:chr_file { ioctl read getattr lock write append };
allow init_t tty_device_t:chr_file { ioctl read getattr lock write append };
allow init_t ttyfile:chr_file { ioctl read getattr lock write append };
allow init_t ptyfile:chr_file { ioctl read getattr lock write append };

# Run system executables.

#line 6132
allow init_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6132


# Run /etc/X11/prefdm.

#line 6135
allow init_t etc_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6135


# for initrd pivot_root seems to access this
dontaudit init_t file_labels_t:dir { search };
#DESC Klogd - Kernel log daemon
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: klogd
#

#################################
#
# Rules for the klogd_t domain.
#

#line 6149

#line 6149
type klogd_t, domain, privlog , privmem;
#line 6149
type klogd_exec_t, file_type, sysadmfile, exec_type;
#line 6149

#line 6149
role system_r types klogd_t;
#line 6149

#line 6149

#line 6149

#line 6149

#line 6149
#
#line 6149
# Allow the process to transition to the new domain.
#line 6149
#
#line 6149
allow initrc_t klogd_t:process transition;
#line 6149

#line 6149
#
#line 6149
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6149
#
#line 6149
dontaudit initrc_t klogd_t:process noatsecure;
#line 6149

#line 6149
#
#line 6149
# Allow the process to execute the program.
#line 6149
# 
#line 6149
allow initrc_t klogd_exec_t:file { read { getattr execute } };
#line 6149

#line 6149
#
#line 6149
# Allow the process to reap the new domain.
#line 6149
#
#line 6149
allow klogd_t initrc_t:process sigchld;
#line 6149

#line 6149
#
#line 6149
# Allow the new domain to inherit and use file 
#line 6149
# descriptions from the creating process and vice versa.
#line 6149
#
#line 6149
allow klogd_t initrc_t:fd use;
#line 6149
allow initrc_t klogd_t:fd use;
#line 6149

#line 6149
#
#line 6149
# Allow the new domain to write back to the old domain via a pipe.
#line 6149
#
#line 6149
allow klogd_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 6149

#line 6149
#
#line 6149
# Allow the new domain to read and execute the program.
#line 6149
#
#line 6149
allow klogd_t klogd_exec_t:file { read getattr lock execute ioctl };
#line 6149

#line 6149
#
#line 6149
# Allow the new domain to be entered via the program.
#line 6149
#
#line 6149
allow klogd_t klogd_exec_t:file entrypoint;
#line 6149

#line 6149
type_transition initrc_t klogd_exec_t:process klogd_t;
#line 6149

#line 6149

#line 6149
# Inherit and use descriptors from init.
#line 6149
allow klogd_t init_t:fd use;
#line 6149
allow klogd_t init_t:process sigchld;
#line 6149
allow klogd_t privfd:fd use;
#line 6149
allow klogd_t newrole_t:process sigchld;
#line 6149
allow klogd_t self:process { { sigchld sigkill sigstop signull signal } fork };
#line 6149

#line 6149

#line 6149
allow klogd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6149
allow klogd_t lib_t:lnk_file { read getattr lock ioctl };
#line 6149
allow klogd_t ld_so_t:file { read getattr lock execute ioctl };
#line 6149
allow klogd_t ld_so_t:file execute_no_trans;
#line 6149
allow klogd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6149
allow klogd_t shlib_t:file { read getattr lock execute ioctl };
#line 6149
allow klogd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6149
allow klogd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6149
allow klogd_t device_t:dir search;
#line 6149
allow klogd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6149

#line 6149

#line 6149
allow klogd_t { self proc_t }:dir { read getattr lock search ioctl };
#line 6149
allow klogd_t { self proc_t }:lnk_file read;
#line 6149

#line 6149
allow klogd_t device_t:dir { getattr search };
#line 6149
allow klogd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6149
allow klogd_t console_device_t:chr_file { ioctl read getattr lock write append };
#line 6149
allow klogd_t initrc_devpts_t:chr_file { ioctl read getattr lock write append };
#line 6149

#line 6149
# Create pid file.
#line 6149
allow klogd_t var_t:dir { getattr search };
#line 6149

#line 6149
type klogd_var_run_t alias var_run_klogd_t, file_type, sysadmfile, pidfile;
#line 6149

#line 6149

#line 6149

#line 6149

#line 6149
#
#line 6149
# Allow the process to modify the directory.
#line 6149
#
#line 6149
allow klogd_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6149

#line 6149
#
#line 6149
# Allow the process to create the file.
#line 6149
#
#line 6149

#line 6149

#line 6149
allow klogd_t klogd_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 6149

#line 6149

#line 6149
type_transition klogd_t var_run_t:file klogd_var_run_t;
#line 6149

#line 6149

#line 6149

#line 6149
allow klogd_t klogd_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6149

#line 6149

#line 6149
allow klogd_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 6149

#line 6149
# for daemons that look at /root on startup
#line 6149
dontaudit klogd_t sysadm_home_dir_t:dir search;
#line 6149

#line 6149
# for df
#line 6149
allow klogd_t fs_type:filesystem getattr;
#line 6149

#line 6149

#line 6149
allow klogd_t etc_t:lnk_file read;
#line 6149

#line 6149
allow klogd_t locale_t:dir { read getattr lock search ioctl };
#line 6149
allow klogd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6149

#line 6149

#line 6149



#line 6151
type klogd_tmp_t, file_type, sysadmfile, tmpfile ;
#line 6151

#line 6151

#line 6151

#line 6151

#line 6151
#
#line 6151
# Allow the process to modify the directory.
#line 6151
#
#line 6151
allow klogd_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6151

#line 6151
#
#line 6151
# Allow the process to create the file.
#line 6151
#
#line 6151

#line 6151
allow klogd_t klogd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6151
allow klogd_t klogd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6151

#line 6151

#line 6151
type_transition klogd_t tmp_t:dir klogd_tmp_t;
#line 6151
type_transition klogd_t tmp_t:{ file lnk_file sock_file fifo_file } klogd_tmp_t;
#line 6151

#line 6151

#line 6151

#line 6151

allow klogd_t proc_t:dir { read getattr lock search ioctl };
allow klogd_t proc_t:lnk_file { read getattr lock ioctl };
allow klogd_t proc_t:file { getattr read };
allow klogd_t self:dir { read getattr lock search ioctl };
allow klogd_t self:lnk_file { read getattr lock ioctl };

# read /etc/nsswitch.conf
allow klogd_t etc_t:lnk_file read;
allow klogd_t etc_t:file { read getattr lock ioctl };


#line 6162
allow klogd_t etc_t:lnk_file read;
#line 6162

#line 6162
allow klogd_t locale_t:dir { read getattr lock search ioctl };
#line 6162
allow klogd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6162

#line 6162


allow klogd_t etc_runtime_t:file { getattr read };

# Create unix sockets
allow klogd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Use the sys_admin and sys_rawio capabilities.
allow klogd_t klogd_t:capability { sys_admin sys_rawio };

# Read /proc/kmsg and /dev/mem.
allow klogd_t proc_kmsg_t:file { read getattr lock ioctl };
allow klogd_t memory_device_t:chr_file { read getattr lock ioctl };

# Control syslog and console logging
allow klogd_t kernel_t:system { syslog_mod syslog_console };

# Read /boot/System.map*
allow klogd_t system_map_t:file { read getattr lock ioctl };
allow klogd_t boot_t:dir { read getattr lock search ioctl };
#DESC Ldconfig - Configure dynamic linker bindings
#
# Author:  Russell Coker <russell@coker.com.au>
# X-Debian-Packages: libc6
#

#################################
#
# Rules for the ldconfig_t domain.
#
type ldconfig_t, domain, privlog;
type ldconfig_exec_t, file_type, sysadmfile, exec_type;

role sysadm_r types ldconfig_t;
role system_r types ldconfig_t;


#line 6198

#line 6198

#line 6198
#
#line 6198
# Allow the process to transition to the new domain.
#line 6198
#
#line 6198
allow { sysadm_t initrc_t dpkg_t } ldconfig_t:process transition;
#line 6198

#line 6198
#
#line 6198
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6198
#
#line 6198
dontaudit { sysadm_t initrc_t dpkg_t } ldconfig_t:process noatsecure;
#line 6198

#line 6198
#
#line 6198
# Allow the process to execute the program.
#line 6198
# 
#line 6198
allow { sysadm_t initrc_t dpkg_t } ldconfig_exec_t:file { read { getattr execute } };
#line 6198

#line 6198
#
#line 6198
# Allow the process to reap the new domain.
#line 6198
#
#line 6198
allow ldconfig_t { sysadm_t initrc_t dpkg_t }:process sigchld;
#line 6198

#line 6198
#
#line 6198
# Allow the new domain to inherit and use file 
#line 6198
# descriptions from the creating process and vice versa.
#line 6198
#
#line 6198
allow ldconfig_t { sysadm_t initrc_t dpkg_t }:fd use;
#line 6198
allow { sysadm_t initrc_t dpkg_t } ldconfig_t:fd use;
#line 6198

#line 6198
#
#line 6198
# Allow the new domain to write back to the old domain via a pipe.
#line 6198
#
#line 6198
allow ldconfig_t { sysadm_t initrc_t dpkg_t }:fifo_file { ioctl read getattr lock write append };
#line 6198

#line 6198
#
#line 6198
# Allow the new domain to read and execute the program.
#line 6198
#
#line 6198
allow ldconfig_t ldconfig_exec_t:file { read getattr lock execute ioctl };
#line 6198

#line 6198
#
#line 6198
# Allow the new domain to be entered via the program.
#line 6198
#
#line 6198
allow ldconfig_t ldconfig_exec_t:file entrypoint;
#line 6198

#line 6198
type_transition { sysadm_t initrc_t dpkg_t } ldconfig_exec_t:process ldconfig_t;
#line 6198

dontaudit ldconfig_t device_t:dir search;
allow ldconfig_t { initrc_devpts_t { sysadm_tty_device_t sysadm_devpts_t } }:chr_file { ioctl read getattr lock write append };
allow ldconfig_t privfd:fd use;


#line 6203
allow ldconfig_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6203
allow ldconfig_t lib_t:lnk_file { read getattr lock ioctl };
#line 6203
allow ldconfig_t ld_so_t:file { read getattr lock execute ioctl };
#line 6203
allow ldconfig_t ld_so_t:file execute_no_trans;
#line 6203
allow ldconfig_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6203
allow ldconfig_t shlib_t:file { read getattr lock execute ioctl };
#line 6203
allow ldconfig_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6203
allow ldconfig_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6203
allow ldconfig_t device_t:dir search;
#line 6203
allow ldconfig_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6203



#line 6205

#line 6205

#line 6205

#line 6205
#
#line 6205
# Allow the process to modify the directory.
#line 6205
#
#line 6205
allow ldconfig_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6205

#line 6205
#
#line 6205
# Allow the process to create the file.
#line 6205
#
#line 6205

#line 6205
allow ldconfig_t ld_so_cache_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6205
allow ldconfig_t ld_so_cache_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6205

#line 6205

#line 6205
type_transition ldconfig_t etc_t:dir ld_so_cache_t;
#line 6205
type_transition ldconfig_t etc_t:{ file lnk_file sock_file fifo_file } ld_so_cache_t;
#line 6205

#line 6205

#line 6205


#line 6206

#line 6206

#line 6206

#line 6206
#
#line 6206
# Allow the process to modify the directory.
#line 6206
#
#line 6206
allow ldconfig_t lib_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6206

#line 6206
#
#line 6206
# Allow the process to create the file.
#line 6206
#
#line 6206

#line 6206
allow ldconfig_t shlib_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6206
allow ldconfig_t shlib_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6206

#line 6206

#line 6206
type_transition ldconfig_t lib_t:dir shlib_t;
#line 6206
type_transition ldconfig_t lib_t:{ file lnk_file sock_file fifo_file } shlib_t;
#line 6206

#line 6206

#line 6206

# allow removing mis-labelled links
allow ldconfig_t lib_t:lnk_file unlink;

allow ldconfig_t userdomain:fd use;
allow ldconfig_t etc_t:file { getattr read };
allow ldconfig_t etc_t:lnk_file read;

allow ldconfig_t fs_t:filesystem getattr;
#DESC LoadPolicy - SELinux policy loading utilities
#
# Authors:  Frank Mayer, mayerf@tresys.com
# X-Debian-Packages: policycoreutils
#

###########################
# load_policy_t is the domain type for load_policy 
# load_policy_exec_t is the file type for the executable


type load_policy_t, domain;
role sysadm_r types load_policy_t;

type load_policy_exec_t, file_type, exec_type, sysadmfile;

##########################
# 
# Rules


#line 6235

#line 6235

#line 6235
#
#line 6235
# Allow the process to transition to the new domain.
#line 6235
#
#line 6235
allow sysadm_t load_policy_t:process transition;
#line 6235

#line 6235
#
#line 6235
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6235
#
#line 6235
dontaudit sysadm_t load_policy_t:process noatsecure;
#line 6235

#line 6235
#
#line 6235
# Allow the process to execute the program.
#line 6235
# 
#line 6235
allow sysadm_t load_policy_exec_t:file { read { getattr execute } };
#line 6235

#line 6235
#
#line 6235
# Allow the process to reap the new domain.
#line 6235
#
#line 6235
allow load_policy_t sysadm_t:process sigchld;
#line 6235

#line 6235
#
#line 6235
# Allow the new domain to inherit and use file 
#line 6235
# descriptions from the creating process and vice versa.
#line 6235
#
#line 6235
allow load_policy_t sysadm_t:fd use;
#line 6235
allow sysadm_t load_policy_t:fd use;
#line 6235

#line 6235
#
#line 6235
# Allow the new domain to write back to the old domain via a pipe.
#line 6235
#
#line 6235
allow load_policy_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6235

#line 6235
#
#line 6235
# Allow the new domain to read and execute the program.
#line 6235
#
#line 6235
allow load_policy_t load_policy_exec_t:file { read getattr lock execute ioctl };
#line 6235

#line 6235
#
#line 6235
# Allow the new domain to be entered via the program.
#line 6235
#
#line 6235
allow load_policy_t load_policy_exec_t:file entrypoint;
#line 6235

#line 6235
type_transition sysadm_t load_policy_exec_t:process load_policy_t;
#line 6235

#line 6239

#line 6239
role system_r types load_policy_t;
#line 6239

#line 6239

#line 6239

#line 6239
#
#line 6239
# Allow the process to transition to the new domain.
#line 6239
#
#line 6239
allow dpkg_t load_policy_t:process transition;
#line 6239

#line 6239
#
#line 6239
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6239
#
#line 6239
dontaudit dpkg_t load_policy_t:process noatsecure;
#line 6239

#line 6239
#
#line 6239
# Allow the process to execute the program.
#line 6239
# 
#line 6239
allow dpkg_t load_policy_exec_t:file { read { getattr execute } };
#line 6239

#line 6239
#
#line 6239
# Allow the process to reap the new domain.
#line 6239
#
#line 6239
allow load_policy_t dpkg_t:process sigchld;
#line 6239

#line 6239
#
#line 6239
# Allow the new domain to inherit and use file 
#line 6239
# descriptions from the creating process and vice versa.
#line 6239
#
#line 6239
allow load_policy_t dpkg_t:fd use;
#line 6239
allow dpkg_t load_policy_t:fd use;
#line 6239

#line 6239
#
#line 6239
# Allow the new domain to write back to the old domain via a pipe.
#line 6239
#
#line 6239
allow load_policy_t dpkg_t:fifo_file { ioctl read getattr lock write append };
#line 6239

#line 6239
#
#line 6239
# Allow the new domain to read and execute the program.
#line 6239
#
#line 6239
allow load_policy_t load_policy_exec_t:file { read getattr lock execute ioctl };
#line 6239

#line 6239
#
#line 6239
# Allow the new domain to be entered via the program.
#line 6239
#
#line 6239
allow load_policy_t load_policy_exec_t:file entrypoint;
#line 6239

#line 6239
type_transition dpkg_t load_policy_exec_t:process load_policy_t;
#line 6239

#line 6239


# Reload the policy configuration (sysadm_t no longer has this ability)

#line 6242
allow load_policy_t security_t:dir { read search getattr };
#line 6242
allow load_policy_t security_t:file { getattr read write };
#line 6242
allow load_policy_t security_t:security load_policy;
#line 6242
auditallow load_policy_t security_t:security load_policy;
#line 6242



###########################
# constrain from where load_policy can load a policy, specifically 
# policy_config_t files 
#

# only allow read of policy config files
allow load_policy_t policy_config_t:dir { read getattr lock search ioctl };
allow load_policy_t policy_config_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };

# directory search permissions for path to binary policy files
allow load_policy_t root_t:dir search;
allow load_policy_t etc_t:dir search;

# Read the devpts root directory (needed?)  
allow load_policy_t devpts_t:dir { read getattr lock search ioctl };

# Other access
allow load_policy_t { initrc_devpts_t { sysadm_tty_device_t sysadm_devpts_t } }:chr_file { read write ioctl getattr };

#line 6263
allow load_policy_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6263
allow load_policy_t lib_t:lnk_file { read getattr lock ioctl };
#line 6263
allow load_policy_t ld_so_t:file { read getattr lock execute ioctl };
#line 6263
allow load_policy_t ld_so_t:file execute_no_trans;
#line 6263
allow load_policy_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6263
allow load_policy_t shlib_t:file { read getattr lock execute ioctl };
#line 6263
allow load_policy_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6263
allow load_policy_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6263
allow load_policy_t device_t:dir search;
#line 6263
allow load_policy_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6263

allow load_policy_t self:capability dac_override;

allow load_policy_t { initrc_t privfd }:fd use;

allow load_policy_t fs_t:filesystem getattr;

allow load_policy_t sysadm_tmp_t:file { getattr write } ;

#DESC Login - Local/remote login utilities
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# Macroised by Russell Coker <russell@coker.com.au>
# X-Debian-Packages: login
#

#################################
# 
# Rules for the local_login_t domain
# and the remote_login_t domain.
#

# $1 is the name of the domain (local or remote)
# I added "mlstrustedreader, mlstrustedwriter, mlstrustedobject" to
# remote_login_t, not sure if this is right
#line 6365


#################################
#
# Rules for the local_login_t domain.
#
# local_login_t is the domain of a login process 
# spawned by getty.
#
# remote_login_t is the domain of a login process 
# spawned by rlogind.
#
# login_exec_t is the type of the login program
#
type login_exec_t, file_type, sysadmfile, exec_type;


#line 6381
type local_login_t, domain, privuser, privrole, privlog, auth, privowner, mlstrustedreader, mlstrustedwriter, mlstrustedobject, privfd;
#line 6381
role system_r types local_login_t;
#line 6381

#line 6381

#line 6381
# Access other processes in the same domain.
#line 6381
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 6381
# separately if desired.
#line 6381
allow local_login_t self:process ~{ptrace setexec setfscreate};
#line 6381

#line 6381
# Access /proc/PID files for processes in the same domain.
#line 6381
allow local_login_t self:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381

#line 6381
# Access file descriptions, pipes, and sockets
#line 6381
# created by processes in the same domain.
#line 6381
allow local_login_t self:fd *;
#line 6381
allow local_login_t self:fifo_file { ioctl read getattr lock write append };
#line 6381
allow local_login_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6381
allow local_login_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 6381

#line 6381
# Allow the domain to communicate with other processes in the same domain.
#line 6381
allow local_login_t self:unix_dgram_socket sendto;
#line 6381
allow local_login_t self:unix_stream_socket connectto;
#line 6381

#line 6381
# Access System V IPC objects created by processes in the same domain.
#line 6381
allow local_login_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 6381
allow local_login_t self:msg  { send receive };
#line 6381
allow local_login_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 6381
allow local_login_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 6381
;
#line 6381

#line 6381
# Read system information files in /proc.
#line 6381
allow local_login_t proc_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381

#line 6381

#line 6381
# Read /.
#line 6381
allow local_login_t root_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381

#line 6381
# Read /home.
#line 6381
allow local_login_t home_root_t:dir { read getattr lock search ioctl };
#line 6381

#line 6381
# Read /usr.
#line 6381
allow local_login_t usr_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381

#line 6381
# Read bin and sbin directories.
#line 6381
allow local_login_t bin_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381
allow local_login_t sbin_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381

#line 6381

#line 6381
# Read directories and files with the readable_t type.
#line 6381
# This type is a general type for "world"-readable files.
#line 6381
allow local_login_t readable_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t readable_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381

#line 6381
# Read /var, /var/spool, /var/log.
#line 6381
allow local_login_t var_t:dir search;
#line 6381
allow local_login_t var_spool_t:dir search;
#line 6381

#line 6381
# Read /etc.
#line 6381
allow local_login_t etc_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6381
allow local_login_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 6381

#line 6381

#line 6381
allow local_login_t etc_t:lnk_file read;
#line 6381

#line 6381
allow local_login_t locale_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6381

#line 6381

#line 6381

#line 6381
# Read executable types.
#line 6381
allow local_login_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 6381

#line 6381
# Read /dev directories and any symbolic links.
#line 6381
allow local_login_t device_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t device_t:lnk_file { read getattr lock ioctl };
#line 6381

#line 6381

#line 6381
allow local_login_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t lib_t:lnk_file { read getattr lock ioctl };
#line 6381
allow local_login_t ld_so_t:file { read getattr lock execute ioctl };
#line 6381
allow local_login_t ld_so_t:file execute_no_trans;
#line 6381
allow local_login_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6381
allow local_login_t shlib_t:file { read getattr lock execute ioctl };
#line 6381
allow local_login_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6381
allow local_login_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6381
allow local_login_t device_t:dir search;
#line 6381
allow local_login_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6381
;
#line 6381

#line 6381

#line 6381
type local_login_tmp_t, file_type, sysadmfile, tmpfile ;
#line 6381

#line 6381

#line 6381

#line 6381

#line 6381
#
#line 6381
# Allow the process to modify the directory.
#line 6381
#
#line 6381
allow local_login_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6381

#line 6381
#
#line 6381
# Allow the process to create the file.
#line 6381
#
#line 6381

#line 6381
allow local_login_t local_login_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6381
allow local_login_t local_login_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6381

#line 6381

#line 6381
type_transition local_login_t tmp_t:dir local_login_tmp_t;
#line 6381
type_transition local_login_t tmp_t:{ file lnk_file sock_file fifo_file } local_login_tmp_t;
#line 6381

#line 6381

#line 6381

#line 6381

#line 6381

#line 6381
# Use capabilities
#line 6381
allow local_login_t self:capability { setuid setgid chown fowner fsetid net_bind_service sys_tty_config dac_override sys_nice sys_resource };
#line 6381

#line 6381
# Run shells in user_t by default.
#line 6381

#line 6381

#line 6381

#line 6381
#
#line 6381
# Allow the process to transition to the new domain.
#line 6381
#
#line 6381
allow local_login_t user_t:process transition;
#line 6381

#line 6381
#
#line 6381
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6381
#
#line 6381
dontaudit local_login_t user_t:process noatsecure;
#line 6381

#line 6381
#
#line 6381
# Allow the process to execute the program.
#line 6381
# 
#line 6381
allow local_login_t shell_exec_t:file { read { getattr execute } };
#line 6381

#line 6381
#
#line 6381
# Allow the process to reap the new domain.
#line 6381
#
#line 6381
allow user_t local_login_t:process sigchld;
#line 6381

#line 6381
#
#line 6381
# Allow the new domain to inherit and use file 
#line 6381
# descriptions from the creating process and vice versa.
#line 6381
#
#line 6381
allow user_t local_login_t:fd use;
#line 6381
allow local_login_t user_t:fd use;
#line 6381

#line 6381
#
#line 6381
# Allow the new domain to write back to the old domain via a pipe.
#line 6381
#
#line 6381
allow user_t local_login_t:fifo_file { ioctl read getattr lock write append };
#line 6381

#line 6381
#
#line 6381
# Allow the new domain to read and execute the program.
#line 6381
#
#line 6381
allow user_t shell_exec_t:file { read getattr lock execute ioctl };
#line 6381

#line 6381
#
#line 6381
# Allow the new domain to be entered via the program.
#line 6381
#
#line 6381
allow user_t shell_exec_t:file entrypoint;
#line 6381

#line 6381
type_transition local_login_t shell_exec_t:process user_t;
#line 6381

#line 6381

#line 6381
# Set exec context.
#line 6381

#line 6381
allow local_login_t self:process setexec;
#line 6381
allow local_login_t proc_t:dir search;
#line 6381
allow local_login_t proc_t:lnk_file read;
#line 6381
allow local_login_t self:dir search;
#line 6381
allow local_login_t self:file { read write };
#line 6381

#line 6381

#line 6381
# Permit login to search the user home directories.
#line 6381
allow local_login_t home_root_t:dir search;
#line 6381
allow local_login_t home_dir_type:dir search;
#line 6381

#line 6381
# Write to /var/run/utmp.
#line 6381
allow local_login_t var_run_t:dir search;
#line 6381
allow local_login_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 6381

#line 6381
# Write to /var/log/wtmp.
#line 6381
allow local_login_t var_log_t:dir search;
#line 6381
allow local_login_t wtmp_t:file { ioctl read getattr lock write append };
#line 6381

#line 6381
# Write to /var/log/lastlog.
#line 6381
allow local_login_t lastlog_t:file { ioctl read getattr lock write append };
#line 6381

#line 6381
# Write to /var/log/btmp
#line 6381
allow local_login_t faillog_t:file { append read write };
#line 6381

#line 6381
# Search for mail spool file.
#line 6381
allow local_login_t mail_spool_t:dir { read getattr lock search ioctl };
#line 6381
allow local_login_t mail_spool_t:file getattr;
#line 6381
allow local_login_t mail_spool_t:lnk_file read;
#line 6381

#line 6381
# Get security policy decisions.
#line 6381

#line 6381
allow local_login_t security_t:dir { read search getattr };
#line 6381
allow local_login_t security_t:file { getattr read write };
#line 6381
allow local_login_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 6381

#line 6381

#line 6381
# allow read access to default_contexts in /etc/security
#line 6381
allow local_login_t default_context_t:file { read getattr lock ioctl };
#line 6381

#line 6381


# But also permit other user domains to be entered by login.

#line 6384

#line 6384
#
#line 6384
# Allow the process to transition to the new domain.
#line 6384
#
#line 6384
allow local_login_t userdomain:process transition;
#line 6384

#line 6384
#
#line 6384
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6384
#
#line 6384
dontaudit local_login_t userdomain:process noatsecure;
#line 6384

#line 6384
#
#line 6384
# Allow the process to execute the program.
#line 6384
# 
#line 6384
allow local_login_t shell_exec_t:file { read { getattr execute } };
#line 6384

#line 6384
#
#line 6384
# Allow the process to reap the new domain.
#line 6384
#
#line 6384
allow userdomain local_login_t:process sigchld;
#line 6384

#line 6384
#
#line 6384
# Allow the new domain to inherit and use file 
#line 6384
# descriptions from the creating process and vice versa.
#line 6384
#
#line 6384
allow userdomain local_login_t:fd use;
#line 6384
allow local_login_t userdomain:fd use;
#line 6384

#line 6384
#
#line 6384
# Allow the new domain to write back to the old domain via a pipe.
#line 6384
#
#line 6384
allow userdomain local_login_t:fifo_file { ioctl read getattr lock write append };
#line 6384

#line 6384
#
#line 6384
# Allow the new domain to read and execute the program.
#line 6384
#
#line 6384
allow userdomain shell_exec_t:file { read getattr lock execute ioctl };
#line 6384

#line 6384
#
#line 6384
# Allow the new domain to be entered via the program.
#line 6384
#
#line 6384
allow userdomain shell_exec_t:file entrypoint;
#line 6384


# Do not audit denied attempts to access devices.
dontaudit local_login_t fixed_disk_device_t:blk_file { getattr setattr };
dontaudit local_login_t removable_device_t:blk_file { getattr setattr };
dontaudit local_login_t device_t:{ chr_file blk_file lnk_file } { getattr setattr };
dontaudit local_login_t misc_device_t:{ chr_file blk_file lnk_file } { getattr setattr };
dontaudit local_login_t framebuf_device_t:{ chr_file blk_file lnk_file } { getattr setattr read };
dontaudit local_login_t apm_bios_t:chr_file { getattr setattr };
dontaudit local_login_t v4l_device_t:{ chr_file blk_file lnk_file } { getattr setattr read };
dontaudit local_login_t v4l_device_t:dir { read search getattr };

# Do not audit denied attempts to access /mnt.
dontaudit local_login_t file_t:dir { read getattr lock search ioctl };


# Create lock file.
allow local_login_t var_lock_t:dir { read getattr lock search ioctl add_name remove_name write };
allow local_login_t var_lock_t:file { create ioctl read getattr lock write setattr append link unlink rename };


# Read and write ttys.
allow local_login_t tty_device_t:chr_file { setattr { ioctl read getattr lock write append } };
allow local_login_t ttyfile:chr_file { setattr { ioctl read getattr lock write append } };

# Relabel ttys.
allow local_login_t tty_device_t:chr_file { getattr relabelfrom relabelto };
allow local_login_t ttyfile:chr_file { getattr relabelfrom relabelto };

#line 6414


# Allow setting of attributes on sound devices.
allow local_login_t sound_device_t:chr_file { getattr setattr };

# Allow access to /var/run/console and /var/run/console.lock.  Need a separate type?
allow local_login_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
allow local_login_t var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };

#################################
#
# Rules for the remote_login_t domain.
#


#line 6428
type remote_login_t, domain, privuser, privrole, privlog, auth, privowner, mlstrustedreader, mlstrustedwriter, mlstrustedobject, privfd;
#line 6428
role system_r types remote_login_t;
#line 6428

#line 6428

#line 6428
# Access other processes in the same domain.
#line 6428
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 6428
# separately if desired.
#line 6428
allow remote_login_t self:process ~{ptrace setexec setfscreate};
#line 6428

#line 6428
# Access /proc/PID files for processes in the same domain.
#line 6428
allow remote_login_t self:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428

#line 6428
# Access file descriptions, pipes, and sockets
#line 6428
# created by processes in the same domain.
#line 6428
allow remote_login_t self:fd *;
#line 6428
allow remote_login_t self:fifo_file { ioctl read getattr lock write append };
#line 6428
allow remote_login_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6428
allow remote_login_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 6428

#line 6428
# Allow the domain to communicate with other processes in the same domain.
#line 6428
allow remote_login_t self:unix_dgram_socket sendto;
#line 6428
allow remote_login_t self:unix_stream_socket connectto;
#line 6428

#line 6428
# Access System V IPC objects created by processes in the same domain.
#line 6428
allow remote_login_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 6428
allow remote_login_t self:msg  { send receive };
#line 6428
allow remote_login_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 6428
allow remote_login_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 6428
;
#line 6428

#line 6428
# Read system information files in /proc.
#line 6428
allow remote_login_t proc_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428

#line 6428

#line 6428
# Read /.
#line 6428
allow remote_login_t root_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428

#line 6428
# Read /home.
#line 6428
allow remote_login_t home_root_t:dir { read getattr lock search ioctl };
#line 6428

#line 6428
# Read /usr.
#line 6428
allow remote_login_t usr_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428

#line 6428
# Read bin and sbin directories.
#line 6428
allow remote_login_t bin_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428
allow remote_login_t sbin_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428

#line 6428

#line 6428
# Read directories and files with the readable_t type.
#line 6428
# This type is a general type for "world"-readable files.
#line 6428
allow remote_login_t readable_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t readable_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428

#line 6428
# Read /var, /var/spool, /var/log.
#line 6428
allow remote_login_t var_t:dir search;
#line 6428
allow remote_login_t var_spool_t:dir search;
#line 6428

#line 6428
# Read /etc.
#line 6428
allow remote_login_t etc_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6428
allow remote_login_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 6428

#line 6428

#line 6428
allow remote_login_t etc_t:lnk_file read;
#line 6428

#line 6428
allow remote_login_t locale_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6428

#line 6428

#line 6428

#line 6428
# Read executable types.
#line 6428
allow remote_login_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 6428

#line 6428
# Read /dev directories and any symbolic links.
#line 6428
allow remote_login_t device_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t device_t:lnk_file { read getattr lock ioctl };
#line 6428

#line 6428

#line 6428
allow remote_login_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t lib_t:lnk_file { read getattr lock ioctl };
#line 6428
allow remote_login_t ld_so_t:file { read getattr lock execute ioctl };
#line 6428
allow remote_login_t ld_so_t:file execute_no_trans;
#line 6428
allow remote_login_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6428
allow remote_login_t shlib_t:file { read getattr lock execute ioctl };
#line 6428
allow remote_login_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6428
allow remote_login_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6428
allow remote_login_t device_t:dir search;
#line 6428
allow remote_login_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6428
;
#line 6428

#line 6428

#line 6428
type remote_login_tmp_t, file_type, sysadmfile, tmpfile ;
#line 6428

#line 6428

#line 6428

#line 6428

#line 6428
#
#line 6428
# Allow the process to modify the directory.
#line 6428
#
#line 6428
allow remote_login_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6428

#line 6428
#
#line 6428
# Allow the process to create the file.
#line 6428
#
#line 6428

#line 6428
allow remote_login_t remote_login_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6428
allow remote_login_t remote_login_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6428

#line 6428

#line 6428
type_transition remote_login_t tmp_t:dir remote_login_tmp_t;
#line 6428
type_transition remote_login_t tmp_t:{ file lnk_file sock_file fifo_file } remote_login_tmp_t;
#line 6428

#line 6428

#line 6428

#line 6428

#line 6428

#line 6428
# Use capabilities
#line 6428
allow remote_login_t self:capability { setuid setgid chown fowner fsetid net_bind_service sys_tty_config dac_override sys_nice sys_resource };
#line 6428

#line 6428
# Run shells in user_t by default.
#line 6428

#line 6428

#line 6428

#line 6428
#
#line 6428
# Allow the process to transition to the new domain.
#line 6428
#
#line 6428
allow remote_login_t user_t:process transition;
#line 6428

#line 6428
#
#line 6428
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6428
#
#line 6428
dontaudit remote_login_t user_t:process noatsecure;
#line 6428

#line 6428
#
#line 6428
# Allow the process to execute the program.
#line 6428
# 
#line 6428
allow remote_login_t shell_exec_t:file { read { getattr execute } };
#line 6428

#line 6428
#
#line 6428
# Allow the process to reap the new domain.
#line 6428
#
#line 6428
allow user_t remote_login_t:process sigchld;
#line 6428

#line 6428
#
#line 6428
# Allow the new domain to inherit and use file 
#line 6428
# descriptions from the creating process and vice versa.
#line 6428
#
#line 6428
allow user_t remote_login_t:fd use;
#line 6428
allow remote_login_t user_t:fd use;
#line 6428

#line 6428
#
#line 6428
# Allow the new domain to write back to the old domain via a pipe.
#line 6428
#
#line 6428
allow user_t remote_login_t:fifo_file { ioctl read getattr lock write append };
#line 6428

#line 6428
#
#line 6428
# Allow the new domain to read and execute the program.
#line 6428
#
#line 6428
allow user_t shell_exec_t:file { read getattr lock execute ioctl };
#line 6428

#line 6428
#
#line 6428
# Allow the new domain to be entered via the program.
#line 6428
#
#line 6428
allow user_t shell_exec_t:file entrypoint;
#line 6428

#line 6428
type_transition remote_login_t shell_exec_t:process user_t;
#line 6428

#line 6428

#line 6428
# Set exec context.
#line 6428

#line 6428
allow remote_login_t self:process setexec;
#line 6428
allow remote_login_t proc_t:dir search;
#line 6428
allow remote_login_t proc_t:lnk_file read;
#line 6428
allow remote_login_t self:dir search;
#line 6428
allow remote_login_t self:file { read write };
#line 6428

#line 6428

#line 6428
# Permit login to search the user home directories.
#line 6428
allow remote_login_t home_root_t:dir search;
#line 6428
allow remote_login_t home_dir_type:dir search;
#line 6428

#line 6428
# Write to /var/run/utmp.
#line 6428
allow remote_login_t var_run_t:dir search;
#line 6428
allow remote_login_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 6428

#line 6428
# Write to /var/log/wtmp.
#line 6428
allow remote_login_t var_log_t:dir search;
#line 6428
allow remote_login_t wtmp_t:file { ioctl read getattr lock write append };
#line 6428

#line 6428
# Write to /var/log/lastlog.
#line 6428
allow remote_login_t lastlog_t:file { ioctl read getattr lock write append };
#line 6428

#line 6428
# Write to /var/log/btmp
#line 6428
allow remote_login_t faillog_t:file { append read write };
#line 6428

#line 6428
# Search for mail spool file.
#line 6428
allow remote_login_t mail_spool_t:dir { read getattr lock search ioctl };
#line 6428
allow remote_login_t mail_spool_t:file getattr;
#line 6428
allow remote_login_t mail_spool_t:lnk_file read;
#line 6428

#line 6428
# Get security policy decisions.
#line 6428

#line 6428
allow remote_login_t security_t:dir { read search getattr };
#line 6428
allow remote_login_t security_t:file { getattr read write };
#line 6428
allow remote_login_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 6428

#line 6428

#line 6428
# allow read access to default_contexts in /etc/security
#line 6428
allow remote_login_t default_context_t:file { read getattr lock ioctl };
#line 6428

#line 6428


# Only permit unprivileged user domains to be entered via rlogin,
# since very weak authentication is used.

#line 6432

#line 6432
#
#line 6432
# Allow the process to transition to the new domain.
#line 6432
#
#line 6432
allow remote_login_t unpriv_userdomain:process transition;
#line 6432

#line 6432
#
#line 6432
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6432
#
#line 6432
dontaudit remote_login_t unpriv_userdomain:process noatsecure;
#line 6432

#line 6432
#
#line 6432
# Allow the process to execute the program.
#line 6432
# 
#line 6432
allow remote_login_t shell_exec_t:file { read { getattr execute } };
#line 6432

#line 6432
#
#line 6432
# Allow the process to reap the new domain.
#line 6432
#
#line 6432
allow unpriv_userdomain remote_login_t:process sigchld;
#line 6432

#line 6432
#
#line 6432
# Allow the new domain to inherit and use file 
#line 6432
# descriptions from the creating process and vice versa.
#line 6432
#
#line 6432
allow unpriv_userdomain remote_login_t:fd use;
#line 6432
allow remote_login_t unpriv_userdomain:fd use;
#line 6432

#line 6432
#
#line 6432
# Allow the new domain to write back to the old domain via a pipe.
#line 6432
#
#line 6432
allow unpriv_userdomain remote_login_t:fifo_file { ioctl read getattr lock write append };
#line 6432

#line 6432
#
#line 6432
# Allow the new domain to read and execute the program.
#line 6432
#
#line 6432
allow unpriv_userdomain shell_exec_t:file { read getattr lock execute ioctl };
#line 6432

#line 6432
#
#line 6432
# Allow the new domain to be entered via the program.
#line 6432
#
#line 6432
allow unpriv_userdomain shell_exec_t:file entrypoint;
#line 6432


allow remote_login_t devpts_t:dir search;

# Use the pty created by rlogind.
#line 6442

allow remote_login_t ptyfile:chr_file { getattr relabelfrom relabelto };

#DESC Logrotate - Rotate log files
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: logrotate
#

#################################
#
# Rules for the logrotate_t domain.
#
# logrotate_t is the domain for the logrotate program.
# logrotate_exec_t is the type of the corresponding program.
#
type logrotate_t, domain, privowner, privmail;
role system_r types logrotate_t;
role sysadm_r types logrotate_t;

#line 6461
allow logrotate_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6461
allow logrotate_t lib_t:lnk_file { read getattr lock ioctl };
#line 6461
allow logrotate_t ld_so_t:file { read getattr lock execute ioctl };
#line 6461
allow logrotate_t ld_so_t:file execute_no_trans;
#line 6461
allow logrotate_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6461
allow logrotate_t shlib_t:file { read getattr lock execute ioctl };
#line 6461
allow logrotate_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6461
allow logrotate_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6461
allow logrotate_t device_t:dir search;
#line 6461
allow logrotate_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6461
;

#line 6462
# Access other processes in the same domain.
#line 6462
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 6462
# separately if desired.
#line 6462
allow logrotate_t self:process ~{ptrace setexec setfscreate};
#line 6462

#line 6462
# Access /proc/PID files for processes in the same domain.
#line 6462
allow logrotate_t self:dir { read getattr lock search ioctl };
#line 6462
allow logrotate_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6462

#line 6462
# Access file descriptions, pipes, and sockets
#line 6462
# created by processes in the same domain.
#line 6462
allow logrotate_t self:fd *;
#line 6462
allow logrotate_t self:fifo_file { ioctl read getattr lock write append };
#line 6462
allow logrotate_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6462
allow logrotate_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 6462

#line 6462
# Allow the domain to communicate with other processes in the same domain.
#line 6462
allow logrotate_t self:unix_dgram_socket sendto;
#line 6462
allow logrotate_t self:unix_stream_socket connectto;
#line 6462

#line 6462
# Access System V IPC objects created by processes in the same domain.
#line 6462
allow logrotate_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 6462
allow logrotate_t self:msg  { send receive };
#line 6462
allow logrotate_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 6462
allow logrotate_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 6462
;
type logrotate_exec_t, file_type, sysadmfile, exec_type;


#line 6465

#line 6465

#line 6465

#line 6465
#
#line 6465
# Allow the process to transition to the new domain.
#line 6465
#
#line 6465
allow system_crond_t logrotate_t:process transition;
#line 6465

#line 6465
#
#line 6465
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6465
#
#line 6465
dontaudit system_crond_t logrotate_t:process noatsecure;
#line 6465

#line 6465
#
#line 6465
# Allow the process to execute the program.
#line 6465
# 
#line 6465
allow system_crond_t logrotate_exec_t:file { read { getattr execute } };
#line 6465

#line 6465
#
#line 6465
# Allow the process to reap the new domain.
#line 6465
#
#line 6465
allow logrotate_t system_crond_t:process sigchld;
#line 6465

#line 6465
#
#line 6465
# Allow the new domain to inherit and use file 
#line 6465
# descriptions from the creating process and vice versa.
#line 6465
#
#line 6465
allow logrotate_t system_crond_t:fd use;
#line 6465
allow system_crond_t logrotate_t:fd use;
#line 6465

#line 6465
#
#line 6465
# Allow the new domain to write back to the old domain via a pipe.
#line 6465
#
#line 6465
allow logrotate_t system_crond_t:fifo_file { ioctl read getattr lock write append };
#line 6465

#line 6465
#
#line 6465
# Allow the new domain to read and execute the program.
#line 6465
#
#line 6465
allow logrotate_t logrotate_exec_t:file { read getattr lock execute ioctl };
#line 6465

#line 6465
#
#line 6465
# Allow the new domain to be entered via the program.
#line 6465
#
#line 6465
allow logrotate_t logrotate_exec_t:file entrypoint;
#line 6465

#line 6465
type_transition system_crond_t logrotate_exec_t:process logrotate_t;
#line 6465

#line 6465
allow logrotate_t crond_t:fifo_file { getattr read write ioctl };
#line 6465
# a rule for privfd may make this obsolete
#line 6465
allow logrotate_t crond_t:fd use;
#line 6465
allow logrotate_t crond_t:process sigchld;
#line 6465


#line 6466

#line 6466

#line 6466
#
#line 6466
# Allow the process to transition to the new domain.
#line 6466
#
#line 6466
allow sysadm_t logrotate_t:process transition;
#line 6466

#line 6466
#
#line 6466
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6466
#
#line 6466
dontaudit sysadm_t logrotate_t:process noatsecure;
#line 6466

#line 6466
#
#line 6466
# Allow the process to execute the program.
#line 6466
# 
#line 6466
allow sysadm_t logrotate_exec_t:file { read { getattr execute } };
#line 6466

#line 6466
#
#line 6466
# Allow the process to reap the new domain.
#line 6466
#
#line 6466
allow logrotate_t sysadm_t:process sigchld;
#line 6466

#line 6466
#
#line 6466
# Allow the new domain to inherit and use file 
#line 6466
# descriptions from the creating process and vice versa.
#line 6466
#
#line 6466
allow logrotate_t sysadm_t:fd use;
#line 6466
allow sysadm_t logrotate_t:fd use;
#line 6466

#line 6466
#
#line 6466
# Allow the new domain to write back to the old domain via a pipe.
#line 6466
#
#line 6466
allow logrotate_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6466

#line 6466
#
#line 6466
# Allow the new domain to read and execute the program.
#line 6466
#
#line 6466
allow logrotate_t logrotate_exec_t:file { read getattr lock execute ioctl };
#line 6466

#line 6466
#
#line 6466
# Allow the new domain to be entered via the program.
#line 6466
#
#line 6466
allow logrotate_t logrotate_exec_t:file entrypoint;
#line 6466

#line 6466
type_transition sysadm_t logrotate_exec_t:process logrotate_t;
#line 6466

allow logrotate_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow logrotate_t devtty_t:chr_file { ioctl read getattr lock write append };

# access files in /etc
allow logrotate_t etc_t:file { getattr read };
allow logrotate_t etc_t:lnk_file read;
allow logrotate_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };

# it should not require this
allow logrotate_t sysadm_home_dir_t:dir { read getattr search };

# create lock files

#line 6479
allow logrotate_t var_lock_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6479
allow logrotate_t var_lock_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6479


# Create temporary files.

#line 6482
type logrotate_tmp_t, file_type, sysadmfile, tmpfile ;
#line 6482

#line 6482

#line 6482

#line 6482

#line 6482
#
#line 6482
# Allow the process to modify the directory.
#line 6482
#
#line 6482
allow logrotate_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6482

#line 6482
#
#line 6482
# Allow the process to create the file.
#line 6482
#
#line 6482

#line 6482
allow logrotate_t logrotate_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6482
allow logrotate_t logrotate_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6482

#line 6482

#line 6482
type_transition logrotate_t tmp_t:dir logrotate_tmp_t;
#line 6482
type_transition logrotate_t tmp_t:{ file lnk_file sock_file fifo_file } logrotate_tmp_t;
#line 6482

#line 6482

#line 6482

#line 6482


# Run helper programs.
allow logrotate_t { bin_t sbin_t }:dir { read getattr lock search ioctl };
allow logrotate_t { bin_t sbin_t }:lnk_file read;

#line 6487
allow logrotate_t { bin_t sbin_t shell_exec_t etc_t ls_exec_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6487
;

# Read PID files.
allow logrotate_t pidfile:file { read getattr lock ioctl };

# Read /proc/PID directories for all domains.
allow logrotate_t proc_t:dir { read getattr lock search ioctl };
allow logrotate_t proc_t:{ file lnk_file } { read getattr lock ioctl };
allow logrotate_t { sysctl_t sysctl_kernel_t }:dir search;
allow logrotate_t sysctl_kernel_t:file { getattr read };
allow logrotate_t domain:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
allow logrotate_t domain:dir { read getattr lock search ioctl };
allow logrotate_t exec_type:file getattr;

# Read /dev directories and any symbolic links.
allow logrotate_t device_t:dir { read getattr lock search ioctl };
allow logrotate_t device_t:lnk_file { read getattr lock ioctl };

# Signal processes.
allow logrotate_t domain:process signal;

# Modify /var/log and other log dirs.
allow logrotate_t var_t:dir { read getattr lock search ioctl };
allow logrotate_t logfile:dir { read getattr lock search ioctl add_name remove_name write };
allow logrotate_t logfile:lnk_file read;

# Create, rename, and truncate log files.
allow logrotate_t logfile:file { create ioctl read getattr lock write setattr append link unlink rename };
allow logrotate_t wtmp_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 6521


# Set a context other than the default one for newly created files.

#line 6524
allow logrotate_t self:process setfscreate;
#line 6524
allow logrotate_t proc_t:dir search;
#line 6524
allow logrotate_t proc_t:lnk_file read;
#line 6524
allow logrotate_t self:dir search;
#line 6524
allow logrotate_t self:file { read write };
#line 6524


# Change ownership on log files.
allow logrotate_t self:capability { chown dac_override kill fsetid fowner };
# for mailx
dontaudit logrotate_t self:capability { setuid setgid };

#line 6533


# Access /var/run
allow logrotate_t var_run_t:dir { read getattr lock search ioctl };

# for /var/lib/logrotate.status and /var/lib/logcheck

#line 6539
type logrotate_var_lib_t alias var_lib_logrotate_t, file_type, sysadmfile;
#line 6539

#line 6539

#line 6539

#line 6539

#line 6539
#
#line 6539
# Allow the process to modify the directory.
#line 6539
#
#line 6539
allow logrotate_t var_lib_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6539

#line 6539
#
#line 6539
# Allow the process to create the file.
#line 6539
#
#line 6539

#line 6539
allow logrotate_t logrotate_var_lib_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6539
allow logrotate_t logrotate_var_lib_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6539

#line 6539

#line 6539
type_transition logrotate_t var_lib_t:dir logrotate_var_lib_t;
#line 6539
type_transition logrotate_t var_lib_t:{ file lnk_file sock_file fifo_file } logrotate_var_lib_t;
#line 6539

#line 6539

#line 6539

#line 6539


# Write to /var/spool/slrnpull - should be moved into its own type.
allow logrotate_t var_spool_t:dir { search write add_name remove_name };
allow logrotate_t var_spool_t:file { rename create setattr unlink };

# Access terminals.
allow logrotate_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { ioctl read getattr lock write append };


# for /var/backups on Debian
#line 6552



#line 6554
allow logrotate_t etc_t:lnk_file read;
#line 6554

#line 6554
allow logrotate_t locale_t:dir { read getattr lock search ioctl };
#line 6554
allow logrotate_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6554

#line 6554


allow logrotate_t fs_t:filesystem getattr;

#line 6557
allow logrotate_t shell_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6557

#DESC Modutil - Dynamic module utilities
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: modutils
#

#################################
#
# Rules for the module utility domains.
#
type modules_dep_t, file_type, sysadmfile;
type modules_conf_t, file_type, sysadmfile;
type modules_object_t, file_type, sysadmfile;


#################################
#
# Rules for the depmod_t domain.
#
type depmod_t, domain;
role system_r types depmod_t;
role sysadm_r types depmod_t;


#line 6581
allow depmod_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6581
allow depmod_t lib_t:lnk_file { read getattr lock ioctl };
#line 6581
allow depmod_t ld_so_t:file { read getattr lock execute ioctl };
#line 6581
allow depmod_t ld_so_t:file execute_no_trans;
#line 6581
allow depmod_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6581
allow depmod_t shlib_t:file { read getattr lock execute ioctl };
#line 6581
allow depmod_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6581
allow depmod_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6581
allow depmod_t device_t:dir search;
#line 6581
allow depmod_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6581


type depmod_exec_t, file_type, exec_type, sysadmfile;

#line 6584

#line 6584

#line 6584
#
#line 6584
# Allow the process to transition to the new domain.
#line 6584
#
#line 6584
allow initrc_t depmod_t:process transition;
#line 6584

#line 6584
#
#line 6584
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6584
#
#line 6584
dontaudit initrc_t depmod_t:process noatsecure;
#line 6584

#line 6584
#
#line 6584
# Allow the process to execute the program.
#line 6584
# 
#line 6584
allow initrc_t depmod_exec_t:file { read { getattr execute } };
#line 6584

#line 6584
#
#line 6584
# Allow the process to reap the new domain.
#line 6584
#
#line 6584
allow depmod_t initrc_t:process sigchld;
#line 6584

#line 6584
#
#line 6584
# Allow the new domain to inherit and use file 
#line 6584
# descriptions from the creating process and vice versa.
#line 6584
#
#line 6584
allow depmod_t initrc_t:fd use;
#line 6584
allow initrc_t depmod_t:fd use;
#line 6584

#line 6584
#
#line 6584
# Allow the new domain to write back to the old domain via a pipe.
#line 6584
#
#line 6584
allow depmod_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 6584

#line 6584
#
#line 6584
# Allow the new domain to read and execute the program.
#line 6584
#
#line 6584
allow depmod_t depmod_exec_t:file { read getattr lock execute ioctl };
#line 6584

#line 6584
#
#line 6584
# Allow the new domain to be entered via the program.
#line 6584
#
#line 6584
allow depmod_t depmod_exec_t:file entrypoint;
#line 6584

#line 6584
type_transition initrc_t depmod_exec_t:process depmod_t;
#line 6584

allow depmod_t { bin_t sbin_t }:dir search;

#line 6586
allow depmod_t depmod_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6586


#line 6587

#line 6587

#line 6587
#
#line 6587
# Allow the process to transition to the new domain.
#line 6587
#
#line 6587
allow sysadm_t depmod_t:process transition;
#line 6587

#line 6587
#
#line 6587
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6587
#
#line 6587
dontaudit sysadm_t depmod_t:process noatsecure;
#line 6587

#line 6587
#
#line 6587
# Allow the process to execute the program.
#line 6587
# 
#line 6587
allow sysadm_t depmod_exec_t:file { read { getattr execute } };
#line 6587

#line 6587
#
#line 6587
# Allow the process to reap the new domain.
#line 6587
#
#line 6587
allow depmod_t sysadm_t:process sigchld;
#line 6587

#line 6587
#
#line 6587
# Allow the new domain to inherit and use file 
#line 6587
# descriptions from the creating process and vice versa.
#line 6587
#
#line 6587
allow depmod_t sysadm_t:fd use;
#line 6587
allow sysadm_t depmod_t:fd use;
#line 6587

#line 6587
#
#line 6587
# Allow the new domain to write back to the old domain via a pipe.
#line 6587
#
#line 6587
allow depmod_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6587

#line 6587
#
#line 6587
# Allow the new domain to read and execute the program.
#line 6587
#
#line 6587
allow depmod_t depmod_exec_t:file { read getattr lock execute ioctl };
#line 6587

#line 6587
#
#line 6587
# Allow the new domain to be entered via the program.
#line 6587
#
#line 6587
allow depmod_t depmod_exec_t:file entrypoint;
#line 6587

#line 6587
type_transition sysadm_t depmod_exec_t:process depmod_t;
#line 6587


# Inherit and use descriptors from init and login programs.
allow depmod_t { init_t privfd }:fd use;

allow depmod_t { etc_t etc_runtime_t }:file { getattr read };
allow depmod_t { device_t proc_t }:dir search;
allow depmod_t proc_t:file { getattr read };
allow depmod_t fs_t:filesystem getattr;

# read system.map
allow depmod_t boot_t:dir search;
allow depmod_t system_map_t:file { getattr read };

# Read conf.modules.
allow depmod_t modules_conf_t:file { read getattr lock ioctl };

# Create modules.dep.

#line 6605

#line 6605

#line 6605

#line 6605
#
#line 6605
# Allow the process to modify the directory.
#line 6605
#
#line 6605
allow depmod_t modules_object_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6605

#line 6605
#
#line 6605
# Allow the process to create the file.
#line 6605
#
#line 6605

#line 6605
allow depmod_t modules_dep_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6605
allow depmod_t modules_dep_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6605

#line 6605

#line 6605
type_transition depmod_t modules_object_t:dir modules_dep_t;
#line 6605
type_transition depmod_t modules_object_t:{ file lnk_file sock_file fifo_file } modules_dep_t;
#line 6605

#line 6605

#line 6605


# Read module objects.
allow depmod_t modules_object_t:dir { read getattr lock search ioctl };
allow depmod_t modules_object_t:{ file lnk_file } { read getattr lock ioctl };

# Access terminals.
allow depmod_t { console_device_t initrc_devpts_t { sysadm_tty_device_t sysadm_devpts_t } }:chr_file { ioctl read getattr lock write append };


# Read System.map from home directories.
allow depmod_t { home_root_t user_home_dir_type sysadm_home_dir_t }:dir { read getattr lock search ioctl };

#line 6617
allow depmod_t { user_home_type sysadm_home_t }:dir { read getattr lock search ioctl };
#line 6617
allow depmod_t { user_home_type sysadm_home_t }:{ file lnk_file } { read getattr lock ioctl };
#line 6617


#################################
#
# Rules for the insmod_t domain.
#

type insmod_t, domain, privlog;
role system_r types insmod_t;
role sysadm_r types insmod_t;


#line 6628
allow insmod_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6628
allow insmod_t lib_t:lnk_file { read getattr lock ioctl };
#line 6628
allow insmod_t ld_so_t:file { read getattr lock execute ioctl };
#line 6628
allow insmod_t ld_so_t:file execute_no_trans;
#line 6628
allow insmod_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6628
allow insmod_t shlib_t:file { read getattr lock execute ioctl };
#line 6628
allow insmod_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6628
allow insmod_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6628
allow insmod_t device_t:dir search;
#line 6628
allow insmod_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6628


#line 6629
allow insmod_t etc_t:lnk_file read;
#line 6629

#line 6629
allow insmod_t locale_t:dir { read getattr lock search ioctl };
#line 6629
allow insmod_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6629

#line 6629


allow insmod_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

type insmod_exec_t, file_type, exec_type, sysadmfile;

#line 6634

#line 6634

#line 6634
#
#line 6634
# Allow the process to transition to the new domain.
#line 6634
#
#line 6634
allow { initrc_t kernel_t } insmod_t:process transition;
#line 6634

#line 6634
#
#line 6634
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6634
#
#line 6634
dontaudit { initrc_t kernel_t } insmod_t:process noatsecure;
#line 6634

#line 6634
#
#line 6634
# Allow the process to execute the program.
#line 6634
# 
#line 6634
allow { initrc_t kernel_t } insmod_exec_t:file { read { getattr execute } };
#line 6634

#line 6634
#
#line 6634
# Allow the process to reap the new domain.
#line 6634
#
#line 6634
allow insmod_t { initrc_t kernel_t }:process sigchld;
#line 6634

#line 6634
#
#line 6634
# Allow the new domain to inherit and use file 
#line 6634
# descriptions from the creating process and vice versa.
#line 6634
#
#line 6634
allow insmod_t { initrc_t kernel_t }:fd use;
#line 6634
allow { initrc_t kernel_t } insmod_t:fd use;
#line 6634

#line 6634
#
#line 6634
# Allow the new domain to write back to the old domain via a pipe.
#line 6634
#
#line 6634
allow insmod_t { initrc_t kernel_t }:fifo_file { ioctl read getattr lock write append };
#line 6634

#line 6634
#
#line 6634
# Allow the new domain to read and execute the program.
#line 6634
#
#line 6634
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 6634

#line 6634
#
#line 6634
# Allow the new domain to be entered via the program.
#line 6634
#
#line 6634
allow insmod_t insmod_exec_t:file entrypoint;
#line 6634

#line 6634
type_transition { initrc_t kernel_t } insmod_exec_t:process insmod_t;
#line 6634

allow initrc_t insmod_exec_t:lnk_file read;

#line 6636
allow insmod_t insmod_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6636

allow insmod_t { bin_t sbin_t }:dir search;
allow insmod_t sbin_t:lnk_file read;
allow insmod_t init_t:fd use;

#line 6640

#line 6640

#line 6640
#
#line 6640
# Allow the process to transition to the new domain.
#line 6640
#
#line 6640
allow sysadm_t insmod_t:process transition;
#line 6640

#line 6640
#
#line 6640
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6640
#
#line 6640
dontaudit sysadm_t insmod_t:process noatsecure;
#line 6640

#line 6640
#
#line 6640
# Allow the process to execute the program.
#line 6640
# 
#line 6640
allow sysadm_t insmod_exec_t:file { read { getattr execute } };
#line 6640

#line 6640
#
#line 6640
# Allow the process to reap the new domain.
#line 6640
#
#line 6640
allow insmod_t sysadm_t:process sigchld;
#line 6640

#line 6640
#
#line 6640
# Allow the new domain to inherit and use file 
#line 6640
# descriptions from the creating process and vice versa.
#line 6640
#
#line 6640
allow insmod_t sysadm_t:fd use;
#line 6640
allow sysadm_t insmod_t:fd use;
#line 6640

#line 6640
#
#line 6640
# Allow the new domain to write back to the old domain via a pipe.
#line 6640
#
#line 6640
allow insmod_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6640

#line 6640
#
#line 6640
# Allow the new domain to read and execute the program.
#line 6640
#
#line 6640
allow insmod_t insmod_exec_t:file { read getattr lock execute ioctl };
#line 6640

#line 6640
#
#line 6640
# Allow the new domain to be entered via the program.
#line 6640
#
#line 6640
allow insmod_t insmod_exec_t:file entrypoint;
#line 6640

#line 6640
type_transition sysadm_t insmod_exec_t:process insmod_t;
#line 6640


# Read module config and dependency information
allow insmod_t fs_t:filesystem getattr;
allow insmod_t { modules_conf_t modules_dep_t }:file { getattr read };

# Read module objects.
allow insmod_t modules_object_t:dir { read getattr lock search ioctl };
allow insmod_t modules_object_t:{ file lnk_file } { read getattr lock ioctl };

allow insmod_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Use the sys_module capability.
allow insmod_t self:capability { sys_module kill };

# for ipv6
allow insmod_t self:capability net_raw;

# Inherit and use descriptors from init.
allow insmod_t privfd:fd use;

# I do not know why insmod sends signals or what signals it sends
allow insmod_t domain:process signal;

# Rules for /proc/sys/kernel/tainted
allow insmod_t { proc_t sysctl_t sysctl_kernel_t }:dir search;
allow insmod_t sysctl_kernel_t:file { setattr { ioctl read getattr lock write append } };
allow insmod_t sysctl_t:file write;
allow insmod_t proc_t:file { getattr read };

# /var/log/ksymoops/*
allow insmod_t var_t:dir { read getattr lock search ioctl };
allow insmod_t var_log_t:dir { read getattr lock search ioctl };
allow insmod_t var_log_ksyms_t:file { create ioctl read getattr lock write setattr append link unlink rename };
allow insmod_t var_log_ksyms_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };

#line 6678

#line 6678

#line 6678
allow system_crond_t var_log_ksyms_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6678
allow system_crond_t var_log_ksyms_t:{ file lnk_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6678

#line 6678


#line 6682


allow insmod_t device_t:dir search;
allow insmod_t etc_runtime_t:file { getattr read };
allow insmod_t self:process { fork { sigchld sigkill sigstop signull signal } };

# Access terminals.
allow insmod_t { initrc_devpts_t { sysadm_tty_device_t sysadm_devpts_t } }:chr_file { ioctl read getattr lock write append };



#line 6692
allow insmod_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6692
allow insmod_t lib_t:lnk_file { read getattr lock ioctl };
#line 6692
allow insmod_t ld_so_t:file { read getattr lock execute ioctl };
#line 6692
allow insmod_t ld_so_t:file execute_no_trans;
#line 6692
allow insmod_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6692
allow insmod_t shlib_t:file { read getattr lock execute ioctl };
#line 6692
allow insmod_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6692
allow insmod_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6692
allow insmod_t device_t:dir search;
#line 6692
allow insmod_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6692


# for when /var is not mounted early in the boot
dontaudit insmod_t file_t:dir search;

#################################
#
# Rules for the update_modules_t domain.
#
type update_modules_t, domain, privlog;
type update_modules_exec_t, file_type, exec_type, sysadmfile;

role system_r types update_modules_t;
role sysadm_r types update_modules_t;


#line 6707

#line 6707

#line 6707
#
#line 6707
# Allow the process to transition to the new domain.
#line 6707
#
#line 6707
allow { initrc_t sysadm_t } update_modules_t:process transition;
#line 6707

#line 6707
#
#line 6707
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6707
#
#line 6707
dontaudit { initrc_t sysadm_t } update_modules_t:process noatsecure;
#line 6707

#line 6707
#
#line 6707
# Allow the process to execute the program.
#line 6707
# 
#line 6707
allow { initrc_t sysadm_t } update_modules_exec_t:file { read { getattr execute } };
#line 6707

#line 6707
#
#line 6707
# Allow the process to reap the new domain.
#line 6707
#
#line 6707
allow update_modules_t { initrc_t sysadm_t }:process sigchld;
#line 6707

#line 6707
#
#line 6707
# Allow the new domain to inherit and use file 
#line 6707
# descriptions from the creating process and vice versa.
#line 6707
#
#line 6707
allow update_modules_t { initrc_t sysadm_t }:fd use;
#line 6707
allow { initrc_t sysadm_t } update_modules_t:fd use;
#line 6707

#line 6707
#
#line 6707
# Allow the new domain to write back to the old domain via a pipe.
#line 6707
#
#line 6707
allow update_modules_t { initrc_t sysadm_t }:fifo_file { ioctl read getattr lock write append };
#line 6707

#line 6707
#
#line 6707
# Allow the new domain to read and execute the program.
#line 6707
#
#line 6707
allow update_modules_t update_modules_exec_t:file { read getattr lock execute ioctl };
#line 6707

#line 6707
#
#line 6707
# Allow the new domain to be entered via the program.
#line 6707
#
#line 6707
allow update_modules_t update_modules_exec_t:file entrypoint;
#line 6707

#line 6707
type_transition { initrc_t sysadm_t } update_modules_exec_t:process update_modules_t;
#line 6707

allow update_modules_t privfd:fd use;
allow update_modules_t init_t:fd use;

allow update_modules_t device_t:dir { getattr search };
allow update_modules_t { console_device_t devtty_t }:chr_file { ioctl read getattr lock write append };
allow update_modules_t { initrc_devpts_t { sysadm_tty_device_t sysadm_devpts_t } }:chr_file { ioctl read getattr lock write append };

dontaudit update_modules_t sysadm_home_dir_t:dir search;


#line 6717
allow update_modules_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6717
allow update_modules_t lib_t:lnk_file { read getattr lock ioctl };
#line 6717
allow update_modules_t ld_so_t:file { read getattr lock execute ioctl };
#line 6717
allow update_modules_t ld_so_t:file execute_no_trans;
#line 6717
allow update_modules_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6717
allow update_modules_t shlib_t:file { read getattr lock execute ioctl };
#line 6717
allow update_modules_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6717
allow update_modules_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6717
allow update_modules_t device_t:dir search;
#line 6717
allow update_modules_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6717

allow update_modules_t self:process { fork sigchld };
allow update_modules_t self:fifo_file { ioctl read getattr lock write append };
allow update_modules_t modules_dep_t:file { ioctl read getattr lock write append };

#line 6721

#line 6721

#line 6721

#line 6721
#
#line 6721
# Allow the process to modify the directory.
#line 6721
#
#line 6721
allow update_modules_t modules_object_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6721

#line 6721
#
#line 6721
# Allow the process to create the file.
#line 6721
#
#line 6721

#line 6721
allow update_modules_t modules_conf_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6721
allow update_modules_t modules_conf_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6721

#line 6721

#line 6721
type_transition update_modules_t modules_object_t:dir modules_conf_t;
#line 6721
type_transition update_modules_t modules_object_t:{ file lnk_file sock_file fifo_file } modules_conf_t;
#line 6721

#line 6721

#line 6721


#line 6722

#line 6722

#line 6722
#
#line 6722
# Allow the process to transition to the new domain.
#line 6722
#
#line 6722
allow update_modules_t depmod_t:process transition;
#line 6722

#line 6722
#
#line 6722
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6722
#
#line 6722
dontaudit update_modules_t depmod_t:process noatsecure;
#line 6722

#line 6722
#
#line 6722
# Allow the process to execute the program.
#line 6722
# 
#line 6722
allow update_modules_t depmod_exec_t:file { read { getattr execute } };
#line 6722

#line 6722
#
#line 6722
# Allow the process to reap the new domain.
#line 6722
#
#line 6722
allow depmod_t update_modules_t:process sigchld;
#line 6722

#line 6722
#
#line 6722
# Allow the new domain to inherit and use file 
#line 6722
# descriptions from the creating process and vice versa.
#line 6722
#
#line 6722
allow depmod_t update_modules_t:fd use;
#line 6722
allow update_modules_t depmod_t:fd use;
#line 6722

#line 6722
#
#line 6722
# Allow the new domain to write back to the old domain via a pipe.
#line 6722
#
#line 6722
allow depmod_t update_modules_t:fifo_file { ioctl read getattr lock write append };
#line 6722

#line 6722
#
#line 6722
# Allow the new domain to read and execute the program.
#line 6722
#
#line 6722
allow depmod_t depmod_exec_t:file { read getattr lock execute ioctl };
#line 6722

#line 6722
#
#line 6722
# Allow the new domain to be entered via the program.
#line 6722
#
#line 6722
allow depmod_t depmod_exec_t:file entrypoint;
#line 6722

#line 6722
type_transition update_modules_t depmod_exec_t:process depmod_t;
#line 6722


#line 6723
allow update_modules_t { shell_exec_t bin_t sbin_t update_modules_exec_t etc_t }:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6723

allow update_modules_t bin_t:lnk_file read;
allow update_modules_t { sbin_t bin_t }:dir search;
allow update_modules_t { etc_t etc_runtime_t }:file { read getattr lock ioctl };
allow update_modules_t etc_t:lnk_file read;
allow update_modules_t fs_t:filesystem getattr;

allow update_modules_t proc_t:dir search;
allow update_modules_t proc_t:file { read getattr lock ioctl };
allow update_modules_t { self proc_t }:lnk_file read;
allow update_modules_t { sysctl_t sysctl_kernel_t }:dir search;
allow update_modules_t sysctl_kernel_t:file { getattr read };
allow update_modules_t self:dir search;
allow update_modules_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };


#line 6738

#line 6738

#line 6738

#line 6738
#
#line 6738
# Allow the process to modify the directory.
#line 6738
#
#line 6738
allow update_modules_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6738

#line 6738
#
#line 6738
# Allow the process to create the file.
#line 6738
#
#line 6738

#line 6738
allow update_modules_t modules_conf_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6738
allow update_modules_t modules_conf_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6738

#line 6738

#line 6738
type_transition update_modules_t etc_t:dir modules_conf_t;
#line 6738
type_transition update_modules_t etc_t:{ file lnk_file sock_file fifo_file } modules_conf_t;
#line 6738

#line 6738

#line 6738


# for when /etc/modules.conf gets the wrong type
allow update_modules_t etc_t:file unlink;


#line 6743
type update_modules_tmp_t, file_type, sysadmfile, tmpfile ;
#line 6743

#line 6743

#line 6743

#line 6743

#line 6743
#
#line 6743
# Allow the process to modify the directory.
#line 6743
#
#line 6743
allow update_modules_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6743

#line 6743
#
#line 6743
# Allow the process to create the file.
#line 6743
#
#line 6743

#line 6743
allow update_modules_t update_modules_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6743
allow update_modules_t update_modules_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6743

#line 6743

#line 6743
type_transition update_modules_t tmp_t:dir update_modules_tmp_t;
#line 6743
type_transition update_modules_t tmp_t:{ file lnk_file sock_file fifo_file } update_modules_tmp_t;
#line 6743

#line 6743

#line 6743

#line 6743

#DESC Mount - Filesystem mount utilities
#
# Macros for mount
#
# Author:  Brian May <bam@snoopy.apana.org.au>
# X-Debian-Packages: mount
#
# based on the work of:
#          Mark Westerman mark.westerman@csoconline.com
#

type mount_exec_t, file_type, sysadmfile, exec_type;


#line 6757
#
#line 6757
# Rules for the mount_t domain, used by the sysadm_t domain.
#line 6757
#
#line 6757
# mount_t is the domain for the mount process.
#line 6757
#
#line 6757
# This macro will not be included by all users and it may be included twice if
#line 6757
# called from other macros, so we need protection for this do not call this
#line 6757
# macro if mount_def is defined
#line 6757

#line 6757
#
#line 6757
type mount_t, domain, privlog;
#line 6757

#line 6757

#line 6757
allow mount_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6757
allow mount_t lib_t:lnk_file { read getattr lock ioctl };
#line 6757
allow mount_t ld_so_t:file { read getattr lock execute ioctl };
#line 6757
allow mount_t ld_so_t:file execute_no_trans;
#line 6757
allow mount_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6757
allow mount_t shlib_t:file { read getattr lock execute ioctl };
#line 6757
allow mount_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6757
allow mount_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6757
allow mount_t device_t:dir search;
#line 6757
allow mount_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6757

#line 6757

#line 6757
# when mount is run by sysadm goto mount_t domain
#line 6757

#line 6757

#line 6757

#line 6757
#
#line 6757
# Allow the process to transition to the new domain.
#line 6757
#
#line 6757
allow sysadm_t mount_t:process transition;
#line 6757

#line 6757
#
#line 6757
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6757
#
#line 6757
dontaudit sysadm_t mount_t:process noatsecure;
#line 6757

#line 6757
#
#line 6757
# Allow the process to execute the program.
#line 6757
# 
#line 6757
allow sysadm_t mount_exec_t:file { read { getattr execute } };
#line 6757

#line 6757
#
#line 6757
# Allow the process to reap the new domain.
#line 6757
#
#line 6757
allow mount_t sysadm_t:process sigchld;
#line 6757

#line 6757
#
#line 6757
# Allow the new domain to inherit and use file 
#line 6757
# descriptions from the creating process and vice versa.
#line 6757
#
#line 6757
allow mount_t sysadm_t:fd use;
#line 6757
allow sysadm_t mount_t:fd use;
#line 6757

#line 6757
#
#line 6757
# Allow the new domain to write back to the old domain via a pipe.
#line 6757
#
#line 6757
allow mount_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6757

#line 6757
#
#line 6757
# Allow the new domain to read and execute the program.
#line 6757
#
#line 6757
allow mount_t mount_exec_t:file { read getattr lock execute ioctl };
#line 6757

#line 6757
#
#line 6757
# Allow the new domain to be entered via the program.
#line 6757
#
#line 6757
allow mount_t mount_exec_t:file entrypoint;
#line 6757

#line 6757
type_transition sysadm_t mount_exec_t:process mount_t;
#line 6757

#line 6757

#line 6757
allow mount_t proc_t:dir search;
#line 6757

#line 6757

#line 6757
type mount_tmp_t, file_type, sysadmfile, tmpfile ;
#line 6757

#line 6757

#line 6757

#line 6757

#line 6757
#
#line 6757
# Allow the process to modify the directory.
#line 6757
#
#line 6757
allow mount_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6757

#line 6757
#
#line 6757
# Allow the process to create the file.
#line 6757
#
#line 6757

#line 6757
allow mount_t mount_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 6757
allow mount_t mount_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 6757

#line 6757

#line 6757
type_transition mount_t tmp_t:dir mount_tmp_t;
#line 6757
type_transition mount_t tmp_t:{ file lnk_file sock_file fifo_file } mount_tmp_t;
#line 6757

#line 6757

#line 6757

#line 6757

#line 6757

#line 6757
# Use capabilities.
#line 6757
allow mount_t self:capability { net_bind_service sys_rawio sys_admin };
#line 6757

#line 6757
# Create and modify /etc/mtab.
#line 6757

#line 6757

#line 6757

#line 6757

#line 6757
#
#line 6757
# Allow the process to modify the directory.
#line 6757
#
#line 6757
allow mount_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 6757

#line 6757
#
#line 6757
# Allow the process to create the file.
#line 6757
#
#line 6757

#line 6757

#line 6757
allow mount_t etc_runtime_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 6757

#line 6757

#line 6757
type_transition mount_t etc_t:file etc_runtime_t;
#line 6757

#line 6757

#line 6757

#line 6757

#line 6757
allow mount_t etc_t:file { getattr read };
#line 6757

#line 6757

#line 6757
allow mount_t etc_t:lnk_file read;
#line 6757

#line 6757
allow mount_t locale_t:dir { read getattr lock search ioctl };
#line 6757
allow mount_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6757

#line 6757

#line 6757

#line 6757
# Access the terminal.
#line 6757
allow mount_t sysadm_tty_device_t:chr_file { getattr read write ioctl };
#line 6757
allow mount_t sysadm_devpts_t:chr_file { getattr read write };
#line 6757

#line 6757

role sysadm_r types mount_t;
role system_r types mount_t;


#line 6761

#line 6761

#line 6761
#
#line 6761
# Allow the process to transition to the new domain.
#line 6761
#
#line 6761
allow initrc_t mount_t:process transition;
#line 6761

#line 6761
#
#line 6761
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6761
#
#line 6761
dontaudit initrc_t mount_t:process noatsecure;
#line 6761

#line 6761
#
#line 6761
# Allow the process to execute the program.
#line 6761
# 
#line 6761
allow initrc_t mount_exec_t:file { read { getattr execute } };
#line 6761

#line 6761
#
#line 6761
# Allow the process to reap the new domain.
#line 6761
#
#line 6761
allow mount_t initrc_t:process sigchld;
#line 6761

#line 6761
#
#line 6761
# Allow the new domain to inherit and use file 
#line 6761
# descriptions from the creating process and vice versa.
#line 6761
#
#line 6761
allow mount_t initrc_t:fd use;
#line 6761
allow initrc_t mount_t:fd use;
#line 6761

#line 6761
#
#line 6761
# Allow the new domain to write back to the old domain via a pipe.
#line 6761
#
#line 6761
allow mount_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 6761

#line 6761
#
#line 6761
# Allow the new domain to read and execute the program.
#line 6761
#
#line 6761
allow mount_t mount_exec_t:file { read getattr lock execute ioctl };
#line 6761

#line 6761
#
#line 6761
# Allow the new domain to be entered via the program.
#line 6761
#
#line 6761
allow mount_t mount_exec_t:file entrypoint;
#line 6761

#line 6761
type_transition initrc_t mount_exec_t:process mount_t;
#line 6761

allow mount_t init_t:fd use;
allow mount_t privfd:fd use;

allow mount_t self:capability { ipc_lock dac_override };
allow mount_t self:process { fork { sigchld sigkill sigstop signull signal } };

allow mount_t file_type:dir search;

# Access disk devices.
allow mount_t fixed_disk_device_t:{ chr_file blk_file } { ioctl read getattr lock write append };
allow mount_t removable_device_t:{ chr_file blk_file } { ioctl read getattr lock write append };

# Mount, remount and unmount file systems.
allow mount_t fs_type:filesystem { mount remount unmount getattr };
allow mount_t file_t:dir mounton;
allow mount_t usr_t:dir mounton;
allow mount_t proc_t:dir mounton;
allow mount_t root_t:dir mounton;
allow mount_t home_root_t:dir mounton;
# On some RedHat systems, /boot is a mount point
allow mount_t boot_t:dir mounton;
allow mount_t device_t:dir mounton;
#line 6786

allow mount_t root_t:filesystem unmount;


#DESC Netutils - Network utilities
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil>
# X-Debian-Packages: netbase iputils arping tcpdump 
#

#
# Rules for the netutils_t domain.
# This domain is for network utilities that require access to
# special protocol families.
#
type netutils_t, domain, privlog;
type netutils_exec_t, file_type, sysadmfile, exec_type;
role system_r types netutils_t;
role sysadm_r types netutils_t;


#line 6806
allow netutils_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6806
allow netutils_t lib_t:lnk_file { read getattr lock ioctl };
#line 6806
allow netutils_t ld_so_t:file { read getattr lock execute ioctl };
#line 6806
allow netutils_t ld_so_t:file execute_no_trans;
#line 6806
allow netutils_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6806
allow netutils_t shlib_t:file { read getattr lock execute ioctl };
#line 6806
allow netutils_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6806
allow netutils_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6806
allow netutils_t device_t:dir search;
#line 6806
allow netutils_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6806


#line 6807
#
#line 6807
# Allow the domain to create and use UDP and TCP sockets.
#line 6807
# Other kinds of sockets must be separately authorized for use.
#line 6807
allow netutils_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6807
allow netutils_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 6807

#line 6807
#
#line 6807
# Allow the domain to send UDP packets.
#line 6807
# Since the destination sockets type is unknown, the generic
#line 6807
# any_socket_t type is used as a placeholder.
#line 6807
#
#line 6807
allow netutils_t any_socket_t:udp_socket sendto;
#line 6807

#line 6807
#
#line 6807
# Allow the domain to send using any network interface.
#line 6807
# netif_type is a type attribute for all network interface types.
#line 6807
#
#line 6807
allow netutils_t netif_type:netif { tcp_send udp_send rawip_send };
#line 6807

#line 6807
#
#line 6807
# Allow packets sent by the domain to be received on any network interface.
#line 6807
#
#line 6807
allow netutils_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 6807

#line 6807
#
#line 6807
# Allow the domain to receive packets from any network interface.
#line 6807
# netmsg_type is a type attribute for all default message types.
#line 6807
#
#line 6807
allow netutils_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 6807

#line 6807
#
#line 6807
# Allow the domain to initiate or accept TCP connections 
#line 6807
# on any network interface.
#line 6807
#
#line 6807
allow netutils_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 6807

#line 6807
#
#line 6807
# Receive resets from the TCP reset socket.
#line 6807
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 6807
#
#line 6807
allow netutils_t tcp_socket_t:tcp_socket recvfrom;
#line 6807

#line 6807
dontaudit netutils_t tcp_socket_t:tcp_socket connectto;
#line 6807

#line 6807
#
#line 6807
# Allow the domain to send to any node.
#line 6807
# node_type is a type attribute for all node types.
#line 6807
#
#line 6807
allow netutils_t node_type:node { tcp_send udp_send rawip_send };
#line 6807

#line 6807
#
#line 6807
# Allow packets sent by the domain to be received from any node.
#line 6807
#
#line 6807
allow netutils_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 6807

#line 6807
#
#line 6807
# Allow the domain to send NFS client requests via the socket
#line 6807
# created by mount.
#line 6807
#
#line 6807
allow netutils_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6807

#line 6807
#
#line 6807
# Bind to the default port type.
#line 6807
# Other port types must be separately authorized.
#line 6807
#
#line 6807
allow netutils_t port_t:udp_socket name_bind;
#line 6807
allow netutils_t port_t:tcp_socket name_bind;
#line 6807



#line 6809

#line 6809

#line 6809
#
#line 6809
# Allow the process to transition to the new domain.
#line 6809
#
#line 6809
allow initrc_t netutils_t:process transition;
#line 6809

#line 6809
#
#line 6809
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6809
#
#line 6809
dontaudit initrc_t netutils_t:process noatsecure;
#line 6809

#line 6809
#
#line 6809
# Allow the process to execute the program.
#line 6809
# 
#line 6809
allow initrc_t netutils_exec_t:file { read { getattr execute } };
#line 6809

#line 6809
#
#line 6809
# Allow the process to reap the new domain.
#line 6809
#
#line 6809
allow netutils_t initrc_t:process sigchld;
#line 6809

#line 6809
#
#line 6809
# Allow the new domain to inherit and use file 
#line 6809
# descriptions from the creating process and vice versa.
#line 6809
#
#line 6809
allow netutils_t initrc_t:fd use;
#line 6809
allow initrc_t netutils_t:fd use;
#line 6809

#line 6809
#
#line 6809
# Allow the new domain to write back to the old domain via a pipe.
#line 6809
#
#line 6809
allow netutils_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 6809

#line 6809
#
#line 6809
# Allow the new domain to read and execute the program.
#line 6809
#
#line 6809
allow netutils_t netutils_exec_t:file { read getattr lock execute ioctl };
#line 6809

#line 6809
#
#line 6809
# Allow the new domain to be entered via the program.
#line 6809
#
#line 6809
allow netutils_t netutils_exec_t:file entrypoint;
#line 6809

#line 6809
type_transition initrc_t netutils_exec_t:process netutils_t;
#line 6809


#line 6810

#line 6810

#line 6810
#
#line 6810
# Allow the process to transition to the new domain.
#line 6810
#
#line 6810
allow sysadm_t netutils_t:process transition;
#line 6810

#line 6810
#
#line 6810
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6810
#
#line 6810
dontaudit sysadm_t netutils_t:process noatsecure;
#line 6810

#line 6810
#
#line 6810
# Allow the process to execute the program.
#line 6810
# 
#line 6810
allow sysadm_t netutils_exec_t:file { read { getattr execute } };
#line 6810

#line 6810
#
#line 6810
# Allow the process to reap the new domain.
#line 6810
#
#line 6810
allow netutils_t sysadm_t:process sigchld;
#line 6810

#line 6810
#
#line 6810
# Allow the new domain to inherit and use file 
#line 6810
# descriptions from the creating process and vice versa.
#line 6810
#
#line 6810
allow netutils_t sysadm_t:fd use;
#line 6810
allow sysadm_t netutils_t:fd use;
#line 6810

#line 6810
#
#line 6810
# Allow the new domain to write back to the old domain via a pipe.
#line 6810
#
#line 6810
allow netutils_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 6810

#line 6810
#
#line 6810
# Allow the new domain to read and execute the program.
#line 6810
#
#line 6810
allow netutils_t netutils_exec_t:file { read getattr lock execute ioctl };
#line 6810

#line 6810
#
#line 6810
# Allow the new domain to be entered via the program.
#line 6810
#
#line 6810
allow netutils_t netutils_exec_t:file entrypoint;
#line 6810

#line 6810
type_transition sysadm_t netutils_exec_t:process netutils_t;
#line 6810


# Inherit and use descriptors from init.
allow netutils_t init_t:fd use;

allow netutils_t self:process { fork { sigchld sigkill sigstop signull signal } };

# Perform network administration operations and have raw access to the network.
allow netutils_t self:capability { net_admin net_raw setuid setgid };

# Create and use netlink sockets.
allow netutils_t self:netlink_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Create and use packet sockets.
allow netutils_t self:packet_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Create and use UDP sockets.
allow netutils_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Create and use TCP sockets.
allow netutils_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

allow netutils_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };

# Read certain files in /etc
allow netutils_t etc_t:file { read getattr lock ioctl };
allow netutils_t resolv_conf_t:file { read getattr lock ioctl };

#line 6837
allow netutils_t etc_t:lnk_file read;
#line 6837

#line 6837
allow netutils_t locale_t:dir { read getattr lock search ioctl };
#line 6837
allow netutils_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6837

#line 6837


allow netutils_t fs_t:filesystem getattr;

# Access terminals.
allow netutils_t privfd:fd use;
allow netutils_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { ioctl read getattr lock write append };

#DESC Newrole - SELinux utility to run a shell with a new role
#
# Authors:  Anthony Colatrella (NSA) 
# Maintained by Stephen Smalley <sds@epoch.ncsc.mil>
# X-Debian-Packages: policycoreutils
#

#################################
#
# Rules for the newrole_t domain.
#
# newrole_t is the domain for the newrole program.
# newrole_exec_t is the type of the newrole executable.
#
type newrole_t, domain, privrole, privowner, privlog, auth, privfd;

#line 6860
role user_r types newrole_t;
#line 6860
role staff_r types newrole_t;
#line 6860

role sysadm_r types newrole_t;


#line 6863
# Access other processes in the same domain.
#line 6863
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 6863
# separately if desired.
#line 6863
allow newrole_t self:process ~{ptrace setexec setfscreate};
#line 6863

#line 6863
# Access /proc/PID files for processes in the same domain.
#line 6863
allow newrole_t self:dir { read getattr lock search ioctl };
#line 6863
allow newrole_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 6863

#line 6863
# Access file descriptions, pipes, and sockets
#line 6863
# created by processes in the same domain.
#line 6863
allow newrole_t self:fd *;
#line 6863
allow newrole_t self:fifo_file { ioctl read getattr lock write append };
#line 6863
allow newrole_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 6863
allow newrole_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 6863

#line 6863
# Allow the domain to communicate with other processes in the same domain.
#line 6863
allow newrole_t self:unix_dgram_socket sendto;
#line 6863
allow newrole_t self:unix_stream_socket connectto;
#line 6863

#line 6863
# Access System V IPC objects created by processes in the same domain.
#line 6863
allow newrole_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 6863
allow newrole_t self:msg  { send receive };
#line 6863
allow newrole_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 6863
allow newrole_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 6863
;


#line 6865
allow newrole_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 6865
allow newrole_t lib_t:lnk_file { read getattr lock ioctl };
#line 6865
allow newrole_t ld_so_t:file { read getattr lock execute ioctl };
#line 6865
allow newrole_t ld_so_t:file execute_no_trans;
#line 6865
allow newrole_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 6865
allow newrole_t shlib_t:file { read getattr lock execute ioctl };
#line 6865
allow newrole_t shlib_t:lnk_file { read getattr lock ioctl };
#line 6865
allow newrole_t ld_so_cache_t:file { read getattr lock ioctl };
#line 6865
allow newrole_t device_t:dir search;
#line 6865
allow newrole_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 6865


#line 6866
allow newrole_t etc_t:lnk_file read;
#line 6866

#line 6866
allow newrole_t locale_t:dir { read getattr lock search ioctl };
#line 6866
allow newrole_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 6866

#line 6866


type newrole_exec_t, file_type, exec_type, sysadmfile;

#line 6869

#line 6869

#line 6869
#
#line 6869
# Allow the process to transition to the new domain.
#line 6869
#
#line 6869
allow userdomain newrole_t:process transition;
#line 6869

#line 6869
#
#line 6869
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6869
#
#line 6869
dontaudit userdomain newrole_t:process noatsecure;
#line 6869

#line 6869
#
#line 6869
# Allow the process to execute the program.
#line 6869
# 
#line 6869
allow userdomain newrole_exec_t:file { read { getattr execute } };
#line 6869

#line 6869
#
#line 6869
# Allow the process to reap the new domain.
#line 6869
#
#line 6869
allow newrole_t userdomain:process sigchld;
#line 6869

#line 6869
#
#line 6869
# Allow the new domain to inherit and use file 
#line 6869
# descriptions from the creating process and vice versa.
#line 6869
#
#line 6869
allow newrole_t userdomain:fd use;
#line 6869
allow userdomain newrole_t:fd use;
#line 6869

#line 6869
#
#line 6869
# Allow the new domain to write back to the old domain via a pipe.
#line 6869
#
#line 6869
allow newrole_t userdomain:fifo_file { ioctl read getattr lock write append };
#line 6869

#line 6869
#
#line 6869
# Allow the new domain to read and execute the program.
#line 6869
#
#line 6869
allow newrole_t newrole_exec_t:file { read getattr lock execute ioctl };
#line 6869

#line 6869
#
#line 6869
# Allow the new domain to be entered via the program.
#line 6869
#
#line 6869
allow newrole_t newrole_exec_t:file entrypoint;
#line 6869

#line 6869
type_transition userdomain newrole_exec_t:process newrole_t;
#line 6869


allow newrole_t shadow_t:file { read getattr };

# Inherit descriptors from the current session.
allow newrole_t privfd:fd use;

# Execute /sbin/pwdb_chkpwd to check the password.
allow newrole_t sbin_t:dir { read getattr lock search ioctl };

#line 6878
allow newrole_t chkpwd_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 6878


# Execute shells
allow newrole_t bin_t:dir { read getattr lock search ioctl };
allow newrole_t bin_t:lnk_file read;
allow newrole_t shell_exec_t:file { read getattr lock ioctl };

# Allow newrole_t to transition to user domains.

#line 6886

#line 6886
#
#line 6886
# Allow the process to transition to the new domain.
#line 6886
#
#line 6886
allow newrole_t userdomain:process transition;
#line 6886

#line 6886
#
#line 6886
# Do not audit when glibc secure mode is enabled upon the transition.
#line 6886
#
#line 6886
dontaudit newrole_t userdomain:process noatsecure;
#line 6886

#line 6886
#
#line 6886
# Allow the process to execute the program.
#line 6886
# 
#line 6886
allow newrole_t shell_exec_t:file { read { getattr execute } };
#line 6886

#line 6886
#
#line 6886
# Allow the process to reap the new domain.
#line 6886
#
#line 6886
allow userdomain newrole_t:process sigchld;
#line 6886

#line 6886
#
#line 6886
# Allow the new domain to inherit and use file 
#line 6886
# descriptions from the creating process and vice versa.
#line 6886
#
#line 6886
allow userdomain newrole_t:fd use;
#line 6886
allow newrole_t userdomain:fd use;
#line 6886

#line 6886
#
#line 6886
# Allow the new domain to write back to the old domain via a pipe.
#line 6886
#
#line 6886
allow userdomain newrole_t:fifo_file { ioctl read getattr lock write append };
#line 6886

#line 6886
#
#line 6886
# Allow the new domain to read and execute the program.
#line 6886
#
#line 6886
allow userdomain shell_exec_t:file { read getattr lock execute ioctl };
#line 6886

#line 6886
#
#line 6886
# Allow the new domain to be entered via the program.
#line 6886
#
#line 6886
allow userdomain shell_exec_t:file entrypoint;
#line 6886


#line 6887
allow newrole_t self:process setexec;
#line 6887
allow newrole_t proc_t:dir search;
#line 6887
allow newrole_t proc_t:lnk_file read;
#line 6887
allow newrole_t self:dir search;
#line 6887
allow newrole_t self:file { read write };
#line 6887


# Use capabilities.
allow newrole_t self:capability { setuid setgid net_bind_service dac_override };

# Write to utmp.
allow newrole_t var_run_t:dir { read getattr lock search ioctl };
allow newrole_t initrc_var_run_t:file { ioctl read getattr lock write append };

# Read the devpts root directory.
allow newrole_t devpts_t:dir { read getattr lock search ioctl };

# Read the /etc/security/default_type file
allow newrole_t etc_t:file { read getattr lock ioctl };

# Read /var.
allow newrole_t var_t:dir { read getattr lock search ioctl };
allow newrole_t var_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };

# Read /dev directories and any symbolic links.
allow newrole_t device_t:dir { read getattr lock search ioctl };

# Relabel terminals.
allow newrole_t { ttyfile ptyfile }:chr_file { relabelfrom relabelto };

# Access terminals.
allow newrole_t { ttyfile ptyfile devtty_t }:chr_file { ioctl read getattr lock write append };


#line 6919

#line 6919
# for /etc/alternatives
#line 6919
allow newrole_t etc_t:lnk_file read;
#line 6919


#
# Allow newrole to obtain contexts to relabel TTYs
#

#line 6924
allow newrole_t security_t:dir { read search getattr };
#line 6924
allow newrole_t security_t:file { getattr read write };
#line 6924
allow newrole_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 6924


allow newrole_t fs_t:filesystem getattr;

# for some PAM modules and for cwd
dontaudit newrole_t { home_root_t home_type }:dir search;

# Failed reads to /proc cause no harm, so do not audit them
dontaudit newrole_t proc_t:dir search;

# for when the network connection is killed
dontaudit unpriv_userdomain newrole_t:process signal;
#DESC Passwd - Password utilities
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: passwd
#

#################################
#
# Rules for the passwd_t domain.
#
# passwd_t is the domain of the passwd process.
# passwd_exec_t is the type of the passwd program.
# This domain and type is also used for chfn and chsh.
#
#line 7002



#line 7004
type passwd_t, domain, privlog, auth, privowner;
#line 7004

#line 7004

#line 7004
# Access other processes in the same domain.
#line 7004
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 7004
# separately if desired.
#line 7004
allow passwd_t self:process ~{ptrace setexec setfscreate};
#line 7004

#line 7004
# Access /proc/PID files for processes in the same domain.
#line 7004
allow passwd_t self:dir { read getattr lock search ioctl };
#line 7004
allow passwd_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7004

#line 7004
# Access file descriptions, pipes, and sockets
#line 7004
# created by processes in the same domain.
#line 7004
allow passwd_t self:fd *;
#line 7004
allow passwd_t self:fifo_file { ioctl read getattr lock write append };
#line 7004
allow passwd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7004
allow passwd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7004

#line 7004
# Allow the domain to communicate with other processes in the same domain.
#line 7004
allow passwd_t self:unix_dgram_socket sendto;
#line 7004
allow passwd_t self:unix_stream_socket connectto;
#line 7004

#line 7004
# Access System V IPC objects created by processes in the same domain.
#line 7004
allow passwd_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 7004
allow passwd_t self:msg  { send receive };
#line 7004
allow passwd_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 7004
allow passwd_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 7004
;
#line 7004

#line 7004
allow passwd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7004
allow passwd_t lib_t:lnk_file { read getattr lock ioctl };
#line 7004
allow passwd_t ld_so_t:file { read getattr lock execute ioctl };
#line 7004
allow passwd_t ld_so_t:file execute_no_trans;
#line 7004
allow passwd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7004
allow passwd_t shlib_t:file { read getattr lock execute ioctl };
#line 7004
allow passwd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7004
allow passwd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7004
allow passwd_t device_t:dir search;
#line 7004
allow passwd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7004
;
#line 7004

#line 7004
# Inherit and use descriptors from login.
#line 7004
allow passwd_t privfd:fd use;
#line 7004

#line 7004

#line 7004

#line 7004
allow passwd_t etc_t:lnk_file read;
#line 7004

#line 7004
allow passwd_t locale_t:dir { read getattr lock search ioctl };
#line 7004
allow passwd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 7004

#line 7004

#line 7004

#line 7004
allow passwd_t fs_t:filesystem getattr;
#line 7004

#line 7004
# allow checking if a shell is executable
#line 7004
allow passwd_t shell_exec_t:file execute;
#line 7004

#line 7004
# Obtain contexts
#line 7004

#line 7004
allow passwd_t security_t:dir { read search getattr };
#line 7004
allow passwd_t security_t:file { getattr read write };
#line 7004
allow passwd_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7004

#line 7004

#line 7004
# Update /etc/shadow and /etc/passwd
#line 7004

#line 7004

#line 7004

#line 7004

#line 7004
#
#line 7004
# Allow the process to modify the directory.
#line 7004
#
#line 7004
allow passwd_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7004

#line 7004
#
#line 7004
# Allow the process to create the file.
#line 7004
#
#line 7004

#line 7004

#line 7004
allow passwd_t shadow_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7004

#line 7004

#line 7004
type_transition passwd_t etc_t:file shadow_t;
#line 7004

#line 7004

#line 7004

#line 7004
allow passwd_t etc_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7004
allow passwd_t { etc_t shadow_t }:file { relabelfrom relabelto };
#line 7004

#line 7004
allow passwd_t self:process setfscreate;
#line 7004
allow passwd_t proc_t:dir search;
#line 7004
allow passwd_t proc_t:lnk_file read;
#line 7004
allow passwd_t self:dir search;
#line 7004
allow passwd_t self:file { read write };
#line 7004

#line 7004

#line 7004
# Allow etc_t symlinks for /etc/alternatives on Debian.
#line 7004
allow passwd_t etc_t:lnk_file read;
#line 7004

#line 7004
# Use capabilities.
#line 7004
allow passwd_t self:capability { chown dac_override fsetid setuid sys_resource };
#line 7004

#line 7004
# Access terminals.
#line 7004
allow passwd_t { ttyfile ptyfile }:chr_file { ioctl read getattr lock write append };
#line 7004
allow passwd_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 7004

#line 7004
dontaudit passwd_t devpts_t:dir getattr;
#line 7004

#line 7004
# /usr/bin/passwd asks for w access to utmp, but it will operate
#line 7004
# correctly without it.  Do not audit write denials to utmp.
#line 7004
dontaudit passwd_t initrc_var_run_t:file { read write };
#line 7004

#line 7004
# user generally runs this from their home directory, so do not audit a search
#line 7004
# on user home dir
#line 7004
dontaudit passwd_t { user_home_dir_type user_home_type }:dir search;
#line 7004

#line 7004
# When the wrong current passwd is entered, passwd, for some reason, 
#line 7004
# attempts to access /proc and /dev, but fails appropriately. So dont
#line 7004
# audit those denials.
#line 7004
dontaudit passwd_t { proc_t device_t }:dir { search read };
#line 7004

#line 7004
allow passwd_t device_t:dir getattr;
#line 7004


#line 7005
type sysadm_passwd_t, domain, privlog, auth, privowner;
#line 7005

#line 7005

#line 7005
# Access other processes in the same domain.
#line 7005
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 7005
# separately if desired.
#line 7005
allow sysadm_passwd_t self:process ~{ptrace setexec setfscreate};
#line 7005

#line 7005
# Access /proc/PID files for processes in the same domain.
#line 7005
allow sysadm_passwd_t self:dir { read getattr lock search ioctl };
#line 7005
allow sysadm_passwd_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7005

#line 7005
# Access file descriptions, pipes, and sockets
#line 7005
# created by processes in the same domain.
#line 7005
allow sysadm_passwd_t self:fd *;
#line 7005
allow sysadm_passwd_t self:fifo_file { ioctl read getattr lock write append };
#line 7005
allow sysadm_passwd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7005
allow sysadm_passwd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7005

#line 7005
# Allow the domain to communicate with other processes in the same domain.
#line 7005
allow sysadm_passwd_t self:unix_dgram_socket sendto;
#line 7005
allow sysadm_passwd_t self:unix_stream_socket connectto;
#line 7005

#line 7005
# Access System V IPC objects created by processes in the same domain.
#line 7005
allow sysadm_passwd_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 7005
allow sysadm_passwd_t self:msg  { send receive };
#line 7005
allow sysadm_passwd_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 7005
allow sysadm_passwd_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 7005
;
#line 7005

#line 7005
allow sysadm_passwd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7005
allow sysadm_passwd_t lib_t:lnk_file { read getattr lock ioctl };
#line 7005
allow sysadm_passwd_t ld_so_t:file { read getattr lock execute ioctl };
#line 7005
allow sysadm_passwd_t ld_so_t:file execute_no_trans;
#line 7005
allow sysadm_passwd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7005
allow sysadm_passwd_t shlib_t:file { read getattr lock execute ioctl };
#line 7005
allow sysadm_passwd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7005
allow sysadm_passwd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7005
allow sysadm_passwd_t device_t:dir search;
#line 7005
allow sysadm_passwd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7005
;
#line 7005

#line 7005
# Inherit and use descriptors from login.
#line 7005
allow sysadm_passwd_t privfd:fd use;
#line 7005

#line 7005

#line 7005

#line 7005
allow sysadm_passwd_t etc_t:lnk_file read;
#line 7005

#line 7005
allow sysadm_passwd_t locale_t:dir { read getattr lock search ioctl };
#line 7005
allow sysadm_passwd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 7005

#line 7005

#line 7005

#line 7005
allow sysadm_passwd_t fs_t:filesystem getattr;
#line 7005

#line 7005
# allow checking if a shell is executable
#line 7005
allow sysadm_passwd_t shell_exec_t:file execute;
#line 7005

#line 7005
# Obtain contexts
#line 7005

#line 7005
allow sysadm_passwd_t security_t:dir { read search getattr };
#line 7005
allow sysadm_passwd_t security_t:file { getattr read write };
#line 7005
allow sysadm_passwd_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7005

#line 7005

#line 7005
# Update /etc/shadow and /etc/passwd
#line 7005

#line 7005

#line 7005

#line 7005

#line 7005
#
#line 7005
# Allow the process to modify the directory.
#line 7005
#
#line 7005
allow sysadm_passwd_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7005

#line 7005
#
#line 7005
# Allow the process to create the file.
#line 7005
#
#line 7005

#line 7005

#line 7005
allow sysadm_passwd_t shadow_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7005

#line 7005

#line 7005
type_transition sysadm_passwd_t etc_t:file shadow_t;
#line 7005

#line 7005

#line 7005

#line 7005
allow sysadm_passwd_t etc_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7005
allow sysadm_passwd_t { etc_t shadow_t }:file { relabelfrom relabelto };
#line 7005

#line 7005
allow sysadm_passwd_t self:process setfscreate;
#line 7005
allow sysadm_passwd_t proc_t:dir search;
#line 7005
allow sysadm_passwd_t proc_t:lnk_file read;
#line 7005
allow sysadm_passwd_t self:dir search;
#line 7005
allow sysadm_passwd_t self:file { read write };
#line 7005

#line 7005

#line 7005
# Allow etc_t symlinks for /etc/alternatives on Debian.
#line 7005
allow sysadm_passwd_t etc_t:lnk_file read;
#line 7005

#line 7005
# Use capabilities.
#line 7005
allow sysadm_passwd_t self:capability { chown dac_override fsetid setuid sys_resource };
#line 7005

#line 7005
# Access terminals.
#line 7005
allow sysadm_passwd_t { ttyfile ptyfile }:chr_file { ioctl read getattr lock write append };
#line 7005
allow sysadm_passwd_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 7005

#line 7005
dontaudit sysadm_passwd_t devpts_t:dir getattr;
#line 7005

#line 7005
# /usr/bin/passwd asks for w access to utmp, but it will operate
#line 7005
# correctly without it.  Do not audit write denials to utmp.
#line 7005
dontaudit sysadm_passwd_t initrc_var_run_t:file { read write };
#line 7005

#line 7005
# user generally runs this from their home directory, so do not audit a search
#line 7005
# on user home dir
#line 7005
dontaudit sysadm_passwd_t { user_home_dir_type user_home_type }:dir search;
#line 7005

#line 7005
# When the wrong current passwd is entered, passwd, for some reason, 
#line 7005
# attempts to access /proc and /dev, but fails appropriately. So dont
#line 7005
# audit those denials.
#line 7005
dontaudit sysadm_passwd_t { proc_t device_t }:dir { search read };
#line 7005

#line 7005
allow sysadm_passwd_t device_t:dir getattr;
#line 7005



#line 7007
role user_r types passwd_t;
#line 7007
role staff_r types passwd_t;
#line 7007

role sysadm_r types passwd_t;
role sysadm_r types sysadm_passwd_t;

type admin_passwd_exec_t, file_type, sysadmfile;
type passwd_exec_t, file_type, sysadmfile, exec_type;


#line 7014

#line 7014

#line 7014
#
#line 7014
# Allow the process to transition to the new domain.
#line 7014
#
#line 7014
allow userdomain passwd_t:process transition;
#line 7014

#line 7014
#
#line 7014
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7014
#
#line 7014
dontaudit userdomain passwd_t:process noatsecure;
#line 7014

#line 7014
#
#line 7014
# Allow the process to execute the program.
#line 7014
# 
#line 7014
allow userdomain passwd_exec_t:file { read { getattr execute } };
#line 7014

#line 7014
#
#line 7014
# Allow the process to reap the new domain.
#line 7014
#
#line 7014
allow passwd_t userdomain:process sigchld;
#line 7014

#line 7014
#
#line 7014
# Allow the new domain to inherit and use file 
#line 7014
# descriptions from the creating process and vice versa.
#line 7014
#
#line 7014
allow passwd_t userdomain:fd use;
#line 7014
allow userdomain passwd_t:fd use;
#line 7014

#line 7014
#
#line 7014
# Allow the new domain to write back to the old domain via a pipe.
#line 7014
#
#line 7014
allow passwd_t userdomain:fifo_file { ioctl read getattr lock write append };
#line 7014

#line 7014
#
#line 7014
# Allow the new domain to read and execute the program.
#line 7014
#
#line 7014
allow passwd_t passwd_exec_t:file { read getattr lock execute ioctl };
#line 7014

#line 7014
#
#line 7014
# Allow the new domain to be entered via the program.
#line 7014
#
#line 7014
allow passwd_t passwd_exec_t:file entrypoint;
#line 7014

#line 7014
type_transition userdomain passwd_exec_t:process passwd_t;
#line 7014


#line 7015

#line 7015

#line 7015
#
#line 7015
# Allow the process to transition to the new domain.
#line 7015
#
#line 7015
allow sysadm_t sysadm_passwd_t:process transition;
#line 7015

#line 7015
#
#line 7015
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7015
#
#line 7015
dontaudit sysadm_t sysadm_passwd_t:process noatsecure;
#line 7015

#line 7015
#
#line 7015
# Allow the process to execute the program.
#line 7015
# 
#line 7015
allow sysadm_t admin_passwd_exec_t:file { read { getattr execute } };
#line 7015

#line 7015
#
#line 7015
# Allow the process to reap the new domain.
#line 7015
#
#line 7015
allow sysadm_passwd_t sysadm_t:process sigchld;
#line 7015

#line 7015
#
#line 7015
# Allow the new domain to inherit and use file 
#line 7015
# descriptions from the creating process and vice versa.
#line 7015
#
#line 7015
allow sysadm_passwd_t sysadm_t:fd use;
#line 7015
allow sysadm_t sysadm_passwd_t:fd use;
#line 7015

#line 7015
#
#line 7015
# Allow the new domain to write back to the old domain via a pipe.
#line 7015
#
#line 7015
allow sysadm_passwd_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 7015

#line 7015
#
#line 7015
# Allow the new domain to read and execute the program.
#line 7015
#
#line 7015
allow sysadm_passwd_t admin_passwd_exec_t:file { read getattr lock execute ioctl };
#line 7015

#line 7015
#
#line 7015
# Allow the new domain to be entered via the program.
#line 7015
#
#line 7015
allow sysadm_passwd_t admin_passwd_exec_t:file entrypoint;
#line 7015

#line 7015
type_transition sysadm_t admin_passwd_exec_t:process sysadm_passwd_t;
#line 7015


#line 7024

#line 7024
dontaudit passwd_t var_t:dir search;
#line 7024


# allow vipw to exec the editor
allow sysadm_passwd_t { root_t usr_t bin_t }:dir search;
allow sysadm_passwd_t { bin_t }:lnk_file read;

#line 7029
allow sysadm_passwd_t bin_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 7029


# allow vipw to create temporary files under /var/tmp/vi.recover
allow sysadm_passwd_t var_t:dir search;

#line 7033
type sysadm_passwd_tmp_t, file_type, sysadmfile, tmpfile ;
#line 7033

#line 7033

#line 7033

#line 7033

#line 7033
#
#line 7033
# Allow the process to modify the directory.
#line 7033
#
#line 7033
allow sysadm_passwd_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7033

#line 7033
#
#line 7033
# Allow the process to create the file.
#line 7033
#
#line 7033

#line 7033
allow sysadm_passwd_t sysadm_passwd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 7033
allow sysadm_passwd_t sysadm_passwd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 7033

#line 7033

#line 7033
type_transition sysadm_passwd_t tmp_t:dir sysadm_passwd_tmp_t;
#line 7033
type_transition sysadm_passwd_t tmp_t:{ file lnk_file sock_file fifo_file } sysadm_passwd_tmp_t;
#line 7033

#line 7033

#line 7033

#line 7033

# for vipw - vi looks in the root home directory for config
dontaudit sysadm_passwd_t sysadm_home_dir_t:dir { getattr search };
# for /etc/alternatives/vi
allow sysadm_passwd_t etc_t:lnk_file read;

# for nscd lookups
dontaudit sysadm_passwd_t var_run_t:dir search;

# make sure that getcon succeeds
allow passwd_t user_t:dir search;
allow passwd_t user_t:file read;
allow passwd_t user_t:process getattr;

#DESC Setfiles - SELinux filesystem labeling utilities
#
# Authors:  Russell Coker <russell@coker.com.au>
# X-Debian-Packages: policycoreutils
#

#################################
#
# Rules for the setfiles_t domain.
#
# setfiles_exec_t is the type of the setfiles executable.
#
type setfiles_t, domain, privlog, privowner;
type setfiles_exec_t, file_type, sysadmfile, exec_type;

role system_r types setfiles_t;
role sysadm_r types setfiles_t;

allow setfiles_t initrc_devpts_t:chr_file { read write ioctl };
allow setfiles_t { sysadm_tty_device_t sysadm_devpts_t }:chr_file { read write ioctl };


#line 7068

#line 7068

#line 7068
#
#line 7068
# Allow the process to transition to the new domain.
#line 7068
#
#line 7068
allow { initrc_t sysadm_t } setfiles_t:process transition;
#line 7068

#line 7068
#
#line 7068
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7068
#
#line 7068
dontaudit { initrc_t sysadm_t } setfiles_t:process noatsecure;
#line 7068

#line 7068
#
#line 7068
# Allow the process to execute the program.
#line 7068
# 
#line 7068
allow { initrc_t sysadm_t } setfiles_exec_t:file { read { getattr execute } };
#line 7068

#line 7068
#
#line 7068
# Allow the process to reap the new domain.
#line 7068
#
#line 7068
allow setfiles_t { initrc_t sysadm_t }:process sigchld;
#line 7068

#line 7068
#
#line 7068
# Allow the new domain to inherit and use file 
#line 7068
# descriptions from the creating process and vice versa.
#line 7068
#
#line 7068
allow setfiles_t { initrc_t sysadm_t }:fd use;
#line 7068
allow { initrc_t sysadm_t } setfiles_t:fd use;
#line 7068

#line 7068
#
#line 7068
# Allow the new domain to write back to the old domain via a pipe.
#line 7068
#
#line 7068
allow setfiles_t { initrc_t sysadm_t }:fifo_file { ioctl read getattr lock write append };
#line 7068

#line 7068
#
#line 7068
# Allow the new domain to read and execute the program.
#line 7068
#
#line 7068
allow setfiles_t setfiles_exec_t:file { read getattr lock execute ioctl };
#line 7068

#line 7068
#
#line 7068
# Allow the new domain to be entered via the program.
#line 7068
#
#line 7068
allow setfiles_t setfiles_exec_t:file entrypoint;
#line 7068

#line 7068
type_transition { initrc_t sysadm_t } setfiles_exec_t:process setfiles_t;
#line 7068

allow setfiles_t init_t:fd use;
allow setfiles_t privfd:fd use;


#line 7072
allow setfiles_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7072
allow setfiles_t lib_t:lnk_file { read getattr lock ioctl };
#line 7072
allow setfiles_t ld_so_t:file { read getattr lock execute ioctl };
#line 7072
allow setfiles_t ld_so_t:file execute_no_trans;
#line 7072
allow setfiles_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7072
allow setfiles_t shlib_t:file { read getattr lock execute ioctl };
#line 7072
allow setfiles_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7072
allow setfiles_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7072
allow setfiles_t device_t:dir search;
#line 7072
allow setfiles_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7072

allow setfiles_t self:capability { dac_override dac_read_search };

# for upgrading glibc - without this the glibc upgrade scripts will put things
# in a state such that setfiles can not be run!
allow setfiles_t lib_t:file execute;

# Get security policy decisions.

#line 7080
allow setfiles_t security_t:dir { read search getattr };
#line 7080
allow setfiles_t security_t:file { getattr read write };
#line 7080
allow setfiles_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7080


allow setfiles_t policy_src_t:file { read getattr lock ioctl };

allow setfiles_t file_type:dir { read getattr lock search ioctl };
allow setfiles_t file_type:lnk_file { read getattr lock ioctl };
allow setfiles_t file_type:{ dir file lnk_file sock_file fifo_file chr_file blk_file } { getattr relabelfrom relabelto };

allow setfiles_t file_labels_t:dir { read getattr lock search ioctl add_name remove_name write };
allow setfiles_t file_labels_t:file { create ioctl read getattr lock write setattr append link unlink rename };

allow setfiles_t fs_t:filesystem getattr;
allow setfiles_t fs_type:dir { read getattr lock search ioctl };

allow setfiles_t etc_runtime_t:file read;
allow setfiles_t etc_t:file read;
allow setfiles_t proc_t:file { getattr read };

# for config files in a home directory
allow setfiles_t home_type:file { read getattr lock ioctl };
#DESC SSH - SSH daemon
#
# Authors:  Anthony Colatrella (NSA) <amcolat@epoch.ncsc.mil>
# Modified by: Russell Coker <russell@coker.com.au>
# X-Debian-Packages: ssh
#

type ssh_port_t, port_type;

#line 7183

#################################
#
# Rules for the sshd_t domain, et al.
#
# sshd_t is the domain for the sshd program.
# sshd_login_t is the domain for sshds login spawn
# sshd_exec_t is the type of the sshd executable.
# sshd_key_t is the type of the ssh private key files
#

#line 7193
type sshd_t, domain, privuser, privrole, privlog, privowner, privfd, auth;
#line 7193
role system_r types sshd_t;
#line 7193

#line 7193
allow sshd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t lib_t:lnk_file { read getattr lock ioctl };
#line 7193
allow sshd_t ld_so_t:file { read getattr lock execute ioctl };
#line 7193
allow sshd_t ld_so_t:file execute_no_trans;
#line 7193
allow sshd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7193
allow sshd_t shlib_t:file { read getattr lock execute ioctl };
#line 7193
allow sshd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7193
allow sshd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7193
allow sshd_t device_t:dir search;
#line 7193
allow sshd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7193

#line 7193
allow sshd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7193
allow sshd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7193
allow sshd_t self:fifo_file { ioctl read getattr lock write append };
#line 7193
allow sshd_t self:process { fork sigchld setsched };
#line 7193

#line 7193
# Read system information files in /proc.
#line 7193
allow sshd_t proc_t:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7193
allow sshd_t self:dir search;
#line 7193
allow sshd_t self:lnk_file read;
#line 7193

#line 7193
# Get attributes of file systems.
#line 7193
allow sshd_t fs_type:filesystem getattr;
#line 7193

#line 7193

#line 7193
# Read /.
#line 7193
allow sshd_t root_t:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7193

#line 7193
# Read /home.
#line 7193
allow sshd_t home_root_t:dir { read getattr lock search ioctl };
#line 7193

#line 7193
# Read /usr.
#line 7193
allow sshd_t usr_t:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7193

#line 7193
# Read bin and sbin directories.
#line 7193
allow sshd_t bin_t:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7193
allow sshd_t sbin_t:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7193

#line 7193

#line 7193
# Read /var.
#line 7193
allow sshd_t var_t:dir search;
#line 7193

#line 7193
# Read /var/log.
#line 7193
allow sshd_t var_log_t:dir search;
#line 7193

#line 7193
# Read /etc.
#line 7193
allow sshd_t etc_t:dir { read getattr lock search ioctl };
#line 7193
allow sshd_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7193
allow sshd_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 7193
allow sshd_t resolv_conf_t:{ file lnk_file } { read getattr lock ioctl };
#line 7193

#line 7193
# Read the linker, shared library, and executable types.
#line 7193
allow sshd_t ld_so_t:{ file lnk_file } { read getattr lock ioctl };
#line 7193
allow sshd_t shlib_t:{ file lnk_file } { read getattr lock ioctl };
#line 7193
allow sshd_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 7193

#line 7193
# Read and write /dev/tty and /dev/null.
#line 7193
allow sshd_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 7193
allow sshd_t { null_device_t zero_device_t }:chr_file { ioctl read getattr lock write append };
#line 7193

#line 7193
# Read /dev/random and /dev/zero.
#line 7193
allow sshd_t random_device_t:chr_file { read getattr lock ioctl };
#line 7193

#line 7193

#line 7193
#
#line 7193
# Allow the domain to create and use UDP and TCP sockets.
#line 7193
# Other kinds of sockets must be separately authorized for use.
#line 7193
allow sshd_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7193
allow sshd_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7193

#line 7193
#
#line 7193
# Allow the domain to send UDP packets.
#line 7193
# Since the destination sockets type is unknown, the generic
#line 7193
# any_socket_t type is used as a placeholder.
#line 7193
#
#line 7193
allow sshd_t any_socket_t:udp_socket sendto;
#line 7193

#line 7193
#
#line 7193
# Allow the domain to send using any network interface.
#line 7193
# netif_type is a type attribute for all network interface types.
#line 7193
#
#line 7193
allow sshd_t netif_type:netif { tcp_send udp_send rawip_send };
#line 7193

#line 7193
#
#line 7193
# Allow packets sent by the domain to be received on any network interface.
#line 7193
#
#line 7193
allow sshd_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 7193

#line 7193
#
#line 7193
# Allow the domain to receive packets from any network interface.
#line 7193
# netmsg_type is a type attribute for all default message types.
#line 7193
#
#line 7193
allow sshd_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 7193

#line 7193
#
#line 7193
# Allow the domain to initiate or accept TCP connections 
#line 7193
# on any network interface.
#line 7193
#
#line 7193
allow sshd_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 7193

#line 7193
#
#line 7193
# Receive resets from the TCP reset socket.
#line 7193
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 7193
#
#line 7193
allow sshd_t tcp_socket_t:tcp_socket recvfrom;
#line 7193

#line 7193
dontaudit sshd_t tcp_socket_t:tcp_socket connectto;
#line 7193

#line 7193
#
#line 7193
# Allow the domain to send to any node.
#line 7193
# node_type is a type attribute for all node types.
#line 7193
#
#line 7193
allow sshd_t node_type:node { tcp_send udp_send rawip_send };
#line 7193

#line 7193
#
#line 7193
# Allow packets sent by the domain to be received from any node.
#line 7193
#
#line 7193
allow sshd_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 7193

#line 7193
#
#line 7193
# Allow the domain to send NFS client requests via the socket
#line 7193
# created by mount.
#line 7193
#
#line 7193
allow sshd_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7193

#line 7193
#
#line 7193
# Bind to the default port type.
#line 7193
# Other port types must be separately authorized.
#line 7193
#
#line 7193
allow sshd_t port_t:udp_socket name_bind;
#line 7193
allow sshd_t port_t:tcp_socket name_bind;
#line 7193

#line 7193

#line 7193
allow sshd_t self:capability { chown dac_override fowner fsetid setgid setuid net_bind_service sys_tty_config };
#line 7193
allow sshd_t shadow_t:file { read getattr };
#line 7193
allow sshd_t { home_root_t user_home_dir_type sysadm_home_dir_t }:dir { search getattr };
#line 7193

#line 7193
# Run shells in user_t by default
#line 7193

#line 7193

#line 7193

#line 7193
#
#line 7193
# Allow the process to transition to the new domain.
#line 7193
#
#line 7193
allow sshd_t user_t:process transition;
#line 7193

#line 7193
#
#line 7193
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7193
#
#line 7193
dontaudit sshd_t user_t:process noatsecure;
#line 7193

#line 7193
#
#line 7193
# Allow the process to execute the program.
#line 7193
# 
#line 7193
allow sshd_t shell_exec_t:file { read { getattr execute } };
#line 7193

#line 7193
#
#line 7193
# Allow the process to reap the new domain.
#line 7193
#
#line 7193
allow user_t sshd_t:process sigchld;
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to inherit and use file 
#line 7193
# descriptions from the creating process and vice versa.
#line 7193
#
#line 7193
allow user_t sshd_t:fd use;
#line 7193
allow sshd_t user_t:fd use;
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to write back to the old domain via a pipe.
#line 7193
#
#line 7193
allow user_t sshd_t:fifo_file { ioctl read getattr lock write append };
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to read and execute the program.
#line 7193
#
#line 7193
allow user_t shell_exec_t:file { read getattr lock execute ioctl };
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to be entered via the program.
#line 7193
#
#line 7193
allow user_t shell_exec_t:file entrypoint;
#line 7193

#line 7193
type_transition sshd_t shell_exec_t:process user_t;
#line 7193

#line 7193

#line 7193

#line 7193
#
#line 7193
# Allow the process to transition to the new domain.
#line 7193
#
#line 7193
allow sshd_t unpriv_userdomain:process transition;
#line 7193

#line 7193
#
#line 7193
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7193
#
#line 7193
dontaudit sshd_t unpriv_userdomain:process noatsecure;
#line 7193

#line 7193
#
#line 7193
# Allow the process to execute the program.
#line 7193
# 
#line 7193
allow sshd_t shell_exec_t:file { read { getattr execute } };
#line 7193

#line 7193
#
#line 7193
# Allow the process to reap the new domain.
#line 7193
#
#line 7193
allow unpriv_userdomain sshd_t:process sigchld;
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to inherit and use file 
#line 7193
# descriptions from the creating process and vice versa.
#line 7193
#
#line 7193
allow unpriv_userdomain sshd_t:fd use;
#line 7193
allow sshd_t unpriv_userdomain:fd use;
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to write back to the old domain via a pipe.
#line 7193
#
#line 7193
allow unpriv_userdomain sshd_t:fifo_file { ioctl read getattr lock write append };
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to read and execute the program.
#line 7193
#
#line 7193
allow unpriv_userdomain shell_exec_t:file { read getattr lock execute ioctl };
#line 7193

#line 7193
#
#line 7193
# Allow the new domain to be entered via the program.
#line 7193
#
#line 7193
allow unpriv_userdomain shell_exec_t:file entrypoint;
#line 7193

#line 7193

#line 7193
# Set exec context.
#line 7193

#line 7193
allow sshd_t self:process setexec;
#line 7193
allow sshd_t proc_t:dir search;
#line 7193
allow sshd_t proc_t:lnk_file read;
#line 7193
allow sshd_t self:dir search;
#line 7193
allow sshd_t self:file { read write };
#line 7193

#line 7193

#line 7193
# Allow shells to be run in sysadm_t as well.
#line 7193
# Commented out.  Use newrole rather than directly entering sysadm_t.
#line 7193
#domain_trans(sshd_t, shell_exec_t, sysadm_t)
#line 7193

#line 7193
# Update utmp.
#line 7193
allow sshd_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 7193

#line 7193
# Update wtmp.
#line 7193
allow sshd_t wtmp_t:file { ioctl read getattr lock write append };
#line 7193

#line 7193
# Get security policy decisions.
#line 7193

#line 7193
allow sshd_t security_t:dir { read search getattr };
#line 7193
allow sshd_t security_t:file { getattr read write };
#line 7193
allow sshd_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7193

#line 7193

#line 7193
# Allow read access to login context
#line 7193
allow sshd_t default_context_t:file { read getattr lock ioctl };
#line 7193

#line 7193

type sshd_exec_t, file_type, exec_type, sysadmfile;


#line 7196

#line 7196

#line 7196
#
#line 7196
# Allow the process to transition to the new domain.
#line 7196
#
#line 7196
allow initrc_t sshd_t:process transition;
#line 7196

#line 7196
#
#line 7196
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7196
#
#line 7196
dontaudit initrc_t sshd_t:process noatsecure;
#line 7196

#line 7196
#
#line 7196
# Allow the process to execute the program.
#line 7196
# 
#line 7196
allow initrc_t sshd_exec_t:file { read { getattr execute } };
#line 7196

#line 7196
#
#line 7196
# Allow the process to reap the new domain.
#line 7196
#
#line 7196
allow sshd_t initrc_t:process sigchld;
#line 7196

#line 7196
#
#line 7196
# Allow the new domain to inherit and use file 
#line 7196
# descriptions from the creating process and vice versa.
#line 7196
#
#line 7196
allow sshd_t initrc_t:fd use;
#line 7196
allow initrc_t sshd_t:fd use;
#line 7196

#line 7196
#
#line 7196
# Allow the new domain to write back to the old domain via a pipe.
#line 7196
#
#line 7196
allow sshd_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 7196

#line 7196
#
#line 7196
# Allow the new domain to read and execute the program.
#line 7196
#
#line 7196
allow sshd_t sshd_exec_t:file { read getattr lock execute ioctl };
#line 7196

#line 7196
#
#line 7196
# Allow the new domain to be entered via the program.
#line 7196
#
#line 7196
allow sshd_t sshd_exec_t:file entrypoint;
#line 7196

#line 7196
type_transition initrc_t sshd_exec_t:process sshd_t;
#line 7196


#line 7202


#line 7212

#line 7212
allow sshd_t initrc_devpts_t:chr_file { ioctl read getattr lock write append };
#line 7212
allow sshd_t ssh_port_t:tcp_socket name_bind;
#line 7212

#line 7212

#line 7212

#line 7212
#
#line 7212
# Allow the process to transition to the new domain.
#line 7212
#
#line 7212
allow initrc_t sshd_t:process transition;
#line 7212

#line 7212
#
#line 7212
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7212
#
#line 7212
dontaudit initrc_t sshd_t:process noatsecure;
#line 7212

#line 7212
#
#line 7212
# Allow the process to execute the program.
#line 7212
# 
#line 7212
allow initrc_t sshd_exec_t:file { read { getattr execute } };
#line 7212

#line 7212
#
#line 7212
# Allow the process to reap the new domain.
#line 7212
#
#line 7212
allow sshd_t initrc_t:process sigchld;
#line 7212

#line 7212
#
#line 7212
# Allow the new domain to inherit and use file 
#line 7212
# descriptions from the creating process and vice versa.
#line 7212
#
#line 7212
allow sshd_t initrc_t:fd use;
#line 7212
allow initrc_t sshd_t:fd use;
#line 7212

#line 7212
#
#line 7212
# Allow the new domain to write back to the old domain via a pipe.
#line 7212
#
#line 7212
allow sshd_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 7212

#line 7212
#
#line 7212
# Allow the new domain to read and execute the program.
#line 7212
#
#line 7212
allow sshd_t sshd_exec_t:file { read getattr lock execute ioctl };
#line 7212

#line 7212
#
#line 7212
# Allow the new domain to be entered via the program.
#line 7212
#
#line 7212
allow sshd_t sshd_exec_t:file entrypoint;
#line 7212

#line 7212
type_transition initrc_t sshd_exec_t:process sshd_t;
#line 7212

#line 7212


#line 7213
type sshd_login_t, domain, privuser, privrole, privlog, privowner, privfd, auth;
#line 7213
role system_r types sshd_login_t;
#line 7213

#line 7213
allow sshd_login_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t lib_t:lnk_file { read getattr lock ioctl };
#line 7213
allow sshd_login_t ld_so_t:file { read getattr lock execute ioctl };
#line 7213
allow sshd_login_t ld_so_t:file execute_no_trans;
#line 7213
allow sshd_login_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7213
allow sshd_login_t shlib_t:file { read getattr lock execute ioctl };
#line 7213
allow sshd_login_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7213
allow sshd_login_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7213
allow sshd_login_t device_t:dir search;
#line 7213
allow sshd_login_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7213

#line 7213
allow sshd_login_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7213
allow sshd_login_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7213
allow sshd_login_t self:fifo_file { ioctl read getattr lock write append };
#line 7213
allow sshd_login_t self:process { fork sigchld setsched };
#line 7213

#line 7213
# Read system information files in /proc.
#line 7213
allow sshd_login_t proc_t:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t proc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7213
allow sshd_login_t self:dir search;
#line 7213
allow sshd_login_t self:lnk_file read;
#line 7213

#line 7213
# Get attributes of file systems.
#line 7213
allow sshd_login_t fs_type:filesystem getattr;
#line 7213

#line 7213

#line 7213
# Read /.
#line 7213
allow sshd_login_t root_t:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t root_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7213

#line 7213
# Read /home.
#line 7213
allow sshd_login_t home_root_t:dir { read getattr lock search ioctl };
#line 7213

#line 7213
# Read /usr.
#line 7213
allow sshd_login_t usr_t:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t usr_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7213

#line 7213
# Read bin and sbin directories.
#line 7213
allow sshd_login_t bin_t:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t bin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7213
allow sshd_login_t sbin_t:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t sbin_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7213

#line 7213

#line 7213
# Read /var.
#line 7213
allow sshd_login_t var_t:dir search;
#line 7213

#line 7213
# Read /var/log.
#line 7213
allow sshd_login_t var_log_t:dir search;
#line 7213

#line 7213
# Read /etc.
#line 7213
allow sshd_login_t etc_t:dir { read getattr lock search ioctl };
#line 7213
allow sshd_login_t etc_t:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7213
allow sshd_login_t etc_runtime_t:{ file lnk_file } { read getattr lock ioctl };
#line 7213
allow sshd_login_t resolv_conf_t:{ file lnk_file } { read getattr lock ioctl };
#line 7213

#line 7213
# Read the linker, shared library, and executable types.
#line 7213
allow sshd_login_t ld_so_t:{ file lnk_file } { read getattr lock ioctl };
#line 7213
allow sshd_login_t shlib_t:{ file lnk_file } { read getattr lock ioctl };
#line 7213
allow sshd_login_t exec_type:{ file lnk_file } { read getattr lock ioctl };
#line 7213

#line 7213
# Read and write /dev/tty and /dev/null.
#line 7213
allow sshd_login_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 7213
allow sshd_login_t { null_device_t zero_device_t }:chr_file { ioctl read getattr lock write append };
#line 7213

#line 7213
# Read /dev/random and /dev/zero.
#line 7213
allow sshd_login_t random_device_t:chr_file { read getattr lock ioctl };
#line 7213

#line 7213

#line 7213
#
#line 7213
# Allow the domain to create and use UDP and TCP sockets.
#line 7213
# Other kinds of sockets must be separately authorized for use.
#line 7213
allow sshd_login_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7213
allow sshd_login_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7213

#line 7213
#
#line 7213
# Allow the domain to send UDP packets.
#line 7213
# Since the destination sockets type is unknown, the generic
#line 7213
# any_socket_t type is used as a placeholder.
#line 7213
#
#line 7213
allow sshd_login_t any_socket_t:udp_socket sendto;
#line 7213

#line 7213
#
#line 7213
# Allow the domain to send using any network interface.
#line 7213
# netif_type is a type attribute for all network interface types.
#line 7213
#
#line 7213
allow sshd_login_t netif_type:netif { tcp_send udp_send rawip_send };
#line 7213

#line 7213
#
#line 7213
# Allow packets sent by the domain to be received on any network interface.
#line 7213
#
#line 7213
allow sshd_login_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 7213

#line 7213
#
#line 7213
# Allow the domain to receive packets from any network interface.
#line 7213
# netmsg_type is a type attribute for all default message types.
#line 7213
#
#line 7213
allow sshd_login_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 7213

#line 7213
#
#line 7213
# Allow the domain to initiate or accept TCP connections 
#line 7213
# on any network interface.
#line 7213
#
#line 7213
allow sshd_login_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 7213

#line 7213
#
#line 7213
# Receive resets from the TCP reset socket.
#line 7213
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 7213
#
#line 7213
allow sshd_login_t tcp_socket_t:tcp_socket recvfrom;
#line 7213

#line 7213
dontaudit sshd_login_t tcp_socket_t:tcp_socket connectto;
#line 7213

#line 7213
#
#line 7213
# Allow the domain to send to any node.
#line 7213
# node_type is a type attribute for all node types.
#line 7213
#
#line 7213
allow sshd_login_t node_type:node { tcp_send udp_send rawip_send };
#line 7213

#line 7213
#
#line 7213
# Allow packets sent by the domain to be received from any node.
#line 7213
#
#line 7213
allow sshd_login_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 7213

#line 7213
#
#line 7213
# Allow the domain to send NFS client requests via the socket
#line 7213
# created by mount.
#line 7213
#
#line 7213
allow sshd_login_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7213

#line 7213
#
#line 7213
# Bind to the default port type.
#line 7213
# Other port types must be separately authorized.
#line 7213
#
#line 7213
allow sshd_login_t port_t:udp_socket name_bind;
#line 7213
allow sshd_login_t port_t:tcp_socket name_bind;
#line 7213

#line 7213

#line 7213
allow sshd_login_t self:capability { chown dac_override fowner fsetid setgid setuid net_bind_service sys_tty_config };
#line 7213
allow sshd_login_t shadow_t:file { read getattr };
#line 7213
allow sshd_login_t { home_root_t user_home_dir_type sysadm_home_dir_t }:dir { search getattr };
#line 7213

#line 7213
# Run shells in user_t by default
#line 7213

#line 7213

#line 7213

#line 7213
#
#line 7213
# Allow the process to transition to the new domain.
#line 7213
#
#line 7213
allow sshd_login_t user_t:process transition;
#line 7213

#line 7213
#
#line 7213
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7213
#
#line 7213
dontaudit sshd_login_t user_t:process noatsecure;
#line 7213

#line 7213
#
#line 7213
# Allow the process to execute the program.
#line 7213
# 
#line 7213
allow sshd_login_t shell_exec_t:file { read { getattr execute } };
#line 7213

#line 7213
#
#line 7213
# Allow the process to reap the new domain.
#line 7213
#
#line 7213
allow user_t sshd_login_t:process sigchld;
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to inherit and use file 
#line 7213
# descriptions from the creating process and vice versa.
#line 7213
#
#line 7213
allow user_t sshd_login_t:fd use;
#line 7213
allow sshd_login_t user_t:fd use;
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to write back to the old domain via a pipe.
#line 7213
#
#line 7213
allow user_t sshd_login_t:fifo_file { ioctl read getattr lock write append };
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to read and execute the program.
#line 7213
#
#line 7213
allow user_t shell_exec_t:file { read getattr lock execute ioctl };
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to be entered via the program.
#line 7213
#
#line 7213
allow user_t shell_exec_t:file entrypoint;
#line 7213

#line 7213
type_transition sshd_login_t shell_exec_t:process user_t;
#line 7213

#line 7213

#line 7213

#line 7213
#
#line 7213
# Allow the process to transition to the new domain.
#line 7213
#
#line 7213
allow sshd_login_t unpriv_userdomain:process transition;
#line 7213

#line 7213
#
#line 7213
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7213
#
#line 7213
dontaudit sshd_login_t unpriv_userdomain:process noatsecure;
#line 7213

#line 7213
#
#line 7213
# Allow the process to execute the program.
#line 7213
# 
#line 7213
allow sshd_login_t shell_exec_t:file { read { getattr execute } };
#line 7213

#line 7213
#
#line 7213
# Allow the process to reap the new domain.
#line 7213
#
#line 7213
allow unpriv_userdomain sshd_login_t:process sigchld;
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to inherit and use file 
#line 7213
# descriptions from the creating process and vice versa.
#line 7213
#
#line 7213
allow unpriv_userdomain sshd_login_t:fd use;
#line 7213
allow sshd_login_t unpriv_userdomain:fd use;
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to write back to the old domain via a pipe.
#line 7213
#
#line 7213
allow unpriv_userdomain sshd_login_t:fifo_file { ioctl read getattr lock write append };
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to read and execute the program.
#line 7213
#
#line 7213
allow unpriv_userdomain shell_exec_t:file { read getattr lock execute ioctl };
#line 7213

#line 7213
#
#line 7213
# Allow the new domain to be entered via the program.
#line 7213
#
#line 7213
allow unpriv_userdomain shell_exec_t:file entrypoint;
#line 7213

#line 7213

#line 7213
# Set exec context.
#line 7213

#line 7213
allow sshd_login_t self:process setexec;
#line 7213
allow sshd_login_t proc_t:dir search;
#line 7213
allow sshd_login_t proc_t:lnk_file read;
#line 7213
allow sshd_login_t self:dir search;
#line 7213
allow sshd_login_t self:file { read write };
#line 7213

#line 7213

#line 7213
# Allow shells to be run in sysadm_t as well.
#line 7213
# Commented out.  Use newrole rather than directly entering sysadm_t.
#line 7213
#domain_trans(sshd_login_t, shell_exec_t, sysadm_t)
#line 7213

#line 7213
# Update utmp.
#line 7213
allow sshd_login_t initrc_var_run_t:file { ioctl read getattr lock write append };
#line 7213

#line 7213
# Update wtmp.
#line 7213
allow sshd_login_t wtmp_t:file { ioctl read getattr lock write append };
#line 7213

#line 7213
# Get security policy decisions.
#line 7213

#line 7213
allow sshd_login_t security_t:dir { read search getattr };
#line 7213
allow sshd_login_t security_t:file { getattr read write };
#line 7213
allow sshd_login_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7213

#line 7213

#line 7213
# Allow read access to login context
#line 7213
allow sshd_login_t default_context_t:file { read getattr lock ioctl };
#line 7213

#line 7213


type sshd_key_t, file_type, sysadmfile;

# so a tunnel can point to another ssh tunnel...

#line 7218
allow sshd_t sshd_t:tcp_socket { connectto recvfrom };
#line 7218
allow sshd_t sshd_t:tcp_socket { acceptfrom recvfrom };
#line 7218
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 7218
allow sshd_t tcp_socket_t:tcp_socket { recvfrom };
#line 7218


type sshd_tmp_t, file_type, sysadmfile, tmpfile;

#line 7221

#line 7221

#line 7221

#line 7221
#
#line 7221
# Allow the process to modify the directory.
#line 7221
#
#line 7221
allow sshd_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7221

#line 7221
#
#line 7221
# Allow the process to create the file.
#line 7221
#
#line 7221

#line 7221
allow sshd_t sshd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 7221
allow sshd_t sshd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 7221

#line 7221

#line 7221
type_transition sshd_t tmp_t:dir sshd_tmp_t;
#line 7221
type_transition sshd_t tmp_t:{ file lnk_file sock_file fifo_file } sshd_tmp_t;
#line 7221

#line 7221

#line 7221


# Inherit and use descriptors from init.
allow sshd_t init_t:fd use;

# Can create ptys

#line 7227

#line 7227
# Access the pty master multiplexer.
#line 7227
allow sshd_t ptmx_t:chr_file { ioctl read getattr lock write append };
#line 7227

#line 7227

#line 7227
allow sshd_t devpts_t:filesystem getattr;
#line 7227

#line 7227
# allow searching /dev/pts
#line 7227
allow sshd_t devpts_t:dir { getattr read search };
#line 7227

#line 7227
type sshd_devpts_t, file_type, sysadmfile, ptyfile ;
#line 7227

#line 7227
# Allow the pty to be associated with the file system.
#line 7227
allow sshd_devpts_t devpts_t:filesystem associate;
#line 7227

#line 7227
# Label pty files with a derived type.
#line 7227
type_transition sshd_t devpts_t:chr_file sshd_devpts_t;
#line 7227

#line 7227
# Read and write my pty files.
#line 7227
allow sshd_t sshd_devpts_t:chr_file { setattr { ioctl read getattr lock write append } };
#line 7227


# Execute Login

#line 7230

#line 7230

#line 7230
#
#line 7230
# Allow the process to transition to the new domain.
#line 7230
#
#line 7230
allow sshd_t sshd_login_t:process transition;
#line 7230

#line 7230
#
#line 7230
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7230
#
#line 7230
dontaudit sshd_t sshd_login_t:process noatsecure;
#line 7230

#line 7230
#
#line 7230
# Allow the process to execute the program.
#line 7230
# 
#line 7230
allow sshd_t login_exec_t:file { read { getattr execute } };
#line 7230

#line 7230
#
#line 7230
# Allow the process to reap the new domain.
#line 7230
#
#line 7230
allow sshd_login_t sshd_t:process sigchld;
#line 7230

#line 7230
#
#line 7230
# Allow the new domain to inherit and use file 
#line 7230
# descriptions from the creating process and vice versa.
#line 7230
#
#line 7230
allow sshd_login_t sshd_t:fd use;
#line 7230
allow sshd_t sshd_login_t:fd use;
#line 7230

#line 7230
#
#line 7230
# Allow the new domain to write back to the old domain via a pipe.
#line 7230
#
#line 7230
allow sshd_login_t sshd_t:fifo_file { ioctl read getattr lock write append };
#line 7230

#line 7230
#
#line 7230
# Allow the new domain to read and execute the program.
#line 7230
#
#line 7230
allow sshd_login_t login_exec_t:file { read getattr lock execute ioctl };
#line 7230

#line 7230
#
#line 7230
# Allow the new domain to be entered via the program.
#line 7230
#
#line 7230
allow sshd_login_t login_exec_t:file entrypoint;
#line 7230

#line 7230
type_transition sshd_t login_exec_t:process sshd_login_t;
#line 7230


# Use capabilities.
allow sshd_t self:capability { sys_chroot sys_resource };

# Create /var/run/sshd.pid

#line 7236
type sshd_var_run_t alias var_run_sshd_t, file_type, sysadmfile, pidfile;
#line 7236

#line 7236

#line 7236

#line 7236

#line 7236
#
#line 7236
# Allow the process to modify the directory.
#line 7236
#
#line 7236
allow sshd_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7236

#line 7236
#
#line 7236
# Allow the process to create the file.
#line 7236
#
#line 7236

#line 7236

#line 7236
allow sshd_t sshd_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7236

#line 7236

#line 7236
type_transition sshd_t var_run_t:file sshd_var_run_t;
#line 7236

#line 7236

#line 7236

#line 7236
allow sshd_t sshd_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7236


# Access key files
allow sshd_t sshd_key_t:file { getattr read };

# Update /var/log/lastlog.
allow sshd_t lastlog_t:file { ioctl read getattr lock write append };



# Signal the user domains.
allow sshd_t unpriv_userdomain:process signal;

# Relabel and access ptys created by sshd
allow sshd_t sshd_devpts_t:chr_file { setattr getattr relabelfrom relabelto };
allow sshd_t userpty_type:chr_file { setattr relabelto { ioctl read getattr lock write append } };

# for when the network connection breaks after running newrole -r sysadm_r
dontaudit sshd_t sysadm_devpts_t:chr_file setattr;

#################################
#
# Rules for the sshd_login_t domain
#
# sshd_login_t is the domain of a login process
# spawned by sshd

# Use the pty created by sshd
allow sshd_login_t sshd_devpts_t:chr_file { setattr { ioctl read getattr lock write append } };

# Write to /var/log/lastlog
allow sshd_login_t lastlog_t:file { ioctl read getattr lock write append };

# Relabel ptys created by sshd
allow sshd_login_t sshd_devpts_t:chr_file { relabelfrom relabelto };
allow sshd_login_t userpty_type:chr_file { getattr relabelfrom relabelto };


#line 7273
allow sshd_t etc_t:lnk_file read;
#line 7273

#line 7273
allow sshd_t locale_t:dir { read getattr lock search ioctl };
#line 7273
allow sshd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 7273

#line 7273


# Allow checking users mail at login
allow sshd_t { var_spool_t mail_spool_t }:dir search;
allow sshd_t mail_spool_t:lnk_file read;
allow sshd_t mail_spool_t:file getattr;
#
# Author:  Stephen Smalley <sds@epoch.ncsc.mil>
#

# Type for the ssh executable.
type ssh_exec_t, file_type, exec_type, sysadmfile;

allow sysadm_ssh_t user_home_dir_type:dir search;

#line 7287
allow sysadm_ssh_t user_home_ssh_t:dir { read getattr lock search ioctl };
#line 7287
allow sysadm_ssh_t user_home_ssh_t:{ file lnk_file } { read getattr lock ioctl };
#line 7287


# Everything else is in the ssh_domain macro in
# macros/program/ssh_macros.te.

#DESC Su - Run shells with substitute user and group
#
# Domains for the su program.
# X-Debian-Packages: login

#
# su_exec_t is the type of the su executable.
#
type su_exec_t, file_type, sysadmfile, exec_type;

allow sysadm_su_t user_home_dir_type:dir search;

# Everything else is in the su_domain macro in
# macros/program/su_macros.te.
#DESC Syslogd - System log daemon
#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
# X-Debian-Packages: sysklogd
#

#################################
#
# Rules for the syslogd_t domain.
#
# syslogd_t is the domain of syslogd.
# syslogd_exec_t is the type of the syslogd executable.
# devlog_t is the type of the Unix domain socket created 
# by syslogd.
#

#line 7321

#line 7321
type syslogd_t, domain, privlog ;
#line 7321
type syslogd_exec_t, file_type, sysadmfile, exec_type;
#line 7321

#line 7321
role system_r types syslogd_t;
#line 7321

#line 7321

#line 7321

#line 7321

#line 7321
#
#line 7321
# Allow the process to transition to the new domain.
#line 7321
#
#line 7321
allow initrc_t syslogd_t:process transition;
#line 7321

#line 7321
#
#line 7321
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7321
#
#line 7321
dontaudit initrc_t syslogd_t:process noatsecure;
#line 7321

#line 7321
#
#line 7321
# Allow the process to execute the program.
#line 7321
# 
#line 7321
allow initrc_t syslogd_exec_t:file { read { getattr execute } };
#line 7321

#line 7321
#
#line 7321
# Allow the process to reap the new domain.
#line 7321
#
#line 7321
allow syslogd_t initrc_t:process sigchld;
#line 7321

#line 7321
#
#line 7321
# Allow the new domain to inherit and use file 
#line 7321
# descriptions from the creating process and vice versa.
#line 7321
#
#line 7321
allow syslogd_t initrc_t:fd use;
#line 7321
allow initrc_t syslogd_t:fd use;
#line 7321

#line 7321
#
#line 7321
# Allow the new domain to write back to the old domain via a pipe.
#line 7321
#
#line 7321
allow syslogd_t initrc_t:fifo_file { ioctl read getattr lock write append };
#line 7321

#line 7321
#
#line 7321
# Allow the new domain to read and execute the program.
#line 7321
#
#line 7321
allow syslogd_t syslogd_exec_t:file { read getattr lock execute ioctl };
#line 7321

#line 7321
#
#line 7321
# Allow the new domain to be entered via the program.
#line 7321
#
#line 7321
allow syslogd_t syslogd_exec_t:file entrypoint;
#line 7321

#line 7321
type_transition initrc_t syslogd_exec_t:process syslogd_t;
#line 7321

#line 7321

#line 7321
# Inherit and use descriptors from init.
#line 7321
allow syslogd_t init_t:fd use;
#line 7321
allow syslogd_t init_t:process sigchld;
#line 7321
allow syslogd_t privfd:fd use;
#line 7321
allow syslogd_t newrole_t:process sigchld;
#line 7321
allow syslogd_t self:process { { sigchld sigkill sigstop signull signal } fork };
#line 7321

#line 7321

#line 7321
allow syslogd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7321
allow syslogd_t lib_t:lnk_file { read getattr lock ioctl };
#line 7321
allow syslogd_t ld_so_t:file { read getattr lock execute ioctl };
#line 7321
allow syslogd_t ld_so_t:file execute_no_trans;
#line 7321
allow syslogd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7321
allow syslogd_t shlib_t:file { read getattr lock execute ioctl };
#line 7321
allow syslogd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7321
allow syslogd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7321
allow syslogd_t device_t:dir search;
#line 7321
allow syslogd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7321

#line 7321

#line 7321
allow syslogd_t { self proc_t }:dir { read getattr lock search ioctl };
#line 7321
allow syslogd_t { self proc_t }:lnk_file read;
#line 7321

#line 7321
allow syslogd_t device_t:dir { getattr search };
#line 7321
allow syslogd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7321
allow syslogd_t console_device_t:chr_file { ioctl read getattr lock write append };
#line 7321
allow syslogd_t initrc_devpts_t:chr_file { ioctl read getattr lock write append };
#line 7321

#line 7321
# Create pid file.
#line 7321
allow syslogd_t var_t:dir { getattr search };
#line 7321

#line 7321
type syslogd_var_run_t alias var_run_syslogd_t, file_type, sysadmfile, pidfile;
#line 7321

#line 7321

#line 7321

#line 7321

#line 7321
#
#line 7321
# Allow the process to modify the directory.
#line 7321
#
#line 7321
allow syslogd_t var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7321

#line 7321
#
#line 7321
# Allow the process to create the file.
#line 7321
#
#line 7321

#line 7321

#line 7321
allow syslogd_t syslogd_var_run_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7321

#line 7321

#line 7321
type_transition syslogd_t var_run_t:file syslogd_var_run_t;
#line 7321

#line 7321

#line 7321

#line 7321
allow syslogd_t syslogd_var_run_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7321

#line 7321

#line 7321
allow syslogd_t devtty_t:chr_file { ioctl read getattr lock write append };
#line 7321

#line 7321
# for daemons that look at /root on startup
#line 7321
dontaudit syslogd_t sysadm_home_dir_t:dir search;
#line 7321

#line 7321
# for df
#line 7321
allow syslogd_t fs_type:filesystem getattr;
#line 7321

#line 7321

#line 7321
allow syslogd_t etc_t:lnk_file read;
#line 7321

#line 7321
allow syslogd_t locale_t:dir { read getattr lock search ioctl };
#line 7321
allow syslogd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 7321

#line 7321

#line 7321


# can_network is for the UDP socket

#line 7324
#
#line 7324
# Allow the domain to create and use UDP and TCP sockets.
#line 7324
# Other kinds of sockets must be separately authorized for use.
#line 7324
allow syslogd_t self:udp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7324
allow syslogd_t self:tcp_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7324

#line 7324
#
#line 7324
# Allow the domain to send UDP packets.
#line 7324
# Since the destination sockets type is unknown, the generic
#line 7324
# any_socket_t type is used as a placeholder.
#line 7324
#
#line 7324
allow syslogd_t any_socket_t:udp_socket sendto;
#line 7324

#line 7324
#
#line 7324
# Allow the domain to send using any network interface.
#line 7324
# netif_type is a type attribute for all network interface types.
#line 7324
#
#line 7324
allow syslogd_t netif_type:netif { tcp_send udp_send rawip_send };
#line 7324

#line 7324
#
#line 7324
# Allow packets sent by the domain to be received on any network interface.
#line 7324
#
#line 7324
allow syslogd_t netif_type:netif { tcp_recv udp_recv rawip_recv };
#line 7324

#line 7324
#
#line 7324
# Allow the domain to receive packets from any network interface.
#line 7324
# netmsg_type is a type attribute for all default message types.
#line 7324
#
#line 7324
allow syslogd_t netmsg_type:{ udp_socket tcp_socket rawip_socket } recvfrom;
#line 7324

#line 7324
#
#line 7324
# Allow the domain to initiate or accept TCP connections 
#line 7324
# on any network interface.
#line 7324
#
#line 7324
allow syslogd_t netmsg_type:tcp_socket { connectto acceptfrom };
#line 7324

#line 7324
#
#line 7324
# Receive resets from the TCP reset socket.
#line 7324
# The TCP reset socket is labeled with the tcp_socket_t type.
#line 7324
#
#line 7324
allow syslogd_t tcp_socket_t:tcp_socket recvfrom;
#line 7324

#line 7324
dontaudit syslogd_t tcp_socket_t:tcp_socket connectto;
#line 7324

#line 7324
#
#line 7324
# Allow the domain to send to any node.
#line 7324
# node_type is a type attribute for all node types.
#line 7324
#
#line 7324
allow syslogd_t node_type:node { tcp_send udp_send rawip_send };
#line 7324

#line 7324
#
#line 7324
# Allow packets sent by the domain to be received from any node.
#line 7324
#
#line 7324
allow syslogd_t node_type:node { tcp_recv udp_recv rawip_recv };
#line 7324

#line 7324
#
#line 7324
# Allow the domain to send NFS client requests via the socket
#line 7324
# created by mount.
#line 7324
#
#line 7324
allow syslogd_t mount_t:udp_socket { ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7324

#line 7324
#
#line 7324
# Bind to the default port type.
#line 7324
# Other port types must be separately authorized.
#line 7324
#
#line 7324
allow syslogd_t port_t:udp_socket name_bind;
#line 7324
allow syslogd_t port_t:tcp_socket name_bind;
#line 7324


type devlog_t, file_type, sysadmfile;

# if something can log to syslog they should be able to log to the console
allow privlog console_device_t:chr_file { ioctl read write getattr };


#line 7331
type syslogd_tmp_t, file_type, sysadmfile, tmpfile ;
#line 7331

#line 7331

#line 7331

#line 7331

#line 7331
#
#line 7331
# Allow the process to modify the directory.
#line 7331
#
#line 7331
allow syslogd_t tmp_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7331

#line 7331
#
#line 7331
# Allow the process to create the file.
#line 7331
#
#line 7331

#line 7331
allow syslogd_t syslogd_tmp_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 7331
allow syslogd_t syslogd_tmp_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 7331

#line 7331

#line 7331
type_transition syslogd_t tmp_t:dir syslogd_tmp_t;
#line 7331
type_transition syslogd_t tmp_t:{ file lnk_file sock_file fifo_file } syslogd_tmp_t;
#line 7331

#line 7331

#line 7331

#line 7331


# read files in /etc
allow syslogd_t etc_t:file { read getattr lock ioctl };
allow syslogd_t resolv_conf_t:{ file lnk_file } { read getattr lock ioctl };

# Use capabilities.
allow syslogd_t syslogd_t:capability { net_bind_service dac_override };

# Inherit and use descriptors from init.
allow syslogd_t init_t:fd use;
allow syslogd_t { initrc_devpts_t console_device_t }:chr_file { read write };

# Modify/create log files.

#line 7345
allow syslogd_t var_log_t:dir { read getattr search add_name write };
#line 7345
allow syslogd_t var_log_t:file { create ioctl getattr setattr append link };
#line 7345


# Create and bind to /dev/log or /var/run/log.

#line 7348

#line 7348

#line 7348

#line 7348
#
#line 7348
# Allow the process to modify the directory.
#line 7348
#
#line 7348
allow syslogd_t { device_t var_run_t }:dir { read getattr lock search ioctl add_name remove_name write };
#line 7348

#line 7348
#
#line 7348
# Allow the process to create the file.
#line 7348
#
#line 7348

#line 7348

#line 7348
allow syslogd_t devlog_t:sock_file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7348

#line 7348

#line 7348
type_transition syslogd_t { device_t var_run_t }:sock_file devlog_t;
#line 7348

#line 7348

#line 7348

allow syslogd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow syslogd_t self:unix_dgram_socket { sendto };
allow syslogd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
allow syslogd_t self:unix_stream_socket { listen accept };
allow syslogd_t devlog_t:unix_stream_socket name_bind;
allow syslogd_t devlog_t:unix_dgram_socket name_bind;

# Domains with the privlog attribute may log to syslogd.
allow privlog devlog_t:sock_file { ioctl read getattr lock write append };

#line 7358
allow privlog syslogd_t:unix_dgram_socket sendto;
#line 7358


#line 7359
allow privlog syslogd_t:unix_stream_socket connectto;
#line 7359

# allow /dev/log to be a link elsewhere for chroot setup
allow privlog devlog_t:lnk_file read;

#line 7366

#line 7366
# Write to the cron log.
#line 7366
allow syslogd_t cron_log_t:file { ioctl read getattr lock write append };
#line 7366


#line 7370

#line 7370
allow logrotate_t syslogd_exec_t:file { read getattr lock ioctl };
#line 7370


# uncomment this to allow syslogd to log to virtual consoles
#allow syslogd_t tty_device_t:chr_file rw_file_perms;

# Allow access to /proc/kmsg for syslog-ng
#allow syslogd_t proc_t:dir search;
#allow syslogd_t proc_kmsg_t:file { getattr read };

#DESC Tmpreaper - Monitor and maintain temporary files
#
# Author:  Russell Coker <russell@coker.com.au>
# X-Debian-Packages: tmpreaper
#

#################################
#
# Rules for the tmpreaper_t domain.
#
type tmpreaper_t, domain, privlog;
type tmpreaper_exec_t, file_type, sysadmfile, exec_type;

role system_r types tmpreaper_t;


#line 7394

#line 7394

#line 7394

#line 7394
#
#line 7394
# Allow the process to transition to the new domain.
#line 7394
#
#line 7394
allow system_crond_t tmpreaper_t:process transition;
#line 7394

#line 7394
#
#line 7394
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7394
#
#line 7394
dontaudit system_crond_t tmpreaper_t:process noatsecure;
#line 7394

#line 7394
#
#line 7394
# Allow the process to execute the program.
#line 7394
# 
#line 7394
allow system_crond_t tmpreaper_exec_t:file { read { getattr execute } };
#line 7394

#line 7394
#
#line 7394
# Allow the process to reap the new domain.
#line 7394
#
#line 7394
allow tmpreaper_t system_crond_t:process sigchld;
#line 7394

#line 7394
#
#line 7394
# Allow the new domain to inherit and use file 
#line 7394
# descriptions from the creating process and vice versa.
#line 7394
#
#line 7394
allow tmpreaper_t system_crond_t:fd use;
#line 7394
allow system_crond_t tmpreaper_t:fd use;
#line 7394

#line 7394
#
#line 7394
# Allow the new domain to write back to the old domain via a pipe.
#line 7394
#
#line 7394
allow tmpreaper_t system_crond_t:fifo_file { ioctl read getattr lock write append };
#line 7394

#line 7394
#
#line 7394
# Allow the new domain to read and execute the program.
#line 7394
#
#line 7394
allow tmpreaper_t tmpreaper_exec_t:file { read getattr lock execute ioctl };
#line 7394

#line 7394
#
#line 7394
# Allow the new domain to be entered via the program.
#line 7394
#
#line 7394
allow tmpreaper_t tmpreaper_exec_t:file entrypoint;
#line 7394

#line 7394
type_transition system_crond_t tmpreaper_exec_t:process tmpreaper_t;
#line 7394

#line 7394
allow tmpreaper_t crond_t:fifo_file { getattr read write ioctl };
#line 7394
# a rule for privfd may make this obsolete
#line 7394
allow tmpreaper_t crond_t:fd use;
#line 7394
allow tmpreaper_t crond_t:process sigchld;
#line 7394


#line 7395
allow tmpreaper_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7395
allow tmpreaper_t lib_t:lnk_file { read getattr lock ioctl };
#line 7395
allow tmpreaper_t ld_so_t:file { read getattr lock execute ioctl };
#line 7395
allow tmpreaper_t ld_so_t:file execute_no_trans;
#line 7395
allow tmpreaper_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7395
allow tmpreaper_t shlib_t:file { read getattr lock execute ioctl };
#line 7395
allow tmpreaper_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7395
allow tmpreaper_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7395
allow tmpreaper_t device_t:dir search;
#line 7395
allow tmpreaper_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7395

allow tmpreaper_t tmpfile:dir { { read getattr lock search ioctl add_name remove_name write } rmdir };
allow tmpreaper_t tmpfile:{ file lnk_file sock_file fifo_file } { getattr unlink };
allow tmpreaper_t home_type:{ file lnk_file sock_file fifo_file } { getattr unlink };
allow tmpreaper_t self:process { fork sigchld };
allow tmpreaper_t self:capability { dac_override dac_read_search fowner };
allow tmpreaper_t fs_t:filesystem getattr;


#line 7403
allow tmpreaper_t etc_t:dir { read getattr lock search ioctl };
#line 7403
allow tmpreaper_t etc_t:{ file lnk_file } { read getattr lock ioctl };
#line 7403

allow tmpreaper_t var_t:dir { getattr search };

#line 7405
allow tmpreaper_t var_lib_t:dir { read getattr lock search ioctl };
#line 7405
allow tmpreaper_t var_lib_t:{ file lnk_file } { read getattr lock ioctl };
#line 7405

allow tmpreaper_t device_t:dir { getattr search };
allow tmpreaper_t random_device_t:chr_file { getattr read };

#line 7412

#DESC Useradd - Manage system user accounts
#
# Authors:  Chris Vance <cvance@tislabs.com>  David Caplan <dac@tresys.com>
#           Russell Coker <russell@coker.com.au>
# X-Debian-Packages: passwd
#

#################################
#
# Rules for the useradd_t and groupadd_t domains.
#
# useradd_t is the domain of the useradd/userdel programs.
# groupadd_t is for adding groups (can not create home dirs)
#
#line 7471


#line 7472
type useradd_t, domain, privlog, auth, privowner;
#line 7472
role sysadm_r types useradd_t;
#line 7472

#line 7472

#line 7472
# Access other processes in the same domain.
#line 7472
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 7472
# separately if desired.
#line 7472
allow useradd_t self:process ~{ptrace setexec setfscreate};
#line 7472

#line 7472
# Access /proc/PID files for processes in the same domain.
#line 7472
allow useradd_t self:dir { read getattr lock search ioctl };
#line 7472
allow useradd_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7472

#line 7472
# Access file descriptions, pipes, and sockets
#line 7472
# created by processes in the same domain.
#line 7472
allow useradd_t self:fd *;
#line 7472
allow useradd_t self:fifo_file { ioctl read getattr lock write append };
#line 7472
allow useradd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7472
allow useradd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7472

#line 7472
# Allow the domain to communicate with other processes in the same domain.
#line 7472
allow useradd_t self:unix_dgram_socket sendto;
#line 7472
allow useradd_t self:unix_stream_socket connectto;
#line 7472

#line 7472
# Access System V IPC objects created by processes in the same domain.
#line 7472
allow useradd_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 7472
allow useradd_t self:msg  { send receive };
#line 7472
allow useradd_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 7472
allow useradd_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 7472

#line 7472

#line 7472
allow useradd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7472
allow useradd_t lib_t:lnk_file { read getattr lock ioctl };
#line 7472
allow useradd_t ld_so_t:file { read getattr lock execute ioctl };
#line 7472
allow useradd_t ld_so_t:file execute_no_trans;
#line 7472
allow useradd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7472
allow useradd_t shlib_t:file { read getattr lock execute ioctl };
#line 7472
allow useradd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7472
allow useradd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7472
allow useradd_t device_t:dir search;
#line 7472
allow useradd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7472

#line 7472

#line 7472
type useradd_exec_t, file_type, sysadmfile, exec_type;
#line 7472

#line 7472

#line 7472

#line 7472
#
#line 7472
# Allow the process to transition to the new domain.
#line 7472
#
#line 7472
allow sysadm_t useradd_t:process transition;
#line 7472

#line 7472
#
#line 7472
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7472
#
#line 7472
dontaudit sysadm_t useradd_t:process noatsecure;
#line 7472

#line 7472
#
#line 7472
# Allow the process to execute the program.
#line 7472
# 
#line 7472
allow sysadm_t useradd_exec_t:file { read { getattr execute } };
#line 7472

#line 7472
#
#line 7472
# Allow the process to reap the new domain.
#line 7472
#
#line 7472
allow useradd_t sysadm_t:process sigchld;
#line 7472

#line 7472
#
#line 7472
# Allow the new domain to inherit and use file 
#line 7472
# descriptions from the creating process and vice versa.
#line 7472
#
#line 7472
allow useradd_t sysadm_t:fd use;
#line 7472
allow sysadm_t useradd_t:fd use;
#line 7472

#line 7472
#
#line 7472
# Allow the new domain to write back to the old domain via a pipe.
#line 7472
#
#line 7472
allow useradd_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 7472

#line 7472
#
#line 7472
# Allow the new domain to read and execute the program.
#line 7472
#
#line 7472
allow useradd_t useradd_exec_t:file { read getattr lock execute ioctl };
#line 7472

#line 7472
#
#line 7472
# Allow the new domain to be entered via the program.
#line 7472
#
#line 7472
allow useradd_t useradd_exec_t:file entrypoint;
#line 7472

#line 7472
type_transition sysadm_t useradd_exec_t:process useradd_t;
#line 7472

#line 7472

#line 7472
# Use capabilities.
#line 7472
allow useradd_t self:capability { dac_override chown };
#line 7472

#line 7472
# Allow access to context for shadow file
#line 7472

#line 7472
allow useradd_t security_t:dir { read search getattr };
#line 7472
allow useradd_t security_t:file { getattr read write };
#line 7472
allow useradd_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7472

#line 7472

#line 7472
# Inherit and use descriptors from login.
#line 7472
allow useradd_t privfd:fd use;
#line 7472

#line 7472
# Update /etc/shadow and /etc/passwd
#line 7472

#line 7472

#line 7472

#line 7472

#line 7472
#
#line 7472
# Allow the process to modify the directory.
#line 7472
#
#line 7472
allow useradd_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7472

#line 7472
#
#line 7472
# Allow the process to create the file.
#line 7472
#
#line 7472

#line 7472

#line 7472
allow useradd_t shadow_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7472

#line 7472

#line 7472
type_transition useradd_t etc_t:file shadow_t;
#line 7472

#line 7472

#line 7472

#line 7472
allow useradd_t etc_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7472

#line 7472
# Set fscreate context.
#line 7472

#line 7472
allow useradd_t self:process setfscreate;
#line 7472
allow useradd_t proc_t:dir search;
#line 7472
allow useradd_t proc_t:lnk_file read;
#line 7472
allow useradd_t self:dir search;
#line 7472
allow useradd_t self:file { read write };
#line 7472

#line 7472

#line 7472
allow useradd_t { etc_t shadow_t }:file { relabelfrom relabelto };
#line 7472

#line 7472

#line 7472
allow useradd_t etc_t:lnk_file read;
#line 7472

#line 7472
allow useradd_t locale_t:dir { read getattr lock search ioctl };
#line 7472
allow useradd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 7472

#line 7472

#line 7472

#line 7472
# useradd/userdel request read/write for /var/log/lastlog, and read of /dev, 
#line 7472
# but will operate without them.
#line 7472
dontaudit useradd_t { device_t var_t var_log_t }:dir search;
#line 7472

#line 7472
# For userdel and groupadd
#line 7472
allow useradd_t fs_t:filesystem getattr;
#line 7472

#line 7472
# Access terminals.
#line 7472
allow useradd_t ttyfile:chr_file { ioctl read getattr lock write append };
#line 7472
allow useradd_t ptyfile:chr_file { ioctl read getattr lock write append };
#line 7472

#line 7472

#line 7472
# for when /root is the cwd
#line 7472
dontaudit useradd_t sysadm_home_dir_t:dir search;
#line 7472


# Add/remove user home directories

#line 7475

#line 7475

#line 7475

#line 7475
#
#line 7475
# Allow the process to modify the directory.
#line 7475
#
#line 7475
allow useradd_t home_root_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7475

#line 7475
#
#line 7475
# Allow the process to create the file.
#line 7475
#
#line 7475

#line 7475

#line 7475
allow useradd_t user_home_dir_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 7475

#line 7475

#line 7475
type_transition useradd_t home_root_t:dir user_home_dir_t;
#line 7475

#line 7475

#line 7475


#line 7476

#line 7476

#line 7476

#line 7476
#
#line 7476
# Allow the process to modify the directory.
#line 7476
#
#line 7476
allow useradd_t user_home_dir_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7476

#line 7476
#
#line 7476
# Allow the process to create the file.
#line 7476
#
#line 7476

#line 7476
allow useradd_t user_home_t:{ file lnk_file sock_file fifo_file } { create ioctl read getattr lock write setattr append link unlink rename };
#line 7476
allow useradd_t user_home_t:dir { create read getattr lock setattr link unlink rename search add_name remove_name reparent write rmdir };
#line 7476

#line 7476

#line 7476
type_transition useradd_t user_home_dir_t:dir user_home_t;
#line 7476
type_transition useradd_t user_home_dir_t:{ file lnk_file sock_file fifo_file } user_home_t;
#line 7476

#line 7476

#line 7476


# create/delete mail spool file in /var/mail
allow useradd_t var_spool_t:dir search;
allow useradd_t mail_spool_t:dir { search write add_name remove_name };
allow useradd_t mail_spool_t:file { create setattr getattr unlink };
# /var/mail is a link to /var/spool/mail
allow useradd_t mail_spool_t:lnk_file read;

allow useradd_t self:capability { fowner fsetid setuid sys_resource };

#line 7486
allow useradd_t shell_exec_t:file { { read getattr lock execute ioctl } execute_no_trans };
#line 7486


# /usr/bin/userdel locks the user being deleted, allow write access to utmp
allow useradd_t initrc_var_run_t:file { read write lock };


#line 7491
type groupadd_t, domain, privlog, auth, privowner;
#line 7491
role sysadm_r types groupadd_t;
#line 7491

#line 7491

#line 7491
# Access other processes in the same domain.
#line 7491
# Omits ptrace, setexec, and setfscreate.  These must be granted 
#line 7491
# separately if desired.
#line 7491
allow groupadd_t self:process ~{ptrace setexec setfscreate};
#line 7491

#line 7491
# Access /proc/PID files for processes in the same domain.
#line 7491
allow groupadd_t self:dir { read getattr lock search ioctl };
#line 7491
allow groupadd_t self:{ file lnk_file sock_file fifo_file } { read getattr lock ioctl };
#line 7491

#line 7491
# Access file descriptions, pipes, and sockets
#line 7491
# created by processes in the same domain.
#line 7491
allow groupadd_t self:fd *;
#line 7491
allow groupadd_t self:fifo_file { ioctl read getattr lock write append };
#line 7491
allow groupadd_t self:unix_dgram_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown };
#line 7491
allow groupadd_t self:unix_stream_socket { create ioctl read getattr write setattr append bind connect getopt setopt shutdown listen accept };
#line 7491

#line 7491
# Allow the domain to communicate with other processes in the same domain.
#line 7491
allow groupadd_t self:unix_dgram_socket sendto;
#line 7491
allow groupadd_t self:unix_stream_socket connectto;
#line 7491

#line 7491
# Access System V IPC objects created by processes in the same domain.
#line 7491
allow groupadd_t self:sem  { associate getattr setattr create destroy read write unix_read unix_write };
#line 7491
allow groupadd_t self:msg  { send receive };
#line 7491
allow groupadd_t self:msgq { associate getattr setattr create destroy read write enqueue unix_read unix_write };
#line 7491
allow groupadd_t self:shm  { associate getattr setattr create destroy read write lock unix_read unix_write };
#line 7491

#line 7491

#line 7491
allow groupadd_t { root_t usr_t lib_t etc_t }:dir { read getattr lock search ioctl };
#line 7491
allow groupadd_t lib_t:lnk_file { read getattr lock ioctl };
#line 7491
allow groupadd_t ld_so_t:file { read getattr lock execute ioctl };
#line 7491
allow groupadd_t ld_so_t:file execute_no_trans;
#line 7491
allow groupadd_t ld_so_t:lnk_file { read getattr lock ioctl };
#line 7491
allow groupadd_t shlib_t:file { read getattr lock execute ioctl };
#line 7491
allow groupadd_t shlib_t:lnk_file { read getattr lock ioctl };
#line 7491
allow groupadd_t ld_so_cache_t:file { read getattr lock ioctl };
#line 7491
allow groupadd_t device_t:dir search;
#line 7491
allow groupadd_t null_device_t:chr_file { ioctl read getattr lock write append };
#line 7491

#line 7491

#line 7491
type groupadd_exec_t, file_type, sysadmfile, exec_type;
#line 7491

#line 7491

#line 7491

#line 7491
#
#line 7491
# Allow the process to transition to the new domain.
#line 7491
#
#line 7491
allow sysadm_t groupadd_t:process transition;
#line 7491

#line 7491
#
#line 7491
# Do not audit when glibc secure mode is enabled upon the transition.
#line 7491
#
#line 7491
dontaudit sysadm_t groupadd_t:process noatsecure;
#line 7491

#line 7491
#
#line 7491
# Allow the process to execute the program.
#line 7491
# 
#line 7491
allow sysadm_t groupadd_exec_t:file { read { getattr execute } };
#line 7491

#line 7491
#
#line 7491
# Allow the process to reap the new domain.
#line 7491
#
#line 7491
allow groupadd_t sysadm_t:process sigchld;
#line 7491

#line 7491
#
#line 7491
# Allow the new domain to inherit and use file 
#line 7491
# descriptions from the creating process and vice versa.
#line 7491
#
#line 7491
allow groupadd_t sysadm_t:fd use;
#line 7491
allow sysadm_t groupadd_t:fd use;
#line 7491

#line 7491
#
#line 7491
# Allow the new domain to write back to the old domain via a pipe.
#line 7491
#
#line 7491
allow groupadd_t sysadm_t:fifo_file { ioctl read getattr lock write append };
#line 7491

#line 7491
#
#line 7491
# Allow the new domain to read and execute the program.
#line 7491
#
#line 7491
allow groupadd_t groupadd_exec_t:file { read getattr lock execute ioctl };
#line 7491

#line 7491
#
#line 7491
# Allow the new domain to be entered via the program.
#line 7491
#
#line 7491
allow groupadd_t groupadd_exec_t:file entrypoint;
#line 7491

#line 7491
type_transition sysadm_t groupadd_exec_t:process groupadd_t;
#line 7491

#line 7491

#line 7491
# Use capabilities.
#line 7491
allow groupadd_t self:capability { dac_override chown };
#line 7491

#line 7491
# Allow access to context for shadow file
#line 7491

#line 7491
allow groupadd_t security_t:dir { read search getattr };
#line 7491
allow groupadd_t security_t:file { getattr read write };
#line 7491
allow groupadd_t security_t:security { check_context compute_av compute_create compute_relabel compute_user };
#line 7491

#line 7491

#line 7491
# Inherit and use descriptors from login.
#line 7491
allow groupadd_t privfd:fd use;
#line 7491

#line 7491
# Update /etc/shadow and /etc/passwd
#line 7491

#line 7491

#line 7491

#line 7491

#line 7491
#
#line 7491
# Allow the process to modify the directory.
#line 7491
#
#line 7491
allow groupadd_t etc_t:dir { read getattr lock search ioctl add_name remove_name write };
#line 7491

#line 7491
#
#line 7491
# Allow the process to create the file.
#line 7491
#
#line 7491

#line 7491

#line 7491
allow groupadd_t shadow_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7491

#line 7491

#line 7491
type_transition groupadd_t etc_t:file shadow_t;
#line 7491

#line 7491

#line 7491

#line 7491
allow groupadd_t etc_t:file { create ioctl read getattr lock write setattr append link unlink rename };
#line 7491

#line 7491
# Set fscreate context.
#line 7491

#line 7491
allow groupadd_t self:process setfscreate;
#line 7491
allow groupadd_t proc_t:dir search;
#line 7491
allow groupadd_t proc_t:lnk_file read;
#line 7491
allow groupadd_t self:dir search;
#line 7491
allow groupadd_t self:file { read write };
#line 7491

#line 7491

#line 7491
allow groupadd_t { etc_t shadow_t }:file { relabelfrom relabelto };
#line 7491

#line 7491

#line 7491
allow groupadd_t etc_t:lnk_file read;
#line 7491

#line 7491
allow groupadd_t locale_t:dir { read getattr lock search ioctl };
#line 7491
allow groupadd_t locale_t:{ file lnk_file } { read getattr lock ioctl };
#line 7491

#line 7491

#line 7491

#line 7491
# useradd/userdel request read/write for /var/log/lastlog, and read of /dev, 
#line 7491
# but will operate without them.
#line 7491
dontaudit groupadd_t { device_t var_t var_log_t }:dir search;
#line 7491

#line 7491
# For userdel and groupadd
#line 7491
allow groupadd_t fs_t:filesystem getattr;
#line 7491

#line 7491
# Access terminals.
#line 7491
allow groupadd_t ttyfile:chr_file { ioctl read getattr lock write append };
#line 7491
allow groupadd_t ptyfile:chr_file { ioctl read getattr lock write append };
#line 7491

#line 7491

#line 7491
# for when /root is the cwd
#line 7491
dontaudit groupadd_t sysadm_home_dir_t:dir search;
#line 7491


# some apps ask for these accesses, but seems to work regardless
dontaudit groupadd_t initrc_var_run_t:file write;
dontaudit groupadd_t var_run_t:dir search;
dontaudit groupadd_t self:capability fsetid;
##############################
#
# Assertions for the type enforcement (TE) configuration.
#

#
# Authors:  Stephen Smalley <sds@epoch.ncsc.mil> and Timothy Fraser  
#

##################################
#
# Access vector assertions.
#
# An access vector assertion specifies permissions that should not be in
# an access vector based on a source type, a target type, and a class.
# If any of the specified permissions are in the corresponding access
# vector, then the policy compiler will reject the policy configuration.
# Currently, there is only one kind of access vector assertion, neverallow, 
# but support for the other kinds of vectors could be easily added.  Access 
# vector assertions use the same syntax as access vector rules.
#

#
# Verify that every type that can be entered by
# a domain is also tagged as a domain.
#
neverallow domain ~domain:process transition;

#
# Verify that only the insmod_t, ifconfig_t, and kernel_t domains 
# have the sys_module capability.
#
neverallow ~{ insmod_t ifconfig_t kernel_t } self:capability sys_module;

#
# Verify that executable types, the system dynamic loaders, and the
# system shared libraries can only be modified by administrators.
#
neverallow ~{ldconfig_t admin} { exec_type ld_so_t shlib_t }:file { write append unlink rename };

#
# Verify that other system software can only be modified by administrators.
#
neverallow ~{ldconfig_t admin} { lib_t bin_t sbin_t }:dir { add_name remove_name rename };
neverallow ~admin { lib_t bin_t sbin_t }:file { write append unlink rename };

#
# Verify that only certain domains have access to the raw disk devices.
#
neverallow ~{  fsadm_t mount_t } fixed_disk_device_t:{ chr_file blk_file } { read write append };

#
# Verify that only the X server and klogd have access to memory devices.
#
neverallow ~privmem memory_device_t:{ chr_file blk_file } { read write append };

#
# Verify that /proc/kmsg is only accessible to klogd.
#
neverallow ~klogd_t proc_kmsg_t:file ~{ getattr };

#
# Verify that /proc/kcore is inaccessible.
#
neverallow * proc_kcore_t:file ~{ getattr };

#
# Verify that sysctl variables are only changeable
# by initrc and administrators.
#
neverallow ~{ initrc_t admin kernel_t insmod_t } sysctl_t:file { write append };
neverallow ~{ initrc_t admin } sysctl_fs_t:file { write append };
neverallow ~{ init_t initrc_t admin kernel_t insmod_t } sysctl_kernel_t:file { write append };
neverallow ~{ initrc_t admin } sysctl_net_t:file { write append };
neverallow ~{ initrc_t admin } sysctl_net_unix_t:file { write append };
neverallow ~{ initrc_t admin } sysctl_vm_t:file { write append };
neverallow ~{ initrc_t admin } sysctl_dev_t:file { write append };
neverallow ~{ initrc_t admin } sysctl_modprobe_t:file { write append };

#
# Verify that certain domains are limited to only being
# entered by their entrypoint types and to only executing
# the dynamic loader without a transition to another domain.
#

#line 7586



#line 7588
    neverallow getty_t ~getty_exec_t:file entrypoint; neverallow getty_t ~{ getty_exec_t ld_so_t }:file execute_no_trans;

#line 7589
    neverallow klogd_t ~klogd_exec_t:file entrypoint; neverallow klogd_t ~{ klogd_exec_t ld_so_t }:file execute_no_trans;




#line 7593
    neverallow syslogd_t ~syslogd_exec_t:file entrypoint; neverallow syslogd_t ~{ syslogd_exec_t ld_so_t }:file execute_no_trans;






#line 7603

#line 7603
neverallow { local_login_t remote_login_t } ~login_exec_t:file entrypoint;
#line 7603
neverallow { local_login_t remote_login_t } ~ld_so_t:file execute_no_trans;
#line 7603


#
# Verify that the passwd domain can only be entered by its
# entrypoint type and can only execute the dynamic loader
# and the ordinary passwd program without a transition to another domain.
#
#line 7612
neverallow passwd_t ~{ admin_passwd_exec_t passwd_exec_t }:file entrypoint;
#line 7612
neverallow passwd_t ~{ ld_so_t }:file execute_no_trans;

#
# Verify that only the admin domains and initrc_t have setenforce.
#
neverallow ~{ admin initrc_t } security_t:security setenforce;

#
# Verify that only the kernel and load_policy_t have load_policy.
#
neverallow ~{ kernel_t load_policy_t } security_t:security load_policy;
#line 1 "rbac"
################################################
#
# Role-based access control (RBAC) configuration.
#

#line 9

#line 13

#line 15

########################################
#
# Role allow rules.
#
# A role allow rule specifies the allowable
# transitions between roles on an execve.
# If no rule is specified, then the change in
# roles will not be permitted.  Additional
# controls over role transitions based on the
# type of the process may be specified through
# the constraints file.
#
# The syntax of a role allow rule is:
# 	allow current_role new_role ;
#

#
# Allow the system_r role to transition 
# into the sysadm_r role.
#
allow system_r sysadm_r;

#
# Allow the user and admin role to transition to httpd_admin_r
#
#line 44


# 
# Allow the admin role to transition to the system
# role for run_init.
#
allow sysadm_r system_r;
